
// http://huonw.github.io/simd-hacks/src/llvmint/lib.rs.html

//! Bindings to (almost) all LLVM intrinsics.
//!
//! Intrinsics are categorised into modules by the architecture on
//! which they are supported (however, see [Platform
//! support](#platform-support) for a caveat), with certain intrinsics
//! available at the top level. These are raw bindings and absolutely
//! zero error checking is performed.
//!
//! # Naming
//!
//! The intrinsics are named entirely for their internal LLVM names,
//! with the `llvm` prefix stripped, `.` replaced by `_` and, if it
//! exists, a platform prefix replaced by being placed in a module of
//! that name. For example, `llvm.adjust.trampoline` becomes
//! `::adjust_trampoline` and `llvm.x86.addcarry.u32` becomes
//! `::x86::addcarry_u32`.
//!
//! Many of these intrinsics have corresponding intrinsics exposed by
//! GCC/Clang in C/C++, these names are listed in
//! [`gcc_names`](gcc_names/index.html).
//!
//! # Platform support
//!
//! An intrinsic being available in a certain module (or at the top
//! level) does not guarantee that it is supported on all CPUs of that
//! architecture (resp. all CPUs), e.g. the `x86::avx512_...`
//! intrinsics are only supported on the very latest x86-64 CPUs, not
//! on older x86 processors.
//!
//! Using an intrinsic in a configuration that is not supported will
//! likely cause LLVM assertions or general badness along those lines.

module mir.llvmint;

version (LDC) {

private alias f32 = float;
private alias f64 = double;
private alias i8 = ubyte;
private alias i16 = ushort;
private alias i32 = uint;
private alias i64 = ulong;

version (ARM)
    version = ARM_Any;

version (AArch64)
    version = ARM_Any;

version (X86)
    version = X86_Any;

version (X86_64)
    version = X86_Any;

// That isn't always true
@safe pure nothrow @nogc:

/// LLVM intrinsics for the x86 architecture.
version (x86_Any) {
    /// The `llvm.x86.int` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.int")
    void _int(i8);
    /// The `llvm.x86.rdtsc` intrinsic; known as `__builtin_ia32_rdtsc` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdtsc")
    i64 rdtsc();
    /// The `llvm.x86.rdtscp` intrinsic; known as `__builtin_ia32_rdtscp` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdtscp")
    i64 rdtscp(i8*);
    /// The `llvm.x86.rdpmc` intrinsic; known as `__builtin_ia32_rdpmc` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdpmc")
    i64 rdpmc(i32);
    /// The `llvm.x86.sse.add.ss` intrinsic; known as `__builtin_ia32_addss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.add.ss")
    __vector(f32[4]) sse_add_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.sub.ss` intrinsic; known as `__builtin_ia32_subss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.sub.ss")
    __vector(f32[4]) sse_sub_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.mul.ss` intrinsic; known as `__builtin_ia32_mulss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.mul.ss")
    __vector(f32[4]) sse_mul_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.div.ss` intrinsic; known as `__builtin_ia32_divss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.div.ss")
    __vector(f32[4]) sse_div_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.sqrt.ss` intrinsic; known as `__builtin_ia32_sqrtss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.sqrt.ss")
    __vector(f32[4]) sse_sqrt_ss(__vector(f32[4]));
    /// The `llvm.x86.sse.sqrt.ps` intrinsic; known as `__builtin_ia32_sqrtps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.sqrt.ps")
    __vector(f32[4]) sse_sqrt_ps(__vector(f32[4]));
    /// The `llvm.x86.sse.rcp.ss` intrinsic; known as `__builtin_ia32_rcpss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.rcp.ss")
    __vector(f32[4]) sse_rcp_ss(__vector(f32[4]));
    /// The `llvm.x86.sse.rcp.ps` intrinsic; known as `__builtin_ia32_rcpps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.rcp.ps")
    __vector(f32[4]) sse_rcp_ps(__vector(f32[4]));
    /// The `llvm.x86.sse.rsqrt.ss` intrinsic; known as `__builtin_ia32_rsqrtss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.rsqrt.ss")
    __vector(f32[4]) sse_rsqrt_ss(__vector(f32[4]));
    /// The `llvm.x86.sse.rsqrt.ps` intrinsic; known as `__builtin_ia32_rsqrtps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.rsqrt.ps")
    __vector(f32[4]) sse_rsqrt_ps(__vector(f32[4]));
    /// The `llvm.x86.sse.min.ss` intrinsic; known as `__builtin_ia32_minss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.min.ss")
    __vector(f32[4]) sse_min_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.min.ps` intrinsic; known as `__builtin_ia32_minps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.min.ps")
    __vector(f32[4]) sse_min_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.max.ss` intrinsic; known as `__builtin_ia32_maxss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.max.ss")
    __vector(f32[4]) sse_max_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.max.ps` intrinsic; known as `__builtin_ia32_maxps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.max.ps")
    __vector(f32[4]) sse_max_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.cmp.ss` intrinsic; known as `__builtin_ia32_cmpss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cmp.ss")
    __vector(f32[4]) sse_cmp_ss(__vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.sse.cmp.ps` intrinsic; known as `__builtin_ia32_cmpps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cmp.ps")
    __vector(f32[4]) sse_cmp_ps(__vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.sse.comieq.ss` intrinsic; known as `__builtin_ia32_comieq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.comieq.ss")
    i32 sse_comieq_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.comilt.ss` intrinsic; known as `__builtin_ia32_comilt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.comilt.ss")
    i32 sse_comilt_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.comile.ss` intrinsic; known as `__builtin_ia32_comile` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.comile.ss")
    i32 sse_comile_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.comigt.ss` intrinsic; known as `__builtin_ia32_comigt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.comigt.ss")
    i32 sse_comigt_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.comige.ss` intrinsic; known as `__builtin_ia32_comige` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.comige.ss")
    i32 sse_comige_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.comineq.ss` intrinsic; known as `__builtin_ia32_comineq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.comineq.ss")
    i32 sse_comineq_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.ucomieq.ss` intrinsic; known as `__builtin_ia32_ucomieq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.ucomieq.ss")
    i32 sse_ucomieq_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.ucomilt.ss` intrinsic; known as `__builtin_ia32_ucomilt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.ucomilt.ss")
    i32 sse_ucomilt_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.ucomile.ss` intrinsic; known as `__builtin_ia32_ucomile` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.ucomile.ss")
    i32 sse_ucomile_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.ucomigt.ss` intrinsic; known as `__builtin_ia32_ucomigt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.ucomigt.ss")
    i32 sse_ucomigt_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.ucomige.ss` intrinsic; known as `__builtin_ia32_ucomige` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.ucomige.ss")
    i32 sse_ucomige_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.ucomineq.ss` intrinsic; known as `__builtin_ia32_ucomineq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.ucomineq.ss")
    i32 sse_ucomineq_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse.cvtss2si` intrinsic; known as `__builtin_ia32_cvtss2si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cvtss2si")
    i32 sse_cvtss2si(__vector(f32[4]));
    /// The `llvm.x86.sse.cvtss2si64` intrinsic; known as `__builtin_ia32_cvtss2si64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cvtss2si64")
    i64 sse_cvtss2si64(__vector(f32[4]));
    /// The `llvm.x86.sse.cvttss2si` intrinsic; known as `__builtin_ia32_cvttss2si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cvttss2si")
    i32 sse_cvttss2si(__vector(f32[4]));
    /// The `llvm.x86.sse.cvttss2si64` intrinsic; known as `__builtin_ia32_cvttss2si64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cvttss2si64")
    i64 sse_cvttss2si64(__vector(f32[4]));
    /// The `llvm.x86.sse.cvtsi2ss` intrinsic; known as `__builtin_ia32_cvtsi2ss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cvtsi2ss")
    __vector(f32[4]) sse_cvtsi2ss(__vector(f32[4]), i32);
    /// The `llvm.x86.sse.cvtsi642ss` intrinsic; known as `__builtin_ia32_cvtsi642ss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.cvtsi642ss")
    __vector(f32[4]) sse_cvtsi642ss(__vector(f32[4]), i64);
    /// The `llvm.x86.sse.storeu.ps` intrinsic; known as `__builtin_ia32_storeups` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.storeu.ps")
    void sse_storeu_ps(i8*, __vector(f32[4]));
    /// The `llvm.x86.sse.sfence` intrinsic; known as `__builtin_ia32_sfence` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.sfence")
    void sse_sfence();
    /// The `llvm.x86.sse.stmxcsr` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.sse.stmxcsr")
    void sse_stmxcsr(i8*);
    /// The `llvm.x86.sse.ldmxcsr` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.sse.ldmxcsr")
    void sse_ldmxcsr(i8*);
    /// The `llvm.x86.sse.movmsk.ps` intrinsic; known as `__builtin_ia32_movmskps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse.movmsk.ps")
    i32 sse_movmsk_ps(__vector(f32[4]));
    /// The `llvm.x86.sse2.add.sd` intrinsic; known as `__builtin_ia32_addsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.add.sd")
    __vector(f64[2]) sse2_add_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.sub.sd` intrinsic; known as `__builtin_ia32_subsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.sub.sd")
    __vector(f64[2]) sse2_sub_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.mul.sd` intrinsic; known as `__builtin_ia32_mulsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.mul.sd")
    __vector(f64[2]) sse2_mul_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.div.sd` intrinsic; known as `__builtin_ia32_divsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.div.sd")
    __vector(f64[2]) sse2_div_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.sqrt.sd` intrinsic; known as `__builtin_ia32_sqrtsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.sqrt.sd")
    __vector(f64[2]) sse2_sqrt_sd(__vector(f64[2]));
    /// The `llvm.x86.sse2.sqrt.pd` intrinsic; known as `__builtin_ia32_sqrtpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.sqrt.pd")
    __vector(f64[2]) sse2_sqrt_pd(__vector(f64[2]));
    /// The `llvm.x86.sse2.min.sd` intrinsic; known as `__builtin_ia32_minsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.min.sd")
    __vector(f64[2]) sse2_min_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.min.pd` intrinsic; known as `__builtin_ia32_minpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.min.pd")
    __vector(f64[2]) sse2_min_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.max.sd` intrinsic; known as `__builtin_ia32_maxsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.max.sd")
    __vector(f64[2]) sse2_max_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.max.pd` intrinsic; known as `__builtin_ia32_maxpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.max.pd")
    __vector(f64[2]) sse2_max_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.cmp.sd` intrinsic; known as `__builtin_ia32_cmpsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cmp.sd")
    __vector(f64[2]) sse2_cmp_sd(__vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.sse2.cmp.pd` intrinsic; known as `__builtin_ia32_cmppd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cmp.pd")
    __vector(f64[2]) sse2_cmp_pd(__vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.sse2.comieq.sd` intrinsic; known as `__builtin_ia32_comisdeq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.comieq.sd")
    i32 sse2_comieq_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.comilt.sd` intrinsic; known as `__builtin_ia32_comisdlt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.comilt.sd")
    i32 sse2_comilt_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.comile.sd` intrinsic; known as `__builtin_ia32_comisdle` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.comile.sd")
    i32 sse2_comile_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.comigt.sd` intrinsic; known as `__builtin_ia32_comisdgt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.comigt.sd")
    i32 sse2_comigt_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.comige.sd` intrinsic; known as `__builtin_ia32_comisdge` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.comige.sd")
    i32 sse2_comige_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.comineq.sd` intrinsic; known as `__builtin_ia32_comisdneq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.comineq.sd")
    i32 sse2_comineq_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.ucomieq.sd` intrinsic; known as `__builtin_ia32_ucomisdeq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.ucomieq.sd")
    i32 sse2_ucomieq_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.ucomilt.sd` intrinsic; known as `__builtin_ia32_ucomisdlt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.ucomilt.sd")
    i32 sse2_ucomilt_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.ucomile.sd` intrinsic; known as `__builtin_ia32_ucomisdle` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.ucomile.sd")
    i32 sse2_ucomile_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.ucomigt.sd` intrinsic; known as `__builtin_ia32_ucomisdgt` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.ucomigt.sd")
    i32 sse2_ucomigt_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.ucomige.sd` intrinsic; known as `__builtin_ia32_ucomisdge` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.ucomige.sd")
    i32 sse2_ucomige_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.ucomineq.sd` intrinsic; known as `__builtin_ia32_ucomisdneq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.ucomineq.sd")
    i32 sse2_ucomineq_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse2.padds.b` intrinsic; known as `__builtin_ia32_paddsb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.padds.b")
    __vector(i8[16]) sse2_padds_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.padds.w` intrinsic; known as `__builtin_ia32_paddsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.padds.w")
    __vector(i16[8]) sse2_padds_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.paddus.b` intrinsic; known as `__builtin_ia32_paddusb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.paddus.b")
    __vector(i8[16]) sse2_paddus_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.paddus.w` intrinsic; known as `__builtin_ia32_paddusw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.paddus.w")
    __vector(i16[8]) sse2_paddus_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.psubs.b` intrinsic; known as `__builtin_ia32_psubsb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psubs.b")
    __vector(i8[16]) sse2_psubs_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.psubs.w` intrinsic; known as `__builtin_ia32_psubsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psubs.w")
    __vector(i16[8]) sse2_psubs_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.psubus.b` intrinsic; known as `__builtin_ia32_psubusb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psubus.b")
    __vector(i8[16]) sse2_psubus_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.psubus.w` intrinsic; known as `__builtin_ia32_psubusw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psubus.w")
    __vector(i16[8]) sse2_psubus_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.pmulhu.w` intrinsic; known as `__builtin_ia32_pmulhuw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmulhu.w")
    __vector(i16[8]) sse2_pmulhu_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.pmulh.w` intrinsic; known as `__builtin_ia32_pmulhw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmulh.w")
    __vector(i16[8]) sse2_pmulh_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.pmulu.dq` intrinsic; known as `__builtin_ia32_pmuludq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmulu.dq")
    __vector(i64[2]) sse2_pmulu_dq(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse2.pmadd.wd` intrinsic; known as `__builtin_ia32_pmaddwd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmadd.wd")
    __vector(i32[4]) sse2_pmadd_wd(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.pavg.b` intrinsic; known as `__builtin_ia32_pavgb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pavg.b")
    __vector(i8[16]) sse2_pavg_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.pavg.w` intrinsic; known as `__builtin_ia32_pavgw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pavg.w")
    __vector(i16[8]) sse2_pavg_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.pmaxu.b` intrinsic; known as `__builtin_ia32_pmaxub128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmaxu.b")
    __vector(i8[16]) sse2_pmaxu_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.pmaxs.w` intrinsic; known as `__builtin_ia32_pmaxsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmaxs.w")
    __vector(i16[8]) sse2_pmaxs_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.pminu.b` intrinsic; known as `__builtin_ia32_pminub128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pminu.b")
    __vector(i8[16]) sse2_pminu_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.pmins.w` intrinsic; known as `__builtin_ia32_pminsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmins.w")
    __vector(i16[8]) sse2_pmins_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.psad.bw` intrinsic; known as `__builtin_ia32_psadbw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psad.bw")
    __vector(i64[2]) sse2_psad_bw(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.psll.w` intrinsic; known as `__builtin_ia32_psllw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psll.w")
    __vector(i16[8]) sse2_psll_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.psll.d` intrinsic; known as `__builtin_ia32_pslld128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psll.d")
    __vector(i32[4]) sse2_psll_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse2.psll.q` intrinsic; known as `__builtin_ia32_psllq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psll.q")
    __vector(i64[2]) sse2_psll_q(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.sse2.psrl.w` intrinsic; known as `__builtin_ia32_psrlw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrl.w")
    __vector(i16[8]) sse2_psrl_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.psrl.d` intrinsic; known as `__builtin_ia32_psrld128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrl.d")
    __vector(i32[4]) sse2_psrl_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse2.psrl.q` intrinsic; known as `__builtin_ia32_psrlq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrl.q")
    __vector(i64[2]) sse2_psrl_q(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.sse2.psra.w` intrinsic; known as `__builtin_ia32_psraw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psra.w")
    __vector(i16[8]) sse2_psra_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.psra.d` intrinsic; known as `__builtin_ia32_psrad128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psra.d")
    __vector(i32[4]) sse2_psra_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse2.pslli.w` intrinsic; known as `__builtin_ia32_psllwi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pslli.w")
    __vector(i16[8]) sse2_pslli_w(__vector(i16[8]), i32);
    /// The `llvm.x86.sse2.pslli.d` intrinsic; known as `__builtin_ia32_pslldi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pslli.d")
    __vector(i32[4]) sse2_pslli_d(__vector(i32[4]), i32);
    /// The `llvm.x86.sse2.pslli.q` intrinsic; known as `__builtin_ia32_psllqi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pslli.q")
    __vector(i64[2]) sse2_pslli_q(__vector(i64[2]), i32);
    /// The `llvm.x86.sse2.psrli.w` intrinsic; known as `__builtin_ia32_psrlwi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrli.w")
    __vector(i16[8]) sse2_psrli_w(__vector(i16[8]), i32);
    /// The `llvm.x86.sse2.psrli.d` intrinsic; known as `__builtin_ia32_psrldi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrli.d")
    __vector(i32[4]) sse2_psrli_d(__vector(i32[4]), i32);
    /// The `llvm.x86.sse2.psrli.q` intrinsic; known as `__builtin_ia32_psrlqi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrli.q")
    __vector(i64[2]) sse2_psrli_q(__vector(i64[2]), i32);
    /// The `llvm.x86.sse2.psrai.w` intrinsic; known as `__builtin_ia32_psrawi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrai.w")
    __vector(i16[8]) sse2_psrai_w(__vector(i16[8]), i32);
    /// The `llvm.x86.sse2.psrai.d` intrinsic; known as `__builtin_ia32_psradi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrai.d")
    __vector(i32[4]) sse2_psrai_d(__vector(i32[4]), i32);
    /// The `llvm.x86.sse2.psll.dq` intrinsic; known as `__builtin_ia32_pslldqi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psll.dq")
    __vector(i64[2]) sse2_psll_dq(__vector(i64[2]), i32);
    /// The `llvm.x86.sse2.psrl.dq` intrinsic; known as `__builtin_ia32_psrldqi128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrl.dq")
    __vector(i64[2]) sse2_psrl_dq(__vector(i64[2]), i32);
    /// The `llvm.x86.sse2.psll.dq.bs` intrinsic; known as `__builtin_ia32_pslldqi128_byteshift` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psll.dq.bs")
    __vector(i64[2]) sse2_psll_dq_bs(__vector(i64[2]), i32);
    /// The `llvm.x86.sse2.psrl.dq.bs` intrinsic; known as `__builtin_ia32_psrldqi128_byteshift` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.psrl.dq.bs")
    __vector(i64[2]) sse2_psrl_dq_bs(__vector(i64[2]), i32);
    /// The `llvm.x86.sse2.cvtdq2pd` intrinsic; known as `__builtin_ia32_cvtdq2pd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtdq2pd")
    __vector(f64[2]) sse2_cvtdq2pd(__vector(i32[4]));
    /// The `llvm.x86.sse2.cvtdq2ps` intrinsic; known as `__builtin_ia32_cvtdq2ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtdq2ps")
    __vector(f32[4]) sse2_cvtdq2ps(__vector(i32[4]));
    /// The `llvm.x86.sse2.cvtpd2dq` intrinsic; known as `__builtin_ia32_cvtpd2dq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtpd2dq")
    __vector(i32[4]) sse2_cvtpd2dq(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvttpd2dq` intrinsic; known as `__builtin_ia32_cvttpd2dq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvttpd2dq")
    __vector(i32[4]) sse2_cvttpd2dq(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvtpd2ps` intrinsic; known as `__builtin_ia32_cvtpd2ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtpd2ps")
    __vector(f32[4]) sse2_cvtpd2ps(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvtps2dq` intrinsic; known as `__builtin_ia32_cvtps2dq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtps2dq")
    __vector(i32[4]) sse2_cvtps2dq(__vector(f32[4]));
    /// The `llvm.x86.sse2.cvttps2dq` intrinsic; known as `__builtin_ia32_cvttps2dq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvttps2dq")
    __vector(i32[4]) sse2_cvttps2dq(__vector(f32[4]));
    /// The `llvm.x86.sse2.cvtps2pd` intrinsic; known as `__builtin_ia32_cvtps2pd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtps2pd")
    __vector(f64[2]) sse2_cvtps2pd(__vector(f32[4]));
    /// The `llvm.x86.sse2.cvtsd2si` intrinsic; known as `__builtin_ia32_cvtsd2si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtsd2si")
    i32 sse2_cvtsd2si(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvtsd2si64` intrinsic; known as `__builtin_ia32_cvtsd2si64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtsd2si64")
    i64 sse2_cvtsd2si64(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvttsd2si` intrinsic; known as `__builtin_ia32_cvttsd2si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvttsd2si")
    i32 sse2_cvttsd2si(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvttsd2si64` intrinsic; known as `__builtin_ia32_cvttsd2si64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvttsd2si64")
    i64 sse2_cvttsd2si64(__vector(f64[2]));
    /// The `llvm.x86.sse2.cvtsi2sd` intrinsic; known as `__builtin_ia32_cvtsi2sd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtsi2sd")
    __vector(f64[2]) sse2_cvtsi2sd(__vector(f64[2]), i32);
    /// The `llvm.x86.sse2.cvtsi642sd` intrinsic; known as `__builtin_ia32_cvtsi642sd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtsi642sd")
    __vector(f64[2]) sse2_cvtsi642sd(__vector(f64[2]), i64);
    /// The `llvm.x86.sse2.cvtsd2ss` intrinsic; known as `__builtin_ia32_cvtsd2ss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtsd2ss")
    __vector(f32[4]) sse2_cvtsd2ss(__vector(f32[4]), __vector(f64[2]));
    /// The `llvm.x86.sse2.cvtss2sd` intrinsic; known as `__builtin_ia32_cvtss2sd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.cvtss2sd")
    __vector(f64[2]) sse2_cvtss2sd(__vector(f64[2]), __vector(f32[4]));
    /// The `llvm.x86.sse2.storeu.pd` intrinsic; known as `__builtin_ia32_storeupd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.storeu.pd")
    void sse2_storeu_pd(i8*, __vector(f64[2]));
    /// The `llvm.x86.sse2.storeu.dq` intrinsic; known as `__builtin_ia32_storedqu` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.storeu.dq")
    void sse2_storeu_dq(i8*, __vector(i8[16]));
    /// The `llvm.x86.sse2.storel.dq` intrinsic; known as `__builtin_ia32_storelv4si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.storel.dq")
    void sse2_storel_dq(i8*, __vector(i32[4]));
    /// The `llvm.x86.sse2.packsswb.128` intrinsic; known as `__builtin_ia32_packsswb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.packsswb.128")
    __vector(i8[16]) sse2_packsswb_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.packssdw.128` intrinsic; known as `__builtin_ia32_packssdw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.packssdw.128")
    __vector(i16[8]) sse2_packssdw_128(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse2.packuswb.128` intrinsic; known as `__builtin_ia32_packuswb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.packuswb.128")
    __vector(i8[16]) sse2_packuswb_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse2.movmsk.pd` intrinsic; known as `__builtin_ia32_movmskpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.movmsk.pd")
    i32 sse2_movmsk_pd(__vector(f64[2]));
    /// The `llvm.x86.sse2.pmovmskb.128` intrinsic; known as `__builtin_ia32_pmovmskb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pmovmskb.128")
    i32 sse2_pmovmskb_128(__vector(i8[16]));
    /// The `llvm.x86.sse2.maskmov.dqu` intrinsic; known as `__builtin_ia32_maskmovdqu` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.maskmov.dqu")
    void sse2_maskmov_dqu(__vector(i8[16]), __vector(i8[16]), i8*);
    /// The `llvm.x86.sse2.clflush` intrinsic; known as `__builtin_ia32_clflush` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.clflush")
    void sse2_clflush(i8*);
    /// The `llvm.x86.sse2.lfence` intrinsic; known as `__builtin_ia32_lfence` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.lfence")
    void sse2_lfence();
    /// The `llvm.x86.sse2.mfence` intrinsic; known as `__builtin_ia32_mfence` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.mfence")
    void sse2_mfence();
    /// The `llvm.x86.sse2.pause` intrinsic; known as `__builtin_ia32_pause` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pause")
    void sse2_pause();
    /// The `llvm.x86.sse3.addsub.ps` intrinsic; known as `__builtin_ia32_addsubps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.addsub.ps")
    __vector(f32[4]) sse3_addsub_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse3.addsub.pd` intrinsic; known as `__builtin_ia32_addsubpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.addsub.pd")
    __vector(f64[2]) sse3_addsub_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse3.hadd.ps` intrinsic; known as `__builtin_ia32_haddps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.hadd.ps")
    __vector(f32[4]) sse3_hadd_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse3.hadd.pd` intrinsic; known as `__builtin_ia32_haddpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.hadd.pd")
    __vector(f64[2]) sse3_hadd_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse3.hsub.ps` intrinsic; known as `__builtin_ia32_hsubps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.hsub.ps")
    __vector(f32[4]) sse3_hsub_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse3.hsub.pd` intrinsic; known as `__builtin_ia32_hsubpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.hsub.pd")
    __vector(f64[2]) sse3_hsub_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse3.ldu.dq` intrinsic; known as `__builtin_ia32_lddqu` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.ldu.dq")
    __vector(i8[16]) sse3_ldu_dq(i8*);
    /// The `llvm.x86.sse3.monitor` intrinsic; known as `__builtin_ia32_monitor` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.monitor")
    void sse3_monitor(i8*, i32, i32);
    /// The `llvm.x86.sse3.mwait` intrinsic; known as `__builtin_ia32_mwait` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse3.mwait")
    void sse3_mwait(i32, i32);
    /// The `llvm.x86.ssse3.phadd.w.128` intrinsic; known as `__builtin_ia32_phaddw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.phadd.w.128")
    __vector(i16[8]) ssse3_phadd_w_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.ssse3.phadd.d.128` intrinsic; known as `__builtin_ia32_phaddd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.phadd.d.128")
    __vector(i32[4]) ssse3_phadd_d_128(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.ssse3.phadd.sw.128` intrinsic; known as `__builtin_ia32_phaddsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.phadd.sw.128")
    __vector(i16[8]) ssse3_phadd_sw_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.ssse3.phsub.w.128` intrinsic; known as `__builtin_ia32_phsubw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.phsub.w.128")
    __vector(i16[8]) ssse3_phsub_w_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.ssse3.phsub.d.128` intrinsic; known as `__builtin_ia32_phsubd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.phsub.d.128")
    __vector(i32[4]) ssse3_phsub_d_128(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.ssse3.phsub.sw.128` intrinsic; known as `__builtin_ia32_phsubsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.phsub.sw.128")
    __vector(i16[8]) ssse3_phsub_sw_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.ssse3.pmadd.ub.sw.128` intrinsic; known as `__builtin_ia32_pmaddubsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.pmadd.ub.sw.128")
    __vector(i16[8]) ssse3_pmadd_ub_sw_128(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.ssse3.pmul.hr.sw.128` intrinsic; known as `__builtin_ia32_pmulhrsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.pmul.hr.sw.128")
    __vector(i16[8]) ssse3_pmul_hr_sw_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.ssse3.pshuf.b.128` intrinsic; known as `__builtin_ia32_pshufb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.pshuf.b.128")
    __vector(i8[16]) ssse3_pshuf_b_128(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse2.pshuf.d` intrinsic; known as `__builtin_ia32_pshufd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pshuf.d")
    __vector(i32[4]) sse2_pshuf_d(__vector(i32[4]), i8);
    /// The `llvm.x86.sse2.pshufl.w` intrinsic; known as `__builtin_ia32_pshuflw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pshufl.w")
    __vector(i16[8]) sse2_pshufl_w(__vector(i16[8]), i8);
    /// The `llvm.x86.sse2.pshufh.w` intrinsic; known as `__builtin_ia32_pshufhw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse2.pshufh.w")
    __vector(i16[8]) sse2_pshufh_w(__vector(i16[8]), i8);
    /// The `llvm.x86.ssse3.psign.b.128` intrinsic; known as `__builtin_ia32_psignb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.psign.b.128")
    __vector(i8[16]) ssse3_psign_b_128(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.ssse3.psign.w.128` intrinsic; known as `__builtin_ia32_psignw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.psign.w.128")
    __vector(i16[8]) ssse3_psign_w_128(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.ssse3.psign.d.128` intrinsic; known as `__builtin_ia32_psignd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.psign.d.128")
    __vector(i32[4]) ssse3_psign_d_128(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.ssse3.pabs.b.128` intrinsic; known as `__builtin_ia32_pabsb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.pabs.b.128")
    __vector(i8[16]) ssse3_pabs_b_128(__vector(i8[16]));
    /// The `llvm.x86.ssse3.pabs.w.128` intrinsic; known as `__builtin_ia32_pabsw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.pabs.w.128")
    __vector(i16[8]) ssse3_pabs_w_128(__vector(i16[8]));
    /// The `llvm.x86.ssse3.pabs.d.128` intrinsic; known as `__builtin_ia32_pabsd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.ssse3.pabs.d.128")
    __vector(i32[4]) ssse3_pabs_d_128(__vector(i32[4]));
    /// The `llvm.x86.sse41.round.ss` intrinsic; known as `__builtin_ia32_roundss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.round.ss")
    __vector(f32[4]) sse41_round_ss(__vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.x86.sse41.round.ps` intrinsic; known as `__builtin_ia32_roundps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.round.ps")
    __vector(f32[4]) sse41_round_ps(__vector(f32[4]), i32);
    /// The `llvm.x86.sse41.round.sd` intrinsic; known as `__builtin_ia32_roundsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.round.sd")
    __vector(f64[2]) sse41_round_sd(__vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.x86.sse41.round.pd` intrinsic; known as `__builtin_ia32_roundpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.round.pd")
    __vector(f64[2]) sse41_round_pd(__vector(f64[2]), i32);
    /// The `llvm.x86.sse41.pmovsxbd` intrinsic; known as `__builtin_ia32_pmovsxbd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovsxbd")
    __vector(i32[4]) sse41_pmovsxbd(__vector(i8[16]));
    /// The `llvm.x86.sse41.pmovsxbq` intrinsic; known as `__builtin_ia32_pmovsxbq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovsxbq")
    __vector(i64[2]) sse41_pmovsxbq(__vector(i8[16]));
    /// The `llvm.x86.sse41.pmovsxbw` intrinsic; known as `__builtin_ia32_pmovsxbw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovsxbw")
    __vector(i16[8]) sse41_pmovsxbw(__vector(i8[16]));
    /// The `llvm.x86.sse41.pmovsxdq` intrinsic; known as `__builtin_ia32_pmovsxdq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovsxdq")
    __vector(i64[2]) sse41_pmovsxdq(__vector(i32[4]));
    /// The `llvm.x86.sse41.pmovsxwd` intrinsic; known as `__builtin_ia32_pmovsxwd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovsxwd")
    __vector(i32[4]) sse41_pmovsxwd(__vector(i16[8]));
    /// The `llvm.x86.sse41.pmovsxwq` intrinsic; known as `__builtin_ia32_pmovsxwq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovsxwq")
    __vector(i64[2]) sse41_pmovsxwq(__vector(i16[8]));
    /// The `llvm.x86.sse41.pmovzxbd` intrinsic; known as `__builtin_ia32_pmovzxbd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovzxbd")
    __vector(i32[4]) sse41_pmovzxbd(__vector(i8[16]));
    /// The `llvm.x86.sse41.pmovzxbq` intrinsic; known as `__builtin_ia32_pmovzxbq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovzxbq")
    __vector(i64[2]) sse41_pmovzxbq(__vector(i8[16]));
    /// The `llvm.x86.sse41.pmovzxbw` intrinsic; known as `__builtin_ia32_pmovzxbw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovzxbw")
    __vector(i16[8]) sse41_pmovzxbw(__vector(i8[16]));
    /// The `llvm.x86.sse41.pmovzxdq` intrinsic; known as `__builtin_ia32_pmovzxdq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovzxdq")
    __vector(i64[2]) sse41_pmovzxdq(__vector(i32[4]));
    /// The `llvm.x86.sse41.pmovzxwd` intrinsic; known as `__builtin_ia32_pmovzxwd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovzxwd")
    __vector(i32[4]) sse41_pmovzxwd(__vector(i16[8]));
    /// The `llvm.x86.sse41.pmovzxwq` intrinsic; known as `__builtin_ia32_pmovzxwq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmovzxwq")
    __vector(i64[2]) sse41_pmovzxwq(__vector(i16[8]));
    /// The `llvm.x86.sse41.phminposuw` intrinsic; known as `__builtin_ia32_phminposuw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.phminposuw")
    __vector(i16[8]) sse41_phminposuw(__vector(i16[8]));
    /// The `llvm.x86.sse41.pmaxsb` intrinsic; known as `__builtin_ia32_pmaxsb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmaxsb")
    __vector(i8[16]) sse41_pmaxsb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse41.pmaxsd` intrinsic; known as `__builtin_ia32_pmaxsd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmaxsd")
    __vector(i32[4]) sse41_pmaxsd(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse41.pmaxud` intrinsic; known as `__builtin_ia32_pmaxud128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmaxud")
    __vector(i32[4]) sse41_pmaxud(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse41.pmaxuw` intrinsic; known as `__builtin_ia32_pmaxuw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmaxuw")
    __vector(i16[8]) sse41_pmaxuw(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.sse41.pminsb` intrinsic; known as `__builtin_ia32_pminsb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pminsb")
    __vector(i8[16]) sse41_pminsb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse41.pminsd` intrinsic; known as `__builtin_ia32_pminsd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pminsd")
    __vector(i32[4]) sse41_pminsd(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse41.pminud` intrinsic; known as `__builtin_ia32_pminud128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pminud")
    __vector(i32[4]) sse41_pminud(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse41.pminuw` intrinsic; known as `__builtin_ia32_pminuw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pminuw")
    __vector(i16[8]) sse41_pminuw(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.aesni.aesimc` intrinsic; known as `__builtin_ia32_aesimc128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.aesni.aesimc")
    __vector(i64[2]) aesni_aesimc(__vector(i64[2]));
    /// The `llvm.x86.aesni.aesenc` intrinsic; known as `__builtin_ia32_aesenc128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.aesni.aesenc")
    __vector(i64[2]) aesni_aesenc(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.aesni.aesenclast` intrinsic; known as `__builtin_ia32_aesenclast128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.aesni.aesenclast")
    __vector(i64[2]) aesni_aesenclast(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.aesni.aesdec` intrinsic; known as `__builtin_ia32_aesdec128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.aesni.aesdec")
    __vector(i64[2]) aesni_aesdec(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.aesni.aesdeclast` intrinsic; known as `__builtin_ia32_aesdeclast128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.aesni.aesdeclast")
    __vector(i64[2]) aesni_aesdeclast(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.aesni.aeskeygenassist` intrinsic; known as `__builtin_ia32_aeskeygenassist128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.aesni.aeskeygenassist")
    __vector(i64[2]) aesni_aeskeygenassist(__vector(i64[2]), i8);
    /// The `llvm.x86.pclmulqdq` intrinsic; known as `__builtin_ia32_pclmulqdq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.pclmulqdq")
    __vector(i64[2]) pclmulqdq(__vector(i64[2]), __vector(i64[2]), i8);
    /// The `llvm.x86.sse41.packusdw` intrinsic; known as `__builtin_ia32_packusdw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.packusdw")
    __vector(i16[8]) sse41_packusdw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse41.pmuldq` intrinsic; known as `__builtin_ia32_pmuldq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pmuldq")
    __vector(i64[2]) sse41_pmuldq(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sse41.pextrb` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pextrb")
    i32 sse41_pextrb(__vector(i8[16]), i32);
    /// The `llvm.x86.sse41.pextrd` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pextrd")
    i32 sse41_pextrd(__vector(i32[4]), i32);
    /// The `llvm.x86.sse41.pextrq` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pextrq")
    i64 sse41_pextrq(__vector(i64[2]), i32);
    /// The `llvm.x86.sse41.extractps` intrinsic; known as `__builtin_ia32_extractps128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.extractps")
    i32 sse41_extractps(__vector(f32[4]), i32);
    /// The `llvm.x86.sse41.insertps` intrinsic; known as `__builtin_ia32_insertps128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.insertps")
    __vector(f32[4]) sse41_insertps(__vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.sse41.pblendvb` intrinsic; known as `__builtin_ia32_pblendvb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pblendvb")
    __vector(i8[16]) sse41_pblendvb(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.sse41.pblendw` intrinsic; known as `__builtin_ia32_pblendw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.pblendw")
    __vector(i16[8]) sse41_pblendw(__vector(i16[8]), __vector(i16[8]), i8);
    /// The `llvm.x86.sse41.blendpd` intrinsic; known as `__builtin_ia32_blendpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.blendpd")
    __vector(f64[2]) sse41_blendpd(__vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.sse41.blendps` intrinsic; known as `__builtin_ia32_blendps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.blendps")
    __vector(f32[4]) sse41_blendps(__vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.sse41.blendvpd` intrinsic; known as `__builtin_ia32_blendvpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.blendvpd")
    __vector(f64[2]) sse41_blendvpd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.sse41.blendvps` intrinsic; known as `__builtin_ia32_blendvps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.blendvps")
    __vector(f32[4]) sse41_blendvps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.sse41.dppd` intrinsic; known as `__builtin_ia32_dppd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.dppd")
    __vector(f64[2]) sse41_dppd(__vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.sse41.dpps` intrinsic; known as `__builtin_ia32_dpps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.dpps")
    __vector(f32[4]) sse41_dpps(__vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.sse41.mpsadbw` intrinsic; known as `__builtin_ia32_mpsadbw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.mpsadbw")
    __vector(i16[8]) sse41_mpsadbw(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse41.movntdqa` intrinsic; known as `__builtin_ia32_movntdqa` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.movntdqa")
    __vector(i64[2]) sse41_movntdqa(i8*);
    /// The `llvm.x86.sse41.ptestz` intrinsic; known as `__builtin_ia32_ptestz128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.ptestz")
    i32 sse41_ptestz(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.sse41.ptestc` intrinsic; known as `__builtin_ia32_ptestc128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.ptestc")
    i32 sse41_ptestc(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.sse41.ptestnzc` intrinsic; known as `__builtin_ia32_ptestnzc128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse41.ptestnzc")
    i32 sse41_ptestnzc(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.sse42.crc32.32.8` intrinsic; known as `__builtin_ia32_crc32qi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.crc32.32.8")
    i32 sse42_crc32_32_8(i32, i8);
    /// The `llvm.x86.sse42.crc32.32.16` intrinsic; known as `__builtin_ia32_crc32hi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.crc32.32.16")
    i32 sse42_crc32_32_16(i32, i16);
    /// The `llvm.x86.sse42.crc32.32.32` intrinsic; known as `__builtin_ia32_crc32si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.crc32.32.32")
    i32 sse42_crc32_32_32(i32, i32);
    /// The `llvm.x86.sse42.crc32.64.64` intrinsic; known as `__builtin_ia32_crc32di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.crc32.64.64")
    i64 sse42_crc32_64_64(i64, i64);
    /// The `llvm.x86.sse42.pcmpistrm128` intrinsic; known as `__builtin_ia32_pcmpistrm128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistrm128")
    __vector(i8[16]) sse42_pcmpistrm128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpistri128` intrinsic; known as `__builtin_ia32_pcmpistri128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistri128")
    i32 sse42_pcmpistri128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpistria128` intrinsic; known as `__builtin_ia32_pcmpistria128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistria128")
    i32 sse42_pcmpistria128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpistric128` intrinsic; known as `__builtin_ia32_pcmpistric128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistric128")
    i32 sse42_pcmpistric128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpistrio128` intrinsic; known as `__builtin_ia32_pcmpistrio128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistrio128")
    i32 sse42_pcmpistrio128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpistris128` intrinsic; known as `__builtin_ia32_pcmpistris128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistris128")
    i32 sse42_pcmpistris128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpistriz128` intrinsic; known as `__builtin_ia32_pcmpistriz128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpistriz128")
    i32 sse42_pcmpistriz128(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.sse42.pcmpestrm128` intrinsic; known as `__builtin_ia32_pcmpestrm128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestrm128")
    __vector(i8[16]) sse42_pcmpestrm128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse42.pcmpestri128` intrinsic; known as `__builtin_ia32_pcmpestri128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestri128")
    i32 sse42_pcmpestri128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse42.pcmpestria128` intrinsic; known as `__builtin_ia32_pcmpestria128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestria128")
    i32 sse42_pcmpestria128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse42.pcmpestric128` intrinsic; known as `__builtin_ia32_pcmpestric128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestric128")
    i32 sse42_pcmpestric128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse42.pcmpestrio128` intrinsic; known as `__builtin_ia32_pcmpestrio128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestrio128")
    i32 sse42_pcmpestrio128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse42.pcmpestris128` intrinsic; known as `__builtin_ia32_pcmpestris128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestris128")
    i32 sse42_pcmpestris128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse42.pcmpestriz128` intrinsic; known as `__builtin_ia32_pcmpestriz128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse42.pcmpestriz128")
    i32 sse42_pcmpestriz128(__vector(i8[16]), i32, __vector(i8[16]), i32, i8);
    /// The `llvm.x86.sse4a.extrqi` intrinsic; known as `__builtin_ia32_extrqi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse4a.extrqi")
    __vector(i64[2]) sse4a_extrqi(__vector(i64[2]), i8, i8);
    /// The `llvm.x86.sse4a.extrq` intrinsic; known as `__builtin_ia32_extrq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse4a.extrq")
    __vector(i64[2]) sse4a_extrq(__vector(i64[2]), __vector(i8[16]));
    /// The `llvm.x86.sse4a.insertqi` intrinsic; known as `__builtin_ia32_insertqi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse4a.insertqi")
    __vector(i64[2]) sse4a_insertqi(__vector(i64[2]), __vector(i64[2]), i8, i8);
    /// The `llvm.x86.sse4a.insertq` intrinsic; known as `__builtin_ia32_insertq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse4a.insertq")
    __vector(i64[2]) sse4a_insertq(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.sse4a.movnt.ss` intrinsic; known as `__builtin_ia32_movntss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse4a.movnt.ss")
    void sse4a_movnt_ss(i8*, __vector(f32[4]));
    /// The `llvm.x86.sse4a.movnt.sd` intrinsic; known as `__builtin_ia32_movntsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sse4a.movnt.sd")
    void sse4a_movnt_sd(i8*, __vector(f64[2]));
    /// The `llvm.x86.avx.addsub.pd.256` intrinsic; known as `__builtin_ia32_addsubpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.addsub.pd.256")
    __vector(f64[4]) avx_addsub_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.addsub.ps.256` intrinsic; known as `__builtin_ia32_addsubps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.addsub.ps.256")
    __vector(f32[8]) avx_addsub_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.max.pd.256` intrinsic; known as `__builtin_ia32_maxpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.max.pd.256")
    __vector(f64[4]) avx_max_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.max.ps.256` intrinsic; known as `__builtin_ia32_maxps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.max.ps.256")
    __vector(f32[8]) avx_max_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.min.pd.256` intrinsic; known as `__builtin_ia32_minpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.min.pd.256")
    __vector(f64[4]) avx_min_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.min.ps.256` intrinsic; known as `__builtin_ia32_minps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.min.ps.256")
    __vector(f32[8]) avx_min_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.sqrt.pd.256` intrinsic; known as `__builtin_ia32_sqrtpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.sqrt.pd.256")
    __vector(f64[4]) avx_sqrt_pd_256(__vector(f64[4]));
    /// The `llvm.x86.avx.sqrt.ps.256` intrinsic; known as `__builtin_ia32_sqrtps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.sqrt.ps.256")
    __vector(f32[8]) avx_sqrt_ps_256(__vector(f32[8]));
    /// The `llvm.x86.avx.rsqrt.ps.256` intrinsic; known as `__builtin_ia32_rsqrtps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.rsqrt.ps.256")
    __vector(f32[8]) avx_rsqrt_ps_256(__vector(f32[8]));
    /// The `llvm.x86.avx.rcp.ps.256` intrinsic; known as `__builtin_ia32_rcpps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.rcp.ps.256")
    __vector(f32[8]) avx_rcp_ps_256(__vector(f32[8]));
    /// The `llvm.x86.avx.round.pd.256` intrinsic; known as `__builtin_ia32_roundpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.round.pd.256")
    __vector(f64[4]) avx_round_pd_256(__vector(f64[4]), i32);
    /// The `llvm.x86.avx.round.ps.256` intrinsic; known as `__builtin_ia32_roundps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.round.ps.256")
    __vector(f32[8]) avx_round_ps_256(__vector(f32[8]), i32);
    /// The `llvm.x86.avx.hadd.pd.256` intrinsic; known as `__builtin_ia32_haddpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.hadd.pd.256")
    __vector(f64[4]) avx_hadd_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.hsub.ps.256` intrinsic; known as `__builtin_ia32_hsubps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.hsub.ps.256")
    __vector(f32[8]) avx_hsub_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.hsub.pd.256` intrinsic; known as `__builtin_ia32_hsubpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.hsub.pd.256")
    __vector(f64[4]) avx_hsub_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.hadd.ps.256` intrinsic; known as `__builtin_ia32_haddps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.hadd.ps.256")
    __vector(f32[8]) avx_hadd_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.vpermilvar.pd` intrinsic; known as `__builtin_ia32_vpermilvarpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vpermilvar.pd")
    __vector(f64[2]) avx_vpermilvar_pd(__vector(f64[2]), __vector(i64[2]));
    /// The `llvm.x86.avx.vpermilvar.ps` intrinsic; known as `__builtin_ia32_vpermilvarps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vpermilvar.ps")
    __vector(f32[4]) avx_vpermilvar_ps(__vector(f32[4]), __vector(i32[4]));
    /// The `llvm.x86.avx.vpermilvar.pd.256` intrinsic; known as `__builtin_ia32_vpermilvarpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vpermilvar.pd.256")
    __vector(f64[4]) avx_vpermilvar_pd_256(__vector(f64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx.vpermilvar.ps.256` intrinsic; known as `__builtin_ia32_vpermilvarps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vpermilvar.ps.256")
    __vector(f32[8]) avx_vpermilvar_ps_256(__vector(f32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx.vperm2f128.pd.256` intrinsic; known as `__builtin_ia32_vperm2f128_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vperm2f128.pd.256")
    __vector(f64[4]) avx_vperm2f128_pd_256(__vector(f64[4]), __vector(f64[4]), i8);
    /// The `llvm.x86.avx.vperm2f128.ps.256` intrinsic; known as `__builtin_ia32_vperm2f128_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vperm2f128.ps.256")
    __vector(f32[8]) avx_vperm2f128_ps_256(__vector(f32[8]), __vector(f32[8]), i8);
    /// The `llvm.x86.avx.vperm2f128.si.256` intrinsic; known as `__builtin_ia32_vperm2f128_si256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vperm2f128.si.256")
    __vector(i32[8]) avx_vperm2f128_si_256(__vector(i32[8]), __vector(i32[8]), i8);
    /// The `llvm.x86.avx512.mask.vpermt.d.512` intrinsic; known as `__builtin_ia32_vpermt2vard512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.vpermt.d.512")
    __vector(i32[16]) avx512_mask_vpermt_d_512(__vector(i32[16]), __vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.vpermt.q.512` intrinsic; known as `__builtin_ia32_vpermt2varq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.vpermt.q.512")
    __vector(i64[8]) avx512_mask_vpermt_q_512(__vector(i64[8]), __vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.vpermt.ps.512` intrinsic; known as `__builtin_ia32_vpermt2varps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.vpermt.ps.512")
    __vector(f32[16]) avx512_mask_vpermt_ps_512(__vector(i32[16]), __vector(f32[16]), __vector(f32[16]), i16);
    /// The `llvm.x86.avx512.mask.vpermt.pd.512` intrinsic; known as `__builtin_ia32_vpermt2varpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.vpermt.pd.512")
    __vector(f64[8]) avx512_mask_vpermt_pd_512(__vector(i64[8]), __vector(f64[8]), __vector(f64[8]), i8);
    /// The `llvm.x86.avx.blend.pd.256` intrinsic; known as `__builtin_ia32_blendpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.blend.pd.256")
    __vector(f64[4]) avx_blend_pd_256(__vector(f64[4]), __vector(f64[4]), i8);
    /// The `llvm.x86.avx.blend.ps.256` intrinsic; known as `__builtin_ia32_blendps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.blend.ps.256")
    __vector(f32[8]) avx_blend_ps_256(__vector(f32[8]), __vector(f32[8]), i8);
    /// The `llvm.x86.avx.blendv.pd.256` intrinsic; known as `__builtin_ia32_blendvpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.blendv.pd.256")
    __vector(f64[4]) avx_blendv_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.blendv.ps.256` intrinsic; known as `__builtin_ia32_blendvps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.blendv.ps.256")
    __vector(f32[8]) avx_blendv_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.dp.ps.256` intrinsic; known as `__builtin_ia32_dpps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.dp.ps.256")
    __vector(f32[8]) avx_dp_ps_256(__vector(f32[8]), __vector(f32[8]), i8);
    /// The `llvm.x86.avx.cmp.pd.256` intrinsic; known as `__builtin_ia32_cmppd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cmp.pd.256")
    __vector(f64[4]) avx_cmp_pd_256(__vector(f64[4]), __vector(f64[4]), i8);
    /// The `llvm.x86.avx.cmp.ps.256` intrinsic; known as `__builtin_ia32_cmpps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cmp.ps.256")
    __vector(f32[8]) avx_cmp_ps_256(__vector(f32[8]), __vector(f32[8]), i8);
    /// The `llvm.x86.avx.vextractf128.pd.256` intrinsic; known as `__builtin_ia32_vextractf128_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vextractf128.pd.256")
    __vector(f64[2]) avx_vextractf128_pd_256(__vector(f64[4]), i8);
    /// The `llvm.x86.avx.vextractf128.ps.256` intrinsic; known as `__builtin_ia32_vextractf128_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vextractf128.ps.256")
    __vector(f32[4]) avx_vextractf128_ps_256(__vector(f32[8]), i8);
    /// The `llvm.x86.avx.vextractf128.si.256` intrinsic; known as `__builtin_ia32_vextractf128_si256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vextractf128.si.256")
    __vector(i32[4]) avx_vextractf128_si_256(__vector(i32[8]), i8);
    /// The `llvm.x86.avx.vinsertf128.pd.256` intrinsic; known as `__builtin_ia32_vinsertf128_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vinsertf128.pd.256")
    __vector(f64[4]) avx_vinsertf128_pd_256(__vector(f64[4]), __vector(f64[2]), i8);
    /// The `llvm.x86.avx.vinsertf128.ps.256` intrinsic; known as `__builtin_ia32_vinsertf128_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vinsertf128.ps.256")
    __vector(f32[8]) avx_vinsertf128_ps_256(__vector(f32[8]), __vector(f32[4]), i8);
    /// The `llvm.x86.avx.vinsertf128.si.256` intrinsic; known as `__builtin_ia32_vinsertf128_si256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vinsertf128.si.256")
    __vector(i32[8]) avx_vinsertf128_si_256(__vector(i32[8]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx.cvtdq2.pd.256` intrinsic; known as `__builtin_ia32_cvtdq2pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvtdq2.pd.256")
    __vector(f64[4]) avx_cvtdq2_pd_256(__vector(i32[4]));
    /// The `llvm.x86.avx.cvtdq2.ps.256` intrinsic; known as `__builtin_ia32_cvtdq2ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvtdq2.ps.256")
    __vector(f32[8]) avx_cvtdq2_ps_256(__vector(i32[8]));
    /// The `llvm.x86.avx.cvt.pd2.ps.256` intrinsic; known as `__builtin_ia32_cvtpd2ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvt.pd2.ps.256")
    __vector(f32[4]) avx_cvt_pd2_ps_256(__vector(f64[4]));
    /// The `llvm.x86.avx.cvt.ps2dq.256` intrinsic; known as `__builtin_ia32_cvtps2dq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvt.ps2dq.256")
    __vector(i32[8]) avx_cvt_ps2dq_256(__vector(f32[8]));
    /// The `llvm.x86.avx.cvt.ps2.pd.256` intrinsic; known as `__builtin_ia32_cvtps2pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvt.ps2.pd.256")
    __vector(f64[4]) avx_cvt_ps2_pd_256(__vector(f32[4]));
    /// The `llvm.x86.avx.cvtt.pd2dq.256` intrinsic; known as `__builtin_ia32_cvttpd2dq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvtt.pd2dq.256")
    __vector(i32[4]) avx_cvtt_pd2dq_256(__vector(f64[4]));
    /// The `llvm.x86.avx.cvt.pd2dq.256` intrinsic; known as `__builtin_ia32_cvtpd2dq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvt.pd2dq.256")
    __vector(i32[4]) avx_cvt_pd2dq_256(__vector(f64[4]));
    /// The `llvm.x86.avx.cvtt.ps2dq.256` intrinsic; known as `__builtin_ia32_cvttps2dq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.cvtt.ps2dq.256")
    __vector(i32[8]) avx_cvtt_ps2dq_256(__vector(f32[8]));
    /// The `llvm.x86.avx.vtestz.pd` intrinsic; known as `__builtin_ia32_vtestzpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestz.pd")
    i32 avx_vtestz_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.avx.vtestc.pd` intrinsic; known as `__builtin_ia32_vtestcpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestc.pd")
    i32 avx_vtestc_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.avx.vtestnzc.pd` intrinsic; known as `__builtin_ia32_vtestnzcpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestnzc.pd")
    i32 avx_vtestnzc_pd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.avx.vtestz.ps` intrinsic; known as `__builtin_ia32_vtestzps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestz.ps")
    i32 avx_vtestz_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.avx.vtestc.ps` intrinsic; known as `__builtin_ia32_vtestcps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestc.ps")
    i32 avx_vtestc_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.avx.vtestnzc.ps` intrinsic; known as `__builtin_ia32_vtestnzcps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestnzc.ps")
    i32 avx_vtestnzc_ps(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.avx.vtestz.pd.256` intrinsic; known as `__builtin_ia32_vtestzpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestz.pd.256")
    i32 avx_vtestz_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.vtestc.pd.256` intrinsic; known as `__builtin_ia32_vtestcpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestc.pd.256")
    i32 avx_vtestc_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.vtestnzc.pd.256` intrinsic; known as `__builtin_ia32_vtestnzcpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestnzc.pd.256")
    i32 avx_vtestnzc_pd_256(__vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.vtestz.ps.256` intrinsic; known as `__builtin_ia32_vtestzps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestz.ps.256")
    i32 avx_vtestz_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.vtestc.ps.256` intrinsic; known as `__builtin_ia32_vtestcps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestc.ps.256")
    i32 avx_vtestc_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.vtestnzc.ps.256` intrinsic; known as `__builtin_ia32_vtestnzcps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vtestnzc.ps.256")
    i32 avx_vtestnzc_ps_256(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx.ptestz.256` intrinsic; known as `__builtin_ia32_ptestz256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.ptestz.256")
    i32 avx_ptestz_256(__vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx.ptestc.256` intrinsic; known as `__builtin_ia32_ptestc256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.ptestc.256")
    i32 avx_ptestc_256(__vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx.ptestnzc.256` intrinsic; known as `__builtin_ia32_ptestnzc256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.ptestnzc.256")
    i32 avx_ptestnzc_256(__vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx512.mask.ptestm.d.512` intrinsic; known as `__builtin_ia32_ptestmd512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.ptestm.d.512")
    i16 avx512_mask_ptestm_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.ptestm.q.512` intrinsic; known as `__builtin_ia32_ptestmq512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.ptestm.q.512")
    i8 avx512_mask_ptestm_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx.movmsk.pd.256` intrinsic; known as `__builtin_ia32_movmskpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.movmsk.pd.256")
    i32 avx_movmsk_pd_256(__vector(f64[4]));
    /// The `llvm.x86.avx.movmsk.ps.256` intrinsic; known as `__builtin_ia32_movmskps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.movmsk.ps.256")
    i32 avx_movmsk_ps_256(__vector(f32[8]));
    /// The `llvm.x86.avx.vzeroall` intrinsic; known as `__builtin_ia32_vzeroall` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vzeroall")
    void avx_vzeroall();
    /// The `llvm.x86.avx.vzeroupper` intrinsic; known as `__builtin_ia32_vzeroupper` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vzeroupper")
    void avx_vzeroupper();
    /// The `llvm.x86.avx.vbroadcastf128.pd.256` intrinsic; known as `__builtin_ia32_vbroadcastf128_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vbroadcastf128.pd.256")
    __vector(f64[4]) avx_vbroadcastf128_pd_256(i8*);
    /// The `llvm.x86.avx.vbroadcastf128.ps.256` intrinsic; known as `__builtin_ia32_vbroadcastf128_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.vbroadcastf128.ps.256")
    __vector(f32[8]) avx_vbroadcastf128_ps_256(i8*);
    /// The `llvm.x86.avx.ldu.dq.256` intrinsic; known as `__builtin_ia32_lddqu256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.ldu.dq.256")
    __vector(i8[32]) avx_ldu_dq_256(i8*);
    /// The `llvm.x86.avx.storeu.pd.256` intrinsic; known as `__builtin_ia32_storeupd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.storeu.pd.256")
    void avx_storeu_pd_256(i8*, __vector(f64[4]));
    /// The `llvm.x86.avx.storeu.ps.256` intrinsic; known as `__builtin_ia32_storeups256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.storeu.ps.256")
    void avx_storeu_ps_256(i8*, __vector(f32[8]));
    /// The `llvm.x86.avx.storeu.dq.256` intrinsic; known as `__builtin_ia32_storedqu256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.storeu.dq.256")
    void avx_storeu_dq_256(i8*, __vector(i8[32]));
    /// The `llvm.x86.avx.maskload.pd` intrinsic; known as `__builtin_ia32_maskloadpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskload.pd")
    __vector(f64[2]) avx_maskload_pd(i8*, __vector(f64[2]));
    /// The `llvm.x86.avx.maskload.ps` intrinsic; known as `__builtin_ia32_maskloadps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskload.ps")
    __vector(f32[4]) avx_maskload_ps(i8*, __vector(f32[4]));
    /// The `llvm.x86.avx.maskload.pd.256` intrinsic; known as `__builtin_ia32_maskloadpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskload.pd.256")
    __vector(f64[4]) avx_maskload_pd_256(i8*, __vector(f64[4]));
    /// The `llvm.x86.avx.maskload.ps.256` intrinsic; known as `__builtin_ia32_maskloadps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskload.ps.256")
    __vector(f32[8]) avx_maskload_ps_256(i8*, __vector(f32[8]));
    /// The `llvm.x86.avx512.mask.loadu.ps.512` intrinsic; known as `__builtin_ia32_loadups512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.loadu.ps.512")
    __vector(f32[16]) avx512_mask_loadu_ps_512(i8*, __vector(f32[16]), i16);
    /// The `llvm.x86.avx512.mask.loadu.pd.512` intrinsic; known as `__builtin_ia32_loadupd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.loadu.pd.512")
    __vector(f64[8]) avx512_mask_loadu_pd_512(i8*, __vector(f64[8]), i8);
    /// The `llvm.x86.avx.maskstore.pd` intrinsic; known as `__builtin_ia32_maskstorepd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskstore.pd")
    void avx_maskstore_pd(i8*, __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.avx.maskstore.ps` intrinsic; known as `__builtin_ia32_maskstoreps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskstore.ps")
    void avx_maskstore_ps(i8*, __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.avx.maskstore.pd.256` intrinsic; known as `__builtin_ia32_maskstorepd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskstore.pd.256")
    void avx_maskstore_pd_256(i8*, __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.avx.maskstore.ps.256` intrinsic; known as `__builtin_ia32_maskstoreps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx.maskstore.ps.256")
    void avx_maskstore_ps_256(i8*, __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx512.mask.storeu.ps.512` intrinsic; known as `__builtin_ia32_storeups512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.storeu.ps.512")
    void avx512_mask_storeu_ps_512(i8*, __vector(f32[16]), i16);
    /// The `llvm.x86.avx512.mask.storeu.pd.512` intrinsic; known as `__builtin_ia32_storeupd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.storeu.pd.512")
    void avx512_mask_storeu_pd_512(i8*, __vector(f64[8]), i8);
    /// The `llvm.x86.avx512.mask.store.ss` intrinsic; known as `__builtin_ia32_storess_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.store.ss")
    void avx512_mask_store_ss(i8*, __vector(f32[4]), i8);
    /// The `llvm.x86.avx2.padds.b` intrinsic; known as `__builtin_ia32_paddsb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.padds.b")
    __vector(i8[32]) avx2_padds_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.padds.w` intrinsic; known as `__builtin_ia32_paddsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.padds.w")
    __vector(i16[16]) avx2_padds_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.paddus.b` intrinsic; known as `__builtin_ia32_paddusb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.paddus.b")
    __vector(i8[32]) avx2_paddus_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.paddus.w` intrinsic; known as `__builtin_ia32_paddusw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.paddus.w")
    __vector(i16[16]) avx2_paddus_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.psubs.b` intrinsic; known as `__builtin_ia32_psubsb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psubs.b")
    __vector(i8[32]) avx2_psubs_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.psubs.w` intrinsic; known as `__builtin_ia32_psubsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psubs.w")
    __vector(i16[16]) avx2_psubs_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.psubus.b` intrinsic; known as `__builtin_ia32_psubusb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psubus.b")
    __vector(i8[32]) avx2_psubus_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.psubus.w` intrinsic; known as `__builtin_ia32_psubusw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psubus.w")
    __vector(i16[16]) avx2_psubus_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmulhu.w` intrinsic; known as `__builtin_ia32_pmulhuw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmulhu.w")
    __vector(i16[16]) avx2_pmulhu_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmulh.w` intrinsic; known as `__builtin_ia32_pmulhw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmulh.w")
    __vector(i16[16]) avx2_pmulh_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmulu.dq` intrinsic; known as `__builtin_ia32_pmuludq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmulu.dq")
    __vector(i64[4]) avx2_pmulu_dq(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pmul.dq` intrinsic; known as `__builtin_ia32_pmuldq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmul.dq")
    __vector(i64[4]) avx2_pmul_dq(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pmadd.wd` intrinsic; known as `__builtin_ia32_pmaddwd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmadd.wd")
    __vector(i32[8]) avx2_pmadd_wd(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pavg.b` intrinsic; known as `__builtin_ia32_pavgb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pavg.b")
    __vector(i8[32]) avx2_pavg_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.pavg.w` intrinsic; known as `__builtin_ia32_pavgw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pavg.w")
    __vector(i16[16]) avx2_pavg_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.psad.bw` intrinsic; known as `__builtin_ia32_psadbw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psad.bw")
    __vector(i64[4]) avx2_psad_bw(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx512.mask.pmulu.dq.512` intrinsic; known as `__builtin_ia32_pmuludq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmulu.dq.512")
    __vector(i64[8]) avx512_mask_pmulu_dq_512(__vector(i32[16]), __vector(i32[16]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pmul.dq.512` intrinsic; known as `__builtin_ia32_pmuldq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmul.dq.512")
    __vector(i64[8]) avx512_mask_pmul_dq_512(__vector(i32[16]), __vector(i32[16]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx2.pmaxu.b` intrinsic; known as `__builtin_ia32_pmaxub256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmaxu.b")
    __vector(i8[32]) avx2_pmaxu_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.pmaxu.w` intrinsic; known as `__builtin_ia32_pmaxuw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmaxu.w")
    __vector(i16[16]) avx2_pmaxu_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmaxu.d` intrinsic; known as `__builtin_ia32_pmaxud256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmaxu.d")
    __vector(i32[8]) avx2_pmaxu_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pmaxs.b` intrinsic; known as `__builtin_ia32_pmaxsb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmaxs.b")
    __vector(i8[32]) avx2_pmaxs_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.pmaxs.w` intrinsic; known as `__builtin_ia32_pmaxsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmaxs.w")
    __vector(i16[16]) avx2_pmaxs_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmaxs.d` intrinsic; known as `__builtin_ia32_pmaxsd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmaxs.d")
    __vector(i32[8]) avx2_pmaxs_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pminu.b` intrinsic; known as `__builtin_ia32_pminub256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pminu.b")
    __vector(i8[32]) avx2_pminu_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.pminu.w` intrinsic; known as `__builtin_ia32_pminuw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pminu.w")
    __vector(i16[16]) avx2_pminu_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pminu.d` intrinsic; known as `__builtin_ia32_pminud256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pminu.d")
    __vector(i32[8]) avx2_pminu_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pmins.b` intrinsic; known as `__builtin_ia32_pminsb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmins.b")
    __vector(i8[32]) avx2_pmins_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.pmins.w` intrinsic; known as `__builtin_ia32_pminsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmins.w")
    __vector(i16[16]) avx2_pmins_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmins.d` intrinsic; known as `__builtin_ia32_pminsd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmins.d")
    __vector(i32[8]) avx2_pmins_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx512.mask.pmaxu.d.512` intrinsic; known as `__builtin_ia32_pmaxud512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmaxu.d.512")
    __vector(i32[16]) avx512_mask_pmaxu_d_512(__vector(i32[16]), __vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pmaxs.d.512` intrinsic; known as `__builtin_ia32_pmaxsd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmaxs.d.512")
    __vector(i32[16]) avx512_mask_pmaxs_d_512(__vector(i32[16]), __vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pmaxu.q.512` intrinsic; known as `__builtin_ia32_pmaxuq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmaxu.q.512")
    __vector(i64[8]) avx512_mask_pmaxu_q_512(__vector(i64[8]), __vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pmaxs.q.512` intrinsic; known as `__builtin_ia32_pmaxsq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmaxs.q.512")
    __vector(i64[8]) avx512_mask_pmaxs_q_512(__vector(i64[8]), __vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pminu.d.512` intrinsic; known as `__builtin_ia32_pminud512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pminu.d.512")
    __vector(i32[16]) avx512_mask_pminu_d_512(__vector(i32[16]), __vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pmins.d.512` intrinsic; known as `__builtin_ia32_pminsd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmins.d.512")
    __vector(i32[16]) avx512_mask_pmins_d_512(__vector(i32[16]), __vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pminu.q.512` intrinsic; known as `__builtin_ia32_pminuq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pminu.q.512")
    __vector(i64[8]) avx512_mask_pminu_q_512(__vector(i64[8]), __vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pmins.q.512` intrinsic; known as `__builtin_ia32_pminsq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pmins.q.512")
    __vector(i64[8]) avx512_mask_pmins_q_512(__vector(i64[8]), __vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx2.psll.w` intrinsic; known as `__builtin_ia32_psllw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psll.w")
    __vector(i16[16]) avx2_psll_w(__vector(i16[16]), __vector(i16[8]));
    /// The `llvm.x86.avx2.psll.d` intrinsic; known as `__builtin_ia32_pslld256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psll.d")
    __vector(i32[8]) avx2_psll_d(__vector(i32[8]), __vector(i32[4]));
    /// The `llvm.x86.avx2.psll.q` intrinsic; known as `__builtin_ia32_psllq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psll.q")
    __vector(i64[4]) avx2_psll_q(__vector(i64[4]), __vector(i64[2]));
    /// The `llvm.x86.avx2.psrl.w` intrinsic; known as `__builtin_ia32_psrlw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrl.w")
    __vector(i16[16]) avx2_psrl_w(__vector(i16[16]), __vector(i16[8]));
    /// The `llvm.x86.avx2.psrl.d` intrinsic; known as `__builtin_ia32_psrld256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrl.d")
    __vector(i32[8]) avx2_psrl_d(__vector(i32[8]), __vector(i32[4]));
    /// The `llvm.x86.avx2.psrl.q` intrinsic; known as `__builtin_ia32_psrlq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrl.q")
    __vector(i64[4]) avx2_psrl_q(__vector(i64[4]), __vector(i64[2]));
    /// The `llvm.x86.avx2.psra.w` intrinsic; known as `__builtin_ia32_psraw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psra.w")
    __vector(i16[16]) avx2_psra_w(__vector(i16[16]), __vector(i16[8]));
    /// The `llvm.x86.avx2.psra.d` intrinsic; known as `__builtin_ia32_psrad256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psra.d")
    __vector(i32[8]) avx2_psra_d(__vector(i32[8]), __vector(i32[4]));
    /// The `llvm.x86.avx2.pslli.w` intrinsic; known as `__builtin_ia32_psllwi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pslli.w")
    __vector(i16[16]) avx2_pslli_w(__vector(i16[16]), i32);
    /// The `llvm.x86.avx2.pslli.d` intrinsic; known as `__builtin_ia32_pslldi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pslli.d")
    __vector(i32[8]) avx2_pslli_d(__vector(i32[8]), i32);
    /// The `llvm.x86.avx2.pslli.q` intrinsic; known as `__builtin_ia32_psllqi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pslli.q")
    __vector(i64[4]) avx2_pslli_q(__vector(i64[4]), i32);
    /// The `llvm.x86.avx2.psrli.w` intrinsic; known as `__builtin_ia32_psrlwi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrli.w")
    __vector(i16[16]) avx2_psrli_w(__vector(i16[16]), i32);
    /// The `llvm.x86.avx2.psrli.d` intrinsic; known as `__builtin_ia32_psrldi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrli.d")
    __vector(i32[8]) avx2_psrli_d(__vector(i32[8]), i32);
    /// The `llvm.x86.avx2.psrli.q` intrinsic; known as `__builtin_ia32_psrlqi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrli.q")
    __vector(i64[4]) avx2_psrli_q(__vector(i64[4]), i32);
    /// The `llvm.x86.avx2.psrai.w` intrinsic; known as `__builtin_ia32_psrawi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrai.w")
    __vector(i16[16]) avx2_psrai_w(__vector(i16[16]), i32);
    /// The `llvm.x86.avx2.psrai.d` intrinsic; known as `__builtin_ia32_psradi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrai.d")
    __vector(i32[8]) avx2_psrai_d(__vector(i32[8]), i32);
    /// The `llvm.x86.avx2.psll.dq` intrinsic; known as `__builtin_ia32_pslldqi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psll.dq")
    __vector(i64[4]) avx2_psll_dq(__vector(i64[4]), i32);
    /// The `llvm.x86.avx2.psrl.dq` intrinsic; known as `__builtin_ia32_psrldqi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrl.dq")
    __vector(i64[4]) avx2_psrl_dq(__vector(i64[4]), i32);
    /// The `llvm.x86.avx2.psll.dq.bs` intrinsic; known as `__builtin_ia32_pslldqi256_byteshift` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psll.dq.bs")
    __vector(i64[4]) avx2_psll_dq_bs(__vector(i64[4]), i32);
    /// The `llvm.x86.avx2.psrl.dq.bs` intrinsic; known as `__builtin_ia32_psrldqi256_byteshift` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrl.dq.bs")
    __vector(i64[4]) avx2_psrl_dq_bs(__vector(i64[4]), i32);
    /// The `llvm.x86.avx2.packsswb` intrinsic; known as `__builtin_ia32_packsswb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.packsswb")
    __vector(i8[32]) avx2_packsswb(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.packssdw` intrinsic; known as `__builtin_ia32_packssdw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.packssdw")
    __vector(i16[16]) avx2_packssdw(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.packuswb` intrinsic; known as `__builtin_ia32_packuswb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.packuswb")
    __vector(i8[32]) avx2_packuswb(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.packusdw` intrinsic; known as `__builtin_ia32_packusdw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.packusdw")
    __vector(i16[16]) avx2_packusdw(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pabs.b` intrinsic; known as `__builtin_ia32_pabsb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pabs.b")
    __vector(i8[32]) avx2_pabs_b(__vector(i8[32]));
    /// The `llvm.x86.avx2.pabs.w` intrinsic; known as `__builtin_ia32_pabsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pabs.w")
    __vector(i16[16]) avx2_pabs_w(__vector(i16[16]));
    /// The `llvm.x86.avx2.pabs.d` intrinsic; known as `__builtin_ia32_pabsd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pabs.d")
    __vector(i32[8]) avx2_pabs_d(__vector(i32[8]));
    /// The `llvm.x86.avx512.mask.pabs.d.512` intrinsic; known as `__builtin_ia32_pabsd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pabs.d.512")
    __vector(i32[16]) avx512_mask_pabs_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pabs.q.512` intrinsic; known as `__builtin_ia32_pabsq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pabs.q.512")
    __vector(i64[8]) avx512_mask_pabs_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx2.phadd.w` intrinsic; known as `__builtin_ia32_phaddw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.phadd.w")
    __vector(i16[16]) avx2_phadd_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.phadd.d` intrinsic; known as `__builtin_ia32_phaddd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.phadd.d")
    __vector(i32[8]) avx2_phadd_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.phadd.sw` intrinsic; known as `__builtin_ia32_phaddsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.phadd.sw")
    __vector(i16[16]) avx2_phadd_sw(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.phsub.w` intrinsic; known as `__builtin_ia32_phsubw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.phsub.w")
    __vector(i16[16]) avx2_phsub_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.phsub.d` intrinsic; known as `__builtin_ia32_phsubd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.phsub.d")
    __vector(i32[8]) avx2_phsub_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.phsub.sw` intrinsic; known as `__builtin_ia32_phsubsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.phsub.sw")
    __vector(i16[16]) avx2_phsub_sw(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmadd.ub.sw` intrinsic; known as `__builtin_ia32_pmaddubsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmadd.ub.sw")
    __vector(i16[16]) avx2_pmadd_ub_sw(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.psign.b` intrinsic; known as `__builtin_ia32_psignb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psign.b")
    __vector(i8[32]) avx2_psign_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.psign.w` intrinsic; known as `__builtin_ia32_psignw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psign.w")
    __vector(i16[16]) avx2_psign_w(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.psign.d` intrinsic; known as `__builtin_ia32_psignd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psign.d")
    __vector(i32[8]) avx2_psign_d(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.pmul.hr.sw` intrinsic; known as `__builtin_ia32_pmulhrsw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmul.hr.sw")
    __vector(i16[16]) avx2_pmul_hr_sw(__vector(i16[16]), __vector(i16[16]));
    /// The `llvm.x86.avx2.pmovsxbd` intrinsic; known as `__builtin_ia32_pmovsxbd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovsxbd")
    __vector(i32[8]) avx2_pmovsxbd(__vector(i8[16]));
    /// The `llvm.x86.avx2.pmovsxbq` intrinsic; known as `__builtin_ia32_pmovsxbq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovsxbq")
    __vector(i64[4]) avx2_pmovsxbq(__vector(i8[16]));
    /// The `llvm.x86.avx2.pmovsxbw` intrinsic; known as `__builtin_ia32_pmovsxbw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovsxbw")
    __vector(i16[16]) avx2_pmovsxbw(__vector(i8[16]));
    /// The `llvm.x86.avx2.pmovsxdq` intrinsic; known as `__builtin_ia32_pmovsxdq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovsxdq")
    __vector(i64[4]) avx2_pmovsxdq(__vector(i32[4]));
    /// The `llvm.x86.avx2.pmovsxwd` intrinsic; known as `__builtin_ia32_pmovsxwd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovsxwd")
    __vector(i32[8]) avx2_pmovsxwd(__vector(i16[8]));
    /// The `llvm.x86.avx2.pmovsxwq` intrinsic; known as `__builtin_ia32_pmovsxwq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovsxwq")
    __vector(i64[4]) avx2_pmovsxwq(__vector(i16[8]));
    /// The `llvm.x86.avx2.pmovzxbd` intrinsic; known as `__builtin_ia32_pmovzxbd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovzxbd")
    __vector(i32[8]) avx2_pmovzxbd(__vector(i8[16]));
    /// The `llvm.x86.avx2.pmovzxbq` intrinsic; known as `__builtin_ia32_pmovzxbq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovzxbq")
    __vector(i64[4]) avx2_pmovzxbq(__vector(i8[16]));
    /// The `llvm.x86.avx2.pmovzxbw` intrinsic; known as `__builtin_ia32_pmovzxbw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovzxbw")
    __vector(i16[16]) avx2_pmovzxbw(__vector(i8[16]));
    /// The `llvm.x86.avx2.pmovzxdq` intrinsic; known as `__builtin_ia32_pmovzxdq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovzxdq")
    __vector(i64[4]) avx2_pmovzxdq(__vector(i32[4]));
    /// The `llvm.x86.avx2.pmovzxwd` intrinsic; known as `__builtin_ia32_pmovzxwd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovzxwd")
    __vector(i32[8]) avx2_pmovzxwd(__vector(i16[8]));
    /// The `llvm.x86.avx2.pmovzxwq` intrinsic; known as `__builtin_ia32_pmovzxwq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovzxwq")
    __vector(i64[4]) avx2_pmovzxwq(__vector(i16[8]));
    /// The `llvm.x86.avx2.pblendvb` intrinsic; known as `__builtin_ia32_pblendvb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pblendvb")
    __vector(i8[32]) avx2_pblendvb(__vector(i8[32]), __vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.pblendw` intrinsic; known as `__builtin_ia32_pblendw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pblendw")
    __vector(i16[16]) avx2_pblendw(__vector(i16[16]), __vector(i16[16]), i8);
    /// The `llvm.x86.avx2.pblendd.128` intrinsic; known as `__builtin_ia32_pblendd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pblendd.128")
    __vector(i32[4]) avx2_pblendd_128(__vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx2.pblendd.256` intrinsic; known as `__builtin_ia32_pblendd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pblendd.256")
    __vector(i32[8]) avx2_pblendd_256(__vector(i32[8]), __vector(i32[8]), i8);
    /// The `llvm.x86.avx2.vbroadcast.ss.ps` intrinsic; known as `__builtin_ia32_vbroadcastss_ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vbroadcast.ss.ps")
    __vector(f32[4]) avx2_vbroadcast_ss_ps(__vector(f32[4]));
    /// The `llvm.x86.avx2.vbroadcast.sd.pd.256` intrinsic; known as `__builtin_ia32_vbroadcastsd_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vbroadcast.sd.pd.256")
    __vector(f64[4]) avx2_vbroadcast_sd_pd_256(__vector(f64[2]));
    /// The `llvm.x86.avx2.vbroadcast.ss.ps.256` intrinsic; known as `__builtin_ia32_vbroadcastss_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vbroadcast.ss.ps.256")
    __vector(f32[8]) avx2_vbroadcast_ss_ps_256(__vector(f32[4]));
    /// The `llvm.x86.avx2.vbroadcasti128` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vbroadcasti128")
    __vector(i64[4]) avx2_vbroadcasti128(i8*);
    /// The `llvm.x86.avx2.pbroadcastb.128` intrinsic; known as `__builtin_ia32_pbroadcastb128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastb.128")
    __vector(i8[16]) avx2_pbroadcastb_128(__vector(i8[16]));
    /// The `llvm.x86.avx2.pbroadcastb.256` intrinsic; known as `__builtin_ia32_pbroadcastb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastb.256")
    __vector(i8[32]) avx2_pbroadcastb_256(__vector(i8[16]));
    /// The `llvm.x86.avx2.pbroadcastw.128` intrinsic; known as `__builtin_ia32_pbroadcastw128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastw.128")
    __vector(i16[8]) avx2_pbroadcastw_128(__vector(i16[8]));
    /// The `llvm.x86.avx2.pbroadcastw.256` intrinsic; known as `__builtin_ia32_pbroadcastw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastw.256")
    __vector(i16[16]) avx2_pbroadcastw_256(__vector(i16[8]));
    /// The `llvm.x86.avx2.pbroadcastd.128` intrinsic; known as `__builtin_ia32_pbroadcastd128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastd.128")
    __vector(i32[4]) avx2_pbroadcastd_128(__vector(i32[4]));
    /// The `llvm.x86.avx2.pbroadcastd.256` intrinsic; known as `__builtin_ia32_pbroadcastd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastd.256")
    __vector(i32[8]) avx2_pbroadcastd_256(__vector(i32[4]));
    /// The `llvm.x86.avx2.pbroadcastq.128` intrinsic; known as `__builtin_ia32_pbroadcastq128` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastq.128")
    __vector(i64[2]) avx2_pbroadcastq_128(__vector(i64[2]));
    /// The `llvm.x86.avx2.pbroadcastq.256` intrinsic; known as `__builtin_ia32_pbroadcastq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pbroadcastq.256")
    __vector(i64[4]) avx2_pbroadcastq_256(__vector(i64[2]));
    /// The `llvm.x86.avx512.mask.pbroadcast.d.gpr.512` intrinsic; known as `__builtin_ia32_pbroadcastd512_gpr_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pbroadcast.d.gpr.512")
    __vector(i32[16]) avx512_mask_pbroadcast_d_gpr_512(i32, __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pbroadcast.q.gpr.512` intrinsic; known as `__builtin_ia32_pbroadcastq512_gpr_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pbroadcast.q.gpr.512")
    __vector(i64[8]) avx512_mask_pbroadcast_q_gpr_512(i64, __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pbroadcast.q.mem.512` intrinsic; known as `__builtin_ia32_pbroadcastq512_mem_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pbroadcast.q.mem.512")
    __vector(i64[8]) avx512_mask_pbroadcast_q_mem_512(i64, __vector(i64[8]), i8);
    /// The `llvm.x86.avx2.permd` intrinsic; known as `__builtin_ia32_permvarsi256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.permd")
    __vector(i32[8]) avx2_permd(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.permps` intrinsic; known as `__builtin_ia32_permvarsf256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.permps")
    __vector(f32[8]) avx2_permps(__vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.avx2.vperm2i128` intrinsic; known as `__builtin_ia32_permti256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vperm2i128")
    __vector(i64[4]) avx2_vperm2i128(__vector(i64[4]), __vector(i64[4]), i8);
    /// The `llvm.x86.avx2.vextracti128` intrinsic; known as `__builtin_ia32_extract128i256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vextracti128")
    __vector(i64[2]) avx2_vextracti128(__vector(i64[4]), i8);
    /// The `llvm.x86.avx2.vinserti128` intrinsic; known as `__builtin_ia32_insert128i256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.vinserti128")
    __vector(i64[4]) avx2_vinserti128(__vector(i64[4]), __vector(i64[2]), i8);
    /// The `llvm.x86.avx2.maskload.d` intrinsic; known as `__builtin_ia32_maskloadd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskload.d")
    __vector(i32[4]) avx2_maskload_d(i8*, __vector(i32[4]));
    /// The `llvm.x86.avx2.maskload.q` intrinsic; known as `__builtin_ia32_maskloadq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskload.q")
    __vector(i64[2]) avx2_maskload_q(i8*, __vector(i64[2]));
    /// The `llvm.x86.avx2.maskload.d.256` intrinsic; known as `__builtin_ia32_maskloadd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskload.d.256")
    __vector(i32[8]) avx2_maskload_d_256(i8*, __vector(i32[8]));
    /// The `llvm.x86.avx2.maskload.q.256` intrinsic; known as `__builtin_ia32_maskloadq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskload.q.256")
    __vector(i64[4]) avx2_maskload_q_256(i8*, __vector(i64[4]));
    /// The `llvm.x86.avx512.mask.loadu.d.512` intrinsic; known as `__builtin_ia32_loaddqusi512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.loadu.d.512")
    __vector(i32[16]) avx512_mask_loadu_d_512(i8*, __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.loadu.q.512` intrinsic; known as `__builtin_ia32_loaddqudi512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.loadu.q.512")
    __vector(i64[8]) avx512_mask_loadu_q_512(i8*, __vector(i64[8]), i8);
    /// The `llvm.x86.avx2.maskstore.d` intrinsic; known as `__builtin_ia32_maskstored` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskstore.d")
    void avx2_maskstore_d(i8*, __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.avx2.maskstore.q` intrinsic; known as `__builtin_ia32_maskstoreq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskstore.q")
    void avx2_maskstore_q(i8*, __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.avx2.maskstore.d.256` intrinsic; known as `__builtin_ia32_maskstored256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskstore.d.256")
    void avx2_maskstore_d_256(i8*, __vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.maskstore.q.256` intrinsic; known as `__builtin_ia32_maskstoreq256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.maskstore.q.256")
    void avx2_maskstore_q_256(i8*, __vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx512.mask.storeu.d.512` intrinsic; known as `__builtin_ia32_storedqusi512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.storeu.d.512")
    void avx512_mask_storeu_d_512(i8*, __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.storeu.q.512` intrinsic; known as `__builtin_ia32_storedqudi512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.storeu.q.512")
    void avx512_mask_storeu_q_512(i8*, __vector(i64[8]), i8);
    /// The `llvm.x86.avx2.psllv.d` intrinsic; known as `__builtin_ia32_psllv4si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psllv.d")
    __vector(i32[4]) avx2_psllv_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.avx2.psllv.d.256` intrinsic; known as `__builtin_ia32_psllv8si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psllv.d.256")
    __vector(i32[8]) avx2_psllv_d_256(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.psllv.q` intrinsic; known as `__builtin_ia32_psllv2di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psllv.q")
    __vector(i64[2]) avx2_psllv_q(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.avx2.psllv.q.256` intrinsic; known as `__builtin_ia32_psllv4di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psllv.q.256")
    __vector(i64[4]) avx2_psllv_q_256(__vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx2.psrlv.d` intrinsic; known as `__builtin_ia32_psrlv4si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrlv.d")
    __vector(i32[4]) avx2_psrlv_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.avx2.psrlv.d.256` intrinsic; known as `__builtin_ia32_psrlv8si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrlv.d.256")
    __vector(i32[8]) avx2_psrlv_d_256(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.psrlv.q` intrinsic; known as `__builtin_ia32_psrlv2di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrlv.q")
    __vector(i64[2]) avx2_psrlv_q(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.avx2.psrlv.q.256` intrinsic; known as `__builtin_ia32_psrlv4di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrlv.q.256")
    __vector(i64[4]) avx2_psrlv_q_256(__vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.avx2.psrav.d` intrinsic; known as `__builtin_ia32_psrav4si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrav.d")
    __vector(i32[4]) avx2_psrav_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.avx2.psrav.d.256` intrinsic; known as `__builtin_ia32_psrav8si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.psrav.d.256")
    __vector(i32[8]) avx2_psrav_d_256(__vector(i32[8]), __vector(i32[8]));
    /// The `llvm.x86.avx2.gather.d.pd` intrinsic; known as `__builtin_ia32_gatherd_pd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.pd")
    __vector(f64[2]) avx2_gather_d_pd(__vector(f64[2]), i8*, __vector(i32[4]), __vector(f64[2]), i8);
    /// The `llvm.x86.avx2.gather.d.pd.256` intrinsic; known as `__builtin_ia32_gatherd_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.pd.256")
    __vector(f64[4]) avx2_gather_d_pd_256(__vector(f64[4]), i8*, __vector(i32[4]), __vector(f64[4]), i8);
    /// The `llvm.x86.avx2.gather.q.pd` intrinsic; known as `__builtin_ia32_gatherq_pd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.pd")
    __vector(f64[2]) avx2_gather_q_pd(__vector(f64[2]), i8*, __vector(i64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.avx2.gather.q.pd.256` intrinsic; known as `__builtin_ia32_gatherq_pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.pd.256")
    __vector(f64[4]) avx2_gather_q_pd_256(__vector(f64[4]), i8*, __vector(i64[4]), __vector(f64[4]), i8);
    /// The `llvm.x86.avx2.gather.d.ps` intrinsic; known as `__builtin_ia32_gatherd_ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.ps")
    __vector(f32[4]) avx2_gather_d_ps(__vector(f32[4]), i8*, __vector(i32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.avx2.gather.d.ps.256` intrinsic; known as `__builtin_ia32_gatherd_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.ps.256")
    __vector(f32[8]) avx2_gather_d_ps_256(__vector(f32[8]), i8*, __vector(i32[8]), __vector(f32[8]), i8);
    /// The `llvm.x86.avx2.gather.q.ps` intrinsic; known as `__builtin_ia32_gatherq_ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.ps")
    __vector(f32[4]) avx2_gather_q_ps(__vector(f32[4]), i8*, __vector(i64[2]), __vector(f32[4]), i8);
    /// The `llvm.x86.avx2.gather.q.ps.256` intrinsic; known as `__builtin_ia32_gatherq_ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.ps.256")
    __vector(f32[4]) avx2_gather_q_ps_256(__vector(f32[4]), i8*, __vector(i64[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.avx2.gather.d.q` intrinsic; known as `__builtin_ia32_gatherd_q` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.q")
    __vector(i64[2]) avx2_gather_d_q(__vector(i64[2]), i8*, __vector(i32[4]), __vector(i64[2]), i8);
    /// The `llvm.x86.avx2.gather.d.q.256` intrinsic; known as `__builtin_ia32_gatherd_q256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.q.256")
    __vector(i64[4]) avx2_gather_d_q_256(__vector(i64[4]), i8*, __vector(i32[4]), __vector(i64[4]), i8);
    /// The `llvm.x86.avx2.gather.q.q` intrinsic; known as `__builtin_ia32_gatherq_q` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.q")
    __vector(i64[2]) avx2_gather_q_q(__vector(i64[2]), i8*, __vector(i64[2]), __vector(i64[2]), i8);
    /// The `llvm.x86.avx2.gather.q.q.256` intrinsic; known as `__builtin_ia32_gatherq_q256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.q.256")
    __vector(i64[4]) avx2_gather_q_q_256(__vector(i64[4]), i8*, __vector(i64[4]), __vector(i64[4]), i8);
    /// The `llvm.x86.avx2.gather.d.d` intrinsic; known as `__builtin_ia32_gatherd_d` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.d")
    __vector(i32[4]) avx2_gather_d_d(__vector(i32[4]), i8*, __vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx2.gather.d.d.256` intrinsic; known as `__builtin_ia32_gatherd_d256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.d.d.256")
    __vector(i32[8]) avx2_gather_d_d_256(__vector(i32[8]), i8*, __vector(i32[8]), __vector(i32[8]), i8);
    /// The `llvm.x86.avx2.gather.q.d` intrinsic; known as `__builtin_ia32_gatherq_d` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.d")
    __vector(i32[4]) avx2_gather_q_d(__vector(i32[4]), i8*, __vector(i64[2]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx2.gather.q.d.256` intrinsic; known as `__builtin_ia32_gatherq_d256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.gather.q.d.256")
    __vector(i32[4]) avx2_gather_q_d_256(__vector(i32[4]), i8*, __vector(i64[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx2.pmovmskb` intrinsic; known as `__builtin_ia32_pmovmskb256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pmovmskb")
    i32 avx2_pmovmskb(__vector(i8[32]));
    /// The `llvm.x86.avx2.pshuf.b` intrinsic; known as `avx2_pshuf_b` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.pshuf.b")
    __vector(i8[32]) avx2_pshuf_b(__vector(i8[32]), __vector(i8[32]));
    /// The `llvm.x86.avx2.mpsadbw` intrinsic; known as `__builtin_ia32_mpsadbw256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.mpsadbw")
    __vector(i16[16]) avx2_mpsadbw(__vector(i8[32]), __vector(i8[32]), i8);
    /// The `llvm.x86.avx2.movntdqa` intrinsic; known as `__builtin_ia32_movntdqa256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx2.movntdqa")
    __vector(i64[4]) avx2_movntdqa(i8*);
    /// The `llvm.x86.fma.vfmadd.ss` intrinsic; known as `__builtin_ia32_vfmaddss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmadd.ss")
    __vector(f32[4]) fma_vfmadd_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfmadd.sd` intrinsic; known as `__builtin_ia32_vfmaddsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmadd.sd")
    __vector(f64[2]) fma_vfmadd_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfmadd.ps` intrinsic; known as `__builtin_ia32_vfmaddps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmadd.ps")
    __vector(f32[4]) fma_vfmadd_ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfmadd.pd` intrinsic; known as `__builtin_ia32_vfmaddpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmadd.pd")
    __vector(f64[2]) fma_vfmadd_pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfmadd.ps.256` intrinsic; known as `__builtin_ia32_vfmaddps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmadd.ps.256")
    __vector(f32[8]) fma_vfmadd_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.fma.vfmadd.pd.256` intrinsic; known as `__builtin_ia32_vfmaddpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmadd.pd.256")
    __vector(f64[4]) fma_vfmadd_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.fma.mask.vfmadd.ps.512` intrinsic; known as `__builtin_ia32_vfmaddps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmadd.ps.512")
    __vector(f32[16]) fma_mask_vfmadd_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.fma.mask.vfmadd.pd.512` intrinsic; known as `__builtin_ia32_vfmaddpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmadd.pd.512")
    __vector(f64[8]) fma_mask_vfmadd_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.fma.vfmsub.ss` intrinsic; known as `__builtin_ia32_vfmsubss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsub.ss")
    __vector(f32[4]) fma_vfmsub_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfmsub.sd` intrinsic; known as `__builtin_ia32_vfmsubsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsub.sd")
    __vector(f64[2]) fma_vfmsub_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfmsub.ps` intrinsic; known as `__builtin_ia32_vfmsubps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsub.ps")
    __vector(f32[4]) fma_vfmsub_ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfmsub.pd` intrinsic; known as `__builtin_ia32_vfmsubpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsub.pd")
    __vector(f64[2]) fma_vfmsub_pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfmsub.ps.256` intrinsic; known as `__builtin_ia32_vfmsubps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsub.ps.256")
    __vector(f32[8]) fma_vfmsub_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.fma.vfmsub.pd.256` intrinsic; known as `__builtin_ia32_vfmsubpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsub.pd.256")
    __vector(f64[4]) fma_vfmsub_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.fma.mask.vfmsub.ps.512` intrinsic; known as `__builtin_ia32_vfmsubps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmsub.ps.512")
    __vector(f32[16]) fma_mask_vfmsub_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.fma.mask.vfmsub.pd.512` intrinsic; known as `__builtin_ia32_vfmsubpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmsub.pd.512")
    __vector(f64[8]) fma_mask_vfmsub_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.fma.vfnmadd.ss` intrinsic; known as `__builtin_ia32_vfnmaddss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmadd.ss")
    __vector(f32[4]) fma_vfnmadd_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfnmadd.sd` intrinsic; known as `__builtin_ia32_vfnmaddsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmadd.sd")
    __vector(f64[2]) fma_vfnmadd_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfnmadd.ps` intrinsic; known as `__builtin_ia32_vfnmaddps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmadd.ps")
    __vector(f32[4]) fma_vfnmadd_ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfnmadd.pd` intrinsic; known as `__builtin_ia32_vfnmaddpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmadd.pd")
    __vector(f64[2]) fma_vfnmadd_pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfnmadd.ps.256` intrinsic; known as `__builtin_ia32_vfnmaddps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmadd.ps.256")
    __vector(f32[8]) fma_vfnmadd_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.fma.vfnmadd.pd.256` intrinsic; known as `__builtin_ia32_vfnmaddpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmadd.pd.256")
    __vector(f64[4]) fma_vfnmadd_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.fma.mask.vfnmadd.ps.512` intrinsic; known as `__builtin_ia32_vfnmaddps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfnmadd.ps.512")
    __vector(f32[16]) fma_mask_vfnmadd_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.fma.mask.vfnmadd.pd.512` intrinsic; known as `__builtin_ia32_vfnmaddpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfnmadd.pd.512")
    __vector(f64[8]) fma_mask_vfnmadd_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.fma.vfnmsub.ss` intrinsic; known as `__builtin_ia32_vfnmsubss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmsub.ss")
    __vector(f32[4]) fma_vfnmsub_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfnmsub.sd` intrinsic; known as `__builtin_ia32_vfnmsubsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmsub.sd")
    __vector(f64[2]) fma_vfnmsub_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfnmsub.ps` intrinsic; known as `__builtin_ia32_vfnmsubps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmsub.ps")
    __vector(f32[4]) fma_vfnmsub_ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfnmsub.pd` intrinsic; known as `__builtin_ia32_vfnmsubpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmsub.pd")
    __vector(f64[2]) fma_vfnmsub_pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfnmsub.ps.256` intrinsic; known as `__builtin_ia32_vfnmsubps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmsub.ps.256")
    __vector(f32[8]) fma_vfnmsub_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.fma.vfnmsub.pd.256` intrinsic; known as `__builtin_ia32_vfnmsubpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfnmsub.pd.256")
    __vector(f64[4]) fma_vfnmsub_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.fma.mask.vfnmsub.ps.512` intrinsic; known as `__builtin_ia32_vfnmsubps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfnmsub.ps.512")
    __vector(f32[16]) fma_mask_vfnmsub_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.fma.mask.vfnmsub.pd.512` intrinsic; known as `__builtin_ia32_vfnmsubpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfnmsub.pd.512")
    __vector(f64[8]) fma_mask_vfnmsub_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.fma.vfmaddsub.ps` intrinsic; known as `__builtin_ia32_vfmaddsubps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmaddsub.ps")
    __vector(f32[4]) fma_vfmaddsub_ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfmaddsub.pd` intrinsic; known as `__builtin_ia32_vfmaddsubpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmaddsub.pd")
    __vector(f64[2]) fma_vfmaddsub_pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfmaddsub.ps.256` intrinsic; known as `__builtin_ia32_vfmaddsubps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmaddsub.ps.256")
    __vector(f32[8]) fma_vfmaddsub_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.fma.vfmaddsub.pd.256` intrinsic; known as `__builtin_ia32_vfmaddsubpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmaddsub.pd.256")
    __vector(f64[4]) fma_vfmaddsub_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.fma.mask.vfmaddsub.ps.512` intrinsic; known as `__builtin_ia32_vfmaddsubps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmaddsub.ps.512")
    __vector(f32[16]) fma_mask_vfmaddsub_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.fma.mask.vfmaddsub.pd.512` intrinsic; known as `__builtin_ia32_vfmaddsubpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmaddsub.pd.512")
    __vector(f64[8]) fma_mask_vfmaddsub_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.fma.vfmsubadd.ps` intrinsic; known as `__builtin_ia32_vfmsubaddps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsubadd.ps")
    __vector(f32[4]) fma_vfmsubadd_ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.fma.vfmsubadd.pd` intrinsic; known as `__builtin_ia32_vfmsubaddpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsubadd.pd")
    __vector(f64[2]) fma_vfmsubadd_pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.fma.vfmsubadd.ps.256` intrinsic; known as `__builtin_ia32_vfmsubaddps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsubadd.ps.256")
    __vector(f32[8]) fma_vfmsubadd_ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]));
    /// The `llvm.x86.fma.vfmsubadd.pd.256` intrinsic; known as `__builtin_ia32_vfmsubaddpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.vfmsubadd.pd.256")
    __vector(f64[4]) fma_vfmsubadd_pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]));
    /// The `llvm.x86.fma.mask.vfmsubadd.ps.512` intrinsic; known as `__builtin_ia32_vfmsubaddps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmsubadd.ps.512")
    __vector(f32[16]) fma_mask_vfmsubadd_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.fma.mask.vfmsubadd.pd.512` intrinsic; known as `__builtin_ia32_vfmsubaddpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.fma.mask.vfmsubadd.pd.512")
    __vector(f64[8]) fma_mask_vfmsubadd_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.xop.vpermil2pd` intrinsic; known as `__builtin_ia32_vpermil2pd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpermil2pd")
    __vector(f64[2]) xop_vpermil2pd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.xop.vpermil2pd.256` intrinsic; known as `__builtin_ia32_vpermil2pd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpermil2pd.256")
    __vector(f64[4]) xop_vpermil2pd_256(__vector(f64[4]), __vector(f64[4]), __vector(f64[4]), i8);
    /// The `llvm.x86.xop.vpermil2ps` intrinsic; known as `__builtin_ia32_vpermil2ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpermil2ps")
    __vector(f32[4]) xop_vpermil2ps(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.xop.vpermil2ps.256` intrinsic; known as `__builtin_ia32_vpermil2ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpermil2ps.256")
    __vector(f32[8]) xop_vpermil2ps_256(__vector(f32[8]), __vector(f32[8]), __vector(f32[8]), i8);
    /// The `llvm.x86.xop.vfrcz.pd` intrinsic; known as `__builtin_ia32_vfrczpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vfrcz.pd")
    __vector(f64[2]) xop_vfrcz_pd(__vector(f64[2]));
    /// The `llvm.x86.xop.vfrcz.ps` intrinsic; known as `__builtin_ia32_vfrczps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vfrcz.ps")
    __vector(f32[4]) xop_vfrcz_ps(__vector(f32[4]));
    /// The `llvm.x86.xop.vfrcz.sd` intrinsic; known as `__builtin_ia32_vfrczsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vfrcz.sd")
    __vector(f64[2]) xop_vfrcz_sd(__vector(f64[2]));
    /// The `llvm.x86.xop.vfrcz.ss` intrinsic; known as `__builtin_ia32_vfrczss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vfrcz.ss")
    __vector(f32[4]) xop_vfrcz_ss(__vector(f32[4]));
    /// The `llvm.x86.xop.vfrcz.pd.256` intrinsic; known as `__builtin_ia32_vfrczpd256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vfrcz.pd.256")
    __vector(f64[4]) xop_vfrcz_pd_256(__vector(f64[4]));
    /// The `llvm.x86.xop.vfrcz.ps.256` intrinsic; known as `__builtin_ia32_vfrczps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vfrcz.ps.256")
    __vector(f32[8]) xop_vfrcz_ps_256(__vector(f32[8]));
    /// The `llvm.x86.xop.vpcmov` intrinsic; known as `__builtin_ia32_vpcmov` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcmov")
    __vector(i64[2]) xop_vpcmov(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpcmov.256` intrinsic; known as `__builtin_ia32_vpcmov_256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcmov.256")
    __vector(i64[4]) xop_vpcmov_256(__vector(i64[4]), __vector(i64[4]), __vector(i64[4]));
    /// The `llvm.x86.xop.vpcomb` intrinsic; known as `__builtin_ia32_vpcomb` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomb")
    __vector(i8[16]) xop_vpcomb(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.xop.vpcomw` intrinsic; known as `__builtin_ia32_vpcomw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomw")
    __vector(i16[8]) xop_vpcomw(__vector(i16[8]), __vector(i16[8]), i8);
    /// The `llvm.x86.xop.vpcomd` intrinsic; known as `__builtin_ia32_vpcomd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomd")
    __vector(i32[4]) xop_vpcomd(__vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.xop.vpcomq` intrinsic; known as `__builtin_ia32_vpcomq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomq")
    __vector(i64[2]) xop_vpcomq(__vector(i64[2]), __vector(i64[2]), i8);
    /// The `llvm.x86.xop.vpcomub` intrinsic; known as `__builtin_ia32_vpcomub` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomub")
    __vector(i8[16]) xop_vpcomub(__vector(i8[16]), __vector(i8[16]), i8);
    /// The `llvm.x86.xop.vpcomuw` intrinsic; known as `__builtin_ia32_vpcomuw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomuw")
    __vector(i16[8]) xop_vpcomuw(__vector(i16[8]), __vector(i16[8]), i8);
    /// The `llvm.x86.xop.vpcomud` intrinsic; known as `__builtin_ia32_vpcomud` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomud")
    __vector(i32[4]) xop_vpcomud(__vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.xop.vpcomuq` intrinsic; known as `__builtin_ia32_vpcomuq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpcomuq")
    __vector(i64[2]) xop_vpcomuq(__vector(i64[2]), __vector(i64[2]), i8);
    /// The `llvm.x86.xop.vphaddbd` intrinsic; known as `__builtin_ia32_vphaddbd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddbd")
    __vector(i32[4]) xop_vphaddbd(__vector(i8[16]));
    /// The `llvm.x86.xop.vphaddbq` intrinsic; known as `__builtin_ia32_vphaddbq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddbq")
    __vector(i64[2]) xop_vphaddbq(__vector(i8[16]));
    /// The `llvm.x86.xop.vphaddbw` intrinsic; known as `__builtin_ia32_vphaddbw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddbw")
    __vector(i16[8]) xop_vphaddbw(__vector(i8[16]));
    /// The `llvm.x86.xop.vphadddq` intrinsic; known as `__builtin_ia32_vphadddq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphadddq")
    __vector(i64[2]) xop_vphadddq(__vector(i32[4]));
    /// The `llvm.x86.xop.vphaddubd` intrinsic; known as `__builtin_ia32_vphaddubd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddubd")
    __vector(i32[4]) xop_vphaddubd(__vector(i8[16]));
    /// The `llvm.x86.xop.vphaddubq` intrinsic; known as `__builtin_ia32_vphaddubq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddubq")
    __vector(i64[2]) xop_vphaddubq(__vector(i8[16]));
    /// The `llvm.x86.xop.vphaddubw` intrinsic; known as `__builtin_ia32_vphaddubw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddubw")
    __vector(i16[8]) xop_vphaddubw(__vector(i8[16]));
    /// The `llvm.x86.xop.vphaddudq` intrinsic; known as `__builtin_ia32_vphaddudq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddudq")
    __vector(i64[2]) xop_vphaddudq(__vector(i32[4]));
    /// The `llvm.x86.xop.vphadduwd` intrinsic; known as `__builtin_ia32_vphadduwd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphadduwd")
    __vector(i32[4]) xop_vphadduwd(__vector(i16[8]));
    /// The `llvm.x86.xop.vphadduwq` intrinsic; known as `__builtin_ia32_vphadduwq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphadduwq")
    __vector(i64[2]) xop_vphadduwq(__vector(i16[8]));
    /// The `llvm.x86.xop.vphaddwd` intrinsic; known as `__builtin_ia32_vphaddwd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddwd")
    __vector(i32[4]) xop_vphaddwd(__vector(i16[8]));
    /// The `llvm.x86.xop.vphaddwq` intrinsic; known as `__builtin_ia32_vphaddwq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphaddwq")
    __vector(i64[2]) xop_vphaddwq(__vector(i16[8]));
    /// The `llvm.x86.xop.vphsubbw` intrinsic; known as `__builtin_ia32_vphsubbw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphsubbw")
    __vector(i16[8]) xop_vphsubbw(__vector(i8[16]));
    /// The `llvm.x86.xop.vphsubdq` intrinsic; known as `__builtin_ia32_vphsubdq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphsubdq")
    __vector(i64[2]) xop_vphsubdq(__vector(i32[4]));
    /// The `llvm.x86.xop.vphsubwd` intrinsic; known as `__builtin_ia32_vphsubwd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vphsubwd")
    __vector(i32[4]) xop_vphsubwd(__vector(i16[8]));
    /// The `llvm.x86.xop.vpmacsdd` intrinsic; known as `__builtin_ia32_vpmacsdd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacsdd")
    __vector(i32[4]) xop_vpmacsdd(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpmacsdqh` intrinsic; known as `__builtin_ia32_vpmacsdqh` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacsdqh")
    __vector(i64[2]) xop_vpmacsdqh(__vector(i32[4]), __vector(i32[4]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpmacsdql` intrinsic; known as `__builtin_ia32_vpmacsdql` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacsdql")
    __vector(i64[2]) xop_vpmacsdql(__vector(i32[4]), __vector(i32[4]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpmacssdd` intrinsic; known as `__builtin_ia32_vpmacssdd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacssdd")
    __vector(i32[4]) xop_vpmacssdd(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpmacssdqh` intrinsic; known as `__builtin_ia32_vpmacssdqh` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacssdqh")
    __vector(i64[2]) xop_vpmacssdqh(__vector(i32[4]), __vector(i32[4]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpmacssdql` intrinsic; known as `__builtin_ia32_vpmacssdql` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacssdql")
    __vector(i64[2]) xop_vpmacssdql(__vector(i32[4]), __vector(i32[4]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpmacsswd` intrinsic; known as `__builtin_ia32_vpmacsswd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacsswd")
    __vector(i32[4]) xop_vpmacsswd(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpmacssww` intrinsic; known as `__builtin_ia32_vpmacssww` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacssww")
    __vector(i16[8]) xop_vpmacssww(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.xop.vpmacswd` intrinsic; known as `__builtin_ia32_vpmacswd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacswd")
    __vector(i32[4]) xop_vpmacswd(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpmacsww` intrinsic; known as `__builtin_ia32_vpmacsww` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmacsww")
    __vector(i16[8]) xop_vpmacsww(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.xop.vpmadcsswd` intrinsic; known as `__builtin_ia32_vpmadcsswd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmadcsswd")
    __vector(i32[4]) xop_vpmadcsswd(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpmadcswd` intrinsic; known as `__builtin_ia32_vpmadcswd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpmadcswd")
    __vector(i32[4]) xop_vpmadcswd(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpperm` intrinsic; known as `__builtin_ia32_vpperm` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpperm")
    __vector(i8[16]) xop_vpperm(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.xop.vprotb` intrinsic; known as `__builtin_ia32_vprotb` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotb")
    __vector(i8[16]) xop_vprotb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.xop.vprotd` intrinsic; known as `__builtin_ia32_vprotd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotd")
    __vector(i32[4]) xop_vprotd(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.xop.vprotq` intrinsic; known as `__builtin_ia32_vprotq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotq")
    __vector(i64[2]) xop_vprotq(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.xop.vprotw` intrinsic; known as `__builtin_ia32_vprotw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotw")
    __vector(i16[8]) xop_vprotw(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.xop.vprotbi` intrinsic; known as `__builtin_ia32_vprotbi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotbi")
    __vector(i8[16]) xop_vprotbi(__vector(i8[16]), i8);
    /// The `llvm.x86.xop.vprotdi` intrinsic; known as `__builtin_ia32_vprotdi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotdi")
    __vector(i32[4]) xop_vprotdi(__vector(i32[4]), i8);
    /// The `llvm.x86.xop.vprotqi` intrinsic; known as `__builtin_ia32_vprotqi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotqi")
    __vector(i64[2]) xop_vprotqi(__vector(i64[2]), i8);
    /// The `llvm.x86.xop.vprotwi` intrinsic; known as `__builtin_ia32_vprotwi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vprotwi")
    __vector(i16[8]) xop_vprotwi(__vector(i16[8]), i8);
    /// The `llvm.x86.xop.vpshab` intrinsic; known as `__builtin_ia32_vpshab` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshab")
    __vector(i8[16]) xop_vpshab(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.xop.vpshad` intrinsic; known as `__builtin_ia32_vpshad` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshad")
    __vector(i32[4]) xop_vpshad(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpshaq` intrinsic; known as `__builtin_ia32_vpshaq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshaq")
    __vector(i64[2]) xop_vpshaq(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpshaw` intrinsic; known as `__builtin_ia32_vpshaw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshaw")
    __vector(i16[8]) xop_vpshaw(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.xop.vpshlb` intrinsic; known as `__builtin_ia32_vpshlb` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshlb")
    __vector(i8[16]) xop_vpshlb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.x86.xop.vpshld` intrinsic; known as `__builtin_ia32_vpshld` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshld")
    __vector(i32[4]) xop_vpshld(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.xop.vpshlq` intrinsic; known as `__builtin_ia32_vpshlq` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshlq")
    __vector(i64[2]) xop_vpshlq(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.x86.xop.vpshlw` intrinsic; known as `__builtin_ia32_vpshlw` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xop.vpshlw")
    __vector(i16[8]) xop_vpshlw(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.x86.mmx.emms` intrinsic; known as `__builtin_ia32_emms` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.mmx.emms")
    void mmx_emms();
    /// The `llvm.x86.mmx.femms` intrinsic; known as `__builtin_ia32_femms` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.mmx.femms")
    void mmx_femms();
    /// The `llvm.x86.bmi.bextr.32` intrinsic; known as `__builtin_ia32_bextr_u32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.bextr.32")
    i32 bmi_bextr_32(i32, i32);
    /// The `llvm.x86.bmi.bextr.64` intrinsic; known as `__builtin_ia32_bextr_u64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.bextr.64")
    i64 bmi_bextr_64(i64, i64);
    /// The `llvm.x86.bmi.bzhi.32` intrinsic; known as `__builtin_ia32_bzhi_si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.bzhi.32")
    i32 bmi_bzhi_32(i32, i32);
    /// The `llvm.x86.bmi.bzhi.64` intrinsic; known as `__builtin_ia32_bzhi_di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.bzhi.64")
    i64 bmi_bzhi_64(i64, i64);
    /// The `llvm.x86.bmi.pdep.32` intrinsic; known as `__builtin_ia32_pdep_si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.pdep.32")
    i32 bmi_pdep_32(i32, i32);
    /// The `llvm.x86.bmi.pdep.64` intrinsic; known as `__builtin_ia32_pdep_di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.pdep.64")
    i64 bmi_pdep_64(i64, i64);
    /// The `llvm.x86.bmi.pext.32` intrinsic; known as `__builtin_ia32_pext_si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.pext.32")
    i32 bmi_pext_32(i32, i32);
    /// The `llvm.x86.bmi.pext.64` intrinsic; known as `__builtin_ia32_pext_di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.bmi.pext.64")
    i64 bmi_pext_64(i64, i64);
    /// The `llvm.x86.rdfsbase.32` intrinsic; known as `__builtin_ia32_rdfsbase32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdfsbase.32")
    i32 rdfsbase_32();
    /// The `llvm.x86.rdgsbase.32` intrinsic; known as `__builtin_ia32_rdgsbase32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdgsbase.32")
    i32 rdgsbase_32();
    /// The `llvm.x86.rdfsbase.64` intrinsic; known as `__builtin_ia32_rdfsbase64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdfsbase.64")
    i64 rdfsbase_64();
    /// The `llvm.x86.rdgsbase.64` intrinsic; known as `__builtin_ia32_rdgsbase64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.rdgsbase.64")
    i64 rdgsbase_64();
    /// The `llvm.x86.wrfsbase.32` intrinsic; known as `__builtin_ia32_wrfsbase32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.wrfsbase.32")
    void wrfsbase_32(i32);
    /// The `llvm.x86.wrgsbase.32` intrinsic; known as `__builtin_ia32_wrgsbase32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.wrgsbase.32")
    void wrgsbase_32(i32);
    /// The `llvm.x86.wrfsbase.64` intrinsic; known as `__builtin_ia32_wrfsbase64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.wrfsbase.64")
    void wrfsbase_64(i64);
    /// The `llvm.x86.wrgsbase.64` intrinsic; known as `__builtin_ia32_wrgsbase64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.wrgsbase.64")
    void wrgsbase_64(i64);
    /// The `llvm.x86.vcvtph2ps.128` intrinsic; known as `__builtin_ia32_vcvtph2ps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.vcvtph2ps.128")
    __vector(f32[4]) vcvtph2ps_128(__vector(i16[8]));
    /// The `llvm.x86.vcvtph2ps.256` intrinsic; known as `__builtin_ia32_vcvtph2ps256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.vcvtph2ps.256")
    __vector(f32[8]) vcvtph2ps_256(__vector(i16[8]));
    /// The `llvm.x86.vcvtps2ph.128` intrinsic; known as `__builtin_ia32_vcvtps2ph` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.vcvtps2ph.128")
    __vector(i16[8]) vcvtps2ph_128(__vector(f32[4]), i32);
    /// The `llvm.x86.vcvtps2ph.256` intrinsic; known as `__builtin_ia32_vcvtps2ph256` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.vcvtps2ph.256")
    __vector(i16[8]) vcvtps2ph_256(__vector(f32[8]), i32);
    /// The `llvm.x86.avx512.mask.vcvtph2ps.512` intrinsic; known as `__builtin_ia32_vcvtph2ps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.vcvtph2ps.512")
    __vector(f32[16]) avx512_mask_vcvtph2ps_512(__vector(i16[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.vcvtps2ph.512` intrinsic; known as `__builtin_ia32_vcvtps2ph512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.vcvtps2ph.512")
    __vector(i16[16]) avx512_mask_vcvtps2ph_512(__vector(f32[16]), i32, __vector(i16[16]), i16);
    /// The `llvm.x86.tbm.bextri.u32` intrinsic; known as `__builtin_ia32_bextri_u32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.tbm.bextri.u32")
    i32 tbm_bextri_u32(i32, i32);
    /// The `llvm.x86.tbm.bextri.u64` intrinsic; known as `__builtin_ia32_bextri_u64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.tbm.bextri.u64")
    i64 tbm_bextri_u64(i64, i64);
    /// The `llvm.x86.addcarryx.u32` intrinsic; known as `__builtin_ia32_addcarryx_u32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.addcarryx.u32")
    i8 addcarryx_u32(i8, i32, i32, i8*);
    /// The `llvm.x86.addcarryx.u64` intrinsic; known as `__builtin_ia32_addcarryx_u64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.addcarryx.u64")
    i8 addcarryx_u64(i8, i64, i64, i8*);
    /// The `llvm.x86.addcarry.u32` intrinsic; known as `__builtin_ia32_addcarry_u32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.addcarry.u32")
    i8 addcarry_u32(i8, i32, i32, i8*);
    /// The `llvm.x86.addcarry.u64` intrinsic; known as `__builtin_ia32_addcarry_u64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.addcarry.u64")
    i8 addcarry_u64(i8, i64, i64, i8*);
    /// The `llvm.x86.subborrow.u32` intrinsic; known as `__builtin_ia32_subborrow_u32` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.subborrow.u32")
    i8 subborrow_u32(i8, i32, i32, i8*);
    /// The `llvm.x86.subborrow.u64` intrinsic; known as `__builtin_ia32_subborrow_u64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.subborrow.u64")
    i8 subborrow_u64(i8, i64, i64, i8*);
    /// The `llvm.x86.xbegin` intrinsic; known as `__builtin_ia32_xbegin` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xbegin")
    i32 xbegin();
    /// The `llvm.x86.xend` intrinsic; known as `__builtin_ia32_xend` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xend")
    void xend();
    /// The `llvm.x86.xabort` intrinsic; known as `__builtin_ia32_xabort` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xabort")
    void xabort(i8);
    /// The `llvm.x86.xtest` intrinsic; known as `__builtin_ia32_xtest` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.xtest")
    i32 xtest();

    // ======= Mir Addition:
    pragma(LDC_intrinsic, "llvm.x86.avx512.pshuf.b.512")
    __vector(ubyte[64]) avx512_pshuf_b_512(__vector(ubyte[64]), __vector(ubyte[64]));
    // =======

    /// The `llvm.x86.avx512.kand.w` intrinsic; known as `__builtin_ia32_kandhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kand.w")
    i16 avx512_kand_w(i16, i16);
    /// The `llvm.x86.avx512.kandn.w` intrinsic; known as `__builtin_ia32_kandnhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kandn.w")
    i16 avx512_kandn_w(i16, i16);
    /// The `llvm.x86.avx512.knot.w` intrinsic; known as `__builtin_ia32_knothi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.knot.w")
    i16 avx512_knot_w(i16);
    /// The `llvm.x86.avx512.kor.w` intrinsic; known as `__builtin_ia32_korhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kor.w")
    i16 avx512_kor_w(i16, i16);
    /// The `llvm.x86.avx512.kxor.w` intrinsic; known as `__builtin_ia32_kxorhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kxor.w")
    i16 avx512_kxor_w(i16, i16);
    /// The `llvm.x86.avx512.kxnor.w` intrinsic; known as `__builtin_ia32_kxnorhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kxnor.w")
    i16 avx512_kxnor_w(i16, i16);
    /// The `llvm.x86.avx512.kunpck.bw` intrinsic; known as `__builtin_ia32_kunpckhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kunpck.bw")
    i16 avx512_kunpck_bw(i16, i16);
    /// The `llvm.x86.avx512.kortestz.w` intrinsic; known as `__builtin_ia32_kortestzhi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kortestz.w")
    i32 avx512_kortestz_w(i16, i16);
    /// The `llvm.x86.avx512.kortestc.w` intrinsic; known as `__builtin_ia32_kortestchi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.kortestc.w")
    i32 avx512_kortestc_w(i16, i16);
    /// The `llvm.x86.avx512.cvtss2usi` intrinsic; known as `__builtin_ia32_cvtss2usi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtss2usi")
    i32 avx512_cvtss2usi(__vector(f32[4]));
    /// The `llvm.x86.avx512.cvtss2usi64` intrinsic; known as `__builtin_ia32_cvtss2usi64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtss2usi64")
    i64 avx512_cvtss2usi64(__vector(f32[4]));
    /// The `llvm.x86.avx512.cvttss2usi` intrinsic; known as `__builtin_ia32_cvttss2usi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvttss2usi")
    i32 avx512_cvttss2usi(__vector(f32[4]));
    /// The `llvm.x86.avx512.cvttss2usi64` intrinsic; known as `__builtin_ia32_cvttss2usi64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvttss2usi64")
    i64 avx512_cvttss2usi64(__vector(f32[4]));
    /// The `llvm.x86.avx512.cvtusi2ss` intrinsic; known as `__builtin_ia32_cvtusi2ss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtusi2ss")
    __vector(f32[4]) avx512_cvtusi2ss(__vector(f32[4]), i32);
    /// The `llvm.x86.avx512.cvtusi642ss` intrinsic; known as `__builtin_ia32_cvtusi642ss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtusi642ss")
    __vector(f32[4]) avx512_cvtusi642ss(__vector(f32[4]), i64);
    /// The `llvm.x86.avx512.cvtsd2usi` intrinsic; known as `__builtin_ia32_cvtsd2usi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtsd2usi")
    i32 avx512_cvtsd2usi(__vector(f64[2]));
    /// The `llvm.x86.avx512.cvtsd2usi64` intrinsic; known as `__builtin_ia32_cvtsd2usi64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtsd2usi64")
    i64 avx512_cvtsd2usi64(__vector(f64[2]));
    /// The `llvm.x86.avx512.cvttsd2usi` intrinsic; known as `__builtin_ia32_cvttsd2usi` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvttsd2usi")
    i32 avx512_cvttsd2usi(__vector(f64[2]));
    /// The `llvm.x86.avx512.cvttsd2usi64` intrinsic; known as `__builtin_ia32_cvttsd2usi64` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvttsd2usi64")
    i64 avx512_cvttsd2usi64(__vector(f64[2]));
    /// The `llvm.x86.avx512.cvtusi2sd` intrinsic; known as `__builtin_ia32_cvtusi2sd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtusi2sd")
    __vector(f64[2]) avx512_cvtusi2sd(__vector(f64[2]), i32);
    /// The `llvm.x86.avx512.cvtusi642sd` intrinsic; known as `__builtin_ia32_cvtusi642sd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.cvtusi642sd")
    __vector(f64[2]) avx512_cvtusi642sd(__vector(f64[2]), i64);
    /// The `llvm.x86.avx512.mask.cvttps2dq.512` intrinsic; known as `__builtin_ia32_cvttps2dq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvttps2dq.512")
    __vector(i32[16]) avx512_mask_cvttps2dq_512(__vector(f32[16]), __vector(i32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.cvttps2udq.512` intrinsic; known as `__builtin_ia32_cvttps2udq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvttps2udq.512")
    __vector(i32[16]) avx512_mask_cvttps2udq_512(__vector(f32[16]), __vector(i32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.cvttpd2dq.512` intrinsic; known as `__builtin_ia32_cvttpd2dq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvttpd2dq.512")
    __vector(i32[8]) avx512_mask_cvttpd2dq_512(__vector(f64[8]), __vector(i32[8]), i8, i32);
    /// The `llvm.x86.avx512.mask.cvttpd2udq.512` intrinsic; known as `__builtin_ia32_cvttpd2udq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvttpd2udq.512")
    __vector(i32[8]) avx512_mask_cvttpd2udq_512(__vector(f64[8]), __vector(i32[8]), i8, i32);
    /// The `llvm.x86.avx512.mask.rndscale.ps.512` intrinsic; known as `__builtin_ia32_rndscaleps_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.rndscale.ps.512")
    __vector(f32[16]) avx512_mask_rndscale_ps_512(__vector(f32[16]), i32, __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.rndscale.pd.512` intrinsic; known as `__builtin_ia32_rndscalepd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.rndscale.pd.512")
    __vector(f64[8]) avx512_mask_rndscale_pd_512(__vector(f64[8]), i32, __vector(f64[8]), i8, i32);
    /// The `llvm.x86.avx512.mask.cvtps2dq.512` intrinsic; known as `__builtin_ia32_cvtps2dq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtps2dq.512")
    __vector(i32[16]) avx512_mask_cvtps2dq_512(__vector(f32[16]), __vector(i32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.cvtpd2dq.512` intrinsic; known as `__builtin_ia32_cvtpd2dq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtpd2dq.512")
    __vector(i32[8]) avx512_mask_cvtpd2dq_512(__vector(f64[8]), __vector(i32[8]), i8, i32);
    /// The `llvm.x86.avx512.mask.cvtps2udq.512` intrinsic; known as `__builtin_ia32_cvtps2udq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtps2udq.512")
    __vector(i32[16]) avx512_mask_cvtps2udq_512(__vector(f32[16]), __vector(i32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.cvtpd2udq.512` intrinsic; known as `__builtin_ia32_cvtpd2udq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtpd2udq.512")
    __vector(i32[8]) avx512_mask_cvtpd2udq_512(__vector(f64[8]), __vector(i32[8]), i8, i32);
    /// The `llvm.x86.avx512.mask.cvtdq2ps.512` intrinsic; known as `__builtin_ia32_cvtdq2ps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtdq2ps.512")
    __vector(f32[16]) avx512_mask_cvtdq2ps_512(__vector(i32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.cvtdq2pd.512` intrinsic; known as `__builtin_ia32_cvtdq2pd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtdq2pd.512")
    __vector(f64[8]) avx512_mask_cvtdq2pd_512(__vector(i32[8]), __vector(f64[8]), i8);
    /// The `llvm.x86.avx512.mask.cvtudq2ps.512` intrinsic; known as `__builtin_ia32_cvtudq2ps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtudq2ps.512")
    __vector(f32[16]) avx512_mask_cvtudq2ps_512(__vector(i32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.cvtudq2pd.512` intrinsic; known as `__builtin_ia32_cvtudq2pd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtudq2pd.512")
    __vector(f64[8]) avx512_mask_cvtudq2pd_512(__vector(i32[8]), __vector(f64[8]), i8);
    /// The `llvm.x86.avx512.mask.cvtpd2ps.512` intrinsic; known as `__builtin_ia32_cvtpd2ps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cvtpd2ps.512")
    __vector(f32[8]) avx512_mask_cvtpd2ps_512(__vector(f64[8]), __vector(f32[8]), i8, i32);
    /// The `llvm.x86.avx512.vbroadcast.ss.512` intrinsic; known as `__builtin_ia32_vbroadcastss512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.vbroadcast.ss.512")
    __vector(f32[16]) avx512_vbroadcast_ss_512(i8*);
    /// The `llvm.x86.avx512.vbroadcast.ss.ps.512` intrinsic; known as `__builtin_ia32_vbroadcastss_ps512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.vbroadcast.ss.ps.512")
    __vector(f32[16]) avx512_vbroadcast_ss_ps_512(__vector(f32[4]));
    /// The `llvm.x86.avx512.vbroadcast.sd.512` intrinsic; known as `__builtin_ia32_vbroadcastsd512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.vbroadcast.sd.512")
    __vector(f64[8]) avx512_vbroadcast_sd_512(i8*);
    /// The `llvm.x86.avx512.vbroadcast.sd.pd.512` intrinsic; known as `__builtin_ia32_vbroadcastsd_pd512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.vbroadcast.sd.pd.512")
    __vector(f64[8]) avx512_vbroadcast_sd_pd_512(__vector(f64[2]));
    /// The `llvm.x86.avx512.pbroadcastd.512` intrinsic; known as `__builtin_ia32_pbroadcastd512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pbroadcastd.512")
    __vector(i32[16]) avx512_pbroadcastd_512(__vector(i32[4]));
    /// The `llvm.x86.avx512.pbroadcastd.i32.512` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pbroadcastd.i32.512")
    __vector(i32[16]) avx512_pbroadcastd_i32_512(i32);
    /// The `llvm.x86.avx512.pbroadcastq.512` intrinsic; known as `__builtin_ia32_pbroadcastq512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pbroadcastq.512")
    __vector(i64[8]) avx512_pbroadcastq_512(__vector(i64[2]));
    /// The `llvm.x86.avx512.pbroadcastq.i64.512` intrinsic.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pbroadcastq.i64.512")
    __vector(i64[8]) avx512_pbroadcastq_i64_512(i64);
    /// The `llvm.x86.avx512.pmovzxbq` intrinsic; known as `__builtin_ia32_pmovzxbq512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pmovzxbq")
    __vector(i64[8]) avx512_pmovzxbq(__vector(i8[16]));
    /// The `llvm.x86.avx512.pmovzxwd` intrinsic; known as `__builtin_ia32_pmovzxwd512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pmovzxwd")
    __vector(i32[16]) avx512_pmovzxwd(__vector(i16[16]));
    /// The `llvm.x86.avx512.pmovzxbd` intrinsic; known as `__builtin_ia32_pmovzxbd512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pmovzxbd")
    __vector(i32[16]) avx512_pmovzxbd(__vector(i8[16]));
    /// The `llvm.x86.avx512.pmovzxwq` intrinsic; known as `__builtin_ia32_pmovzxwq512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pmovzxwq")
    __vector(i64[8]) avx512_pmovzxwq(__vector(i16[8]));
    /// The `llvm.x86.avx512.pmovzxdq` intrinsic; known as `__builtin_ia32_pmovzxdq512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.pmovzxdq")
    __vector(i64[8]) avx512_pmovzxdq(__vector(i32[8]));
    /// The `llvm.x86.avx512.mask.max.ps.512` intrinsic; known as `__builtin_ia32_maxps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.max.ps.512")
    __vector(f32[16]) avx512_mask_max_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.max.pd.512` intrinsic; known as `__builtin_ia32_maxpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.max.pd.512")
    __vector(f64[8]) avx512_mask_max_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.avx512.mask.min.ps.512` intrinsic; known as `__builtin_ia32_minps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.min.ps.512")
    __vector(f32[16]) avx512_mask_min_ps_512(__vector(f32[16]), __vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.mask.min.pd.512` intrinsic; known as `__builtin_ia32_minpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.min.pd.512")
    __vector(f64[8]) avx512_mask_min_pd_512(__vector(f64[8]), __vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.avx512.rndscale.ss` intrinsic; known as `__builtin_ia32_rndscaless` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rndscale.ss")
    __vector(f32[4]) avx512_rndscale_ss(__vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.x86.avx512.rndscale.sd` intrinsic; known as `__builtin_ia32_rndscalesd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rndscale.sd")
    __vector(f64[2]) avx512_rndscale_sd(__vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.x86.avx512.sqrt.ss` intrinsic; known as `__builtin_ia32_sqrtrndss` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.sqrt.ss")
    __vector(f32[4]) avx512_sqrt_ss(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.x86.avx512.sqrt.sd` intrinsic; known as `__builtin_ia32_sqrtrndsd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.sqrt.sd")
    __vector(f64[2]) avx512_sqrt_sd(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.x86.avx512.sqrt.pd.512` intrinsic; known as `__builtin_ia32_sqrtpd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.sqrt.pd.512")
    __vector(f64[8]) avx512_sqrt_pd_512(__vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.avx512.sqrt.ps.512` intrinsic; known as `__builtin_ia32_sqrtps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.sqrt.ps.512")
    __vector(f32[16]) avx512_sqrt_ps_512(__vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.rsqrt14.ss` intrinsic; known as `__builtin_ia32_rsqrt14ss_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt14.ss")
    __vector(f32[4]) avx512_rsqrt14_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.avx512.rsqrt14.sd` intrinsic; known as `__builtin_ia32_rsqrt14sd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt14.sd")
    __vector(f64[2]) avx512_rsqrt14_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.avx512.rsqrt14.pd.512` intrinsic; known as `__builtin_ia32_rsqrt14pd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt14.pd.512")
    __vector(f64[8]) avx512_rsqrt14_pd_512(__vector(f64[8]), __vector(f64[8]), i8);
    /// The `llvm.x86.avx512.rsqrt14.ps.512` intrinsic; known as `__builtin_ia32_rsqrt14ps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt14.ps.512")
    __vector(f32[16]) avx512_rsqrt14_ps_512(__vector(f32[16]), __vector(f32[16]), i16);
    /// The `llvm.x86.avx512.rcp14.ss` intrinsic; known as `__builtin_ia32_rcp14ss_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp14.ss")
    __vector(f32[4]) avx512_rcp14_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i8);
    /// The `llvm.x86.avx512.rcp14.sd` intrinsic; known as `__builtin_ia32_rcp14sd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp14.sd")
    __vector(f64[2]) avx512_rcp14_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i8);
    /// The `llvm.x86.avx512.rcp14.pd.512` intrinsic; known as `__builtin_ia32_rcp14pd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp14.pd.512")
    __vector(f64[8]) avx512_rcp14_pd_512(__vector(f64[8]), __vector(f64[8]), i8);
    /// The `llvm.x86.avx512.rcp14.ps.512` intrinsic; known as `__builtin_ia32_rcp14ps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp14.ps.512")
    __vector(f32[16]) avx512_rcp14_ps_512(__vector(f32[16]), __vector(f32[16]), i16);
    /// The `llvm.x86.avx512.rcp28.ps` intrinsic; known as `__builtin_ia32_rcp28ps_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp28.ps")
    __vector(f32[16]) avx512_rcp28_ps(__vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.rcp28.pd` intrinsic; known as `__builtin_ia32_rcp28pd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp28.pd")
    __vector(f64[8]) avx512_rcp28_pd(__vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.avx512.rcp28.ss` intrinsic; known as `__builtin_ia32_rcp28ss_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp28.ss")
    __vector(f32[4]) avx512_rcp28_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i8, i32);
    /// The `llvm.x86.avx512.rcp28.sd` intrinsic; known as `__builtin_ia32_rcp28sd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rcp28.sd")
    __vector(f64[2]) avx512_rcp28_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i8, i32);
    /// The `llvm.x86.avx512.rsqrt28.ps` intrinsic; known as `__builtin_ia32_rsqrt28ps_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt28.ps")
    __vector(f32[16]) avx512_rsqrt28_ps(__vector(f32[16]), __vector(f32[16]), i16, i32);
    /// The `llvm.x86.avx512.rsqrt28.pd` intrinsic; known as `__builtin_ia32_rsqrt28pd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt28.pd")
    __vector(f64[8]) avx512_rsqrt28_pd(__vector(f64[8]), __vector(f64[8]), i8, i32);
    /// The `llvm.x86.avx512.rsqrt28.ss` intrinsic; known as `__builtin_ia32_rsqrt28ss_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt28.ss")
    __vector(f32[4]) avx512_rsqrt28_ss(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i8, i32);
    /// The `llvm.x86.avx512.rsqrt28.sd` intrinsic; known as `__builtin_ia32_rsqrt28sd_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.rsqrt28.sd")
    __vector(f64[2]) avx512_rsqrt28_sd(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i8, i32);
    /// The `llvm.x86.avx512.psll.dq` intrinsic; known as `__builtin_ia32_pslldqi512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.psll.dq")
    __vector(i64[8]) avx512_psll_dq(__vector(i64[8]), i32);
    /// The `llvm.x86.avx512.psrl.dq` intrinsic; known as `__builtin_ia32_psrldqi512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.psrl.dq")
    __vector(i64[8]) avx512_psrl_dq(__vector(i64[8]), i32);
    /// The `llvm.x86.avx512.psll.dq.bs` intrinsic; known as `__builtin_ia32_pslldqi512_byteshift` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.psll.dq.bs")
    __vector(i64[8]) avx512_psll_dq_bs(__vector(i64[8]), i32);
    /// The `llvm.x86.avx512.psrl.dq.bs` intrinsic; known as `__builtin_ia32_psrldqi512_byteshift` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.psrl.dq.bs")
    __vector(i64[8]) avx512_psrl_dq_bs(__vector(i64[8]), i32);
    /// The `llvm.x86.avx512.gather.dpd.512` intrinsic; known as `__builtin_ia32_gathersiv8df` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.dpd.512")
    __vector(f64[8]) avx512_gather_dpd_512(__vector(f64[8]), i8*, __vector(i32[8]), i8, i32);
    /// The `llvm.x86.avx512.gather.dps.512` intrinsic; known as `__builtin_ia32_gathersiv16sf` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.dps.512")
    __vector(f32[16]) avx512_gather_dps_512(__vector(f32[16]), i8*, __vector(i32[16]), i16, i32);
    /// The `llvm.x86.avx512.gather.qpd.512` intrinsic; known as `__builtin_ia32_gatherdiv8df` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.qpd.512")
    __vector(f64[8]) avx512_gather_qpd_512(__vector(f64[8]), i8*, __vector(i64[8]), i8, i32);
    /// The `llvm.x86.avx512.gather.qps.512` intrinsic; known as `__builtin_ia32_gatherdiv16sf` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.qps.512")
    __vector(f32[8]) avx512_gather_qps_512(__vector(f32[8]), i8*, __vector(i64[8]), i8, i32);
    /// The `llvm.x86.avx512.gather.dpq.512` intrinsic; known as `__builtin_ia32_gathersiv8di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.dpq.512")
    __vector(i64[8]) avx512_gather_dpq_512(__vector(i64[8]), i8*, __vector(i32[8]), i8, i32);
    /// The `llvm.x86.avx512.gather.dpi.512` intrinsic; known as `__builtin_ia32_gathersiv16si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.dpi.512")
    __vector(i32[16]) avx512_gather_dpi_512(__vector(i32[16]), i8*, __vector(i32[16]), i16, i32);
    /// The `llvm.x86.avx512.gather.qpq.512` intrinsic; known as `__builtin_ia32_gatherdiv8di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.qpq.512")
    __vector(i64[8]) avx512_gather_qpq_512(__vector(i64[8]), i8*, __vector(i64[8]), i8, i32);
    /// The `llvm.x86.avx512.gather.qpi.512` intrinsic; known as `__builtin_ia32_gatherdiv16si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gather.qpi.512")
    __vector(i32[8]) avx512_gather_qpi_512(__vector(i32[8]), i8*, __vector(i64[8]), i8, i32);
    /// The `llvm.x86.avx512.scatter.dpd.512` intrinsic; known as `__builtin_ia32_scattersiv8df` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.dpd.512")
    void avx512_scatter_dpd_512(i8*, i8, __vector(i32[8]), __vector(f64[8]), i32);
    /// The `llvm.x86.avx512.scatter.dps.512` intrinsic; known as `__builtin_ia32_scattersiv16sf` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.dps.512")
    void avx512_scatter_dps_512(i8*, i16, __vector(i32[16]), __vector(f32[16]), i32);
    /// The `llvm.x86.avx512.scatter.qpd.512` intrinsic; known as `__builtin_ia32_scatterdiv8df` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.qpd.512")
    void avx512_scatter_qpd_512(i8*, i8, __vector(i64[8]), __vector(f64[8]), i32);
    /// The `llvm.x86.avx512.scatter.qps.512` intrinsic; known as `__builtin_ia32_scatterdiv16sf` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.qps.512")
    void avx512_scatter_qps_512(i8*, i8, __vector(i64[8]), __vector(f32[8]), i32);
    /// The `llvm.x86.avx512.scatter.dpq.512` intrinsic; known as `__builtin_ia32_scattersiv8di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.dpq.512")
    void avx512_scatter_dpq_512(i8*, i8, __vector(i32[8]), __vector(i64[8]), i32);
    /// The `llvm.x86.avx512.scatter.dpi.512` intrinsic; known as `__builtin_ia32_scattersiv16si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.dpi.512")
    void avx512_scatter_dpi_512(i8*, i16, __vector(i32[16]), __vector(i32[16]), i32);
    /// The `llvm.x86.avx512.scatter.qpq.512` intrinsic; known as `__builtin_ia32_scatterdiv8di` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.qpq.512")
    void avx512_scatter_qpq_512(i8*, i8, __vector(i64[8]), __vector(i64[8]), i32);
    /// The `llvm.x86.avx512.scatter.qpi.512` intrinsic; known as `__builtin_ia32_scatterdiv16si` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatter.qpi.512")
    void avx512_scatter_qpi_512(i8*, i8, __vector(i64[8]), __vector(i32[8]), i32);
    /// The `llvm.x86.avx512.gatherpf.dpd.512` intrinsic; known as `__builtin_ia32_gatherpfdpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gatherpf.dpd.512")
    void avx512_gatherpf_dpd_512(i8, __vector(i32[8]), i8*, i32, i32);
    /// The `llvm.x86.avx512.gatherpf.dps.512` intrinsic; known as `__builtin_ia32_gatherpfdps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gatherpf.dps.512")
    void avx512_gatherpf_dps_512(i16, __vector(i32[16]), i8*, i32, i32);
    /// The `llvm.x86.avx512.gatherpf.qpd.512` intrinsic; known as `__builtin_ia32_gatherpfqpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gatherpf.qpd.512")
    void avx512_gatherpf_qpd_512(i8, __vector(i64[8]), i8*, i32, i32);
    /// The `llvm.x86.avx512.gatherpf.qps.512` intrinsic; known as `__builtin_ia32_gatherpfqps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.gatherpf.qps.512")
    void avx512_gatherpf_qps_512(i8, __vector(i64[8]), i8*, i32, i32);
    /// The `llvm.x86.avx512.scatterpf.dpd.512` intrinsic; known as `__builtin_ia32_scatterpfdpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatterpf.dpd.512")
    void avx512_scatterpf_dpd_512(i8, __vector(i32[8]), i8*, i32, i32);
    /// The `llvm.x86.avx512.scatterpf.dps.512` intrinsic; known as `__builtin_ia32_scatterpfdps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatterpf.dps.512")
    void avx512_scatterpf_dps_512(i16, __vector(i32[16]), i8*, i32, i32);
    /// The `llvm.x86.avx512.scatterpf.qpd.512` intrinsic; known as `__builtin_ia32_scatterpfqpd` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatterpf.qpd.512")
    void avx512_scatterpf_qpd_512(i8, __vector(i64[8]), i8*, i32, i32);
    /// The `llvm.x86.avx512.scatterpf.qps.512` intrinsic; known as `__builtin_ia32_scatterpfqps` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.scatterpf.qps.512")
    void avx512_scatterpf_qps_512(i8, __vector(i64[8]), i8*, i32, i32);
    /// The `llvm.x86.avx512.mask.conflict.d.512` intrinsic; known as `__builtin_ia32_vpconflictsi_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.conflict.d.512")
    __vector(i32[16]) avx512_mask_conflict_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.conflict.q.512` intrinsic; known as `__builtin_ia32_vpconflictdi_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.conflict.q.512")
    __vector(i64[8]) avx512_mask_conflict_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.lzcnt.d.512` intrinsic; known as `__builtin_ia32_vplzcntd_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.lzcnt.d.512")
    __vector(i32[16]) avx512_mask_lzcnt_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.lzcnt.q.512` intrinsic; known as `__builtin_ia32_vplzcntq_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.lzcnt.q.512")
    __vector(i64[8]) avx512_mask_lzcnt_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.blend.ps.512` intrinsic; known as `__builtin_ia32_blendmps_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.blend.ps.512")
    __vector(f32[16]) avx512_mask_blend_ps_512(__vector(f32[16]), __vector(f32[16]), i16);
    /// The `llvm.x86.avx512.mask.blend.pd.512` intrinsic; known as `__builtin_ia32_blendmpd_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.blend.pd.512")
    __vector(f64[8]) avx512_mask_blend_pd_512(__vector(f64[8]), __vector(f64[8]), i8);
    /// The `llvm.x86.avx512.mask.blend.d.512` intrinsic; known as `__builtin_ia32_blendmd_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.blend.d.512")
    __vector(i32[16]) avx512_mask_blend_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.blend.q.512` intrinsic; known as `__builtin_ia32_blendmq_512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.blend.q.512")
    __vector(i64[8]) avx512_mask_blend_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.valign.q.512` intrinsic; known as `__builtin_ia32_alignq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.valign.q.512")
    __vector(i64[8]) avx512_mask_valign_q_512(__vector(i64[8]), __vector(i64[8]), i8, __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.valign.d.512` intrinsic; known as `__builtin_ia32_alignd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.valign.d.512")
    __vector(i32[16]) avx512_mask_valign_d_512(__vector(i32[16]), __vector(i32[16]), i8, __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpeq.b.512` intrinsic; known as `__builtin_ia32_pcmpeqb512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.b.512")
    i64 avx512_mask_pcmpeq_b_512(i8x64, i8x64, i64);
    /// The `llvm.x86.avx512.mask.pcmpeq.w.512` intrinsic; known as `__builtin_ia32_pcmpeqw512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.w.512")
    i32 avx512_mask_pcmpeq_w_512(__vector(i16[32]), __vector(i16[32]), i32);
    /// The `llvm.x86.avx512.mask.pcmpeq.d.512` intrinsic; known as `__builtin_ia32_pcmpeqd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.d.512")
    i16 avx512_mask_pcmpeq_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpeq.q.512` intrinsic; known as `__builtin_ia32_pcmpeqq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.q.512")
    i8 avx512_mask_pcmpeq_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pcmpgt.b.512` intrinsic; known as `__builtin_ia32_pcmpgtb512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.b.512")
    i64 avx512_mask_pcmpgt_b_512(i8x64, i8x64, i64);
    /// The `llvm.x86.avx512.mask.pcmpgt.w.512` intrinsic; known as `__builtin_ia32_pcmpgtw512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.w.512")
    i32 avx512_mask_pcmpgt_w_512(__vector(i16[32]), __vector(i16[32]), i32);
    /// The `llvm.x86.avx512.mask.pcmpgt.d.512` intrinsic; known as `__builtin_ia32_pcmpgtd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.d.512")
    i16 avx512_mask_pcmpgt_d_512(__vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpgt.q.512` intrinsic; known as `__builtin_ia32_pcmpgtq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.q.512")
    i8 avx512_mask_pcmpgt_q_512(__vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.mask.pcmpeq.b.256` intrinsic; known as `__builtin_ia32_pcmpeqb256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.b.256")
    i32 avx512_mask_pcmpeq_b_256(__vector(i8[32]), __vector(i8[32]), i32);
    /// The `llvm.x86.avx512.mask.pcmpeq.w.256` intrinsic; known as `__builtin_ia32_pcmpeqw256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.w.256")
    i16 avx512_mask_pcmpeq_w_256(__vector(i16[16]), __vector(i16[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpeq.d.256` intrinsic; known as `__builtin_ia32_pcmpeqd256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.d.256")
    i8 avx512_mask_pcmpeq_d_256(__vector(i32[8]), __vector(i32[8]), i8);
    /// The `llvm.x86.avx512.mask.pcmpeq.q.256` intrinsic; known as `__builtin_ia32_pcmpeqq256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.q.256")
    i8 avx512_mask_pcmpeq_q_256(__vector(i64[4]), __vector(i64[4]), i8);
    /// The `llvm.x86.avx512.mask.pcmpgt.b.256` intrinsic; known as `__builtin_ia32_pcmpgtb256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.b.256")
    i32 avx512_mask_pcmpgt_b_256(__vector(i8[32]), __vector(i8[32]), i32);
    /// The `llvm.x86.avx512.mask.pcmpgt.w.256` intrinsic; known as `__builtin_ia32_pcmpgtw256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.w.256")
    i16 avx512_mask_pcmpgt_w_256(__vector(i16[16]), __vector(i16[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpgt.d.256` intrinsic; known as `__builtin_ia32_pcmpgtd256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.d.256")
    i8 avx512_mask_pcmpgt_d_256(__vector(i32[8]), __vector(i32[8]), i8);
    /// The `llvm.x86.avx512.mask.pcmpgt.q.256` intrinsic; known as `__builtin_ia32_pcmpgtq256_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.q.256")
    i8 avx512_mask_pcmpgt_q_256(__vector(i64[4]), __vector(i64[4]), i8);
    /// The `llvm.x86.avx512.mask.pcmpeq.b.128` intrinsic; known as `__builtin_ia32_pcmpeqb128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.b.128")
    i16 avx512_mask_pcmpeq_b_128(__vector(i8[16]), __vector(i8[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpeq.w.128` intrinsic; known as `__builtin_ia32_pcmpeqw128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.w.128")
    i8 avx512_mask_pcmpeq_w_128(__vector(i16[8]), __vector(i16[8]), i8);
    /// The `llvm.x86.avx512.mask.pcmpeq.d.128` intrinsic; known as `__builtin_ia32_pcmpeqd128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.d.128")
    i8 avx512_mask_pcmpeq_d_128(__vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx512.mask.pcmpeq.q.128` intrinsic; known as `__builtin_ia32_pcmpeqq128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpeq.q.128")
    i8 avx512_mask_pcmpeq_q_128(__vector(i64[2]), __vector(i64[2]), i8);
    /// The `llvm.x86.avx512.mask.pcmpgt.b.128` intrinsic; known as `__builtin_ia32_pcmpgtb128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.b.128")
    i16 avx512_mask_pcmpgt_b_128(__vector(i8[16]), __vector(i8[16]), i16);
    /// The `llvm.x86.avx512.mask.pcmpgt.w.128` intrinsic; known as `__builtin_ia32_pcmpgtw128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.w.128")
    i8 avx512_mask_pcmpgt_w_128(__vector(i16[8]), __vector(i16[8]), i8);
    /// The `llvm.x86.avx512.mask.pcmpgt.d.128` intrinsic; known as `__builtin_ia32_pcmpgtd128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.d.128")
    i8 avx512_mask_pcmpgt_d_128(__vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.avx512.mask.pcmpgt.q.128` intrinsic; known as `__builtin_ia32_pcmpgtq128_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pcmpgt.q.128")
    i8 avx512_mask_pcmpgt_q_128(__vector(i64[2]), __vector(i64[2]), i8);
    /// The `llvm.x86.avx512.mask.cmp.ps.512` intrinsic; known as `__builtin_ia32_cmpps512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cmp.ps.512")
    i16 avx512_mask_cmp_ps_512(__vector(f32[16]), __vector(f32[16]), i32, i16, i32);
    /// The `llvm.x86.avx512.mask.cmp.pd.512` intrinsic; known as `__builtin_ia32_cmppd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.cmp.pd.512")
    i8 avx512_mask_cmp_pd_512(__vector(f64[8]), __vector(f64[8]), i32, i8, i32);
    /// The `llvm.x86.avx512.mask.pand.d.512` intrinsic; known as `__builtin_ia32_pandd512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pand.d.512")
    __vector(i32[16]) avx512_mask_pand_d_512(__vector(i32[16]), __vector(i32[16]), __vector(i32[16]), i16);
    /// The `llvm.x86.avx512.mask.pand.q.512` intrinsic; known as `__builtin_ia32_pandq512_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.mask.pand.q.512")
    __vector(i64[8]) avx512_mask_pand_q_512(__vector(i64[8]), __vector(i64[8]), __vector(i64[8]), i8);
    /// The `llvm.x86.avx512.movntdqa` intrinsic; known as `__builtin_ia32_movntdqa512` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.avx512.movntdqa")
    __vector(i64[8]) avx512_movntdqa(i8*);
    /// The `llvm.x86.sha1rnds4` intrinsic; known as `__builtin_ia32_sha1rnds4` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha1rnds4")
    __vector(i32[4]) sha1rnds4(__vector(i32[4]), __vector(i32[4]), i8);
    /// The `llvm.x86.sha1nexte` intrinsic; known as `__builtin_ia32_sha1nexte` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha1nexte")
    __vector(i32[4]) sha1nexte(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sha1msg1` intrinsic; known as `__builtin_ia32_sha1msg1` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha1msg1")
    __vector(i32[4]) sha1msg1(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sha1msg2` intrinsic; known as `__builtin_ia32_sha1msg2` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha1msg2")
    __vector(i32[4]) sha1msg2(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sha256rnds2` intrinsic; known as `__builtin_ia32_sha256rnds2` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha256rnds2")
    __vector(i32[4]) sha256rnds2(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sha256msg1` intrinsic; known as `__builtin_ia32_sha256msg1` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha256msg1")
    __vector(i32[4]) sha256msg1(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.x86.sha256msg2` intrinsic; known as `__builtin_ia32_sha256msg2` in GCC.
    pragma(LDC_intrinsic, "llvm.x86.sha256msg2")
    __vector(i32[4]) sha256msg2(__vector(i32[4]), __vector(i32[4]));
}

/// LLVM intrinsics for the AMDGPU architecture.
version (AMDGPU) {
    /// The `llvm.AMDGPU.div.fmas.v4f32` intrinsic; known as `__builtin_amdgpu_div_fmas` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fmas.v4f32")
    __vector(f32[4]) div_fmas_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.AMDGPU.div.fmas.f32` intrinsic; known as `__builtin_amdgpu_div_fmas` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fmas.f32")
    f32 div_fmas_f32(f32, f32, f32);
    /// The `llvm.AMDGPU.div.fmas.v2f64` intrinsic; known as `__builtin_amdgpu_div_fmas` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fmas.v2f64")
    __vector(f64[2]) div_fmas_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.AMDGPU.div.fmas.f64` intrinsic; known as `__builtin_amdgpu_div_fmas` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fmas.f64")
    f64 div_fmas_f64(f64, f64, f64);
    /// The `llvm.AMDGPU.div.fixup.v4f32` intrinsic; known as `__builtin_amdgpu_div_fixup` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fixup.v4f32")
    __vector(f32[4]) div_fixup_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.AMDGPU.div.fixup.f32` intrinsic; known as `__builtin_amdgpu_div_fixup` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fixup.f32")
    f32 div_fixup_f32(f32, f32, f32);
    /// The `llvm.AMDGPU.div.fixup.v2f64` intrinsic; known as `__builtin_amdgpu_div_fixup` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fixup.v2f64")
    __vector(f64[2]) div_fixup_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.AMDGPU.div.fixup.f64` intrinsic; known as `__builtin_amdgpu_div_fixup` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.div.fixup.f64")
    f64 div_fixup_f64(f64, f64, f64);
    /// The `llvm.AMDGPU.trig.preop.v4f32` intrinsic; known as `__builtin_amdgpu_trig_preop` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.trig.preop.v4f32")
    __vector(f32[4]) trig_preop_v4f32(__vector(f32[4]), i32);
    /// The `llvm.AMDGPU.trig.preop.f32` intrinsic; known as `__builtin_amdgpu_trig_preop` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.trig.preop.f32")
    f32 trig_preop_f32(f32, i32);
    /// The `llvm.AMDGPU.trig.preop.v2f64` intrinsic; known as `__builtin_amdgpu_trig_preop` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.trig.preop.v2f64")
    __vector(f64[2]) trig_preop_v2f64(__vector(f64[2]), i32);
    /// The `llvm.AMDGPU.trig.preop.f64` intrinsic; known as `__builtin_amdgpu_trig_preop` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.trig.preop.f64")
    f64 trig_preop_f64(f64, i32);
    /// The `llvm.AMDGPU.rcp.v4f32` intrinsic; known as `__builtin_amdgpu_rcp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rcp.v4f32")
    __vector(f32[4]) rcp_v4f32(__vector(f32[4]));
    /// The `llvm.AMDGPU.rcp.f32` intrinsic; known as `__builtin_amdgpu_rcp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rcp.f32")
    f32 rcp_f32(f32);
    /// The `llvm.AMDGPU.rcp.v2f64` intrinsic; known as `__builtin_amdgpu_rcp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rcp.v2f64")
    __vector(f64[2]) rcp_v2f64(__vector(f64[2]));
    /// The `llvm.AMDGPU.rcp.f64` intrinsic; known as `__builtin_amdgpu_rcp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rcp.f64")
    f64 rcp_f64(f64);
    /// The `llvm.AMDGPU.rsq.v4f32` intrinsic; known as `__builtin_amdgpu_rsq` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.v4f32")
    __vector(f32[4]) rsq_v4f32(__vector(f32[4]));
    /// The `llvm.AMDGPU.rsq.f32` intrinsic; known as `__builtin_amdgpu_rsq` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.f32")
    f32 rsq_f32(f32);
    /// The `llvm.AMDGPU.rsq.v2f64` intrinsic; known as `__builtin_amdgpu_rsq` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.v2f64")
    __vector(f64[2]) rsq_v2f64(__vector(f64[2]));
    /// The `llvm.AMDGPU.rsq.f64` intrinsic; known as `__builtin_amdgpu_rsq` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.f64")
    f64 rsq_f64(f64);
    /// The `llvm.AMDGPU.rsq.clamped.v4f32` intrinsic; known as `__builtin_amdgpu_rsq_clamped` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.clamped.v4f32")
    __vector(f32[4]) rsq_clamped_v4f32(__vector(f32[4]));
    /// The `llvm.AMDGPU.rsq.clamped.f32` intrinsic; known as `__builtin_amdgpu_rsq_clamped` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.clamped.f32")
    f32 rsq_clamped_f32(f32);
    /// The `llvm.AMDGPU.rsq.clamped.v2f64` intrinsic; known as `__builtin_amdgpu_rsq_clamped` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.clamped.v2f64")
    __vector(f64[2]) rsq_clamped_v2f64(__vector(f64[2]));
    /// The `llvm.AMDGPU.rsq.clamped.f64` intrinsic; known as `__builtin_amdgpu_rsq_clamped` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.rsq.clamped.f64")
    f64 rsq_clamped_f64(f64);
    /// The `llvm.AMDGPU.ldexp.v4f32` intrinsic; known as `__builtin_amdgpu_ldexp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.ldexp.v4f32")
    __vector(f32[4]) ldexp_v4f32(__vector(f32[4]), i32);
    /// The `llvm.AMDGPU.ldexp.f32` intrinsic; known as `__builtin_amdgpu_ldexp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.ldexp.f32")
    f32 ldexp_f32(f32, i32);
    /// The `llvm.AMDGPU.ldexp.v2f64` intrinsic; known as `__builtin_amdgpu_ldexp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.ldexp.v2f64")
    __vector(f64[2]) ldexp_v2f64(__vector(f64[2]), i32);
    /// The `llvm.AMDGPU.ldexp.f64` intrinsic; known as `__builtin_amdgpu_ldexp` in GCC.
    pragma(LDC_intrinsic, "llvm.AMDGPU.ldexp.f64")
    f64 ldexp_f64(f64, i32);
}

/// LLVM intrinsics for the aarch64 architecture.
version (AArch64) {
    /// The `llvm.aarch64.ldxr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.ldxr.p0i8")
    i64 ldxr_p0i8(i8*);
    /// The `llvm.aarch64.ldaxr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.ldaxr.p0i8")
    i64 ldaxr_p0i8(i8*);
    /// The `llvm.aarch64.stxr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.stxr.p0i8")
    i32 stxr_p0i8(i64, i8*);
    /// The `llvm.aarch64.stlxr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.stlxr.p0i8")
    i32 stlxr_p0i8(i64, i8*);
    /// The `llvm.aarch64.stxp` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.stxp")
    i32 stxp(i64, i64, i8*);
    /// The `llvm.aarch64.stlxp` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.stlxp")
    i32 stlxp(i64, i64, i8*);
    /// The `llvm.aarch64.clrex` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.clrex")
    void clrex();
    /// The `llvm.aarch64.sdiv.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.v16i8")
    __vector(i8[16]) sdiv_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.sdiv.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.i8")
    i8 sdiv_i8(i8, i8);
    /// The `llvm.aarch64.sdiv.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.v8i16")
    __vector(i16[8]) sdiv_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.sdiv.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.i16")
    i16 sdiv_i16(i16, i16);
    /// The `llvm.aarch64.sdiv.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.v4i32")
    __vector(i32[4]) sdiv_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.sdiv.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.i32")
    i32 sdiv_i32(i32, i32);
    /// The `llvm.aarch64.sdiv.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.v2i64")
    __vector(i64[2]) sdiv_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.sdiv.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sdiv.i64")
    i64 sdiv_i64(i64, i64);
    /// The `llvm.aarch64.udiv.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.v16i8")
    __vector(i8[16]) udiv_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.udiv.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.i8")
    i8 udiv_i8(i8, i8);
    /// The `llvm.aarch64.udiv.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.v8i16")
    __vector(i16[8]) udiv_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.udiv.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.i16")
    i16 udiv_i16(i16, i16);
    /// The `llvm.aarch64.udiv.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.v4i32")
    __vector(i32[4]) udiv_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.udiv.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.i32")
    i32 udiv_i32(i32, i32);
    /// The `llvm.aarch64.udiv.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.v2i64")
    __vector(i64[2]) udiv_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.udiv.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.udiv.i64")
    i64 udiv_i64(i64, i64);
    /// The `llvm.aarch64.hint` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.hint")
    void hint(i32);
    /// The `llvm.aarch64.rbit.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.v16i8")
    __vector(i8[16]) rbit_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.rbit.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.i8")
    i8 rbit_i8(i8);
    /// The `llvm.aarch64.rbit.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.v8i16")
    __vector(i16[8]) rbit_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.rbit.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.i16")
    i16 rbit_i16(i16);
    /// The `llvm.aarch64.rbit.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.v4i32")
    __vector(i32[4]) rbit_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.rbit.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.i32")
    i32 rbit_i32(i32);
    /// The `llvm.aarch64.rbit.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.v2i64")
    __vector(i64[2]) rbit_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.rbit.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.rbit.i64")
    i64 rbit_i64(i64);
    /// The `llvm.aarch64.dmb` intrinsic; known as `__builtin_arm_dmb` in GCC.
    pragma(LDC_intrinsic, "llvm.aarch64.dmb")
    void dmb(i32);
    /// The `llvm.aarch64.dsb` intrinsic; known as `__builtin_arm_dsb` in GCC.
    pragma(LDC_intrinsic, "llvm.aarch64.dsb")
    void dsb(i32);
    /// The `llvm.aarch64.isb` intrinsic; known as `__builtin_arm_isb` in GCC.
    pragma(LDC_intrinsic, "llvm.aarch64.isb")
    void isb(i32);
    /// The `llvm.aarch64.neon.saddv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v16i8.v16i8")
    __vector(i8[16]) neon_saddv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v16i8.v8i16")
    __vector(i8[16]) neon_saddv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v16i8.v4i32")
    __vector(i8[16]) neon_saddv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v16i8.v2i64")
    __vector(i8[16]) neon_saddv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v16i8.v4f32")
    __vector(i8[16]) neon_saddv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v16i8.v2f64")
    __vector(i8[16]) neon_saddv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i8.v16i8")
    i8 neon_saddv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i8.v8i16")
    i8 neon_saddv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i8.v4i32")
    i8 neon_saddv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i8.v2i64")
    i8 neon_saddv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i8.v4f32")
    i8 neon_saddv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i8.v2f64")
    i8 neon_saddv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v8i16.v16i8")
    __vector(i16[8]) neon_saddv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v8i16.v8i16")
    __vector(i16[8]) neon_saddv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v8i16.v4i32")
    __vector(i16[8]) neon_saddv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v8i16.v2i64")
    __vector(i16[8]) neon_saddv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v8i16.v4f32")
    __vector(i16[8]) neon_saddv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v8i16.v2f64")
    __vector(i16[8]) neon_saddv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i16.v16i8")
    i16 neon_saddv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i16.v8i16")
    i16 neon_saddv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i16.v4i32")
    i16 neon_saddv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i16.v2i64")
    i16 neon_saddv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i16.v4f32")
    i16 neon_saddv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i16.v2f64")
    i16 neon_saddv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v4i32.v16i8")
    __vector(i32[4]) neon_saddv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v4i32.v8i16")
    __vector(i32[4]) neon_saddv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v4i32.v4i32")
    __vector(i32[4]) neon_saddv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v4i32.v2i64")
    __vector(i32[4]) neon_saddv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v4i32.v4f32")
    __vector(i32[4]) neon_saddv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v4i32.v2f64")
    __vector(i32[4]) neon_saddv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i32.v16i8")
    i32 neon_saddv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i32.v8i16")
    i32 neon_saddv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i32.v4i32")
    i32 neon_saddv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i32.v2i64")
    i32 neon_saddv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i32.v4f32")
    i32 neon_saddv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i32.v2f64")
    i32 neon_saddv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v2i64.v16i8")
    __vector(i64[2]) neon_saddv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v2i64.v8i16")
    __vector(i64[2]) neon_saddv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v2i64.v4i32")
    __vector(i64[2]) neon_saddv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v2i64.v2i64")
    __vector(i64[2]) neon_saddv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v2i64.v4f32")
    __vector(i64[2]) neon_saddv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.v2i64.v2f64")
    __vector(i64[2]) neon_saddv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i64.v16i8")
    i64 neon_saddv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i64.v8i16")
    i64 neon_saddv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i64.v4i32")
    i64 neon_saddv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i64.v2i64")
    i64 neon_saddv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i64.v4f32")
    i64 neon_saddv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddv.i64.v2f64")
    i64 neon_saddv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v16i8.v16i8")
    __vector(i8[16]) neon_uaddv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v16i8.v8i16")
    __vector(i8[16]) neon_uaddv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v16i8.v4i32")
    __vector(i8[16]) neon_uaddv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v16i8.v2i64")
    __vector(i8[16]) neon_uaddv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v16i8.v4f32")
    __vector(i8[16]) neon_uaddv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v16i8.v2f64")
    __vector(i8[16]) neon_uaddv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.i8.v16i8` intrinsic.
    // =====
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v8i8")
    i8 neon_uaddv_i8_v8i8(__vector(i8[8]));
    // =====
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v16i8")
    i8 neon_uaddv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v8i16")
    i8 neon_uaddv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v4i32")
    i8 neon_uaddv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v2i64")
    i8 neon_uaddv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v4f32")
    i8 neon_uaddv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i8.v2f64")
    i8 neon_uaddv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v8i16.v16i8")
    __vector(i16[8]) neon_uaddv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v8i16.v8i16")
    __vector(i16[8]) neon_uaddv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v8i16.v4i32")
    __vector(i16[8]) neon_uaddv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v8i16.v2i64")
    __vector(i16[8]) neon_uaddv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v8i16.v4f32")
    __vector(i16[8]) neon_uaddv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v8i16.v2f64")
    __vector(i16[8]) neon_uaddv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i16.v16i8")
    i16 neon_uaddv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i16.v8i16")
    i16 neon_uaddv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i16.v4i32")
    i16 neon_uaddv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i16.v2i64")
    i16 neon_uaddv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i16.v4f32")
    i16 neon_uaddv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i16.v2f64")
    i16 neon_uaddv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v4i32.v16i8")
    __vector(i32[4]) neon_uaddv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v4i32.v8i16")
    __vector(i32[4]) neon_uaddv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v4i32.v4i32")
    __vector(i32[4]) neon_uaddv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v4i32.v2i64")
    __vector(i32[4]) neon_uaddv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v4i32.v4f32")
    __vector(i32[4]) neon_uaddv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v4i32.v2f64")
    __vector(i32[4]) neon_uaddv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i32.v16i8")
    i32 neon_uaddv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i32.v8i16")
    i32 neon_uaddv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i32.v4i32")
    i32 neon_uaddv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i32.v2i64")
    i32 neon_uaddv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i32.v4f32")
    i32 neon_uaddv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i32.v2f64")
    i32 neon_uaddv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v2i64.v16i8")
    __vector(i64[2]) neon_uaddv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v2i64.v8i16")
    __vector(i64[2]) neon_uaddv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v2i64.v4i32")
    __vector(i64[2]) neon_uaddv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v2i64.v2i64")
    __vector(i64[2]) neon_uaddv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v2i64.v4f32")
    __vector(i64[2]) neon_uaddv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.v2i64.v2f64")
    __vector(i64[2]) neon_uaddv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i64.v16i8")
    i64 neon_uaddv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i64.v8i16")
    i64 neon_uaddv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i64.v4i32")
    i64 neon_uaddv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i64.v2i64")
    i64 neon_uaddv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i64.v4f32")
    i64 neon_uaddv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddv.i64.v2f64")
    i64 neon_uaddv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.faddv.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v4f32.v16i8")
    __vector(f32[4]) neon_faddv_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.faddv.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v4f32.v8i16")
    __vector(f32[4]) neon_faddv_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.faddv.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v4f32.v4i32")
    __vector(f32[4]) neon_faddv_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.faddv.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v4f32.v2i64")
    __vector(f32[4]) neon_faddv_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.faddv.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v4f32.v4f32")
    __vector(f32[4]) neon_faddv_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.faddv.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v4f32.v2f64")
    __vector(f32[4]) neon_faddv_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.faddv.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f32.v16i8")
    f32 neon_faddv_f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.faddv.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f32.v8i16")
    f32 neon_faddv_f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.faddv.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f32.v4i32")
    f32 neon_faddv_f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.faddv.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f32.v2i64")
    f32 neon_faddv_f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.faddv.f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f32.v4f32")
    f32 neon_faddv_f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.faddv.f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f32.v2f64")
    f32 neon_faddv_f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.faddv.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v2f64.v16i8")
    __vector(f64[2]) neon_faddv_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.faddv.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v2f64.v8i16")
    __vector(f64[2]) neon_faddv_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.faddv.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v2f64.v4i32")
    __vector(f64[2]) neon_faddv_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.faddv.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v2f64.v2i64")
    __vector(f64[2]) neon_faddv_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.faddv.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v2f64.v4f32")
    __vector(f64[2]) neon_faddv_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.faddv.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.v2f64.v2f64")
    __vector(f64[2]) neon_faddv_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.faddv.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f64.v16i8")
    f64 neon_faddv_f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.faddv.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f64.v8i16")
    f64 neon_faddv_f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.faddv.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f64.v4i32")
    f64 neon_faddv_f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.faddv.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f64.v2i64")
    f64 neon_faddv_f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.faddv.f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f64.v4f32")
    f64 neon_faddv_f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.faddv.f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.faddv.f64.v2f64")
    f64 neon_faddv_f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v16i8.v16i8")
    __vector(i8[16]) neon_saddlv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v16i8.v8i16")
    __vector(i8[16]) neon_saddlv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v16i8.v4i32")
    __vector(i8[16]) neon_saddlv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v16i8.v2i64")
    __vector(i8[16]) neon_saddlv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v16i8.v4f32")
    __vector(i8[16]) neon_saddlv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v16i8.v2f64")
    __vector(i8[16]) neon_saddlv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i8.v16i8")
    i8 neon_saddlv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i8.v8i16")
    i8 neon_saddlv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i8.v4i32")
    i8 neon_saddlv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i8.v2i64")
    i8 neon_saddlv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i8.v4f32")
    i8 neon_saddlv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i8.v2f64")
    i8 neon_saddlv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v8i16.v16i8")
    __vector(i16[8]) neon_saddlv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v8i16.v8i16")
    __vector(i16[8]) neon_saddlv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v8i16.v4i32")
    __vector(i16[8]) neon_saddlv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v8i16.v2i64")
    __vector(i16[8]) neon_saddlv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v8i16.v4f32")
    __vector(i16[8]) neon_saddlv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v8i16.v2f64")
    __vector(i16[8]) neon_saddlv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i16.v16i8")
    i16 neon_saddlv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i16.v8i16")
    i16 neon_saddlv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i16.v4i32")
    i16 neon_saddlv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i16.v2i64")
    i16 neon_saddlv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i16.v4f32")
    i16 neon_saddlv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i16.v2f64")
    i16 neon_saddlv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v4i32.v16i8")
    __vector(i32[4]) neon_saddlv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v4i32.v8i16")
    __vector(i32[4]) neon_saddlv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v4i32.v4i32")
    __vector(i32[4]) neon_saddlv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v4i32.v2i64")
    __vector(i32[4]) neon_saddlv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v4i32.v4f32")
    __vector(i32[4]) neon_saddlv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v4i32.v2f64")
    __vector(i32[4]) neon_saddlv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i32.v16i8")
    i32 neon_saddlv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i32.v8i16")
    i32 neon_saddlv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i32.v4i32")
    i32 neon_saddlv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i32.v2i64")
    i32 neon_saddlv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i32.v4f32")
    i32 neon_saddlv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i32.v2f64")
    i32 neon_saddlv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v2i64.v16i8")
    __vector(i64[2]) neon_saddlv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v2i64.v8i16")
    __vector(i64[2]) neon_saddlv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v2i64.v4i32")
    __vector(i64[2]) neon_saddlv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v2i64.v2i64")
    __vector(i64[2]) neon_saddlv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v2i64.v4f32")
    __vector(i64[2]) neon_saddlv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.v2i64.v2f64")
    __vector(i64[2]) neon_saddlv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i64.v16i8")
    i64 neon_saddlv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i64.v8i16")
    i64 neon_saddlv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i64.v4i32")
    i64 neon_saddlv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i64.v2i64")
    i64 neon_saddlv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i64.v4f32")
    i64 neon_saddlv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlv.i64.v2f64")
    i64 neon_saddlv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v16i8.v16i8")
    __vector(i8[16]) neon_uaddlv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v16i8.v8i16")
    __vector(i8[16]) neon_uaddlv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v16i8.v4i32")
    __vector(i8[16]) neon_uaddlv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v16i8.v2i64")
    __vector(i8[16]) neon_uaddlv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v16i8.v4f32")
    __vector(i8[16]) neon_uaddlv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v16i8.v2f64")
    __vector(i8[16]) neon_uaddlv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i8.v16i8")
    i8 neon_uaddlv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i8.v8i16")
    i8 neon_uaddlv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i8.v4i32")
    i8 neon_uaddlv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i8.v2i64")
    i8 neon_uaddlv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i8.v4f32")
    i8 neon_uaddlv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i8.v2f64")
    i8 neon_uaddlv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v8i16.v16i8")
    __vector(i16[8]) neon_uaddlv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v8i16.v8i16")
    __vector(i16[8]) neon_uaddlv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v8i16.v4i32")
    __vector(i16[8]) neon_uaddlv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v8i16.v2i64")
    __vector(i16[8]) neon_uaddlv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v8i16.v4f32")
    __vector(i16[8]) neon_uaddlv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v8i16.v2f64")
    __vector(i16[8]) neon_uaddlv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i16.v16i8")
    i16 neon_uaddlv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i16.v8i16")
    i16 neon_uaddlv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i16.v4i32")
    i16 neon_uaddlv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i16.v2i64")
    i16 neon_uaddlv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i16.v4f32")
    i16 neon_uaddlv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i16.v2f64")
    i16 neon_uaddlv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v4i32.v16i8")
    __vector(i32[4]) neon_uaddlv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v4i32.v8i16")
    __vector(i32[4]) neon_uaddlv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v4i32.v4i32")
    __vector(i32[4]) neon_uaddlv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v4i32.v2i64")
    __vector(i32[4]) neon_uaddlv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v4i32.v4f32")
    __vector(i32[4]) neon_uaddlv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v4i32.v2f64")
    __vector(i32[4]) neon_uaddlv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i32.v16i8")
    i32 neon_uaddlv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i32.v8i16")
    i32 neon_uaddlv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i32.v4i32")
    i32 neon_uaddlv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i32.v2i64")
    i32 neon_uaddlv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i32.v4f32")
    i32 neon_uaddlv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i32.v2f64")
    i32 neon_uaddlv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v2i64.v16i8")
    __vector(i64[2]) neon_uaddlv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v2i64.v8i16")
    __vector(i64[2]) neon_uaddlv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v2i64.v4i32")
    __vector(i64[2]) neon_uaddlv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v2i64.v2i64")
    __vector(i64[2]) neon_uaddlv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v2i64.v4f32")
    __vector(i64[2]) neon_uaddlv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.v2i64.v2f64")
    __vector(i64[2]) neon_uaddlv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i64.v16i8")
    i64 neon_uaddlv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i64.v8i16")
    i64 neon_uaddlv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i64.v4i32")
    i64 neon_uaddlv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i64.v2i64")
    i64 neon_uaddlv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i64.v4f32")
    i64 neon_uaddlv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlv.i64.v2f64")
    i64 neon_uaddlv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.shadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shadd.v16i8")
    __vector(i8[16]) neon_shadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.shadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shadd.v8i16")
    __vector(i16[8]) neon_shadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.shadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shadd.v4i32")
    __vector(i32[4]) neon_shadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.shadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shadd.v2i64")
    __vector(i64[2]) neon_shadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.shadd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shadd.v4f32")
    __vector(f32[4]) neon_shadd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.shadd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shadd.v2f64")
    __vector(f64[2]) neon_shadd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.uhadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhadd.v16i8")
    __vector(i8[16]) neon_uhadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uhadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhadd.v8i16")
    __vector(i16[8]) neon_uhadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uhadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhadd.v4i32")
    __vector(i32[4]) neon_uhadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uhadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhadd.v2i64")
    __vector(i64[2]) neon_uhadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uhadd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhadd.v4f32")
    __vector(f32[4]) neon_uhadd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.uhadd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhadd.v2f64")
    __vector(f64[2]) neon_uhadd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.srhadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srhadd.v16i8")
    __vector(i8[16]) neon_srhadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.srhadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srhadd.v8i16")
    __vector(i16[8]) neon_srhadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.srhadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srhadd.v4i32")
    __vector(i32[4]) neon_srhadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.srhadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srhadd.v2i64")
    __vector(i64[2]) neon_srhadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.srhadd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srhadd.v4f32")
    __vector(f32[4]) neon_srhadd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.srhadd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srhadd.v2f64")
    __vector(f64[2]) neon_srhadd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.urhadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urhadd.v16i8")
    __vector(i8[16]) neon_urhadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.urhadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urhadd.v8i16")
    __vector(i16[8]) neon_urhadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.urhadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urhadd.v4i32")
    __vector(i32[4]) neon_urhadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.urhadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urhadd.v2i64")
    __vector(i64[2]) neon_urhadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.urhadd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urhadd.v4f32")
    __vector(f32[4]) neon_urhadd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.urhadd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urhadd.v2f64")
    __vector(f64[2]) neon_urhadd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.sqadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.v16i8")
    __vector(i8[16]) neon_sqadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqadd.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.i8")
    i8 neon_sqadd_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.v8i16")
    __vector(i16[8]) neon_sqadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqadd.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.i16")
    i16 neon_sqadd_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.v4i32")
    __vector(i32[4]) neon_sqadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqadd.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.i32")
    i32 neon_sqadd_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.v2i64")
    __vector(i64[2]) neon_sqadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqadd.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqadd.i64")
    i64 neon_sqadd_i64(i64, i64);
    /// The `llvm.aarch64.neon.suqadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.v16i8")
    __vector(i8[16]) neon_suqadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.suqadd.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.i8")
    i8 neon_suqadd_i8(i8, i8);
    /// The `llvm.aarch64.neon.suqadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.v8i16")
    __vector(i16[8]) neon_suqadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.suqadd.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.i16")
    i16 neon_suqadd_i16(i16, i16);
    /// The `llvm.aarch64.neon.suqadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.v4i32")
    __vector(i32[4]) neon_suqadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.suqadd.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.i32")
    i32 neon_suqadd_i32(i32, i32);
    /// The `llvm.aarch64.neon.suqadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.v2i64")
    __vector(i64[2]) neon_suqadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.suqadd.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.suqadd.i64")
    i64 neon_suqadd_i64(i64, i64);
    /// The `llvm.aarch64.neon.usqadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.v16i8")
    __vector(i8[16]) neon_usqadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.usqadd.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.i8")
    i8 neon_usqadd_i8(i8, i8);
    /// The `llvm.aarch64.neon.usqadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.v8i16")
    __vector(i16[8]) neon_usqadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.usqadd.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.i16")
    i16 neon_usqadd_i16(i16, i16);
    /// The `llvm.aarch64.neon.usqadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.v4i32")
    __vector(i32[4]) neon_usqadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.usqadd.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.i32")
    i32 neon_usqadd_i32(i32, i32);
    /// The `llvm.aarch64.neon.usqadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.v2i64")
    __vector(i64[2]) neon_usqadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.usqadd.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.usqadd.i64")
    i64 neon_usqadd_i64(i64, i64);
    /// The `llvm.aarch64.neon.uqadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.v16i8")
    __vector(i8[16]) neon_uqadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uqadd.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.i8")
    i8 neon_uqadd_i8(i8, i8);
    /// The `llvm.aarch64.neon.uqadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.v8i16")
    __vector(i16[8]) neon_uqadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uqadd.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.i16")
    i16 neon_uqadd_i16(i16, i16);
    /// The `llvm.aarch64.neon.uqadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.v4i32")
    __vector(i32[4]) neon_uqadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uqadd.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.i32")
    i32 neon_uqadd_i32(i32, i32);
    /// The `llvm.aarch64.neon.uqadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.v2i64")
    __vector(i64[2]) neon_uqadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uqadd.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqadd.i64")
    i64 neon_uqadd_i64(i64, i64);
    /// The `llvm.aarch64.neon.sqdmulh.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.v16i8")
    __vector(i8[16]) neon_sqdmulh_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqdmulh.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.i8")
    i8 neon_sqdmulh_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqdmulh.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.v8i16")
    __vector(i16[8]) neon_sqdmulh_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqdmulh.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.i16")
    i16 neon_sqdmulh_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqdmulh.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.v4i32")
    __vector(i32[4]) neon_sqdmulh_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqdmulh.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.i32")
    i32 neon_sqdmulh_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqdmulh.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.v2i64")
    __vector(i64[2]) neon_sqdmulh_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqdmulh.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulh.i64")
    i64 neon_sqdmulh_i64(i64, i64);
    /// The `llvm.aarch64.neon.sqrdmulh.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.v16i8")
    __vector(i8[16]) neon_sqrdmulh_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqrdmulh.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.i8")
    i8 neon_sqrdmulh_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqrdmulh.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.v8i16")
    __vector(i16[8]) neon_sqrdmulh_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqrdmulh.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.i16")
    i16 neon_sqrdmulh_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqrdmulh.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.v4i32")
    __vector(i32[4]) neon_sqrdmulh_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqrdmulh.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.i32")
    i32 neon_sqrdmulh_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqrdmulh.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.v2i64")
    __vector(i64[2]) neon_sqrdmulh_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqrdmulh.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrdmulh.i64")
    i64 neon_sqrdmulh_i64(i64, i64);
    /// The `llvm.aarch64.neon.pmul.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmul.v16i8")
    __vector(i8[16]) neon_pmul_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.pmul.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmul.v8i16")
    __vector(i16[8]) neon_pmul_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.pmul.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmul.v4i32")
    __vector(i32[4]) neon_pmul_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.pmul.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmul.v2i64")
    __vector(i64[2]) neon_pmul_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.pmul.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmul.v4f32")
    __vector(f32[4]) neon_pmul_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.pmul.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmul.v2f64")
    __vector(f64[2]) neon_pmul_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.pmull64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.pmull64")
    __vector(i8[16]) neon_pmull64(i64, i64);
    /// The `llvm.aarch64.neon.fmulx.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmulx.v4f32")
    __vector(f32[4]) neon_fmulx_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fmulx.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmulx.f32")
    f32 neon_fmulx_f32(f32, f32);
    /// The `llvm.aarch64.neon.fmulx.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmulx.v2f64")
    __vector(f64[2]) neon_fmulx_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fmulx.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmulx.f64")
    f64 neon_fmulx_f64(f64, f64);
    /// The `llvm.aarch64.neon.sqdmulls.scalar` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqdmulls.scalar")
    i64 neon_sqdmulls_scalar(i32, i32);
    /// The `llvm.aarch64.neon.shsub.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shsub.v16i8")
    __vector(i8[16]) neon_shsub_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.shsub.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shsub.v8i16")
    __vector(i16[8]) neon_shsub_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.shsub.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shsub.v4i32")
    __vector(i32[4]) neon_shsub_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.shsub.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shsub.v2i64")
    __vector(i64[2]) neon_shsub_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.shsub.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shsub.v4f32")
    __vector(f32[4]) neon_shsub_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.shsub.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.shsub.v2f64")
    __vector(f64[2]) neon_shsub_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.uhsub.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhsub.v16i8")
    __vector(i8[16]) neon_uhsub_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uhsub.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhsub.v8i16")
    __vector(i16[8]) neon_uhsub_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uhsub.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhsub.v4i32")
    __vector(i32[4]) neon_uhsub_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uhsub.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhsub.v2i64")
    __vector(i64[2]) neon_uhsub_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uhsub.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhsub.v4f32")
    __vector(f32[4]) neon_uhsub_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.uhsub.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uhsub.v2f64")
    __vector(f64[2]) neon_uhsub_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.sqsub.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.v16i8")
    __vector(i8[16]) neon_sqsub_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqsub.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.i8")
    i8 neon_sqsub_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqsub.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.v8i16")
    __vector(i16[8]) neon_sqsub_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqsub.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.i16")
    i16 neon_sqsub_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqsub.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.v4i32")
    __vector(i32[4]) neon_sqsub_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqsub.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.i32")
    i32 neon_sqsub_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqsub.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.v2i64")
    __vector(i64[2]) neon_sqsub_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqsub.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqsub.i64")
    i64 neon_sqsub_i64(i64, i64);
    /// The `llvm.aarch64.neon.uqsub.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.v16i8")
    __vector(i8[16]) neon_uqsub_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uqsub.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.i8")
    i8 neon_uqsub_i8(i8, i8);
    /// The `llvm.aarch64.neon.uqsub.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.v8i16")
    __vector(i16[8]) neon_uqsub_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uqsub.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.i16")
    i16 neon_uqsub_i16(i16, i16);
    /// The `llvm.aarch64.neon.uqsub.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.v4i32")
    __vector(i32[4]) neon_uqsub_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uqsub.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.i32")
    i32 neon_uqsub_i32(i32, i32);
    /// The `llvm.aarch64.neon.uqsub.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.v2i64")
    __vector(i64[2]) neon_uqsub_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uqsub.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqsub.i64")
    i64 neon_uqsub_i64(i64, i64);
    /// The `llvm.aarch64.neon.facge.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v16i8.v4f32")
    __vector(i8[16]) neon_facge_v16i8_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v16i8.f32")
    __vector(i8[16]) neon_facge_v16i8_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v16i8.v2f64")
    __vector(i8[16]) neon_facge_v16i8_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v16i8.f64")
    __vector(i8[16]) neon_facge_v16i8_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i8.v4f32")
    i8 neon_facge_i8_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i8.f32")
    i8 neon_facge_i8_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i8.v2f64")
    i8 neon_facge_i8_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i8.f64")
    i8 neon_facge_i8_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v8i16.v4f32")
    __vector(i16[8]) neon_facge_v8i16_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v8i16.f32")
    __vector(i16[8]) neon_facge_v8i16_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v8i16.v2f64")
    __vector(i16[8]) neon_facge_v8i16_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v8i16.f64")
    __vector(i16[8]) neon_facge_v8i16_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i16.v4f32")
    i16 neon_facge_i16_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i16.f32")
    i16 neon_facge_i16_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i16.v2f64")
    i16 neon_facge_i16_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i16.f64")
    i16 neon_facge_i16_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v4i32.v4f32")
    __vector(i32[4]) neon_facge_v4i32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v4i32.f32")
    __vector(i32[4]) neon_facge_v4i32_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v4i32.v2f64")
    __vector(i32[4]) neon_facge_v4i32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v4i32.f64")
    __vector(i32[4]) neon_facge_v4i32_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i32.v4f32")
    i32 neon_facge_i32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i32.f32")
    i32 neon_facge_i32_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i32.v2f64")
    i32 neon_facge_i32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i32.f64")
    i32 neon_facge_i32_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v2i64.v4f32")
    __vector(i64[2]) neon_facge_v2i64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v2i64.f32")
    __vector(i64[2]) neon_facge_v2i64_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v2i64.v2f64")
    __vector(i64[2]) neon_facge_v2i64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.v2i64.f64")
    __vector(i64[2]) neon_facge_v2i64_f64(f64, f64);
    /// The `llvm.aarch64.neon.facge.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i64.v4f32")
    i64 neon_facge_i64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facge.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i64.f32")
    i64 neon_facge_i64_f32(f32, f32);
    /// The `llvm.aarch64.neon.facge.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i64.v2f64")
    i64 neon_facge_i64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facge.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facge.i64.f64")
    i64 neon_facge_i64_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v16i8.v4f32")
    __vector(i8[16]) neon_facgt_v16i8_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v16i8.f32")
    __vector(i8[16]) neon_facgt_v16i8_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v16i8.v2f64")
    __vector(i8[16]) neon_facgt_v16i8_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v16i8.f64")
    __vector(i8[16]) neon_facgt_v16i8_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i8.v4f32")
    i8 neon_facgt_i8_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i8.f32")
    i8 neon_facgt_i8_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i8.v2f64")
    i8 neon_facgt_i8_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i8.f64")
    i8 neon_facgt_i8_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v8i16.v4f32")
    __vector(i16[8]) neon_facgt_v8i16_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v8i16.f32")
    __vector(i16[8]) neon_facgt_v8i16_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v8i16.v2f64")
    __vector(i16[8]) neon_facgt_v8i16_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v8i16.f64")
    __vector(i16[8]) neon_facgt_v8i16_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i16.v4f32")
    i16 neon_facgt_i16_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i16.f32")
    i16 neon_facgt_i16_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i16.v2f64")
    i16 neon_facgt_i16_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i16.f64")
    i16 neon_facgt_i16_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v4i32.v4f32")
    __vector(i32[4]) neon_facgt_v4i32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v4i32.f32")
    __vector(i32[4]) neon_facgt_v4i32_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v4i32.v2f64")
    __vector(i32[4]) neon_facgt_v4i32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v4i32.f64")
    __vector(i32[4]) neon_facgt_v4i32_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i32.v4f32")
    i32 neon_facgt_i32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i32.f32")
    i32 neon_facgt_i32_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i32.v2f64")
    i32 neon_facgt_i32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i32.f64")
    i32 neon_facgt_i32_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v2i64.v4f32")
    __vector(i64[2]) neon_facgt_v2i64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v2i64.f32")
    __vector(i64[2]) neon_facgt_v2i64_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v2i64.v2f64")
    __vector(i64[2]) neon_facgt_v2i64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.v2i64.f64")
    __vector(i64[2]) neon_facgt_v2i64_f64(f64, f64);
    /// The `llvm.aarch64.neon.facgt.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i64.v4f32")
    i64 neon_facgt_i64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.facgt.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i64.f32")
    i64 neon_facgt_i64_f32(f32, f32);
    /// The `llvm.aarch64.neon.facgt.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i64.v2f64")
    i64 neon_facgt_i64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.facgt.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.facgt.i64.f64")
    i64 neon_facgt_i64_f64(f64, f64);
    /// The `llvm.aarch64.neon.sabd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sabd.v16i8")
    __vector(i8[16]) neon_sabd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sabd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sabd.v8i16")
    __vector(i16[8]) neon_sabd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sabd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sabd.v4i32")
    __vector(i32[4]) neon_sabd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sabd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sabd.v2i64")
    __vector(i64[2]) neon_sabd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sabd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sabd.v4f32")
    __vector(f32[4]) neon_sabd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.sabd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sabd.v2f64")
    __vector(f64[2]) neon_sabd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.uabd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uabd.v16i8")
    __vector(i8[16]) neon_uabd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uabd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uabd.v8i16")
    __vector(i16[8]) neon_uabd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uabd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uabd.v4i32")
    __vector(i32[4]) neon_uabd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uabd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uabd.v2i64")
    __vector(i64[2]) neon_uabd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uabd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uabd.v4f32")
    __vector(f32[4]) neon_uabd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.uabd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uabd.v2f64")
    __vector(f64[2]) neon_uabd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fabd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fabd.v16i8")
    __vector(i8[16]) neon_fabd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fabd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fabd.v8i16")
    __vector(i16[8]) neon_fabd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fabd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fabd.v4i32")
    __vector(i32[4]) neon_fabd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fabd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fabd.v2i64")
    __vector(i64[2]) neon_fabd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fabd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fabd.v4f32")
    __vector(f32[4]) neon_fabd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fabd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fabd.v2f64")
    __vector(f64[2]) neon_fabd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.sisd.fabd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sisd.fabd.v4f32")
    __vector(f32[4]) sisd_fabd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.sisd.fabd.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sisd.fabd.f32")
    f32 sisd_fabd_f32(f32, f32);
    /// The `llvm.aarch64.sisd.fabd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sisd.fabd.v2f64")
    __vector(f64[2]) sisd_fabd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.sisd.fabd.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sisd.fabd.f64")
    f64 sisd_fabd_f64(f64, f64);
    /// The `llvm.aarch64.neon.smax.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smax.v16i8")
    __vector(i8[16]) neon_smax_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.smax.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smax.v8i16")
    __vector(i16[8]) neon_smax_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.smax.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smax.v4i32")
    __vector(i32[4]) neon_smax_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.smax.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smax.v2i64")
    __vector(i64[2]) neon_smax_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.smax.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smax.v4f32")
    __vector(f32[4]) neon_smax_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.smax.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smax.v2f64")
    __vector(f64[2]) neon_smax_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.umax.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umax.v16i8")
    __vector(i8[16]) neon_umax_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.umax.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umax.v8i16")
    __vector(i16[8]) neon_umax_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.umax.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umax.v4i32")
    __vector(i32[4]) neon_umax_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.umax.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umax.v2i64")
    __vector(i64[2]) neon_umax_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.umax.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umax.v4f32")
    __vector(f32[4]) neon_umax_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.umax.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umax.v2f64")
    __vector(f64[2]) neon_umax_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fmax.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmax.v16i8")
    __vector(i8[16]) neon_fmax_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fmax.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmax.v8i16")
    __vector(i16[8]) neon_fmax_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fmax.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmax.v4i32")
    __vector(i32[4]) neon_fmax_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fmax.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmax.v2i64")
    __vector(i64[2]) neon_fmax_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fmax.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmax.v4f32")
    __vector(f32[4]) neon_fmax_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fmax.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmax.v2f64")
    __vector(f64[2]) neon_fmax_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxnmp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmp.v16i8")
    __vector(i8[16]) neon_fmaxnmp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxnmp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmp.v8i16")
    __vector(i16[8]) neon_fmaxnmp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxnmp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmp.v4i32")
    __vector(i32[4]) neon_fmaxnmp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxnmp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmp.v2i64")
    __vector(i64[2]) neon_fmaxnmp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxnmp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmp.v4f32")
    __vector(f32[4]) neon_fmaxnmp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxnmp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmp.v2f64")
    __vector(f64[2]) neon_fmaxnmp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v16i8.v16i8")
    __vector(i8[16]) neon_smaxv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v16i8.v8i16")
    __vector(i8[16]) neon_smaxv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v16i8.v4i32")
    __vector(i8[16]) neon_smaxv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v16i8.v2i64")
    __vector(i8[16]) neon_smaxv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v16i8.v4f32")
    __vector(i8[16]) neon_smaxv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v16i8.v2f64")
    __vector(i8[16]) neon_smaxv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i8.v16i8")
    i8 neon_smaxv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i8.v8i16")
    i8 neon_smaxv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i8.v4i32")
    i8 neon_smaxv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i8.v2i64")
    i8 neon_smaxv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i8.v4f32")
    i8 neon_smaxv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i8.v2f64")
    i8 neon_smaxv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v8i16.v16i8")
    __vector(i16[8]) neon_smaxv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v8i16.v8i16")
    __vector(i16[8]) neon_smaxv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v8i16.v4i32")
    __vector(i16[8]) neon_smaxv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v8i16.v2i64")
    __vector(i16[8]) neon_smaxv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v8i16.v4f32")
    __vector(i16[8]) neon_smaxv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v8i16.v2f64")
    __vector(i16[8]) neon_smaxv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i16.v16i8")
    i16 neon_smaxv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i16.v8i16")
    i16 neon_smaxv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i16.v4i32")
    i16 neon_smaxv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i16.v2i64")
    i16 neon_smaxv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i16.v4f32")
    i16 neon_smaxv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i16.v2f64")
    i16 neon_smaxv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v4i32.v16i8")
    __vector(i32[4]) neon_smaxv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v4i32.v8i16")
    __vector(i32[4]) neon_smaxv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v4i32.v4i32")
    __vector(i32[4]) neon_smaxv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v4i32.v2i64")
    __vector(i32[4]) neon_smaxv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v4i32.v4f32")
    __vector(i32[4]) neon_smaxv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v4i32.v2f64")
    __vector(i32[4]) neon_smaxv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i32.v16i8")
    i32 neon_smaxv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i32.v8i16")
    i32 neon_smaxv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i32.v4i32")
    i32 neon_smaxv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i32.v2i64")
    i32 neon_smaxv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i32.v4f32")
    i32 neon_smaxv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i32.v2f64")
    i32 neon_smaxv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v2i64.v16i8")
    __vector(i64[2]) neon_smaxv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v2i64.v8i16")
    __vector(i64[2]) neon_smaxv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v2i64.v4i32")
    __vector(i64[2]) neon_smaxv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v2i64.v2i64")
    __vector(i64[2]) neon_smaxv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v2i64.v4f32")
    __vector(i64[2]) neon_smaxv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.v2i64.v2f64")
    __vector(i64[2]) neon_smaxv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i64.v16i8")
    i64 neon_smaxv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i64.v8i16")
    i64 neon_smaxv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i64.v4i32")
    i64 neon_smaxv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i64.v2i64")
    i64 neon_smaxv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i64.v4f32")
    i64 neon_smaxv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxv.i64.v2f64")
    i64 neon_smaxv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v16i8.v16i8")
    __vector(i8[16]) neon_umaxv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v16i8.v8i16")
    __vector(i8[16]) neon_umaxv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v16i8.v4i32")
    __vector(i8[16]) neon_umaxv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v16i8.v2i64")
    __vector(i8[16]) neon_umaxv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v16i8.v4f32")
    __vector(i8[16]) neon_umaxv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v16i8.v2f64")
    __vector(i8[16]) neon_umaxv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i8.v16i8")
    i8 neon_umaxv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i8.v8i16")
    i8 neon_umaxv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i8.v4i32")
    i8 neon_umaxv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i8.v2i64")
    i8 neon_umaxv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i8.v4f32")
    i8 neon_umaxv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i8.v2f64")
    i8 neon_umaxv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v8i16.v16i8")
    __vector(i16[8]) neon_umaxv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v8i16.v8i16")
    __vector(i16[8]) neon_umaxv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v8i16.v4i32")
    __vector(i16[8]) neon_umaxv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v8i16.v2i64")
    __vector(i16[8]) neon_umaxv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v8i16.v4f32")
    __vector(i16[8]) neon_umaxv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v8i16.v2f64")
    __vector(i16[8]) neon_umaxv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i16.v16i8")
    i16 neon_umaxv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i16.v8i16")
    i16 neon_umaxv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i16.v4i32")
    i16 neon_umaxv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i16.v2i64")
    i16 neon_umaxv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i16.v4f32")
    i16 neon_umaxv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i16.v2f64")
    i16 neon_umaxv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v4i32.v16i8")
    __vector(i32[4]) neon_umaxv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v4i32.v8i16")
    __vector(i32[4]) neon_umaxv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v4i32.v4i32")
    __vector(i32[4]) neon_umaxv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v4i32.v2i64")
    __vector(i32[4]) neon_umaxv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v4i32.v4f32")
    __vector(i32[4]) neon_umaxv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v4i32.v2f64")
    __vector(i32[4]) neon_umaxv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i32.v16i8")
    i32 neon_umaxv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i32.v8i16")
    i32 neon_umaxv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i32.v4i32")
    i32 neon_umaxv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i32.v2i64")
    i32 neon_umaxv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i32.v4f32")
    i32 neon_umaxv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i32.v2f64")
    i32 neon_umaxv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v2i64.v16i8")
    __vector(i64[2]) neon_umaxv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v2i64.v8i16")
    __vector(i64[2]) neon_umaxv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v2i64.v4i32")
    __vector(i64[2]) neon_umaxv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v2i64.v2i64")
    __vector(i64[2]) neon_umaxv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v2i64.v4f32")
    __vector(i64[2]) neon_umaxv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.v2i64.v2f64")
    __vector(i64[2]) neon_umaxv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i64.v16i8")
    i64 neon_umaxv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i64.v8i16")
    i64 neon_umaxv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i64.v4i32")
    i64 neon_umaxv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i64.v2i64")
    i64 neon_umaxv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i64.v4f32")
    i64 neon_umaxv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxv.i64.v2f64")
    i64 neon_umaxv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxv.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v4f32.v16i8")
    __vector(f32[4]) neon_fmaxv_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxv.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v4f32.v8i16")
    __vector(f32[4]) neon_fmaxv_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxv.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v4f32.v4i32")
    __vector(f32[4]) neon_fmaxv_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxv.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v4f32.v2i64")
    __vector(f32[4]) neon_fmaxv_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxv.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v4f32.v4f32")
    __vector(f32[4]) neon_fmaxv_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxv.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v4f32.v2f64")
    __vector(f32[4]) neon_fmaxv_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxv.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f32.v16i8")
    f32 neon_fmaxv_f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxv.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f32.v8i16")
    f32 neon_fmaxv_f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxv.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f32.v4i32")
    f32 neon_fmaxv_f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxv.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f32.v2i64")
    f32 neon_fmaxv_f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxv.f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f32.v4f32")
    f32 neon_fmaxv_f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxv.f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f32.v2f64")
    f32 neon_fmaxv_f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxv.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v2f64.v16i8")
    __vector(f64[2]) neon_fmaxv_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxv.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v2f64.v8i16")
    __vector(f64[2]) neon_fmaxv_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxv.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v2f64.v4i32")
    __vector(f64[2]) neon_fmaxv_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxv.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v2f64.v2i64")
    __vector(f64[2]) neon_fmaxv_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxv.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v2f64.v4f32")
    __vector(f64[2]) neon_fmaxv_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxv.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.v2f64.v2f64")
    __vector(f64[2]) neon_fmaxv_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxv.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f64.v16i8")
    f64 neon_fmaxv_f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxv.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f64.v8i16")
    f64 neon_fmaxv_f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxv.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f64.v4i32")
    f64 neon_fmaxv_f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxv.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f64.v2i64")
    f64 neon_fmaxv_f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxv.f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f64.v4f32")
    f64 neon_fmaxv_f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxv.f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxv.f64.v2f64")
    f64 neon_fmaxv_f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v4f32.v16i8")
    __vector(f32[4]) neon_fmaxnmv_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxnmv.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v4f32.v8i16")
    __vector(f32[4]) neon_fmaxnmv_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxnmv.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v4f32.v4i32")
    __vector(f32[4]) neon_fmaxnmv_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v4f32.v2i64")
    __vector(f32[4]) neon_fmaxnmv_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v4f32.v4f32")
    __vector(f32[4]) neon_fmaxnmv_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v4f32.v2f64")
    __vector(f32[4]) neon_fmaxnmv_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f32.v16i8")
    f32 neon_fmaxnmv_f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxnmv.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f32.v8i16")
    f32 neon_fmaxnmv_f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxnmv.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f32.v4i32")
    f32 neon_fmaxnmv_f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f32.v2i64")
    f32 neon_fmaxnmv_f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f32.v4f32")
    f32 neon_fmaxnmv_f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f32.v2f64")
    f32 neon_fmaxnmv_f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v2f64.v16i8")
    __vector(f64[2]) neon_fmaxnmv_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxnmv.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v2f64.v8i16")
    __vector(f64[2]) neon_fmaxnmv_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxnmv.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v2f64.v4i32")
    __vector(f64[2]) neon_fmaxnmv_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v2f64.v2i64")
    __vector(f64[2]) neon_fmaxnmv_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v2f64.v4f32")
    __vector(f64[2]) neon_fmaxnmv_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.v2f64.v2f64")
    __vector(f64[2]) neon_fmaxnmv_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f64.v16i8")
    f64 neon_fmaxnmv_f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxnmv.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f64.v8i16")
    f64 neon_fmaxnmv_f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxnmv.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f64.v4i32")
    f64 neon_fmaxnmv_f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f64.v2i64")
    f64 neon_fmaxnmv_f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxnmv.f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f64.v4f32")
    f64 neon_fmaxnmv_f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxnmv.f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnmv.f64.v2f64")
    f64 neon_fmaxnmv_f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smin.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smin.v16i8")
    __vector(i8[16]) neon_smin_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.smin.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smin.v8i16")
    __vector(i16[8]) neon_smin_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.smin.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smin.v4i32")
    __vector(i32[4]) neon_smin_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.smin.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smin.v2i64")
    __vector(i64[2]) neon_smin_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.smin.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smin.v4f32")
    __vector(f32[4]) neon_smin_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.smin.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smin.v2f64")
    __vector(f64[2]) neon_smin_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.umin.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umin.v16i8")
    __vector(i8[16]) neon_umin_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.umin.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umin.v8i16")
    __vector(i16[8]) neon_umin_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.umin.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umin.v4i32")
    __vector(i32[4]) neon_umin_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.umin.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umin.v2i64")
    __vector(i64[2]) neon_umin_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.umin.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umin.v4f32")
    __vector(f32[4]) neon_umin_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.umin.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umin.v2f64")
    __vector(f64[2]) neon_umin_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fmin.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmin.v16i8")
    __vector(i8[16]) neon_fmin_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fmin.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmin.v8i16")
    __vector(i16[8]) neon_fmin_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fmin.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmin.v4i32")
    __vector(i32[4]) neon_fmin_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fmin.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmin.v2i64")
    __vector(i64[2]) neon_fmin_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fmin.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmin.v4f32")
    __vector(f32[4]) neon_fmin_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fmin.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmin.v2f64")
    __vector(f64[2]) neon_fmin_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnmp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmp.v16i8")
    __vector(i8[16]) neon_fminnmp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fminnmp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmp.v8i16")
    __vector(i16[8]) neon_fminnmp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fminnmp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmp.v4i32")
    __vector(i32[4]) neon_fminnmp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fminnmp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmp.v2i64")
    __vector(i64[2]) neon_fminnmp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fminnmp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmp.v4f32")
    __vector(f32[4]) neon_fminnmp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fminnmp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmp.v2f64")
    __vector(f64[2]) neon_fminnmp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnm.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnm.v4f32")
    __vector(f32[4]) neon_fminnm_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fminnm.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnm.f32")
    f32 neon_fminnm_f32(f32, f32);
    /// The `llvm.aarch64.neon.fminnm.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnm.v2f64")
    __vector(f64[2]) neon_fminnm_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnm.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnm.f64")
    f64 neon_fminnm_f64(f64, f64);
    /// The `llvm.aarch64.neon.fmaxnm.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnm.v4f32")
    __vector(f32[4]) neon_fmaxnm_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxnm.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnm.f32")
    f32 neon_fmaxnm_f32(f32, f32);
    /// The `llvm.aarch64.neon.fmaxnm.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnm.v2f64")
    __vector(f64[2]) neon_fmaxnm_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxnm.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxnm.f64")
    f64 neon_fmaxnm_f64(f64, f64);
    /// The `llvm.aarch64.neon.sminv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v16i8.v16i8")
    __vector(i8[16]) neon_sminv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v16i8.v8i16")
    __vector(i8[16]) neon_sminv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v16i8.v4i32")
    __vector(i8[16]) neon_sminv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v16i8.v2i64")
    __vector(i8[16]) neon_sminv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v16i8.v4f32")
    __vector(i8[16]) neon_sminv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v16i8.v2f64")
    __vector(i8[16]) neon_sminv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i8.v16i8")
    i8 neon_sminv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i8.v8i16")
    i8 neon_sminv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i8.v4i32")
    i8 neon_sminv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i8.v2i64")
    i8 neon_sminv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i8.v4f32")
    i8 neon_sminv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i8.v2f64")
    i8 neon_sminv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v8i16.v16i8")
    __vector(i16[8]) neon_sminv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v8i16.v8i16")
    __vector(i16[8]) neon_sminv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v8i16.v4i32")
    __vector(i16[8]) neon_sminv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v8i16.v2i64")
    __vector(i16[8]) neon_sminv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v8i16.v4f32")
    __vector(i16[8]) neon_sminv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v8i16.v2f64")
    __vector(i16[8]) neon_sminv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i16.v16i8")
    i16 neon_sminv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i16.v8i16")
    i16 neon_sminv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i16.v4i32")
    i16 neon_sminv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i16.v2i64")
    i16 neon_sminv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i16.v4f32")
    i16 neon_sminv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i16.v2f64")
    i16 neon_sminv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v4i32.v16i8")
    __vector(i32[4]) neon_sminv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v4i32.v8i16")
    __vector(i32[4]) neon_sminv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v4i32.v4i32")
    __vector(i32[4]) neon_sminv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v4i32.v2i64")
    __vector(i32[4]) neon_sminv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v4i32.v4f32")
    __vector(i32[4]) neon_sminv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v4i32.v2f64")
    __vector(i32[4]) neon_sminv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i32.v16i8")
    i32 neon_sminv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i32.v8i16")
    i32 neon_sminv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i32.v4i32")
    i32 neon_sminv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i32.v2i64")
    i32 neon_sminv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i32.v4f32")
    i32 neon_sminv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i32.v2f64")
    i32 neon_sminv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v2i64.v16i8")
    __vector(i64[2]) neon_sminv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v2i64.v8i16")
    __vector(i64[2]) neon_sminv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v2i64.v4i32")
    __vector(i64[2]) neon_sminv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v2i64.v2i64")
    __vector(i64[2]) neon_sminv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v2i64.v4f32")
    __vector(i64[2]) neon_sminv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.v2i64.v2f64")
    __vector(i64[2]) neon_sminv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.sminv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i64.v16i8")
    i64 neon_sminv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sminv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i64.v8i16")
    i64 neon_sminv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sminv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i64.v4i32")
    i64 neon_sminv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sminv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i64.v2i64")
    i64 neon_sminv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sminv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i64.v4f32")
    i64 neon_sminv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.sminv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminv.i64.v2f64")
    i64 neon_sminv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v16i8.v16i8")
    __vector(i8[16]) neon_uminv_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v16i8.v8i16")
    __vector(i8[16]) neon_uminv_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v16i8.v4i32")
    __vector(i8[16]) neon_uminv_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v16i8.v2i64")
    __vector(i8[16]) neon_uminv_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v16i8.v4f32")
    __vector(i8[16]) neon_uminv_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v16i8.v2f64")
    __vector(i8[16]) neon_uminv_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i8.v16i8")
    i8 neon_uminv_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i8.v8i16")
    i8 neon_uminv_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i8.v4i32")
    i8 neon_uminv_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i8.v2i64")
    i8 neon_uminv_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i8.v4f32")
    i8 neon_uminv_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i8.v2f64")
    i8 neon_uminv_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v8i16.v16i8")
    __vector(i16[8]) neon_uminv_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v8i16.v8i16")
    __vector(i16[8]) neon_uminv_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v8i16.v4i32")
    __vector(i16[8]) neon_uminv_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v8i16.v2i64")
    __vector(i16[8]) neon_uminv_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v8i16.v4f32")
    __vector(i16[8]) neon_uminv_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v8i16.v2f64")
    __vector(i16[8]) neon_uminv_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i16.v16i8")
    i16 neon_uminv_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i16.v8i16")
    i16 neon_uminv_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i16.v4i32")
    i16 neon_uminv_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i16.v2i64")
    i16 neon_uminv_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i16.v4f32")
    i16 neon_uminv_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i16.v2f64")
    i16 neon_uminv_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v4i32.v16i8")
    __vector(i32[4]) neon_uminv_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v4i32.v8i16")
    __vector(i32[4]) neon_uminv_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v4i32.v4i32")
    __vector(i32[4]) neon_uminv_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v4i32.v2i64")
    __vector(i32[4]) neon_uminv_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v4i32.v4f32")
    __vector(i32[4]) neon_uminv_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v4i32.v2f64")
    __vector(i32[4]) neon_uminv_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i32.v16i8")
    i32 neon_uminv_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i32.v8i16")
    i32 neon_uminv_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i32.v4i32")
    i32 neon_uminv_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i32.v2i64")
    i32 neon_uminv_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i32.v4f32")
    i32 neon_uminv_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i32.v2f64")
    i32 neon_uminv_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v2i64.v16i8")
    __vector(i64[2]) neon_uminv_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v2i64.v8i16")
    __vector(i64[2]) neon_uminv_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v2i64.v4i32")
    __vector(i64[2]) neon_uminv_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v2i64.v2i64")
    __vector(i64[2]) neon_uminv_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v2i64.v4f32")
    __vector(i64[2]) neon_uminv_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.v2i64.v2f64")
    __vector(i64[2]) neon_uminv_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uminv.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i64.v16i8")
    i64 neon_uminv_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uminv.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i64.v8i16")
    i64 neon_uminv_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uminv.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i64.v4i32")
    i64 neon_uminv_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uminv.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i64.v2i64")
    i64 neon_uminv_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uminv.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i64.v4f32")
    i64 neon_uminv_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uminv.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminv.i64.v2f64")
    i64 neon_uminv_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminv.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v4f32.v16i8")
    __vector(f32[4]) neon_fminv_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminv.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v4f32.v8i16")
    __vector(f32[4]) neon_fminv_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminv.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v4f32.v4i32")
    __vector(f32[4]) neon_fminv_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminv.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v4f32.v2i64")
    __vector(f32[4]) neon_fminv_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminv.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v4f32.v4f32")
    __vector(f32[4]) neon_fminv_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminv.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v4f32.v2f64")
    __vector(f32[4]) neon_fminv_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminv.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f32.v16i8")
    f32 neon_fminv_f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminv.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f32.v8i16")
    f32 neon_fminv_f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminv.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f32.v4i32")
    f32 neon_fminv_f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminv.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f32.v2i64")
    f32 neon_fminv_f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminv.f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f32.v4f32")
    f32 neon_fminv_f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminv.f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f32.v2f64")
    f32 neon_fminv_f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminv.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v2f64.v16i8")
    __vector(f64[2]) neon_fminv_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminv.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v2f64.v8i16")
    __vector(f64[2]) neon_fminv_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminv.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v2f64.v4i32")
    __vector(f64[2]) neon_fminv_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminv.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v2f64.v2i64")
    __vector(f64[2]) neon_fminv_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminv.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v2f64.v4f32")
    __vector(f64[2]) neon_fminv_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminv.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.v2f64.v2f64")
    __vector(f64[2]) neon_fminv_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminv.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f64.v16i8")
    f64 neon_fminv_f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminv.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f64.v8i16")
    f64 neon_fminv_f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminv.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f64.v4i32")
    f64 neon_fminv_f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminv.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f64.v2i64")
    f64 neon_fminv_f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminv.f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f64.v4f32")
    f64 neon_fminv_f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminv.f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminv.f64.v2f64")
    f64 neon_fminv_f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnmv.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v4f32.v16i8")
    __vector(f32[4]) neon_fminnmv_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminnmv.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v4f32.v8i16")
    __vector(f32[4]) neon_fminnmv_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminnmv.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v4f32.v4i32")
    __vector(f32[4]) neon_fminnmv_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminnmv.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v4f32.v2i64")
    __vector(f32[4]) neon_fminnmv_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminnmv.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v4f32.v4f32")
    __vector(f32[4]) neon_fminnmv_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminnmv.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v4f32.v2f64")
    __vector(f32[4]) neon_fminnmv_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnmv.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f32.v16i8")
    f32 neon_fminnmv_f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminnmv.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f32.v8i16")
    f32 neon_fminnmv_f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminnmv.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f32.v4i32")
    f32 neon_fminnmv_f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminnmv.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f32.v2i64")
    f32 neon_fminnmv_f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminnmv.f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f32.v4f32")
    f32 neon_fminnmv_f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminnmv.f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f32.v2f64")
    f32 neon_fminnmv_f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnmv.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v2f64.v16i8")
    __vector(f64[2]) neon_fminnmv_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminnmv.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v2f64.v8i16")
    __vector(f64[2]) neon_fminnmv_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminnmv.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v2f64.v4i32")
    __vector(f64[2]) neon_fminnmv_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminnmv.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v2f64.v2i64")
    __vector(f64[2]) neon_fminnmv_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminnmv.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v2f64.v4f32")
    __vector(f64[2]) neon_fminnmv_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminnmv.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.v2f64.v2f64")
    __vector(f64[2]) neon_fminnmv_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fminnmv.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f64.v16i8")
    f64 neon_fminnmv_f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fminnmv.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f64.v8i16")
    f64 neon_fminnmv_f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fminnmv.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f64.v4i32")
    f64 neon_fminnmv_f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fminnmv.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f64.v2i64")
    f64 neon_fminnmv_f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fminnmv.f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f64.v4f32")
    f64 neon_fminnmv_f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fminnmv.f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminnmv.f64.v2f64")
    f64 neon_fminnmv_f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.addp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.addp.v16i8")
    __vector(i8[16]) neon_addp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.addp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.addp.v8i16")
    __vector(i16[8]) neon_addp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.addp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.addp.v4i32")
    __vector(i32[4]) neon_addp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.addp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.addp.v2i64")
    __vector(i64[2]) neon_addp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.addp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.addp.v4f32")
    __vector(f32[4]) neon_addp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.addp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.addp.v2f64")
    __vector(f64[2]) neon_addp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlp.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v16i8.v16i8")
    __vector(i8[16]) neon_saddlp_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlp.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v16i8.v8i16")
    __vector(i8[16]) neon_saddlp_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlp.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v16i8.v4i32")
    __vector(i8[16]) neon_saddlp_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlp.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v16i8.v2i64")
    __vector(i8[16]) neon_saddlp_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlp.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v16i8.v4f32")
    __vector(i8[16]) neon_saddlp_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlp.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v16i8.v2f64")
    __vector(i8[16]) neon_saddlp_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlp.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v8i16.v16i8")
    __vector(i16[8]) neon_saddlp_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlp.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v8i16.v8i16")
    __vector(i16[8]) neon_saddlp_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlp.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v8i16.v4i32")
    __vector(i16[8]) neon_saddlp_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlp.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v8i16.v2i64")
    __vector(i16[8]) neon_saddlp_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlp.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v8i16.v4f32")
    __vector(i16[8]) neon_saddlp_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlp.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v8i16.v2f64")
    __vector(i16[8]) neon_saddlp_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlp.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4i32.v16i8")
    __vector(i32[4]) neon_saddlp_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlp.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4i32.v8i16")
    __vector(i32[4]) neon_saddlp_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlp.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4i32.v4i32")
    __vector(i32[4]) neon_saddlp_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlp.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4i32.v2i64")
    __vector(i32[4]) neon_saddlp_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlp.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4i32.v4f32")
    __vector(i32[4]) neon_saddlp_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlp.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4i32.v2f64")
    __vector(i32[4]) neon_saddlp_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlp.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2i64.v16i8")
    __vector(i64[2]) neon_saddlp_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlp.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2i64.v8i16")
    __vector(i64[2]) neon_saddlp_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlp.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2i64.v4i32")
    __vector(i64[2]) neon_saddlp_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlp.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2i64.v2i64")
    __vector(i64[2]) neon_saddlp_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlp.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2i64.v4f32")
    __vector(i64[2]) neon_saddlp_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlp.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2i64.v2f64")
    __vector(i64[2]) neon_saddlp_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlp.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4f32.v16i8")
    __vector(f32[4]) neon_saddlp_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlp.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4f32.v8i16")
    __vector(f32[4]) neon_saddlp_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlp.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4f32.v4i32")
    __vector(f32[4]) neon_saddlp_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlp.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4f32.v2i64")
    __vector(f32[4]) neon_saddlp_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlp.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4f32.v4f32")
    __vector(f32[4]) neon_saddlp_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlp.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v4f32.v2f64")
    __vector(f32[4]) neon_saddlp_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.saddlp.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2f64.v16i8")
    __vector(f64[2]) neon_saddlp_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.saddlp.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2f64.v8i16")
    __vector(f64[2]) neon_saddlp_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.saddlp.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2f64.v4i32")
    __vector(f64[2]) neon_saddlp_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.saddlp.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2f64.v2i64")
    __vector(f64[2]) neon_saddlp_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.saddlp.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2f64.v4f32")
    __vector(f64[2]) neon_saddlp_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.saddlp.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.saddlp.v2f64.v2f64")
    __vector(f64[2]) neon_saddlp_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v16i8.v16i8")
    __vector(i8[16]) neon_uaddlp_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlp.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v16i8.v8i16")
    __vector(i8[16]) neon_uaddlp_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlp.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v16i8.v4i32")
    __vector(i8[16]) neon_uaddlp_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v16i8.v2i64")
    __vector(i8[16]) neon_uaddlp_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v16i8.v4f32")
    __vector(i8[16]) neon_uaddlp_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v16i8.v2f64")
    __vector(i8[16]) neon_uaddlp_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v8i16.v16i8")
    __vector(i16[8]) neon_uaddlp_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlp.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v8i16.v8i16")
    __vector(i16[8]) neon_uaddlp_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlp.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v8i16.v4i32")
    __vector(i16[8]) neon_uaddlp_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v8i16.v2i64")
    __vector(i16[8]) neon_uaddlp_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v8i16.v4f32")
    __vector(i16[8]) neon_uaddlp_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v8i16.v2f64")
    __vector(i16[8]) neon_uaddlp_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4i32.v16i8")
    __vector(i32[4]) neon_uaddlp_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlp.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4i32.v8i16")
    __vector(i32[4]) neon_uaddlp_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlp.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4i32.v4i32")
    __vector(i32[4]) neon_uaddlp_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4i32.v2i64")
    __vector(i32[4]) neon_uaddlp_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4i32.v4f32")
    __vector(i32[4]) neon_uaddlp_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4i32.v2f64")
    __vector(i32[4]) neon_uaddlp_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2i64.v16i8")
    __vector(i64[2]) neon_uaddlp_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlp.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2i64.v8i16")
    __vector(i64[2]) neon_uaddlp_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlp.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2i64.v4i32")
    __vector(i64[2]) neon_uaddlp_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2i64.v2i64")
    __vector(i64[2]) neon_uaddlp_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2i64.v4f32")
    __vector(i64[2]) neon_uaddlp_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2i64.v2f64")
    __vector(i64[2]) neon_uaddlp_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4f32.v16i8")
    __vector(f32[4]) neon_uaddlp_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlp.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4f32.v8i16")
    __vector(f32[4]) neon_uaddlp_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlp.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4f32.v4i32")
    __vector(f32[4]) neon_uaddlp_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4f32.v2i64")
    __vector(f32[4]) neon_uaddlp_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4f32.v4f32")
    __vector(f32[4]) neon_uaddlp_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v4f32.v2f64")
    __vector(f32[4]) neon_uaddlp_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2f64.v16i8")
    __vector(f64[2]) neon_uaddlp_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.uaddlp.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2f64.v8i16")
    __vector(f64[2]) neon_uaddlp_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.uaddlp.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2f64.v4i32")
    __vector(f64[2]) neon_uaddlp_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2f64.v2i64")
    __vector(f64[2]) neon_uaddlp_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.uaddlp.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2f64.v4f32")
    __vector(f64[2]) neon_uaddlp_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.uaddlp.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uaddlp.v2f64.v2f64")
    __vector(f64[2]) neon_uaddlp_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.smaxp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxp.v16i8")
    __vector(i8[16]) neon_smaxp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.smaxp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxp.v8i16")
    __vector(i16[8]) neon_smaxp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.smaxp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxp.v4i32")
    __vector(i32[4]) neon_smaxp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.smaxp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxp.v2i64")
    __vector(i64[2]) neon_smaxp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.smaxp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxp.v4f32")
    __vector(f32[4]) neon_smaxp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.smaxp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.smaxp.v2f64")
    __vector(f64[2]) neon_smaxp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.umaxp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxp.v16i8")
    __vector(i8[16]) neon_umaxp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.umaxp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxp.v8i16")
    __vector(i16[8]) neon_umaxp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.umaxp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxp.v4i32")
    __vector(i32[4]) neon_umaxp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.umaxp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxp.v2i64")
    __vector(i64[2]) neon_umaxp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.umaxp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxp.v4f32")
    __vector(f32[4]) neon_umaxp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.umaxp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.umaxp.v2f64")
    __vector(f64[2]) neon_umaxp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fmaxp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxp.v16i8")
    __vector(i8[16]) neon_fmaxp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fmaxp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxp.v8i16")
    __vector(i16[8]) neon_fmaxp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fmaxp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxp.v4i32")
    __vector(i32[4]) neon_fmaxp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fmaxp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxp.v2i64")
    __vector(i64[2]) neon_fmaxp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fmaxp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxp.v4f32")
    __vector(f32[4]) neon_fmaxp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fmaxp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fmaxp.v2f64")
    __vector(f64[2]) neon_fmaxp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.sminp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminp.v16i8")
    __vector(i8[16]) neon_sminp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sminp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminp.v8i16")
    __vector(i16[8]) neon_sminp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sminp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminp.v4i32")
    __vector(i32[4]) neon_sminp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sminp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminp.v2i64")
    __vector(i64[2]) neon_sminp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sminp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminp.v4f32")
    __vector(f32[4]) neon_sminp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.sminp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sminp.v2f64")
    __vector(f64[2]) neon_sminp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.uminp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminp.v16i8")
    __vector(i8[16]) neon_uminp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uminp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminp.v8i16")
    __vector(i16[8]) neon_uminp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uminp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminp.v4i32")
    __vector(i32[4]) neon_uminp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uminp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminp.v2i64")
    __vector(i64[2]) neon_uminp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uminp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminp.v4f32")
    __vector(f32[4]) neon_uminp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.uminp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uminp.v2f64")
    __vector(f64[2]) neon_uminp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.fminp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminp.v16i8")
    __vector(i8[16]) neon_fminp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.fminp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminp.v8i16")
    __vector(i16[8]) neon_fminp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.fminp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminp.v4i32")
    __vector(i32[4]) neon_fminp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.fminp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminp.v2i64")
    __vector(i64[2]) neon_fminp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.fminp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminp.v4f32")
    __vector(f32[4]) neon_fminp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.fminp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fminp.v2f64")
    __vector(f64[2]) neon_fminp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.frecps.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecps.v4f32")
    __vector(f32[4]) neon_frecps_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.frecps.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecps.f32")
    f32 neon_frecps_f32(f32, f32);
    /// The `llvm.aarch64.neon.frecps.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecps.v2f64")
    __vector(f64[2]) neon_frecps_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.frecps.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecps.f64")
    f64 neon_frecps_f64(f64, f64);
    /// The `llvm.aarch64.neon.frsqrts.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrts.v4f32")
    __vector(f32[4]) neon_frsqrts_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.frsqrts.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrts.f32")
    f32 neon_frsqrts_f32(f32, f32);
    /// The `llvm.aarch64.neon.frsqrts.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrts.v2f64")
    __vector(f64[2]) neon_frsqrts_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.frsqrts.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrts.f64")
    f64 neon_frsqrts_f64(f64, f64);
    /// The `llvm.aarch64.neon.frecpx.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpx.v4f32")
    __vector(f32[4]) neon_frecpx_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.frecpx.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpx.f32")
    f32 neon_frecpx_f32(f32);
    /// The `llvm.aarch64.neon.frecpx.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpx.v2f64")
    __vector(f64[2]) neon_frecpx_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.frecpx.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpx.f64")
    f64 neon_frecpx_f64(f64);
    /// The `llvm.aarch64.neon.sqshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.v16i8")
    __vector(i8[16]) neon_sqshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.i8")
    i8 neon_sqshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.v8i16")
    __vector(i16[8]) neon_sqshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.i16")
    i16 neon_sqshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.v4i32")
    __vector(i32[4]) neon_sqshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.i32")
    i32 neon_sqshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.v2i64")
    __vector(i64[2]) neon_sqshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshl.i64")
    i64 neon_sqshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.uqshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.v16i8")
    __vector(i8[16]) neon_uqshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uqshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.i8")
    i8 neon_uqshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.uqshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.v8i16")
    __vector(i16[8]) neon_uqshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uqshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.i16")
    i16 neon_uqshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.uqshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.v4i32")
    __vector(i32[4]) neon_uqshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uqshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.i32")
    i32 neon_uqshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.uqshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.v2i64")
    __vector(i64[2]) neon_uqshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uqshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqshl.i64")
    i64 neon_uqshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.srshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.v16i8")
    __vector(i8[16]) neon_srshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.srshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.i8")
    i8 neon_srshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.srshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.v8i16")
    __vector(i16[8]) neon_srshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.srshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.i16")
    i16 neon_srshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.srshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.v4i32")
    __vector(i32[4]) neon_srshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.srshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.i32")
    i32 neon_srshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.srshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.v2i64")
    __vector(i64[2]) neon_srshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.srshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.srshl.i64")
    i64 neon_srshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.urshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.v16i8")
    __vector(i8[16]) neon_urshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.urshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.i8")
    i8 neon_urshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.urshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.v8i16")
    __vector(i16[8]) neon_urshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.urshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.i16")
    i16 neon_urshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.urshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.v4i32")
    __vector(i32[4]) neon_urshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.urshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.i32")
    i32 neon_urshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.urshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.v2i64")
    __vector(i64[2]) neon_urshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.urshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urshl.i64")
    i64 neon_urshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.sqrshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.v16i8")
    __vector(i8[16]) neon_sqrshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqrshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.i8")
    i8 neon_sqrshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqrshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.v8i16")
    __vector(i16[8]) neon_sqrshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqrshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.i16")
    i16 neon_sqrshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqrshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.v4i32")
    __vector(i32[4]) neon_sqrshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqrshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.i32")
    i32 neon_sqrshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqrshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.v2i64")
    __vector(i64[2]) neon_sqrshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqrshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqrshl.i64")
    i64 neon_sqrshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.uqrshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.v16i8")
    __vector(i8[16]) neon_uqrshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.uqrshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.i8")
    i8 neon_uqrshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.uqrshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.v8i16")
    __vector(i16[8]) neon_uqrshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.uqrshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.i16")
    i16 neon_uqrshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.uqrshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.v4i32")
    __vector(i32[4]) neon_uqrshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.uqrshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.i32")
    i32 neon_uqrshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.uqrshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.v2i64")
    __vector(i64[2]) neon_uqrshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.uqrshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.uqrshl.i64")
    i64 neon_uqrshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.sqshlu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.v16i8")
    __vector(i8[16]) neon_sqshlu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sqshlu.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.i8")
    i8 neon_sqshlu_i8(i8, i8);
    /// The `llvm.aarch64.neon.sqshlu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.v8i16")
    __vector(i16[8]) neon_sqshlu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sqshlu.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.i16")
    i16 neon_sqshlu_i16(i16, i16);
    /// The `llvm.aarch64.neon.sqshlu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.v4i32")
    __vector(i32[4]) neon_sqshlu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sqshlu.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.i32")
    i32 neon_sqshlu_i32(i32, i32);
    /// The `llvm.aarch64.neon.sqshlu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.v2i64")
    __vector(i64[2]) neon_sqshlu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sqshlu.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqshlu.i64")
    i64 neon_sqshlu_i64(i64, i64);
    /// The `llvm.aarch64.neon.sshl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.v16i8")
    __vector(i8[16]) neon_sshl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.sshl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.i8")
    i8 neon_sshl_i8(i8, i8);
    /// The `llvm.aarch64.neon.sshl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.v8i16")
    __vector(i16[8]) neon_sshl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.sshl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.i16")
    i16 neon_sshl_i16(i16, i16);
    /// The `llvm.aarch64.neon.sshl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.v4i32")
    __vector(i32[4]) neon_sshl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.sshl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.i32")
    i32 neon_sshl_i32(i32, i32);
    /// The `llvm.aarch64.neon.sshl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.v2i64")
    __vector(i64[2]) neon_sshl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.sshl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sshl.i64")
    i64 neon_sshl_i64(i64, i64);
    /// The `llvm.aarch64.neon.ushl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.v16i8")
    __vector(i8[16]) neon_ushl_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.ushl.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.i8")
    i8 neon_ushl_i8(i8, i8);
    /// The `llvm.aarch64.neon.ushl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.v8i16")
    __vector(i16[8]) neon_ushl_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.ushl.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.i16")
    i16 neon_ushl_i16(i16, i16);
    /// The `llvm.aarch64.neon.ushl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.v4i32")
    __vector(i32[4]) neon_ushl_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.ushl.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.i32")
    i32 neon_ushl_i32(i32, i32);
    /// The `llvm.aarch64.neon.ushl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.v2i64")
    __vector(i64[2]) neon_ushl_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.ushl.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ushl.i64")
    i64 neon_ushl_i64(i64, i64);
    /// The `llvm.aarch64.neon.vsri.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsri.v16i8")
    __vector(i8[16]) neon_vsri_v16i8(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vsri.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsri.v8i16")
    __vector(i16[8]) neon_vsri_v8i16(__vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vsri.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsri.v4i32")
    __vector(i32[4]) neon_vsri_v4i32(__vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vsri.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsri.v2i64")
    __vector(i64[2]) neon_vsri_v2i64(__vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vsri.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsri.v4f32")
    __vector(f32[4]) neon_vsri_v4f32(__vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vsri.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsri.v2f64")
    __vector(f64[2]) neon_vsri_v2f64(__vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vsli.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsli.v16i8")
    __vector(i8[16]) neon_vsli_v16i8(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vsli.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsli.v8i16")
    __vector(i16[8]) neon_vsli_v8i16(__vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vsli.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsli.v4i32")
    __vector(i32[4]) neon_vsli_v4i32(__vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vsli.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsli.v2i64")
    __vector(i64[2]) neon_vsli_v2i64(__vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vsli.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsli.v4f32")
    __vector(f32[4]) neon_vsli_v4f32(__vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vsli.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vsli.v2f64")
    __vector(f64[2]) neon_vsli_v2f64(__vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.v16i8")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.i8")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.v8i16")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.i16")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.v4i32")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.i32")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.v2i64")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v16i8.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v16i8.i64")
    __vector(i8[16]) neon_scalar_sqxtn_v16i8_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.v16i8")
    i8 neon_scalar_sqxtn_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.i8")
    i8 neon_scalar_sqxtn_i8_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.v8i16")
    i8 neon_scalar_sqxtn_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.i16")
    i8 neon_scalar_sqxtn_i8_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.v4i32")
    i8 neon_scalar_sqxtn_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.i32")
    i8 neon_scalar_sqxtn_i8_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.v2i64")
    i8 neon_scalar_sqxtn_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i8.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i8.i64")
    i8 neon_scalar_sqxtn_i8_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.v16i8")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.i8")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.v8i16")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.i16")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.v4i32")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.i32")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.v2i64")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v8i16.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v8i16.i64")
    __vector(i16[8]) neon_scalar_sqxtn_v8i16_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.v16i8")
    i16 neon_scalar_sqxtn_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.i8")
    i16 neon_scalar_sqxtn_i16_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.v8i16")
    i16 neon_scalar_sqxtn_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.i16")
    i16 neon_scalar_sqxtn_i16_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.v4i32")
    i16 neon_scalar_sqxtn_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.i32")
    i16 neon_scalar_sqxtn_i16_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.v2i64")
    i16 neon_scalar_sqxtn_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i16.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i16.i64")
    i16 neon_scalar_sqxtn_i16_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.v16i8")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.i8")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.v8i16")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.i16")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.v4i32")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.i32")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.v2i64")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v4i32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v4i32.i64")
    __vector(i32[4]) neon_scalar_sqxtn_v4i32_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.v16i8")
    i32 neon_scalar_sqxtn_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.i8")
    i32 neon_scalar_sqxtn_i32_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.v8i16")
    i32 neon_scalar_sqxtn_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.i16")
    i32 neon_scalar_sqxtn_i32_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.v4i32")
    i32 neon_scalar_sqxtn_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.i32")
    i32 neon_scalar_sqxtn_i32_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.v2i64")
    i32 neon_scalar_sqxtn_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i32.i64")
    i32 neon_scalar_sqxtn_i32_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.v16i8")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.i8")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.v8i16")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.i16")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.v4i32")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.i32")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.v2i64")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.v2i64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.v2i64.i64")
    __vector(i64[2]) neon_scalar_sqxtn_v2i64_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.v16i8")
    i64 neon_scalar_sqxtn_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.i8")
    i64 neon_scalar_sqxtn_i64_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.v8i16")
    i64 neon_scalar_sqxtn_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.i16")
    i64 neon_scalar_sqxtn_i64_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.v4i32")
    i64 neon_scalar_sqxtn_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.i32")
    i64 neon_scalar_sqxtn_i64_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.v2i64")
    i64 neon_scalar_sqxtn_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtn.i64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtn.i64.i64")
    i64 neon_scalar_sqxtn_i64_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.v16i8")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.i8")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.v8i16")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.i16")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.v4i32")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.i32")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.v2i64")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v16i8.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v16i8.i64")
    __vector(i8[16]) neon_scalar_uqxtn_v16i8_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.v16i8")
    i8 neon_scalar_uqxtn_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.i8")
    i8 neon_scalar_uqxtn_i8_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.v8i16")
    i8 neon_scalar_uqxtn_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.i16")
    i8 neon_scalar_uqxtn_i8_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.v4i32")
    i8 neon_scalar_uqxtn_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.i32")
    i8 neon_scalar_uqxtn_i8_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.v2i64")
    i8 neon_scalar_uqxtn_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i8.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i8.i64")
    i8 neon_scalar_uqxtn_i8_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.v16i8")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.i8")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.v8i16")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.i16")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.v4i32")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.i32")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.v2i64")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v8i16.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v8i16.i64")
    __vector(i16[8]) neon_scalar_uqxtn_v8i16_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.v16i8")
    i16 neon_scalar_uqxtn_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.i8")
    i16 neon_scalar_uqxtn_i16_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.v8i16")
    i16 neon_scalar_uqxtn_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.i16")
    i16 neon_scalar_uqxtn_i16_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.v4i32")
    i16 neon_scalar_uqxtn_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.i32")
    i16 neon_scalar_uqxtn_i16_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.v2i64")
    i16 neon_scalar_uqxtn_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i16.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i16.i64")
    i16 neon_scalar_uqxtn_i16_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.v16i8")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.i8")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.v8i16")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.i16")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.v4i32")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.i32")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.v2i64")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v4i32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v4i32.i64")
    __vector(i32[4]) neon_scalar_uqxtn_v4i32_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.v16i8")
    i32 neon_scalar_uqxtn_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.i8")
    i32 neon_scalar_uqxtn_i32_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.v8i16")
    i32 neon_scalar_uqxtn_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.i16")
    i32 neon_scalar_uqxtn_i32_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.v4i32")
    i32 neon_scalar_uqxtn_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.i32")
    i32 neon_scalar_uqxtn_i32_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.v2i64")
    i32 neon_scalar_uqxtn_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i32.i64")
    i32 neon_scalar_uqxtn_i32_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.v16i8")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.i8")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.v8i16")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.i16")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.v4i32")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.i32")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.v2i64")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.v2i64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.v2i64.i64")
    __vector(i64[2]) neon_scalar_uqxtn_v2i64_i64(i64);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.v16i8")
    i64 neon_scalar_uqxtn_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.i8")
    i64 neon_scalar_uqxtn_i64_i8(i8);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.v8i16")
    i64 neon_scalar_uqxtn_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.i16")
    i64 neon_scalar_uqxtn_i64_i16(i16);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.v4i32")
    i64 neon_scalar_uqxtn_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.i32")
    i64 neon_scalar_uqxtn_i64_i32(i32);
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.v2i64")
    i64 neon_scalar_uqxtn_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.uqxtn.i64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.uqxtn.i64.i64")
    i64 neon_scalar_uqxtn_i64_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.v16i8")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.i8")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.v8i16")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.i16")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.v4i32")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.i32")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.v2i64")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v16i8.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v16i8.i64")
    __vector(i8[16]) neon_scalar_sqxtun_v16i8_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.v16i8")
    i8 neon_scalar_sqxtun_i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.i8")
    i8 neon_scalar_sqxtun_i8_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.v8i16")
    i8 neon_scalar_sqxtun_i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.i16")
    i8 neon_scalar_sqxtun_i8_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.v4i32")
    i8 neon_scalar_sqxtun_i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.i32")
    i8 neon_scalar_sqxtun_i8_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.v2i64")
    i8 neon_scalar_sqxtun_i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i8.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i8.i64")
    i8 neon_scalar_sqxtun_i8_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.v16i8")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.i8")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.v8i16")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.i16")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.v4i32")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.i32")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.v2i64")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v8i16.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v8i16.i64")
    __vector(i16[8]) neon_scalar_sqxtun_v8i16_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.v16i8")
    i16 neon_scalar_sqxtun_i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.i8")
    i16 neon_scalar_sqxtun_i16_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.v8i16")
    i16 neon_scalar_sqxtun_i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.i16")
    i16 neon_scalar_sqxtun_i16_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.v4i32")
    i16 neon_scalar_sqxtun_i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.i32")
    i16 neon_scalar_sqxtun_i16_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.v2i64")
    i16 neon_scalar_sqxtun_i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i16.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i16.i64")
    i16 neon_scalar_sqxtun_i16_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.v16i8")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.i8")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.v8i16")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.i16")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.v4i32")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.i32")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.v2i64")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v4i32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v4i32.i64")
    __vector(i32[4]) neon_scalar_sqxtun_v4i32_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.v16i8")
    i32 neon_scalar_sqxtun_i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.i8")
    i32 neon_scalar_sqxtun_i32_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.v8i16")
    i32 neon_scalar_sqxtun_i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.i16")
    i32 neon_scalar_sqxtun_i32_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.v4i32")
    i32 neon_scalar_sqxtun_i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.i32")
    i32 neon_scalar_sqxtun_i32_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.v2i64")
    i32 neon_scalar_sqxtun_i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i32.i64")
    i32 neon_scalar_sqxtun_i32_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.v16i8")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.i8")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.v8i16")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.i16")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.v4i32")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.i32")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.v2i64")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.v2i64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.v2i64.i64")
    __vector(i64[2]) neon_scalar_sqxtun_v2i64_i64(i64);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.v16i8")
    i64 neon_scalar_sqxtun_i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.i8")
    i64 neon_scalar_sqxtun_i64_i8(i8);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.v8i16")
    i64 neon_scalar_sqxtun_i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.i16")
    i64 neon_scalar_sqxtun_i64_i16(i16);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.v4i32")
    i64 neon_scalar_sqxtun_i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.i32")
    i64 neon_scalar_sqxtun_i64_i32(i32);
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.v2i64")
    i64 neon_scalar_sqxtun_i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.scalar.sqxtun.i64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.scalar.sqxtun.i64.i64")
    i64 neon_scalar_sqxtun_i64_i64(i64);
    /// The `llvm.aarch64.neon.abs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.v16i8")
    __vector(i8[16]) neon_abs_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.abs.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.i8")
    i8 neon_abs_i8(i8);
    /// The `llvm.aarch64.neon.abs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.v8i16")
    __vector(i16[8]) neon_abs_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.abs.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.i16")
    i16 neon_abs_i16(i16);
    /// The `llvm.aarch64.neon.abs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.v4i32")
    __vector(i32[4]) neon_abs_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.abs.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.i32")
    i32 neon_abs_i32(i32);
    /// The `llvm.aarch64.neon.abs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.v2i64")
    __vector(i64[2]) neon_abs_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.abs.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.abs.i64")
    i64 neon_abs_i64(i64);
    /// The `llvm.aarch64.neon.sqabs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.v16i8")
    __vector(i8[16]) neon_sqabs_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sqabs.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.i8")
    i8 neon_sqabs_i8(i8);
    /// The `llvm.aarch64.neon.sqabs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.v8i16")
    __vector(i16[8]) neon_sqabs_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sqabs.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.i16")
    i16 neon_sqabs_i16(i16);
    /// The `llvm.aarch64.neon.sqabs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.v4i32")
    __vector(i32[4]) neon_sqabs_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sqabs.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.i32")
    i32 neon_sqabs_i32(i32);
    /// The `llvm.aarch64.neon.sqabs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.v2i64")
    __vector(i64[2]) neon_sqabs_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sqabs.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqabs.i64")
    i64 neon_sqabs_i64(i64);
    /// The `llvm.aarch64.neon.sqneg.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.v16i8")
    __vector(i8[16]) neon_sqneg_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.sqneg.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.i8")
    i8 neon_sqneg_i8(i8);
    /// The `llvm.aarch64.neon.sqneg.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.v8i16")
    __vector(i16[8]) neon_sqneg_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.sqneg.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.i16")
    i16 neon_sqneg_i16(i16);
    /// The `llvm.aarch64.neon.sqneg.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.v4i32")
    __vector(i32[4]) neon_sqneg_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.sqneg.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.i32")
    i32 neon_sqneg_i32(i32);
    /// The `llvm.aarch64.neon.sqneg.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.v2i64")
    __vector(i64[2]) neon_sqneg_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.sqneg.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.sqneg.i64")
    i64 neon_sqneg_i64(i64);
    /// The `llvm.aarch64.neon.cls.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.cls.v16i8")
    __vector(i8[16]) neon_cls_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.cls.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.cls.v8i16")
    __vector(i16[8]) neon_cls_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.cls.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.cls.v4i32")
    __vector(i32[4]) neon_cls_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.cls.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.cls.v2i64")
    __vector(i64[2]) neon_cls_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.cls.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.cls.v4f32")
    __vector(f32[4]) neon_cls_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.cls.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.cls.v2f64")
    __vector(f64[2]) neon_cls_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.urecpe.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urecpe.v16i8")
    __vector(i8[16]) neon_urecpe_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.urecpe.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urecpe.v8i16")
    __vector(i16[8]) neon_urecpe_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.urecpe.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urecpe.v4i32")
    __vector(i32[4]) neon_urecpe_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.urecpe.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urecpe.v2i64")
    __vector(i64[2]) neon_urecpe_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.urecpe.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urecpe.v4f32")
    __vector(f32[4]) neon_urecpe_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.urecpe.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.urecpe.v2f64")
    __vector(f64[2]) neon_urecpe_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.frecpe.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpe.v4f32")
    __vector(f32[4]) neon_frecpe_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.frecpe.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpe.f32")
    f32 neon_frecpe_f32(f32);
    /// The `llvm.aarch64.neon.frecpe.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpe.v2f64")
    __vector(f64[2]) neon_frecpe_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.frecpe.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frecpe.f64")
    f64 neon_frecpe_f64(f64);
    /// The `llvm.aarch64.neon.ursqrte.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ursqrte.v16i8")
    __vector(i8[16]) neon_ursqrte_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.ursqrte.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ursqrte.v8i16")
    __vector(i16[8]) neon_ursqrte_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.ursqrte.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ursqrte.v4i32")
    __vector(i32[4]) neon_ursqrte_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.ursqrte.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ursqrte.v2i64")
    __vector(i64[2]) neon_ursqrte_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.ursqrte.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ursqrte.v4f32")
    __vector(f32[4]) neon_ursqrte_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.ursqrte.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.ursqrte.v2f64")
    __vector(f64[2]) neon_ursqrte_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.frsqrte.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrte.v4f32")
    __vector(f32[4]) neon_frsqrte_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.frsqrte.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrte.f32")
    f32 neon_frsqrte_f32(f32);
    /// The `llvm.aarch64.neon.frsqrte.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrte.v2f64")
    __vector(f64[2]) neon_frsqrte_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.frsqrte.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frsqrte.f64")
    f64 neon_frsqrte_f64(f64);
    /// The `llvm.aarch64.neon.rbit.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.rbit.v16i8")
    __vector(i8[16]) neon_rbit_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.rbit.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.rbit.v8i16")
    __vector(i16[8]) neon_rbit_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.rbit.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.rbit.v4i32")
    __vector(i32[4]) neon_rbit_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.rbit.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.rbit.v2i64")
    __vector(i64[2]) neon_rbit_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.rbit.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.rbit.v4f32")
    __vector(f32[4]) neon_rbit_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.rbit.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.rbit.v2f64")
    __vector(f64[2]) neon_rbit_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.vcvtfp2hf` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2hf")
    __vector(i16[4]) neon_vcvtfp2hf(__vector(f32[4]));
    /// The `llvm.aarch64.neon.vcvthf2fp` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvthf2fp")
    __vector(f32[4]) neon_vcvthf2fp(__vector(i16[4]));
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v16i8.f32")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v16i8.f64")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i8.v4f32")
    i8 neon_vcvtfp2fxs_i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i8.f32")
    i8 neon_vcvtfp2fxs_i8_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i8.v2f64")
    i8 neon_vcvtfp2fxs_i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i8.f64")
    i8 neon_vcvtfp2fxs_i8_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v8i16.f32")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v8i16.f64")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i16.v4f32")
    i16 neon_vcvtfp2fxs_i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i16.f32")
    i16 neon_vcvtfp2fxs_i16_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i16.v2f64")
    i16 neon_vcvtfp2fxs_i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i16.f64")
    i16 neon_vcvtfp2fxs_i16_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v4i32.f32")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v4i32.f64")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i32.v4f32")
    i32 neon_vcvtfp2fxs_i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i32.f32")
    i32 neon_vcvtfp2fxs_i32_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i32.v2f64")
    i32 neon_vcvtfp2fxs_i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i32.f64")
    i32 neon_vcvtfp2fxs_i32_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v2i64.f32")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.v2i64.f64")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i64.v4f32")
    i64 neon_vcvtfp2fxs_i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i64.f32")
    i64 neon_vcvtfp2fxs_i64_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i64.v2f64")
    i64 neon_vcvtfp2fxs_i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxs.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxs.i64.f64")
    i64 neon_vcvtfp2fxs_i64_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v16i8.f32")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v16i8.f64")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i8.v4f32")
    i8 neon_vcvtfp2fxu_i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i8.f32")
    i8 neon_vcvtfp2fxu_i8_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i8.v2f64")
    i8 neon_vcvtfp2fxu_i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i8.f64")
    i8 neon_vcvtfp2fxu_i8_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v8i16.f32")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v8i16.f64")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i16.v4f32")
    i16 neon_vcvtfp2fxu_i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i16.f32")
    i16 neon_vcvtfp2fxu_i16_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i16.v2f64")
    i16 neon_vcvtfp2fxu_i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i16.f64")
    i16 neon_vcvtfp2fxu_i16_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v4i32.f32")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v4i32.f64")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i32.v4f32")
    i32 neon_vcvtfp2fxu_i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i32.f32")
    i32 neon_vcvtfp2fxu_i32_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i32.v2f64")
    i32 neon_vcvtfp2fxu_i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i32.f64")
    i32 neon_vcvtfp2fxu_i32_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v2i64.f32")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.v2i64.f64")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i64.v4f32")
    i64 neon_vcvtfp2fxu_i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i64.f32")
    i64 neon_vcvtfp2fxu_i64_f32(f32, i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i64.v2f64")
    i64 neon_vcvtfp2fxu_i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfp2fxu.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfp2fxu.i64.f64")
    i64 neon_vcvtfp2fxu_i64_f64(f64, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.i8")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.i16")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.i32")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v4f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v4f32.i64")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.v16i8")
    f32 neon_vcvtfxs2fp_f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.i8")
    f32 neon_vcvtfxs2fp_f32_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.v8i16")
    f32 neon_vcvtfxs2fp_f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.i16")
    f32 neon_vcvtfxs2fp_f32_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.v4i32")
    f32 neon_vcvtfxs2fp_f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.i32")
    f32 neon_vcvtfxs2fp_f32_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.v2i64")
    f32 neon_vcvtfxs2fp_f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f32.i64")
    f32 neon_vcvtfxs2fp_f32_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.i8")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.i16")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.i32")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.v2f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.v2f64.i64")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.v16i8")
    f64 neon_vcvtfxs2fp_f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.i8")
    f64 neon_vcvtfxs2fp_f64_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.v8i16")
    f64 neon_vcvtfxs2fp_f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.i16")
    f64 neon_vcvtfxs2fp_f64_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.v4i32")
    f64 neon_vcvtfxs2fp_f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.i32")
    f64 neon_vcvtfxs2fp_f64_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.v2i64")
    f64 neon_vcvtfxs2fp_f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxs2fp.f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxs2fp.f64.i64")
    f64 neon_vcvtfxs2fp_f64_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.i8")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.i16")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.i32")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v4f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v4f32.i64")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.v16i8")
    f32 neon_vcvtfxu2fp_f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.i8")
    f32 neon_vcvtfxu2fp_f32_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.v8i16")
    f32 neon_vcvtfxu2fp_f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.i16")
    f32 neon_vcvtfxu2fp_f32_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.v4i32")
    f32 neon_vcvtfxu2fp_f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.i32")
    f32 neon_vcvtfxu2fp_f32_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.v2i64")
    f32 neon_vcvtfxu2fp_f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f32.i64")
    f32 neon_vcvtfxu2fp_f32_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.i8")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.i16")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.i32")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.v2f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.v2f64.i64")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i64(i64, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.v16i8")
    f64 neon_vcvtfxu2fp_f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.i8")
    f64 neon_vcvtfxu2fp_f64_i8(i8, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.v8i16")
    f64 neon_vcvtfxu2fp_f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.i16")
    f64 neon_vcvtfxu2fp_f64_i16(i16, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.v4i32")
    f64 neon_vcvtfxu2fp_f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.i32")
    f64 neon_vcvtfxu2fp_f64_i32(i32, i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.v2i64")
    f64 neon_vcvtfxu2fp_f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.aarch64.neon.vcvtfxu2fp.f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcvtfxu2fp.f64.i64")
    f64 neon_vcvtfxu2fp_f64_i64(i64, i32);
    /// The `llvm.aarch64.neon.fcvtas.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtas_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v16i8.f32")
    __vector(i8[16]) neon_fcvtas_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtas_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v16i8.f64")
    __vector(i8[16]) neon_fcvtas_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i8.v4f32")
    i8 neon_fcvtas_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i8.f32")
    i8 neon_fcvtas_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i8.v2f64")
    i8 neon_fcvtas_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i8.f64")
    i8 neon_fcvtas_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtas_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v8i16.f32")
    __vector(i16[8]) neon_fcvtas_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtas_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v8i16.f64")
    __vector(i16[8]) neon_fcvtas_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i16.v4f32")
    i16 neon_fcvtas_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i16.f32")
    i16 neon_fcvtas_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i16.v2f64")
    i16 neon_fcvtas_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i16.f64")
    i16 neon_fcvtas_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtas_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v4i32.f32")
    __vector(i32[4]) neon_fcvtas_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtas_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v4i32.f64")
    __vector(i32[4]) neon_fcvtas_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i32.v4f32")
    i32 neon_fcvtas_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i32.f32")
    i32 neon_fcvtas_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i32.v2f64")
    i32 neon_fcvtas_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i32.f64")
    i32 neon_fcvtas_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtas_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v2i64.f32")
    __vector(i64[2]) neon_fcvtas_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtas_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.v2i64.f64")
    __vector(i64[2]) neon_fcvtas_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtas.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i64.v4f32")
    i64 neon_fcvtas_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtas.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i64.f32")
    i64 neon_fcvtas_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtas.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i64.v2f64")
    i64 neon_fcvtas_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtas.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtas.i64.f64")
    i64 neon_fcvtas_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtau_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v16i8.f32")
    __vector(i8[16]) neon_fcvtau_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtau_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v16i8.f64")
    __vector(i8[16]) neon_fcvtau_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i8.v4f32")
    i8 neon_fcvtau_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i8.f32")
    i8 neon_fcvtau_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i8.v2f64")
    i8 neon_fcvtau_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i8.f64")
    i8 neon_fcvtau_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtau_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v8i16.f32")
    __vector(i16[8]) neon_fcvtau_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtau_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v8i16.f64")
    __vector(i16[8]) neon_fcvtau_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i16.v4f32")
    i16 neon_fcvtau_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i16.f32")
    i16 neon_fcvtau_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i16.v2f64")
    i16 neon_fcvtau_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i16.f64")
    i16 neon_fcvtau_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtau_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v4i32.f32")
    __vector(i32[4]) neon_fcvtau_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtau_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v4i32.f64")
    __vector(i32[4]) neon_fcvtau_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i32.v4f32")
    i32 neon_fcvtau_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i32.f32")
    i32 neon_fcvtau_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i32.v2f64")
    i32 neon_fcvtau_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i32.f64")
    i32 neon_fcvtau_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtau_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v2i64.f32")
    __vector(i64[2]) neon_fcvtau_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtau_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.v2i64.f64")
    __vector(i64[2]) neon_fcvtau_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtau.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i64.v4f32")
    i64 neon_fcvtau_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtau.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i64.f32")
    i64 neon_fcvtau_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtau.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i64.v2f64")
    i64 neon_fcvtau_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtau.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtau.i64.f64")
    i64 neon_fcvtau_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtms_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v16i8.f32")
    __vector(i8[16]) neon_fcvtms_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtms_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v16i8.f64")
    __vector(i8[16]) neon_fcvtms_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i8.v4f32")
    i8 neon_fcvtms_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i8.f32")
    i8 neon_fcvtms_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i8.v2f64")
    i8 neon_fcvtms_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i8.f64")
    i8 neon_fcvtms_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtms_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v8i16.f32")
    __vector(i16[8]) neon_fcvtms_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtms_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v8i16.f64")
    __vector(i16[8]) neon_fcvtms_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i16.v4f32")
    i16 neon_fcvtms_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i16.f32")
    i16 neon_fcvtms_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i16.v2f64")
    i16 neon_fcvtms_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i16.f64")
    i16 neon_fcvtms_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtms_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v4i32.f32")
    __vector(i32[4]) neon_fcvtms_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtms_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v4i32.f64")
    __vector(i32[4]) neon_fcvtms_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i32.v4f32")
    i32 neon_fcvtms_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i32.f32")
    i32 neon_fcvtms_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i32.v2f64")
    i32 neon_fcvtms_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i32.f64")
    i32 neon_fcvtms_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtms_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v2i64.f32")
    __vector(i64[2]) neon_fcvtms_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtms_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.v2i64.f64")
    __vector(i64[2]) neon_fcvtms_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtms.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i64.v4f32")
    i64 neon_fcvtms_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtms.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i64.f32")
    i64 neon_fcvtms_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtms.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i64.v2f64")
    i64 neon_fcvtms_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtms.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtms.i64.f64")
    i64 neon_fcvtms_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtmu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v16i8.f32")
    __vector(i8[16]) neon_fcvtmu_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtmu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v16i8.f64")
    __vector(i8[16]) neon_fcvtmu_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i8.v4f32")
    i8 neon_fcvtmu_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i8.f32")
    i8 neon_fcvtmu_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i8.v2f64")
    i8 neon_fcvtmu_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i8.f64")
    i8 neon_fcvtmu_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtmu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v8i16.f32")
    __vector(i16[8]) neon_fcvtmu_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtmu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v8i16.f64")
    __vector(i16[8]) neon_fcvtmu_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i16.v4f32")
    i16 neon_fcvtmu_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i16.f32")
    i16 neon_fcvtmu_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i16.v2f64")
    i16 neon_fcvtmu_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i16.f64")
    i16 neon_fcvtmu_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtmu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v4i32.f32")
    __vector(i32[4]) neon_fcvtmu_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtmu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v4i32.f64")
    __vector(i32[4]) neon_fcvtmu_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i32.v4f32")
    i32 neon_fcvtmu_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i32.f32")
    i32 neon_fcvtmu_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i32.v2f64")
    i32 neon_fcvtmu_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i32.f64")
    i32 neon_fcvtmu_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtmu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v2i64.f32")
    __vector(i64[2]) neon_fcvtmu_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtmu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.v2i64.f64")
    __vector(i64[2]) neon_fcvtmu_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtmu.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i64.v4f32")
    i64 neon_fcvtmu_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtmu.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i64.f32")
    i64 neon_fcvtmu_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtmu.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i64.v2f64")
    i64 neon_fcvtmu_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtmu.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtmu.i64.f64")
    i64 neon_fcvtmu_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtns_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v16i8.f32")
    __vector(i8[16]) neon_fcvtns_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtns_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v16i8.f64")
    __vector(i8[16]) neon_fcvtns_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i8.v4f32")
    i8 neon_fcvtns_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i8.f32")
    i8 neon_fcvtns_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i8.v2f64")
    i8 neon_fcvtns_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i8.f64")
    i8 neon_fcvtns_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtns_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v8i16.f32")
    __vector(i16[8]) neon_fcvtns_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtns_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v8i16.f64")
    __vector(i16[8]) neon_fcvtns_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i16.v4f32")
    i16 neon_fcvtns_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i16.f32")
    i16 neon_fcvtns_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i16.v2f64")
    i16 neon_fcvtns_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i16.f64")
    i16 neon_fcvtns_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtns_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v4i32.f32")
    __vector(i32[4]) neon_fcvtns_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtns_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v4i32.f64")
    __vector(i32[4]) neon_fcvtns_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i32.v4f32")
    i32 neon_fcvtns_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i32.f32")
    i32 neon_fcvtns_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i32.v2f64")
    i32 neon_fcvtns_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i32.f64")
    i32 neon_fcvtns_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtns_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v2i64.f32")
    __vector(i64[2]) neon_fcvtns_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtns_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.v2i64.f64")
    __vector(i64[2]) neon_fcvtns_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtns.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i64.v4f32")
    i64 neon_fcvtns_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtns.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i64.f32")
    i64 neon_fcvtns_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtns.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i64.v2f64")
    i64 neon_fcvtns_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtns.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtns.i64.f64")
    i64 neon_fcvtns_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtnu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v16i8.f32")
    __vector(i8[16]) neon_fcvtnu_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtnu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v16i8.f64")
    __vector(i8[16]) neon_fcvtnu_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i8.v4f32")
    i8 neon_fcvtnu_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i8.f32")
    i8 neon_fcvtnu_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i8.v2f64")
    i8 neon_fcvtnu_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i8.f64")
    i8 neon_fcvtnu_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtnu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v8i16.f32")
    __vector(i16[8]) neon_fcvtnu_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtnu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v8i16.f64")
    __vector(i16[8]) neon_fcvtnu_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i16.v4f32")
    i16 neon_fcvtnu_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i16.f32")
    i16 neon_fcvtnu_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i16.v2f64")
    i16 neon_fcvtnu_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i16.f64")
    i16 neon_fcvtnu_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtnu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v4i32.f32")
    __vector(i32[4]) neon_fcvtnu_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtnu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v4i32.f64")
    __vector(i32[4]) neon_fcvtnu_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i32.v4f32")
    i32 neon_fcvtnu_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i32.f32")
    i32 neon_fcvtnu_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i32.v2f64")
    i32 neon_fcvtnu_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i32.f64")
    i32 neon_fcvtnu_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtnu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v2i64.f32")
    __vector(i64[2]) neon_fcvtnu_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtnu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.v2i64.f64")
    __vector(i64[2]) neon_fcvtnu_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtnu.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i64.v4f32")
    i64 neon_fcvtnu_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtnu.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i64.f32")
    i64 neon_fcvtnu_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtnu.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i64.v2f64")
    i64 neon_fcvtnu_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtnu.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtnu.i64.f64")
    i64 neon_fcvtnu_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtps_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v16i8.f32")
    __vector(i8[16]) neon_fcvtps_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtps_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v16i8.f64")
    __vector(i8[16]) neon_fcvtps_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i8.v4f32")
    i8 neon_fcvtps_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i8.f32")
    i8 neon_fcvtps_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i8.v2f64")
    i8 neon_fcvtps_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i8.f64")
    i8 neon_fcvtps_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtps_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v8i16.f32")
    __vector(i16[8]) neon_fcvtps_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtps_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v8i16.f64")
    __vector(i16[8]) neon_fcvtps_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i16.v4f32")
    i16 neon_fcvtps_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i16.f32")
    i16 neon_fcvtps_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i16.v2f64")
    i16 neon_fcvtps_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i16.f64")
    i16 neon_fcvtps_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtps_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v4i32.f32")
    __vector(i32[4]) neon_fcvtps_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtps_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v4i32.f64")
    __vector(i32[4]) neon_fcvtps_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i32.v4f32")
    i32 neon_fcvtps_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i32.f32")
    i32 neon_fcvtps_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i32.v2f64")
    i32 neon_fcvtps_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i32.f64")
    i32 neon_fcvtps_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtps_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v2i64.f32")
    __vector(i64[2]) neon_fcvtps_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtps_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.v2i64.f64")
    __vector(i64[2]) neon_fcvtps_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtps.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i64.v4f32")
    i64 neon_fcvtps_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtps.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i64.f32")
    i64 neon_fcvtps_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtps.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i64.v2f64")
    i64 neon_fcvtps_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtps.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtps.i64.f64")
    i64 neon_fcvtps_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtpu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v16i8.f32")
    __vector(i8[16]) neon_fcvtpu_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtpu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v16i8.f64")
    __vector(i8[16]) neon_fcvtpu_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i8.v4f32")
    i8 neon_fcvtpu_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i8.f32")
    i8 neon_fcvtpu_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i8.v2f64")
    i8 neon_fcvtpu_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i8.f64")
    i8 neon_fcvtpu_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtpu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v8i16.f32")
    __vector(i16[8]) neon_fcvtpu_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtpu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v8i16.f64")
    __vector(i16[8]) neon_fcvtpu_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i16.v4f32")
    i16 neon_fcvtpu_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i16.f32")
    i16 neon_fcvtpu_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i16.v2f64")
    i16 neon_fcvtpu_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i16.f64")
    i16 neon_fcvtpu_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtpu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v4i32.f32")
    __vector(i32[4]) neon_fcvtpu_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtpu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v4i32.f64")
    __vector(i32[4]) neon_fcvtpu_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i32.v4f32")
    i32 neon_fcvtpu_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i32.f32")
    i32 neon_fcvtpu_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i32.v2f64")
    i32 neon_fcvtpu_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i32.f64")
    i32 neon_fcvtpu_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtpu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v2i64.f32")
    __vector(i64[2]) neon_fcvtpu_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtpu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.v2i64.f64")
    __vector(i64[2]) neon_fcvtpu_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtpu.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i64.v4f32")
    i64 neon_fcvtpu_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtpu.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i64.f32")
    i64 neon_fcvtpu_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtpu.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i64.v2f64")
    i64 neon_fcvtpu_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtpu.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtpu.i64.f64")
    i64 neon_fcvtpu_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtzs_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v16i8.f32")
    __vector(i8[16]) neon_fcvtzs_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtzs_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v16i8.f64")
    __vector(i8[16]) neon_fcvtzs_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i8.v4f32")
    i8 neon_fcvtzs_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i8.f32")
    i8 neon_fcvtzs_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i8.v2f64")
    i8 neon_fcvtzs_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i8.f64")
    i8 neon_fcvtzs_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtzs_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v8i16.f32")
    __vector(i16[8]) neon_fcvtzs_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtzs_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v8i16.f64")
    __vector(i16[8]) neon_fcvtzs_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i16.v4f32")
    i16 neon_fcvtzs_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i16.f32")
    i16 neon_fcvtzs_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i16.v2f64")
    i16 neon_fcvtzs_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i16.f64")
    i16 neon_fcvtzs_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtzs_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v4i32.f32")
    __vector(i32[4]) neon_fcvtzs_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtzs_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v4i32.f64")
    __vector(i32[4]) neon_fcvtzs_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i32.v4f32")
    i32 neon_fcvtzs_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i32.f32")
    i32 neon_fcvtzs_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i32.v2f64")
    i32 neon_fcvtzs_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i32.f64")
    i32 neon_fcvtzs_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtzs_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v2i64.f32")
    __vector(i64[2]) neon_fcvtzs_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtzs_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.v2i64.f64")
    __vector(i64[2]) neon_fcvtzs_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzs.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i64.v4f32")
    i64 neon_fcvtzs_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzs.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i64.f32")
    i64 neon_fcvtzs_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzs.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i64.v2f64")
    i64 neon_fcvtzs_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzs.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzs.i64.f64")
    i64 neon_fcvtzs_i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtzu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v16i8.f32")
    __vector(i8[16]) neon_fcvtzu_v16i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtzu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v16i8.f64")
    __vector(i8[16]) neon_fcvtzu_v16i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i8.v4f32")
    i8 neon_fcvtzu_i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i8.f32")
    i8 neon_fcvtzu_i8_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i8.v2f64")
    i8 neon_fcvtzu_i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i8.f64")
    i8 neon_fcvtzu_i8_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtzu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v8i16.f32")
    __vector(i16[8]) neon_fcvtzu_v8i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtzu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v8i16.f64")
    __vector(i16[8]) neon_fcvtzu_v8i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i16.v4f32")
    i16 neon_fcvtzu_i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i16.f32")
    i16 neon_fcvtzu_i16_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i16.v2f64")
    i16 neon_fcvtzu_i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i16.f64")
    i16 neon_fcvtzu_i16_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtzu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v4i32.f32")
    __vector(i32[4]) neon_fcvtzu_v4i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtzu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v4i32.f64")
    __vector(i32[4]) neon_fcvtzu_v4i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i32.v4f32")
    i32 neon_fcvtzu_i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i32.f32")
    i32 neon_fcvtzu_i32_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i32.v2f64")
    i32 neon_fcvtzu_i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i32.f64")
    i32 neon_fcvtzu_i32_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtzu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v2i64.f32")
    __vector(i64[2]) neon_fcvtzu_v2i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtzu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.v2i64.f64")
    __vector(i64[2]) neon_fcvtzu_v2i64_f64(f64);
    /// The `llvm.aarch64.neon.fcvtzu.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i64.v4f32")
    i64 neon_fcvtzu_i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtzu.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i64.f32")
    i64 neon_fcvtzu_i64_f32(f32);
    /// The `llvm.aarch64.neon.fcvtzu.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i64.v2f64")
    i64 neon_fcvtzu_i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtzu.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtzu.i64.f64")
    i64 neon_fcvtzu_i64_f64(f64);
    /// The `llvm.aarch64.neon.frintn.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frintn.v4f32")
    __vector(f32[4]) neon_frintn_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.frintn.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frintn.f32")
    f32 neon_frintn_f32(f32);
    /// The `llvm.aarch64.neon.frintn.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frintn.v2f64")
    __vector(f64[2]) neon_frintn_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.frintn.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.frintn.f64")
    f64 neon_frintn_f64(f64);
    /// The `llvm.aarch64.neon.fcvtxn.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v16i8.v16i8")
    __vector(i8[16]) neon_fcvtxn_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fcvtxn.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v16i8.v8i16")
    __vector(i8[16]) neon_fcvtxn_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fcvtxn.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v16i8.v4i32")
    __vector(i8[16]) neon_fcvtxn_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v16i8.v2i64")
    __vector(i8[16]) neon_fcvtxn_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v16i8.v4f32")
    __vector(i8[16]) neon_fcvtxn_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v16i8.v2f64")
    __vector(i8[16]) neon_fcvtxn_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v8i16.v16i8")
    __vector(i16[8]) neon_fcvtxn_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fcvtxn.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v8i16.v8i16")
    __vector(i16[8]) neon_fcvtxn_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fcvtxn.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v8i16.v4i32")
    __vector(i16[8]) neon_fcvtxn_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v8i16.v2i64")
    __vector(i16[8]) neon_fcvtxn_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v8i16.v4f32")
    __vector(i16[8]) neon_fcvtxn_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v8i16.v2f64")
    __vector(i16[8]) neon_fcvtxn_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4i32.v16i8")
    __vector(i32[4]) neon_fcvtxn_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fcvtxn.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4i32.v8i16")
    __vector(i32[4]) neon_fcvtxn_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fcvtxn.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4i32.v4i32")
    __vector(i32[4]) neon_fcvtxn_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4i32.v2i64")
    __vector(i32[4]) neon_fcvtxn_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4i32.v4f32")
    __vector(i32[4]) neon_fcvtxn_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4i32.v2f64")
    __vector(i32[4]) neon_fcvtxn_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2i64.v16i8")
    __vector(i64[2]) neon_fcvtxn_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fcvtxn.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2i64.v8i16")
    __vector(i64[2]) neon_fcvtxn_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fcvtxn.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2i64.v4i32")
    __vector(i64[2]) neon_fcvtxn_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2i64.v2i64")
    __vector(i64[2]) neon_fcvtxn_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2i64.v4f32")
    __vector(i64[2]) neon_fcvtxn_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2i64.v2f64")
    __vector(i64[2]) neon_fcvtxn_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4f32.v16i8")
    __vector(f32[4]) neon_fcvtxn_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fcvtxn.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4f32.v8i16")
    __vector(f32[4]) neon_fcvtxn_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fcvtxn.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4f32.v4i32")
    __vector(f32[4]) neon_fcvtxn_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4f32.v2i64")
    __vector(f32[4]) neon_fcvtxn_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4f32.v4f32")
    __vector(f32[4]) neon_fcvtxn_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v4f32.v2f64")
    __vector(f32[4]) neon_fcvtxn_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2f64.v16i8")
    __vector(f64[2]) neon_fcvtxn_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.aarch64.neon.fcvtxn.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2f64.v8i16")
    __vector(f64[2]) neon_fcvtxn_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.aarch64.neon.fcvtxn.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2f64.v4i32")
    __vector(f64[2]) neon_fcvtxn_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2f64.v2i64")
    __vector(f64[2]) neon_fcvtxn_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.aarch64.neon.fcvtxn.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2f64.v4f32")
    __vector(f64[2]) neon_fcvtxn_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.aarch64.neon.fcvtxn.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.fcvtxn.v2f64.v2f64")
    __vector(f64[2]) neon_fcvtxn_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.aarch64.sisd.fcvtxn` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.sisd.fcvtxn")
    f32 sisd_fcvtxn(f64);
    /// The `llvm.aarch64.neon.vcopy.lane.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v16i8.v16i8")
    __vector(i8[16]) neon_vcopy_lane_v16i8_v16i8(__vector(i8[16]), i64, __vector(i8[16]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v16i8.v8i16")
    __vector(i8[16]) neon_vcopy_lane_v16i8_v8i16(__vector(i16[8]), i64, __vector(i8[16]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v16i8.v4i32")
    __vector(i8[16]) neon_vcopy_lane_v16i8_v4i32(__vector(i32[4]), i64, __vector(i8[16]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v16i8.v2i64")
    __vector(i8[16]) neon_vcopy_lane_v16i8_v2i64(__vector(i64[2]), i64, __vector(i8[16]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v16i8.v4f32")
    __vector(i8[16]) neon_vcopy_lane_v16i8_v4f32(__vector(f32[4]), i64, __vector(i8[16]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v16i8.v2f64")
    __vector(i8[16]) neon_vcopy_lane_v16i8_v2f64(__vector(f64[2]), i64, __vector(i8[16]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v8i16.v16i8")
    __vector(i16[8]) neon_vcopy_lane_v8i16_v16i8(__vector(i8[16]), i64, __vector(i16[8]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v8i16.v8i16")
    __vector(i16[8]) neon_vcopy_lane_v8i16_v8i16(__vector(i16[8]), i64, __vector(i16[8]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v8i16.v4i32")
    __vector(i16[8]) neon_vcopy_lane_v8i16_v4i32(__vector(i32[4]), i64, __vector(i16[8]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v8i16.v2i64")
    __vector(i16[8]) neon_vcopy_lane_v8i16_v2i64(__vector(i64[2]), i64, __vector(i16[8]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v8i16.v4f32")
    __vector(i16[8]) neon_vcopy_lane_v8i16_v4f32(__vector(f32[4]), i64, __vector(i16[8]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v8i16.v2f64")
    __vector(i16[8]) neon_vcopy_lane_v8i16_v2f64(__vector(f64[2]), i64, __vector(i16[8]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4i32.v16i8")
    __vector(i32[4]) neon_vcopy_lane_v4i32_v16i8(__vector(i8[16]), i64, __vector(i32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4i32.v8i16")
    __vector(i32[4]) neon_vcopy_lane_v4i32_v8i16(__vector(i16[8]), i64, __vector(i32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4i32.v4i32")
    __vector(i32[4]) neon_vcopy_lane_v4i32_v4i32(__vector(i32[4]), i64, __vector(i32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4i32.v2i64")
    __vector(i32[4]) neon_vcopy_lane_v4i32_v2i64(__vector(i64[2]), i64, __vector(i32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4i32.v4f32")
    __vector(i32[4]) neon_vcopy_lane_v4i32_v4f32(__vector(f32[4]), i64, __vector(i32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4i32.v2f64")
    __vector(i32[4]) neon_vcopy_lane_v4i32_v2f64(__vector(f64[2]), i64, __vector(i32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2i64.v16i8")
    __vector(i64[2]) neon_vcopy_lane_v2i64_v16i8(__vector(i8[16]), i64, __vector(i64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2i64.v8i16")
    __vector(i64[2]) neon_vcopy_lane_v2i64_v8i16(__vector(i16[8]), i64, __vector(i64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2i64.v4i32")
    __vector(i64[2]) neon_vcopy_lane_v2i64_v4i32(__vector(i32[4]), i64, __vector(i64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2i64.v2i64")
    __vector(i64[2]) neon_vcopy_lane_v2i64_v2i64(__vector(i64[2]), i64, __vector(i64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2i64.v4f32")
    __vector(i64[2]) neon_vcopy_lane_v2i64_v4f32(__vector(f32[4]), i64, __vector(i64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2i64.v2f64")
    __vector(i64[2]) neon_vcopy_lane_v2i64_v2f64(__vector(f64[2]), i64, __vector(i64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4f32.v16i8")
    __vector(f32[4]) neon_vcopy_lane_v4f32_v16i8(__vector(i8[16]), i64, __vector(f32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4f32.v8i16")
    __vector(f32[4]) neon_vcopy_lane_v4f32_v8i16(__vector(i16[8]), i64, __vector(f32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4f32.v4i32")
    __vector(f32[4]) neon_vcopy_lane_v4f32_v4i32(__vector(i32[4]), i64, __vector(f32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4f32.v2i64")
    __vector(f32[4]) neon_vcopy_lane_v4f32_v2i64(__vector(i64[2]), i64, __vector(f32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4f32.v4f32")
    __vector(f32[4]) neon_vcopy_lane_v4f32_v4f32(__vector(f32[4]), i64, __vector(f32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v4f32.v2f64")
    __vector(f32[4]) neon_vcopy_lane_v4f32_v2f64(__vector(f64[2]), i64, __vector(f32[4]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2f64.v16i8")
    __vector(f64[2]) neon_vcopy_lane_v2f64_v16i8(__vector(i8[16]), i64, __vector(f64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2f64.v8i16")
    __vector(f64[2]) neon_vcopy_lane_v2f64_v8i16(__vector(i16[8]), i64, __vector(f64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2f64.v4i32")
    __vector(f64[2]) neon_vcopy_lane_v2f64_v4i32(__vector(i32[4]), i64, __vector(f64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2f64.v2i64")
    __vector(f64[2]) neon_vcopy_lane_v2f64_v2i64(__vector(i64[2]), i64, __vector(f64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2f64.v4f32")
    __vector(f64[2]) neon_vcopy_lane_v2f64_v4f32(__vector(f32[4]), i64, __vector(f64[2]), i64);
    /// The `llvm.aarch64.neon.vcopy.lane.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.vcopy.lane.v2f64.v2f64")
    __vector(f64[2]) neon_vcopy_lane_v2f64_v2f64(__vector(f64[2]), i64, __vector(f64[2]), i64);
    /// The `llvm.aarch64.neon.st1x2.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x2.v16i8")
    void neon_st1x2_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16])*);
    /// The `llvm.aarch64.neon.st1x2.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x2.v8i16")
    void neon_st1x2_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8])*);
    /// The `llvm.aarch64.neon.st1x2.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x2.v4i32")
    void neon_st1x2_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4])*);
    /// The `llvm.aarch64.neon.st1x2.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x2.v2i64")
    void neon_st1x2_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2])*);
    /// The `llvm.aarch64.neon.st1x2.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x2.v4f32")
    void neon_st1x2_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4])*);
    /// The `llvm.aarch64.neon.st1x2.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x2.v2f64")
    void neon_st1x2_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2])*);
    /// The `llvm.aarch64.neon.st1x3.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x3.v16i8")
    void neon_st1x3_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16])*);
    /// The `llvm.aarch64.neon.st1x3.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x3.v8i16")
    void neon_st1x3_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8])*);
    /// The `llvm.aarch64.neon.st1x3.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x3.v4i32")
    void neon_st1x3_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4])*);
    /// The `llvm.aarch64.neon.st1x3.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x3.v2i64")
    void neon_st1x3_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2])*);
    /// The `llvm.aarch64.neon.st1x3.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x3.v4f32")
    void neon_st1x3_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4])*);
    /// The `llvm.aarch64.neon.st1x3.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x3.v2f64")
    void neon_st1x3_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2])*);
    /// The `llvm.aarch64.neon.st1x4.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x4.v16i8")
    void neon_st1x4_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16])*);
    /// The `llvm.aarch64.neon.st1x4.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x4.v8i16")
    void neon_st1x4_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8])*);
    /// The `llvm.aarch64.neon.st1x4.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x4.v4i32")
    void neon_st1x4_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4])*);
    /// The `llvm.aarch64.neon.st1x4.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x4.v2i64")
    void neon_st1x4_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2])*);
    /// The `llvm.aarch64.neon.st1x4.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x4.v4f32")
    void neon_st1x4_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4])*);
    /// The `llvm.aarch64.neon.st1x4.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st1x4.v2f64")
    void neon_st1x4_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2])*);
    /// The `llvm.aarch64.neon.st2.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2.v16i8")
    void neon_st2_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16])*);
    /// The `llvm.aarch64.neon.st2.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2.v8i16")
    void neon_st2_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8])*);
    /// The `llvm.aarch64.neon.st2.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2.v4i32")
    void neon_st2_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4])*);
    /// The `llvm.aarch64.neon.st2.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2.v2i64")
    void neon_st2_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2])*);
    /// The `llvm.aarch64.neon.st2.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2.v4f32")
    void neon_st2_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4])*);
    /// The `llvm.aarch64.neon.st2.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2.v2f64")
    void neon_st2_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2])*);
    /// The `llvm.aarch64.neon.st3.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3.v16i8")
    void neon_st3_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16])*);
    /// The `llvm.aarch64.neon.st3.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3.v8i16")
    void neon_st3_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8])*);
    /// The `llvm.aarch64.neon.st3.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3.v4i32")
    void neon_st3_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4])*);
    /// The `llvm.aarch64.neon.st3.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3.v2i64")
    void neon_st3_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2])*);
    /// The `llvm.aarch64.neon.st3.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3.v4f32")
    void neon_st3_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4])*);
    /// The `llvm.aarch64.neon.st3.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3.v2f64")
    void neon_st3_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2])*);
    /// The `llvm.aarch64.neon.st4.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4.v16i8")
    void neon_st4_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16])*);
    /// The `llvm.aarch64.neon.st4.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4.v8i16")
    void neon_st4_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8])*);
    /// The `llvm.aarch64.neon.st4.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4.v4i32")
    void neon_st4_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4])*);
    /// The `llvm.aarch64.neon.st4.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4.v2i64")
    void neon_st4_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2])*);
    /// The `llvm.aarch64.neon.st4.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4.v4f32")
    void neon_st4_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4])*);
    /// The `llvm.aarch64.neon.st4.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4.v2f64")
    void neon_st4_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2])*);
    /// The `llvm.aarch64.neon.st2lane.v16i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2lane.v16i8.p0i8")
    void neon_st2lane_v16i8_p0i8(__vector(i8[16]), __vector(i8[16]), i64, i8*);
    /// The `llvm.aarch64.neon.st2lane.v8i16.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2lane.v8i16.p0i8")
    void neon_st2lane_v8i16_p0i8(__vector(i16[8]), __vector(i16[8]), i64, i8*);
    /// The `llvm.aarch64.neon.st2lane.v4i32.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2lane.v4i32.p0i8")
    void neon_st2lane_v4i32_p0i8(__vector(i32[4]), __vector(i32[4]), i64, i8*);
    /// The `llvm.aarch64.neon.st2lane.v2i64.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2lane.v2i64.p0i8")
    void neon_st2lane_v2i64_p0i8(__vector(i64[2]), __vector(i64[2]), i64, i8*);
    /// The `llvm.aarch64.neon.st2lane.v4f32.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2lane.v4f32.p0i8")
    void neon_st2lane_v4f32_p0i8(__vector(f32[4]), __vector(f32[4]), i64, i8*);
    /// The `llvm.aarch64.neon.st2lane.v2f64.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st2lane.v2f64.p0i8")
    void neon_st2lane_v2f64_p0i8(__vector(f64[2]), __vector(f64[2]), i64, i8*);
    /// The `llvm.aarch64.neon.st3lane.v16i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3lane.v16i8.p0i8")
    void neon_st3lane_v16i8_p0i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), i64, i8*);
    /// The `llvm.aarch64.neon.st3lane.v8i16.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3lane.v8i16.p0i8")
    void neon_st3lane_v8i16_p0i8(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]), i64, i8*);
    /// The `llvm.aarch64.neon.st3lane.v4i32.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3lane.v4i32.p0i8")
    void neon_st3lane_v4i32_p0i8(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]), i64, i8*);
    /// The `llvm.aarch64.neon.st3lane.v2i64.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3lane.v2i64.p0i8")
    void neon_st3lane_v2i64_p0i8(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]), i64, i8*);
    /// The `llvm.aarch64.neon.st3lane.v4f32.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3lane.v4f32.p0i8")
    void neon_st3lane_v4f32_p0i8(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i64, i8*);
    /// The `llvm.aarch64.neon.st3lane.v2f64.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st3lane.v2f64.p0i8")
    void neon_st3lane_v2f64_p0i8(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i64, i8*);
    /// The `llvm.aarch64.neon.st4lane.v16i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4lane.v16i8.p0i8")
    void neon_st4lane_v16i8_p0i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), i64, i8*);
    /// The `llvm.aarch64.neon.st4lane.v8i16.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4lane.v8i16.p0i8")
    void neon_st4lane_v8i16_p0i8(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), i64, i8*);
    /// The `llvm.aarch64.neon.st4lane.v4i32.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4lane.v4i32.p0i8")
    void neon_st4lane_v4i32_p0i8(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), i64, i8*);
    /// The `llvm.aarch64.neon.st4lane.v2i64.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4lane.v2i64.p0i8")
    void neon_st4lane_v2i64_p0i8(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), i64, i8*);
    /// The `llvm.aarch64.neon.st4lane.v4f32.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4lane.v4f32.p0i8")
    void neon_st4lane_v4f32_p0i8(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i64, i8*);
    /// The `llvm.aarch64.neon.st4lane.v2f64.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.st4lane.v2f64.p0i8")
    void neon_st4lane_v2f64_p0i8(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i64, i8*);
    /// The `llvm.aarch64.neon.tbl1.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl1.v16i8")
    __vector(i8[16]) neon_tbl1_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbl1.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl1.v8i16")
    __vector(i16[8]) neon_tbl1_v8i16(__vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbl1.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl1.v4i32")
    __vector(i32[4]) neon_tbl1_v4i32(__vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbl1.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl1.v2i64")
    __vector(i64[2]) neon_tbl1_v2i64(__vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbl1.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl1.v4f32")
    __vector(f32[4]) neon_tbl1_v4f32(__vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbl1.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl1.v2f64")
    __vector(f64[2]) neon_tbl1_v2f64(__vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbl2.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl2.v16i8")
    __vector(i8[16]) neon_tbl2_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbl2.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl2.v8i16")
    __vector(i16[8]) neon_tbl2_v8i16(__vector(i8[16]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbl2.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl2.v4i32")
    __vector(i32[4]) neon_tbl2_v4i32(__vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbl2.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl2.v2i64")
    __vector(i64[2]) neon_tbl2_v2i64(__vector(i8[16]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbl2.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl2.v4f32")
    __vector(f32[4]) neon_tbl2_v4f32(__vector(i8[16]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbl2.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl2.v2f64")
    __vector(f64[2]) neon_tbl2_v2f64(__vector(i8[16]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbl3.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl3.v16i8")
    __vector(i8[16]) neon_tbl3_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbl3.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl3.v8i16")
    __vector(i16[8]) neon_tbl3_v8i16(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbl3.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl3.v4i32")
    __vector(i32[4]) neon_tbl3_v4i32(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbl3.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl3.v2i64")
    __vector(i64[2]) neon_tbl3_v2i64(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbl3.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl3.v4f32")
    __vector(f32[4]) neon_tbl3_v4f32(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbl3.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl3.v2f64")
    __vector(f64[2]) neon_tbl3_v2f64(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbl4.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl4.v16i8")
    __vector(i8[16]) neon_tbl4_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbl4.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl4.v8i16")
    __vector(i16[8]) neon_tbl4_v8i16(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbl4.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl4.v4i32")
    __vector(i32[4]) neon_tbl4_v4i32(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbl4.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl4.v2i64")
    __vector(i64[2]) neon_tbl4_v2i64(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbl4.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl4.v4f32")
    __vector(f32[4]) neon_tbl4_v4f32(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbl4.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbl4.v2f64")
    __vector(f64[2]) neon_tbl4_v2f64(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbx1.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx1.v16i8")
    __vector(i8[16]) neon_tbx1_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbx1.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx1.v8i16")
    __vector(i16[8]) neon_tbx1_v8i16(__vector(i16[8]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbx1.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx1.v4i32")
    __vector(i32[4]) neon_tbx1_v4i32(__vector(i32[4]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbx1.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx1.v2i64")
    __vector(i64[2]) neon_tbx1_v2i64(__vector(i64[2]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbx1.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx1.v4f32")
    __vector(f32[4]) neon_tbx1_v4f32(__vector(f32[4]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbx1.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx1.v2f64")
    __vector(f64[2]) neon_tbx1_v2f64(__vector(f64[2]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbx2.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx2.v16i8")
    __vector(i8[16]) neon_tbx2_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbx2.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx2.v8i16")
    __vector(i16[8]) neon_tbx2_v8i16(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbx2.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx2.v4i32")
    __vector(i32[4]) neon_tbx2_v4i32(__vector(i32[4]), __vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbx2.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx2.v2i64")
    __vector(i64[2]) neon_tbx2_v2i64(__vector(i64[2]), __vector(i8[16]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbx2.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx2.v4f32")
    __vector(f32[4]) neon_tbx2_v4f32(__vector(f32[4]), __vector(i8[16]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbx2.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx2.v2f64")
    __vector(f64[2]) neon_tbx2_v2f64(__vector(f64[2]), __vector(i8[16]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbx3.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx3.v16i8")
    __vector(i8[16]) neon_tbx3_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbx3.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx3.v8i16")
    __vector(i16[8]) neon_tbx3_v8i16(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbx3.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx3.v4i32")
    __vector(i32[4]) neon_tbx3_v4i32(__vector(i32[4]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbx3.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx3.v2i64")
    __vector(i64[2]) neon_tbx3_v2i64(__vector(i64[2]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbx3.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx3.v4f32")
    __vector(f32[4]) neon_tbx3_v4f32(__vector(f32[4]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbx3.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx3.v2f64")
    __vector(f64[2]) neon_tbx3_v2f64(__vector(f64[2]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.neon.tbx4.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx4.v16i8")
    __vector(i8[16]) neon_tbx4_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.neon.tbx4.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx4.v8i16")
    __vector(i16[8]) neon_tbx4_v8i16(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i16[8]));
    /// The `llvm.aarch64.neon.tbx4.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx4.v4i32")
    __vector(i32[4]) neon_tbx4_v4i32(__vector(i32[4]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.aarch64.neon.tbx4.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx4.v2i64")
    __vector(i64[2]) neon_tbx4_v2i64(__vector(i64[2]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i64[2]));
    /// The `llvm.aarch64.neon.tbx4.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx4.v4f32")
    __vector(f32[4]) neon_tbx4_v4f32(__vector(f32[4]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f32[4]));
    /// The `llvm.aarch64.neon.tbx4.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.neon.tbx4.v2f64")
    __vector(f64[2]) neon_tbx4_v2f64(__vector(f64[2]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(f64[2]));
    /// The `llvm.aarch64.crypto.aese` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.aese")
    __vector(i8[16]) crypto_aese(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.crypto.aesd` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.aesd")
    __vector(i8[16]) crypto_aesd(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.aarch64.crypto.aesmc` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.aesmc")
    __vector(i8[16]) crypto_aesmc(__vector(i8[16]));
    /// The `llvm.aarch64.crypto.aesimc` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.aesimc")
    __vector(i8[16]) crypto_aesimc(__vector(i8[16]));
    /// The `llvm.aarch64.crypto.sha1c` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha1c")
    __vector(i32[4]) crypto_sha1c(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha1p` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha1p")
    __vector(i32[4]) crypto_sha1p(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha1m` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha1m")
    __vector(i32[4]) crypto_sha1m(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha1h` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha1h")
    i32 crypto_sha1h(i32);
    /// The `llvm.aarch64.crypto.sha1su0` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha1su0")
    __vector(i32[4]) crypto_sha1su0(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha1su1` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha1su1")
    __vector(i32[4]) crypto_sha1su1(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha256h` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha256h")
    __vector(i32[4]) crypto_sha256h(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha256h2` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha256h2")
    __vector(i32[4]) crypto_sha256h2(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha256su0` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha256su0")
    __vector(i32[4]) crypto_sha256su0(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.crypto.sha256su1` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crypto.sha256su1")
    __vector(i32[4]) crypto_sha256su1(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.aarch64.crc32b` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32b")
    i32 crc32b(i32, i32);
    /// The `llvm.aarch64.crc32cb` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32cb")
    i32 crc32cb(i32, i32);
    /// The `llvm.aarch64.crc32h` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32h")
    i32 crc32h(i32, i32);
    /// The `llvm.aarch64.crc32ch` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32ch")
    i32 crc32ch(i32, i32);
    /// The `llvm.aarch64.crc32w` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32w")
    i32 crc32w(i32, i32);
    /// The `llvm.aarch64.crc32cw` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32cw")
    i32 crc32cw(i32, i32);
    /// The `llvm.aarch64.crc32x` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32x")
    i32 crc32x(i32, i64);
    /// The `llvm.aarch64.crc32cx` intrinsic.
    pragma(LDC_intrinsic, "llvm.aarch64.crc32cx")
    i32 crc32cx(i32, i64);
}

/// LLVM intrinsics for the arm architecture.
version (ARM_Any) {
    /// The `llvm.arm.thread.pointer` intrinsic; known as `__builtin_thread_pointer` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.thread.pointer")
    i8* thread_pointer();
    /// The `llvm.arm.qadd` intrinsic; known as `__builtin_arm_qadd` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.qadd")
    i32 qadd(i32, i32);
    /// The `llvm.arm.qsub` intrinsic; known as `__builtin_arm_qsub` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.qsub")
    i32 qsub(i32, i32);
    /// The `llvm.arm.ssat` intrinsic; known as `__builtin_arm_ssat` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.ssat")
    i32 ssat(i32, i32);
    /// The `llvm.arm.usat` intrinsic; known as `__builtin_arm_usat` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.usat")
    i32 usat(i32, i32);
    /// The `llvm.arm.ldrex.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.ldrex.p0i8")
    i32 ldrex_p0i8(i8*);
    /// The `llvm.arm.strex.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.strex.p0i8")
    i32 strex_p0i8(i32, i8*);
    /// The `llvm.arm.ldaex.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.ldaex.p0i8")
    i32 ldaex_p0i8(i8*);
    /// The `llvm.arm.stlex.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.stlex.p0i8")
    i32 stlex_p0i8(i32, i8*);
    /// The `llvm.arm.clrex` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.clrex")
    void clrex();
    /// The `llvm.arm.strexd` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.strexd")
    i32 strexd(i32, i32, i8*);
    /// The `llvm.arm.stlexd` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.stlexd")
    i32 stlexd(i32, i32, i8*);
    /// The `llvm.arm.dmb` intrinsic; known as `__builtin_arm_dmb` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.dmb")
    void dmb(i32);
    /// The `llvm.arm.dsb` intrinsic; known as `__builtin_arm_dsb` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.dsb")
    void dsb(i32);
    /// The `llvm.arm.isb` intrinsic; known as `__builtin_arm_isb` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.isb")
    void isb(i32);
    /// The `llvm.arm.get.fpscr` intrinsic; known as `__builtin_arm_get_fpscr` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.get.fpscr")
    i32 get_fpscr();
    /// The `llvm.arm.set.fpscr` intrinsic; known as `__builtin_arm_set_fpscr` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.set.fpscr")
    void set_fpscr(i32);
    /// The `llvm.arm.vcvtr.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtr.v4f32")
    f32 vcvtr_v4f32(__vector(f32[4]));
    /// The `llvm.arm.vcvtr.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtr.f32")
    f32 vcvtr_f32(f32);
    /// The `llvm.arm.vcvtr.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtr.v2f64")
    f32 vcvtr_v2f64(__vector(f64[2]));
    /// The `llvm.arm.vcvtr.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtr.f64")
    f32 vcvtr_f64(f64);
    /// The `llvm.arm.vcvtru.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtru.v4f32")
    f32 vcvtru_v4f32(__vector(f32[4]));
    /// The `llvm.arm.vcvtru.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtru.f32")
    f32 vcvtru_f32(f32);
    /// The `llvm.arm.vcvtru.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtru.v2f64")
    f32 vcvtru_v2f64(__vector(f64[2]));
    /// The `llvm.arm.vcvtru.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.vcvtru.f64")
    f32 vcvtru_f64(f64);
    /// The `llvm.arm.mcr` intrinsic; known as `__builtin_arm_mcr` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.mcr")
    void mcr(i32, i32, i32, i32, i32, i32);
    /// The `llvm.arm.mcr2` intrinsic; known as `__builtin_arm_mcr2` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.mcr2")
    void mcr2(i32, i32, i32, i32, i32, i32);
    /// The `llvm.arm.mrc` intrinsic; known as `__builtin_arm_mrc` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.mrc")
    i32 mrc(i32, i32, i32, i32, i32);
    /// The `llvm.arm.mrc2` intrinsic; known as `__builtin_arm_mrc2` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.mrc2")
    i32 mrc2(i32, i32, i32, i32, i32);
    /// The `llvm.arm.cdp` intrinsic; known as `__builtin_arm_cdp` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.cdp")
    void cdp(i32, i32, i32, i32, i32, i32);
    /// The `llvm.arm.cdp2` intrinsic; known as `__builtin_arm_cdp2` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.cdp2")
    void cdp2(i32, i32, i32, i32, i32, i32);
    /// The `llvm.arm.mcrr` intrinsic; known as `__builtin_arm_mcrr` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.mcrr")
    void mcrr(i32, i32, i32, i32, i32);
    /// The `llvm.arm.mcrr2` intrinsic; known as `__builtin_arm_mcrr2` in GCC.
    pragma(LDC_intrinsic, "llvm.arm.mcrr2")
    void mcrr2(i32, i32, i32, i32, i32);
    /// The `llvm.arm.crc32b` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.crc32b")
    i32 crc32b(i32, i32);
    /// The `llvm.arm.crc32cb` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.crc32cb")
    i32 crc32cb(i32, i32);
    /// The `llvm.arm.crc32h` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.crc32h")
    i32 crc32h(i32, i32);
    /// The `llvm.arm.crc32ch` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.crc32ch")
    i32 crc32ch(i32, i32);
    /// The `llvm.arm.crc32w` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.crc32w")
    i32 crc32w(i32, i32);
    /// The `llvm.arm.crc32cw` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.crc32cw")
    i32 crc32cw(i32, i32);
    /// The `llvm.arm.hint` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.hint")
    void hint(i32);
    /// The `llvm.arm.dbg` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.dbg")
    void dbg(i32);
    /// The `llvm.arm.rbit` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.rbit")
    i32 rbit(i32);
    /// The `llvm.arm.undefined` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.undefined")
    void undefined(i32);
    /// The `llvm.arm.neon.vhadds.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhadds.v16i8")
    __vector(i8[16]) neon_vhadds_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vhadds.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhadds.v8i16")
    __vector(i16[8]) neon_vhadds_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vhadds.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhadds.v4i32")
    __vector(i32[4]) neon_vhadds_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vhadds.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhadds.v2i64")
    __vector(i64[2]) neon_vhadds_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vhadds.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhadds.v4f32")
    __vector(f32[4]) neon_vhadds_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vhadds.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhadds.v2f64")
    __vector(f64[2]) neon_vhadds_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vhaddu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhaddu.v16i8")
    __vector(i8[16]) neon_vhaddu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vhaddu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhaddu.v8i16")
    __vector(i16[8]) neon_vhaddu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vhaddu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhaddu.v4i32")
    __vector(i32[4]) neon_vhaddu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vhaddu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhaddu.v2i64")
    __vector(i64[2]) neon_vhaddu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vhaddu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhaddu.v4f32")
    __vector(f32[4]) neon_vhaddu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vhaddu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhaddu.v2f64")
    __vector(f64[2]) neon_vhaddu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vrhadds.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhadds.v16i8")
    __vector(i8[16]) neon_vrhadds_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vrhadds.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhadds.v8i16")
    __vector(i16[8]) neon_vrhadds_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vrhadds.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhadds.v4i32")
    __vector(i32[4]) neon_vrhadds_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vrhadds.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhadds.v2i64")
    __vector(i64[2]) neon_vrhadds_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vrhadds.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhadds.v4f32")
    __vector(f32[4]) neon_vrhadds_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vrhadds.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhadds.v2f64")
    __vector(f64[2]) neon_vrhadds_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vrhaddu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhaddu.v16i8")
    __vector(i8[16]) neon_vrhaddu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vrhaddu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhaddu.v8i16")
    __vector(i16[8]) neon_vrhaddu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vrhaddu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhaddu.v4i32")
    __vector(i32[4]) neon_vrhaddu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vrhaddu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhaddu.v2i64")
    __vector(i64[2]) neon_vrhaddu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vrhaddu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhaddu.v4f32")
    __vector(f32[4]) neon_vrhaddu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vrhaddu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrhaddu.v2f64")
    __vector(f64[2]) neon_vrhaddu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqadds.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqadds.v16i8")
    __vector(i8[16]) neon_vqadds_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqadds.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqadds.v8i16")
    __vector(i16[8]) neon_vqadds_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqadds.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqadds.v4i32")
    __vector(i32[4]) neon_vqadds_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqadds.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqadds.v2i64")
    __vector(i64[2]) neon_vqadds_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqadds.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqadds.v4f32")
    __vector(f32[4]) neon_vqadds_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqadds.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqadds.v2f64")
    __vector(f64[2]) neon_vqadds_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqaddu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqaddu.v16i8")
    __vector(i8[16]) neon_vqaddu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqaddu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqaddu.v8i16")
    __vector(i16[8]) neon_vqaddu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqaddu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqaddu.v4i32")
    __vector(i32[4]) neon_vqaddu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqaddu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqaddu.v2i64")
    __vector(i64[2]) neon_vqaddu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqaddu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqaddu.v4f32")
    __vector(f32[4]) neon_vqaddu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqaddu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqaddu.v2f64")
    __vector(f64[2]) neon_vqaddu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vmulp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmulp.v16i8")
    __vector(i8[16]) neon_vmulp_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vmulp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmulp.v8i16")
    __vector(i16[8]) neon_vmulp_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vmulp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmulp.v4i32")
    __vector(i32[4]) neon_vmulp_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vmulp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmulp.v2i64")
    __vector(i64[2]) neon_vmulp_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vmulp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmulp.v4f32")
    __vector(f32[4]) neon_vmulp_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vmulp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmulp.v2f64")
    __vector(f64[2]) neon_vmulp_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqdmulh.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqdmulh.v16i8")
    __vector(i8[16]) neon_vqdmulh_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqdmulh.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqdmulh.v8i16")
    __vector(i16[8]) neon_vqdmulh_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqdmulh.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqdmulh.v4i32")
    __vector(i32[4]) neon_vqdmulh_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqdmulh.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqdmulh.v2i64")
    __vector(i64[2]) neon_vqdmulh_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqdmulh.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqdmulh.v4f32")
    __vector(f32[4]) neon_vqdmulh_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqdmulh.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqdmulh.v2f64")
    __vector(f64[2]) neon_vqdmulh_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqrdmulh.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrdmulh.v16i8")
    __vector(i8[16]) neon_vqrdmulh_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqrdmulh.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrdmulh.v8i16")
    __vector(i16[8]) neon_vqrdmulh_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqrdmulh.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrdmulh.v4i32")
    __vector(i32[4]) neon_vqrdmulh_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqrdmulh.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrdmulh.v2i64")
    __vector(i64[2]) neon_vqrdmulh_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqrdmulh.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrdmulh.v4f32")
    __vector(f32[4]) neon_vqrdmulh_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqrdmulh.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrdmulh.v2f64")
    __vector(f64[2]) neon_vqrdmulh_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vmaxs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxs.v16i8")
    __vector(i8[16]) neon_vmaxs_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vmaxs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxs.v8i16")
    __vector(i16[8]) neon_vmaxs_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vmaxs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxs.v4i32")
    __vector(i32[4]) neon_vmaxs_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vmaxs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxs.v2i64")
    __vector(i64[2]) neon_vmaxs_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vmaxs.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxs.v4f32")
    __vector(f32[4]) neon_vmaxs_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vmaxs.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxs.v2f64")
    __vector(f64[2]) neon_vmaxs_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vmaxu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxu.v16i8")
    __vector(i8[16]) neon_vmaxu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vmaxu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxu.v8i16")
    __vector(i16[8]) neon_vmaxu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vmaxu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxu.v4i32")
    __vector(i32[4]) neon_vmaxu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vmaxu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxu.v2i64")
    __vector(i64[2]) neon_vmaxu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vmaxu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxu.v4f32")
    __vector(f32[4]) neon_vmaxu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vmaxu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxu.v2f64")
    __vector(f64[2]) neon_vmaxu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vmaxnm.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxnm.v16i8")
    __vector(i8[16]) neon_vmaxnm_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vmaxnm.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxnm.v8i16")
    __vector(i16[8]) neon_vmaxnm_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vmaxnm.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxnm.v4i32")
    __vector(i32[4]) neon_vmaxnm_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vmaxnm.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxnm.v2i64")
    __vector(i64[2]) neon_vmaxnm_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vmaxnm.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxnm.v4f32")
    __vector(f32[4]) neon_vmaxnm_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vmaxnm.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmaxnm.v2f64")
    __vector(f64[2]) neon_vmaxnm_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vmins.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmins.v16i8")
    __vector(i8[16]) neon_vmins_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vmins.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmins.v8i16")
    __vector(i16[8]) neon_vmins_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vmins.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmins.v4i32")
    __vector(i32[4]) neon_vmins_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vmins.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmins.v2i64")
    __vector(i64[2]) neon_vmins_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vmins.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmins.v4f32")
    __vector(f32[4]) neon_vmins_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vmins.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vmins.v2f64")
    __vector(f64[2]) neon_vmins_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vminu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminu.v16i8")
    __vector(i8[16]) neon_vminu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vminu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminu.v8i16")
    __vector(i16[8]) neon_vminu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vminu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminu.v4i32")
    __vector(i32[4]) neon_vminu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vminu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminu.v2i64")
    __vector(i64[2]) neon_vminu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vminu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminu.v4f32")
    __vector(f32[4]) neon_vminu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vminu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminu.v2f64")
    __vector(f64[2]) neon_vminu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vminnm.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminnm.v16i8")
    __vector(i8[16]) neon_vminnm_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vminnm.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminnm.v8i16")
    __vector(i16[8]) neon_vminnm_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vminnm.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminnm.v4i32")
    __vector(i32[4]) neon_vminnm_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vminnm.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminnm.v2i64")
    __vector(i64[2]) neon_vminnm_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vminnm.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminnm.v4f32")
    __vector(f32[4]) neon_vminnm_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vminnm.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vminnm.v2f64")
    __vector(f64[2]) neon_vminnm_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vrecps.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecps.v16i8")
    __vector(i8[16]) neon_vrecps_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vrecps.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecps.v8i16")
    __vector(i16[8]) neon_vrecps_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vrecps.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecps.v4i32")
    __vector(i32[4]) neon_vrecps_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vrecps.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecps.v2i64")
    __vector(i64[2]) neon_vrecps_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vrecps.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecps.v4f32")
    __vector(f32[4]) neon_vrecps_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vrecps.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecps.v2f64")
    __vector(f64[2]) neon_vrecps_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vrsqrts.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrts.v16i8")
    __vector(i8[16]) neon_vrsqrts_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vrsqrts.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrts.v8i16")
    __vector(i16[8]) neon_vrsqrts_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vrsqrts.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrts.v4i32")
    __vector(i32[4]) neon_vrsqrts_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vrsqrts.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrts.v2i64")
    __vector(i64[2]) neon_vrsqrts_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vrsqrts.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrts.v4f32")
    __vector(f32[4]) neon_vrsqrts_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vrsqrts.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrts.v2f64")
    __vector(f64[2]) neon_vrsqrts_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vhsubs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubs.v16i8")
    __vector(i8[16]) neon_vhsubs_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vhsubs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubs.v8i16")
    __vector(i16[8]) neon_vhsubs_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vhsubs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubs.v4i32")
    __vector(i32[4]) neon_vhsubs_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vhsubs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubs.v2i64")
    __vector(i64[2]) neon_vhsubs_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vhsubs.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubs.v4f32")
    __vector(f32[4]) neon_vhsubs_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vhsubs.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubs.v2f64")
    __vector(f64[2]) neon_vhsubs_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vhsubu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubu.v16i8")
    __vector(i8[16]) neon_vhsubu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vhsubu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubu.v8i16")
    __vector(i16[8]) neon_vhsubu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vhsubu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubu.v4i32")
    __vector(i32[4]) neon_vhsubu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vhsubu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubu.v2i64")
    __vector(i64[2]) neon_vhsubu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vhsubu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubu.v4f32")
    __vector(f32[4]) neon_vhsubu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vhsubu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vhsubu.v2f64")
    __vector(f64[2]) neon_vhsubu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqsubs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubs.v16i8")
    __vector(i8[16]) neon_vqsubs_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqsubs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubs.v8i16")
    __vector(i16[8]) neon_vqsubs_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqsubs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubs.v4i32")
    __vector(i32[4]) neon_vqsubs_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqsubs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubs.v2i64")
    __vector(i64[2]) neon_vqsubs_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqsubs.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubs.v4f32")
    __vector(f32[4]) neon_vqsubs_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqsubs.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubs.v2f64")
    __vector(f64[2]) neon_vqsubs_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqsubu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubu.v16i8")
    __vector(i8[16]) neon_vqsubu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqsubu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubu.v8i16")
    __vector(i16[8]) neon_vqsubu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqsubu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubu.v4i32")
    __vector(i32[4]) neon_vqsubu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqsubu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubu.v2i64")
    __vector(i64[2]) neon_vqsubu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqsubu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubu.v4f32")
    __vector(f32[4]) neon_vqsubu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqsubu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqsubu.v2f64")
    __vector(f64[2]) neon_vqsubu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacge.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v16i8.v16i8")
    __vector(i8[16]) neon_vacge_v16i8_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacge.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v16i8.v8i16")
    __vector(i8[16]) neon_vacge_v16i8_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacge.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v16i8.v4i32")
    __vector(i8[16]) neon_vacge_v16i8_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacge.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v16i8.v2i64")
    __vector(i8[16]) neon_vacge_v16i8_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacge.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v16i8.v4f32")
    __vector(i8[16]) neon_vacge_v16i8_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacge.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v16i8.v2f64")
    __vector(i8[16]) neon_vacge_v16i8_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacge.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v8i16.v16i8")
    __vector(i16[8]) neon_vacge_v8i16_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacge.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v8i16.v8i16")
    __vector(i16[8]) neon_vacge_v8i16_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacge.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v8i16.v4i32")
    __vector(i16[8]) neon_vacge_v8i16_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacge.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v8i16.v2i64")
    __vector(i16[8]) neon_vacge_v8i16_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacge.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v8i16.v4f32")
    __vector(i16[8]) neon_vacge_v8i16_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacge.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v8i16.v2f64")
    __vector(i16[8]) neon_vacge_v8i16_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacge.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4i32.v16i8")
    __vector(i32[4]) neon_vacge_v4i32_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacge.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4i32.v8i16")
    __vector(i32[4]) neon_vacge_v4i32_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacge.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4i32.v4i32")
    __vector(i32[4]) neon_vacge_v4i32_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacge.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4i32.v2i64")
    __vector(i32[4]) neon_vacge_v4i32_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacge.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4i32.v4f32")
    __vector(i32[4]) neon_vacge_v4i32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacge.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4i32.v2f64")
    __vector(i32[4]) neon_vacge_v4i32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacge.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2i64.v16i8")
    __vector(i64[2]) neon_vacge_v2i64_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacge.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2i64.v8i16")
    __vector(i64[2]) neon_vacge_v2i64_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacge.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2i64.v4i32")
    __vector(i64[2]) neon_vacge_v2i64_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacge.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2i64.v2i64")
    __vector(i64[2]) neon_vacge_v2i64_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacge.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2i64.v4f32")
    __vector(i64[2]) neon_vacge_v2i64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacge.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2i64.v2f64")
    __vector(i64[2]) neon_vacge_v2i64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacge.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4f32.v16i8")
    __vector(f32[4]) neon_vacge_v4f32_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacge.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4f32.v8i16")
    __vector(f32[4]) neon_vacge_v4f32_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacge.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4f32.v4i32")
    __vector(f32[4]) neon_vacge_v4f32_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacge.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4f32.v2i64")
    __vector(f32[4]) neon_vacge_v4f32_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacge.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4f32.v4f32")
    __vector(f32[4]) neon_vacge_v4f32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacge.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v4f32.v2f64")
    __vector(f32[4]) neon_vacge_v4f32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacge.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2f64.v16i8")
    __vector(f64[2]) neon_vacge_v2f64_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacge.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2f64.v8i16")
    __vector(f64[2]) neon_vacge_v2f64_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacge.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2f64.v4i32")
    __vector(f64[2]) neon_vacge_v2f64_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacge.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2f64.v2i64")
    __vector(f64[2]) neon_vacge_v2f64_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacge.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2f64.v4f32")
    __vector(f64[2]) neon_vacge_v2f64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacge.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacge.v2f64.v2f64")
    __vector(f64[2]) neon_vacge_v2f64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacgt.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v16i8.v16i8")
    __vector(i8[16]) neon_vacgt_v16i8_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacgt.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v16i8.v8i16")
    __vector(i8[16]) neon_vacgt_v16i8_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacgt.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v16i8.v4i32")
    __vector(i8[16]) neon_vacgt_v16i8_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacgt.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v16i8.v2i64")
    __vector(i8[16]) neon_vacgt_v16i8_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacgt.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v16i8.v4f32")
    __vector(i8[16]) neon_vacgt_v16i8_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacgt.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v16i8.v2f64")
    __vector(i8[16]) neon_vacgt_v16i8_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacgt.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v8i16.v16i8")
    __vector(i16[8]) neon_vacgt_v8i16_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacgt.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v8i16.v8i16")
    __vector(i16[8]) neon_vacgt_v8i16_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacgt.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v8i16.v4i32")
    __vector(i16[8]) neon_vacgt_v8i16_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacgt.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v8i16.v2i64")
    __vector(i16[8]) neon_vacgt_v8i16_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacgt.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v8i16.v4f32")
    __vector(i16[8]) neon_vacgt_v8i16_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacgt.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v8i16.v2f64")
    __vector(i16[8]) neon_vacgt_v8i16_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacgt.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4i32.v16i8")
    __vector(i32[4]) neon_vacgt_v4i32_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacgt.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4i32.v8i16")
    __vector(i32[4]) neon_vacgt_v4i32_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacgt.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4i32.v4i32")
    __vector(i32[4]) neon_vacgt_v4i32_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacgt.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4i32.v2i64")
    __vector(i32[4]) neon_vacgt_v4i32_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacgt.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4i32.v4f32")
    __vector(i32[4]) neon_vacgt_v4i32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacgt.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4i32.v2f64")
    __vector(i32[4]) neon_vacgt_v4i32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacgt.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2i64.v16i8")
    __vector(i64[2]) neon_vacgt_v2i64_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacgt.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2i64.v8i16")
    __vector(i64[2]) neon_vacgt_v2i64_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacgt.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2i64.v4i32")
    __vector(i64[2]) neon_vacgt_v2i64_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacgt.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2i64.v2i64")
    __vector(i64[2]) neon_vacgt_v2i64_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacgt.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2i64.v4f32")
    __vector(i64[2]) neon_vacgt_v2i64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacgt.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2i64.v2f64")
    __vector(i64[2]) neon_vacgt_v2i64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacgt.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4f32.v16i8")
    __vector(f32[4]) neon_vacgt_v4f32_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacgt.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4f32.v8i16")
    __vector(f32[4]) neon_vacgt_v4f32_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacgt.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4f32.v4i32")
    __vector(f32[4]) neon_vacgt_v4f32_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacgt.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4f32.v2i64")
    __vector(f32[4]) neon_vacgt_v4f32_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacgt.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4f32.v4f32")
    __vector(f32[4]) neon_vacgt_v4f32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacgt.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v4f32.v2f64")
    __vector(f32[4]) neon_vacgt_v4f32_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vacgt.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2f64.v16i8")
    __vector(f64[2]) neon_vacgt_v2f64_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vacgt.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2f64.v8i16")
    __vector(f64[2]) neon_vacgt_v2f64_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vacgt.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2f64.v4i32")
    __vector(f64[2]) neon_vacgt_v2f64_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vacgt.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2f64.v2i64")
    __vector(f64[2]) neon_vacgt_v2f64_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vacgt.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2f64.v4f32")
    __vector(f64[2]) neon_vacgt_v2f64_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vacgt.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vacgt.v2f64.v2f64")
    __vector(f64[2]) neon_vacgt_v2f64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vabds.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabds.v16i8")
    __vector(i8[16]) neon_vabds_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vabds.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabds.v8i16")
    __vector(i16[8]) neon_vabds_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vabds.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabds.v4i32")
    __vector(i32[4]) neon_vabds_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vabds.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabds.v2i64")
    __vector(i64[2]) neon_vabds_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vabds.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabds.v4f32")
    __vector(f32[4]) neon_vabds_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vabds.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabds.v2f64")
    __vector(f64[2]) neon_vabds_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vabdu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabdu.v16i8")
    __vector(i8[16]) neon_vabdu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vabdu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabdu.v8i16")
    __vector(i16[8]) neon_vabdu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vabdu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabdu.v4i32")
    __vector(i32[4]) neon_vabdu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vabdu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabdu.v2i64")
    __vector(i64[2]) neon_vabdu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vabdu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabdu.v4f32")
    __vector(f32[4]) neon_vabdu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vabdu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabdu.v2f64")
    __vector(f64[2]) neon_vabdu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadd.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadd.v16i8")
    __vector(i8[16]) neon_vpadd_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadd.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadd.v8i16")
    __vector(i16[8]) neon_vpadd_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadd.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadd.v4i32")
    __vector(i32[4]) neon_vpadd_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadd.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadd.v2i64")
    __vector(i64[2]) neon_vpadd_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadd.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadd.v4f32")
    __vector(f32[4]) neon_vpadd_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadd.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadd.v2f64")
    __vector(f64[2]) neon_vpadd_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpaddls.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v16i8.v16i8")
    __vector(i8[16]) neon_vpaddls_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddls.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v16i8.v8i16")
    __vector(i8[16]) neon_vpaddls_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddls.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v16i8.v4i32")
    __vector(i8[16]) neon_vpaddls_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddls.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v16i8.v2i64")
    __vector(i8[16]) neon_vpaddls_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddls.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v16i8.v4f32")
    __vector(i8[16]) neon_vpaddls_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddls.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v16i8.v2f64")
    __vector(i8[16]) neon_vpaddls_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddls.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v8i16.v16i8")
    __vector(i16[8]) neon_vpaddls_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddls.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v8i16.v8i16")
    __vector(i16[8]) neon_vpaddls_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddls.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v8i16.v4i32")
    __vector(i16[8]) neon_vpaddls_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddls.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v8i16.v2i64")
    __vector(i16[8]) neon_vpaddls_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddls.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v8i16.v4f32")
    __vector(i16[8]) neon_vpaddls_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddls.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v8i16.v2f64")
    __vector(i16[8]) neon_vpaddls_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddls.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4i32.v16i8")
    __vector(i32[4]) neon_vpaddls_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddls.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4i32.v8i16")
    __vector(i32[4]) neon_vpaddls_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddls.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4i32.v4i32")
    __vector(i32[4]) neon_vpaddls_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddls.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4i32.v2i64")
    __vector(i32[4]) neon_vpaddls_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddls.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4i32.v4f32")
    __vector(i32[4]) neon_vpaddls_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddls.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4i32.v2f64")
    __vector(i32[4]) neon_vpaddls_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddls.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2i64.v16i8")
    __vector(i64[2]) neon_vpaddls_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddls.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2i64.v8i16")
    __vector(i64[2]) neon_vpaddls_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddls.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2i64.v4i32")
    __vector(i64[2]) neon_vpaddls_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddls.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2i64.v2i64")
    __vector(i64[2]) neon_vpaddls_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddls.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2i64.v4f32")
    __vector(i64[2]) neon_vpaddls_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddls.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2i64.v2f64")
    __vector(i64[2]) neon_vpaddls_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddls.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4f32.v16i8")
    __vector(f32[4]) neon_vpaddls_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddls.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4f32.v8i16")
    __vector(f32[4]) neon_vpaddls_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddls.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4f32.v4i32")
    __vector(f32[4]) neon_vpaddls_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddls.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4f32.v2i64")
    __vector(f32[4]) neon_vpaddls_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddls.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4f32.v4f32")
    __vector(f32[4]) neon_vpaddls_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddls.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v4f32.v2f64")
    __vector(f32[4]) neon_vpaddls_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddls.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2f64.v16i8")
    __vector(f64[2]) neon_vpaddls_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddls.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2f64.v8i16")
    __vector(f64[2]) neon_vpaddls_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddls.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2f64.v4i32")
    __vector(f64[2]) neon_vpaddls_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddls.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2f64.v2i64")
    __vector(f64[2]) neon_vpaddls_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddls.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2f64.v4f32")
    __vector(f64[2]) neon_vpaddls_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddls.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddls.v2f64.v2f64")
    __vector(f64[2]) neon_vpaddls_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddlu.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v16i8.v16i8")
    __vector(i8[16]) neon_vpaddlu_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddlu.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v16i8.v8i16")
    __vector(i8[16]) neon_vpaddlu_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddlu.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v16i8.v4i32")
    __vector(i8[16]) neon_vpaddlu_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddlu.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v16i8.v2i64")
    __vector(i8[16]) neon_vpaddlu_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddlu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v16i8.v4f32")
    __vector(i8[16]) neon_vpaddlu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddlu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v16i8.v2f64")
    __vector(i8[16]) neon_vpaddlu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddlu.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v8i16.v16i8")
    __vector(i16[8]) neon_vpaddlu_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddlu.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v8i16.v8i16")
    __vector(i16[8]) neon_vpaddlu_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddlu.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v8i16.v4i32")
    __vector(i16[8]) neon_vpaddlu_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddlu.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v8i16.v2i64")
    __vector(i16[8]) neon_vpaddlu_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddlu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v8i16.v4f32")
    __vector(i16[8]) neon_vpaddlu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddlu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v8i16.v2f64")
    __vector(i16[8]) neon_vpaddlu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddlu.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4i32.v16i8")
    __vector(i32[4]) neon_vpaddlu_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddlu.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4i32.v8i16")
    __vector(i32[4]) neon_vpaddlu_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddlu.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4i32.v4i32")
    __vector(i32[4]) neon_vpaddlu_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddlu.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4i32.v2i64")
    __vector(i32[4]) neon_vpaddlu_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddlu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4i32.v4f32")
    __vector(i32[4]) neon_vpaddlu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddlu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4i32.v2f64")
    __vector(i32[4]) neon_vpaddlu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddlu.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2i64.v16i8")
    __vector(i64[2]) neon_vpaddlu_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddlu.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2i64.v8i16")
    __vector(i64[2]) neon_vpaddlu_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddlu.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2i64.v4i32")
    __vector(i64[2]) neon_vpaddlu_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddlu.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2i64.v2i64")
    __vector(i64[2]) neon_vpaddlu_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddlu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2i64.v4f32")
    __vector(i64[2]) neon_vpaddlu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddlu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2i64.v2f64")
    __vector(i64[2]) neon_vpaddlu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddlu.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4f32.v16i8")
    __vector(f32[4]) neon_vpaddlu_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddlu.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4f32.v8i16")
    __vector(f32[4]) neon_vpaddlu_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddlu.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4f32.v4i32")
    __vector(f32[4]) neon_vpaddlu_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddlu.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4f32.v2i64")
    __vector(f32[4]) neon_vpaddlu_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddlu.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4f32.v4f32")
    __vector(f32[4]) neon_vpaddlu_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddlu.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v4f32.v2f64")
    __vector(f32[4]) neon_vpaddlu_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpaddlu.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2f64.v16i8")
    __vector(f64[2]) neon_vpaddlu_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vpaddlu.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2f64.v8i16")
    __vector(f64[2]) neon_vpaddlu_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vpaddlu.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2f64.v4i32")
    __vector(f64[2]) neon_vpaddlu_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vpaddlu.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2f64.v2i64")
    __vector(f64[2]) neon_vpaddlu_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vpaddlu.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2f64.v4f32")
    __vector(f64[2]) neon_vpaddlu_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vpaddlu.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpaddlu.v2f64.v2f64")
    __vector(f64[2]) neon_vpaddlu_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vpadals.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v16i8.v16i8")
    __vector(i8[16]) neon_vpadals_v16i8_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadals.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v16i8.v8i16")
    __vector(i8[16]) neon_vpadals_v16i8_v8i16(__vector(i8[16]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadals.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v16i8.v4i32")
    __vector(i8[16]) neon_vpadals_v16i8_v4i32(__vector(i8[16]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadals.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v16i8.v2i64")
    __vector(i8[16]) neon_vpadals_v16i8_v2i64(__vector(i8[16]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadals.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v16i8.v4f32")
    __vector(i8[16]) neon_vpadals_v16i8_v4f32(__vector(i8[16]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadals.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v16i8.v2f64")
    __vector(i8[16]) neon_vpadals_v16i8_v2f64(__vector(i8[16]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadals.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v8i16.v16i8")
    __vector(i16[8]) neon_vpadals_v8i16_v16i8(__vector(i16[8]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadals.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v8i16.v8i16")
    __vector(i16[8]) neon_vpadals_v8i16_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadals.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v8i16.v4i32")
    __vector(i16[8]) neon_vpadals_v8i16_v4i32(__vector(i16[8]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadals.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v8i16.v2i64")
    __vector(i16[8]) neon_vpadals_v8i16_v2i64(__vector(i16[8]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadals.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v8i16.v4f32")
    __vector(i16[8]) neon_vpadals_v8i16_v4f32(__vector(i16[8]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadals.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v8i16.v2f64")
    __vector(i16[8]) neon_vpadals_v8i16_v2f64(__vector(i16[8]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadals.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4i32.v16i8")
    __vector(i32[4]) neon_vpadals_v4i32_v16i8(__vector(i32[4]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadals.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4i32.v8i16")
    __vector(i32[4]) neon_vpadals_v4i32_v8i16(__vector(i32[4]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadals.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4i32.v4i32")
    __vector(i32[4]) neon_vpadals_v4i32_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadals.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4i32.v2i64")
    __vector(i32[4]) neon_vpadals_v4i32_v2i64(__vector(i32[4]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadals.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4i32.v4f32")
    __vector(i32[4]) neon_vpadals_v4i32_v4f32(__vector(i32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadals.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4i32.v2f64")
    __vector(i32[4]) neon_vpadals_v4i32_v2f64(__vector(i32[4]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadals.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2i64.v16i8")
    __vector(i64[2]) neon_vpadals_v2i64_v16i8(__vector(i64[2]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadals.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2i64.v8i16")
    __vector(i64[2]) neon_vpadals_v2i64_v8i16(__vector(i64[2]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadals.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2i64.v4i32")
    __vector(i64[2]) neon_vpadals_v2i64_v4i32(__vector(i64[2]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadals.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2i64.v2i64")
    __vector(i64[2]) neon_vpadals_v2i64_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadals.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2i64.v4f32")
    __vector(i64[2]) neon_vpadals_v2i64_v4f32(__vector(i64[2]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadals.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2i64.v2f64")
    __vector(i64[2]) neon_vpadals_v2i64_v2f64(__vector(i64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadals.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4f32.v16i8")
    __vector(f32[4]) neon_vpadals_v4f32_v16i8(__vector(f32[4]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadals.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4f32.v8i16")
    __vector(f32[4]) neon_vpadals_v4f32_v8i16(__vector(f32[4]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadals.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4f32.v4i32")
    __vector(f32[4]) neon_vpadals_v4f32_v4i32(__vector(f32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadals.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4f32.v2i64")
    __vector(f32[4]) neon_vpadals_v4f32_v2i64(__vector(f32[4]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadals.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4f32.v4f32")
    __vector(f32[4]) neon_vpadals_v4f32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadals.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v4f32.v2f64")
    __vector(f32[4]) neon_vpadals_v4f32_v2f64(__vector(f32[4]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadals.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2f64.v16i8")
    __vector(f64[2]) neon_vpadals_v2f64_v16i8(__vector(f64[2]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadals.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2f64.v8i16")
    __vector(f64[2]) neon_vpadals_v2f64_v8i16(__vector(f64[2]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadals.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2f64.v4i32")
    __vector(f64[2]) neon_vpadals_v2f64_v4i32(__vector(f64[2]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadals.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2f64.v2i64")
    __vector(f64[2]) neon_vpadals_v2f64_v2i64(__vector(f64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadals.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2f64.v4f32")
    __vector(f64[2]) neon_vpadals_v2f64_v4f32(__vector(f64[2]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadals.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadals.v2f64.v2f64")
    __vector(f64[2]) neon_vpadals_v2f64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadalu.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v16i8.v16i8")
    __vector(i8[16]) neon_vpadalu_v16i8_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadalu.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v16i8.v8i16")
    __vector(i8[16]) neon_vpadalu_v16i8_v8i16(__vector(i8[16]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadalu.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v16i8.v4i32")
    __vector(i8[16]) neon_vpadalu_v16i8_v4i32(__vector(i8[16]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadalu.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v16i8.v2i64")
    __vector(i8[16]) neon_vpadalu_v16i8_v2i64(__vector(i8[16]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadalu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v16i8.v4f32")
    __vector(i8[16]) neon_vpadalu_v16i8_v4f32(__vector(i8[16]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadalu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v16i8.v2f64")
    __vector(i8[16]) neon_vpadalu_v16i8_v2f64(__vector(i8[16]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadalu.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v8i16.v16i8")
    __vector(i16[8]) neon_vpadalu_v8i16_v16i8(__vector(i16[8]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadalu.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v8i16.v8i16")
    __vector(i16[8]) neon_vpadalu_v8i16_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadalu.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v8i16.v4i32")
    __vector(i16[8]) neon_vpadalu_v8i16_v4i32(__vector(i16[8]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadalu.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v8i16.v2i64")
    __vector(i16[8]) neon_vpadalu_v8i16_v2i64(__vector(i16[8]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadalu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v8i16.v4f32")
    __vector(i16[8]) neon_vpadalu_v8i16_v4f32(__vector(i16[8]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadalu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v8i16.v2f64")
    __vector(i16[8]) neon_vpadalu_v8i16_v2f64(__vector(i16[8]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadalu.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4i32.v16i8")
    __vector(i32[4]) neon_vpadalu_v4i32_v16i8(__vector(i32[4]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadalu.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4i32.v8i16")
    __vector(i32[4]) neon_vpadalu_v4i32_v8i16(__vector(i32[4]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadalu.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4i32.v4i32")
    __vector(i32[4]) neon_vpadalu_v4i32_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadalu.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4i32.v2i64")
    __vector(i32[4]) neon_vpadalu_v4i32_v2i64(__vector(i32[4]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadalu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4i32.v4f32")
    __vector(i32[4]) neon_vpadalu_v4i32_v4f32(__vector(i32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadalu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4i32.v2f64")
    __vector(i32[4]) neon_vpadalu_v4i32_v2f64(__vector(i32[4]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadalu.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2i64.v16i8")
    __vector(i64[2]) neon_vpadalu_v2i64_v16i8(__vector(i64[2]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadalu.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2i64.v8i16")
    __vector(i64[2]) neon_vpadalu_v2i64_v8i16(__vector(i64[2]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadalu.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2i64.v4i32")
    __vector(i64[2]) neon_vpadalu_v2i64_v4i32(__vector(i64[2]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadalu.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2i64.v2i64")
    __vector(i64[2]) neon_vpadalu_v2i64_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadalu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2i64.v4f32")
    __vector(i64[2]) neon_vpadalu_v2i64_v4f32(__vector(i64[2]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadalu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2i64.v2f64")
    __vector(i64[2]) neon_vpadalu_v2i64_v2f64(__vector(i64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadalu.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4f32.v16i8")
    __vector(f32[4]) neon_vpadalu_v4f32_v16i8(__vector(f32[4]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadalu.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4f32.v8i16")
    __vector(f32[4]) neon_vpadalu_v4f32_v8i16(__vector(f32[4]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadalu.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4f32.v4i32")
    __vector(f32[4]) neon_vpadalu_v4f32_v4i32(__vector(f32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadalu.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4f32.v2i64")
    __vector(f32[4]) neon_vpadalu_v4f32_v2i64(__vector(f32[4]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadalu.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4f32.v4f32")
    __vector(f32[4]) neon_vpadalu_v4f32_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadalu.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v4f32.v2f64")
    __vector(f32[4]) neon_vpadalu_v4f32_v2f64(__vector(f32[4]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpadalu.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2f64.v16i8")
    __vector(f64[2]) neon_vpadalu_v2f64_v16i8(__vector(f64[2]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpadalu.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2f64.v8i16")
    __vector(f64[2]) neon_vpadalu_v2f64_v8i16(__vector(f64[2]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpadalu.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2f64.v4i32")
    __vector(f64[2]) neon_vpadalu_v2f64_v4i32(__vector(f64[2]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpadalu.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2f64.v2i64")
    __vector(f64[2]) neon_vpadalu_v2f64_v2i64(__vector(f64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpadalu.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2f64.v4f32")
    __vector(f64[2]) neon_vpadalu_v2f64_v4f32(__vector(f64[2]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpadalu.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpadalu.v2f64.v2f64")
    __vector(f64[2]) neon_vpadalu_v2f64_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpmaxs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxs.v16i8")
    __vector(i8[16]) neon_vpmaxs_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpmaxs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxs.v8i16")
    __vector(i16[8]) neon_vpmaxs_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpmaxs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxs.v4i32")
    __vector(i32[4]) neon_vpmaxs_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpmaxs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxs.v2i64")
    __vector(i64[2]) neon_vpmaxs_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpmaxs.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxs.v4f32")
    __vector(f32[4]) neon_vpmaxs_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpmaxs.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxs.v2f64")
    __vector(f64[2]) neon_vpmaxs_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpmaxu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxu.v16i8")
    __vector(i8[16]) neon_vpmaxu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpmaxu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxu.v8i16")
    __vector(i16[8]) neon_vpmaxu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpmaxu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxu.v4i32")
    __vector(i32[4]) neon_vpmaxu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpmaxu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxu.v2i64")
    __vector(i64[2]) neon_vpmaxu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpmaxu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxu.v4f32")
    __vector(f32[4]) neon_vpmaxu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpmaxu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmaxu.v2f64")
    __vector(f64[2]) neon_vpmaxu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpmins.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmins.v16i8")
    __vector(i8[16]) neon_vpmins_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpmins.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmins.v8i16")
    __vector(i16[8]) neon_vpmins_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpmins.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmins.v4i32")
    __vector(i32[4]) neon_vpmins_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpmins.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmins.v2i64")
    __vector(i64[2]) neon_vpmins_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpmins.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmins.v4f32")
    __vector(f32[4]) neon_vpmins_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpmins.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpmins.v2f64")
    __vector(f64[2]) neon_vpmins_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vpminu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpminu.v16i8")
    __vector(i8[16]) neon_vpminu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vpminu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpminu.v8i16")
    __vector(i16[8]) neon_vpminu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vpminu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpminu.v4i32")
    __vector(i32[4]) neon_vpminu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vpminu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpminu.v2i64")
    __vector(i64[2]) neon_vpminu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vpminu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpminu.v4f32")
    __vector(f32[4]) neon_vpminu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vpminu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vpminu.v2f64")
    __vector(f64[2]) neon_vpminu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vshifts.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshifts.v16i8")
    __vector(i8[16]) neon_vshifts_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vshifts.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshifts.v8i16")
    __vector(i16[8]) neon_vshifts_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vshifts.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshifts.v4i32")
    __vector(i32[4]) neon_vshifts_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vshifts.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshifts.v2i64")
    __vector(i64[2]) neon_vshifts_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vshifts.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshifts.v4f32")
    __vector(f32[4]) neon_vshifts_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vshifts.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshifts.v2f64")
    __vector(f64[2]) neon_vshifts_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vshiftu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftu.v16i8")
    __vector(i8[16]) neon_vshiftu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vshiftu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftu.v8i16")
    __vector(i16[8]) neon_vshiftu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vshiftu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftu.v4i32")
    __vector(i32[4]) neon_vshiftu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vshiftu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftu.v2i64")
    __vector(i64[2]) neon_vshiftu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vshiftu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftu.v4f32")
    __vector(f32[4]) neon_vshiftu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vshiftu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftu.v2f64")
    __vector(f64[2]) neon_vshiftu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vrshifts.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshifts.v16i8")
    __vector(i8[16]) neon_vrshifts_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vrshifts.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshifts.v8i16")
    __vector(i16[8]) neon_vrshifts_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vrshifts.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshifts.v4i32")
    __vector(i32[4]) neon_vrshifts_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vrshifts.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshifts.v2i64")
    __vector(i64[2]) neon_vrshifts_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vrshifts.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshifts.v4f32")
    __vector(f32[4]) neon_vrshifts_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vrshifts.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshifts.v2f64")
    __vector(f64[2]) neon_vrshifts_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vrshiftu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshiftu.v16i8")
    __vector(i8[16]) neon_vrshiftu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vrshiftu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshiftu.v8i16")
    __vector(i16[8]) neon_vrshiftu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vrshiftu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshiftu.v4i32")
    __vector(i32[4]) neon_vrshiftu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vrshiftu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshiftu.v2i64")
    __vector(i64[2]) neon_vrshiftu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vrshiftu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshiftu.v4f32")
    __vector(f32[4]) neon_vrshiftu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vrshiftu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrshiftu.v2f64")
    __vector(f64[2]) neon_vrshiftu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqshifts.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshifts.v16i8")
    __vector(i8[16]) neon_vqshifts_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqshifts.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshifts.v8i16")
    __vector(i16[8]) neon_vqshifts_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqshifts.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshifts.v4i32")
    __vector(i32[4]) neon_vqshifts_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqshifts.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshifts.v2i64")
    __vector(i64[2]) neon_vqshifts_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqshifts.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshifts.v4f32")
    __vector(f32[4]) neon_vqshifts_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqshifts.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshifts.v2f64")
    __vector(f64[2]) neon_vqshifts_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqshiftu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftu.v16i8")
    __vector(i8[16]) neon_vqshiftu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqshiftu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftu.v8i16")
    __vector(i16[8]) neon_vqshiftu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqshiftu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftu.v4i32")
    __vector(i32[4]) neon_vqshiftu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqshiftu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftu.v2i64")
    __vector(i64[2]) neon_vqshiftu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqshiftu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftu.v4f32")
    __vector(f32[4]) neon_vqshiftu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqshiftu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftu.v2f64")
    __vector(f64[2]) neon_vqshiftu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqshiftsu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftsu.v16i8")
    __vector(i8[16]) neon_vqshiftsu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqshiftsu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftsu.v8i16")
    __vector(i16[8]) neon_vqshiftsu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqshiftsu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftsu.v4i32")
    __vector(i32[4]) neon_vqshiftsu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqshiftsu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftsu.v2i64")
    __vector(i64[2]) neon_vqshiftsu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqshiftsu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftsu.v4f32")
    __vector(f32[4]) neon_vqshiftsu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqshiftsu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqshiftsu.v2f64")
    __vector(f64[2]) neon_vqshiftsu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqrshifts.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshifts.v16i8")
    __vector(i8[16]) neon_vqrshifts_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqrshifts.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshifts.v8i16")
    __vector(i16[8]) neon_vqrshifts_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqrshifts.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshifts.v4i32")
    __vector(i32[4]) neon_vqrshifts_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqrshifts.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshifts.v2i64")
    __vector(i64[2]) neon_vqrshifts_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqrshifts.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshifts.v4f32")
    __vector(f32[4]) neon_vqrshifts_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqrshifts.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshifts.v2f64")
    __vector(f64[2]) neon_vqrshifts_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vqrshiftu.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshiftu.v16i8")
    __vector(i8[16]) neon_vqrshiftu_v16i8(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vqrshiftu.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshiftu.v8i16")
    __vector(i16[8]) neon_vqrshiftu_v8i16(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vqrshiftu.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshiftu.v4i32")
    __vector(i32[4]) neon_vqrshiftu_v4i32(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vqrshiftu.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshiftu.v2i64")
    __vector(i64[2]) neon_vqrshiftu_v2i64(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vqrshiftu.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshiftu.v4f32")
    __vector(f32[4]) neon_vqrshiftu_v4f32(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vqrshiftu.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqrshiftu.v2f64")
    __vector(f64[2]) neon_vqrshiftu_v2f64(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vshiftins.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftins.v16i8")
    __vector(i8[16]) neon_vshiftins_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vshiftins.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftins.v8i16")
    __vector(i16[8]) neon_vshiftins_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vshiftins.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftins.v4i32")
    __vector(i32[4]) neon_vshiftins_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vshiftins.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftins.v2i64")
    __vector(i64[2]) neon_vshiftins_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vshiftins.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftins.v4f32")
    __vector(f32[4]) neon_vshiftins_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vshiftins.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vshiftins.v2f64")
    __vector(f64[2]) neon_vshiftins_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.vabs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabs.v16i8")
    __vector(i8[16]) neon_vabs_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vabs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabs.v8i16")
    __vector(i16[8]) neon_vabs_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vabs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabs.v4i32")
    __vector(i32[4]) neon_vabs_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vabs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabs.v2i64")
    __vector(i64[2]) neon_vabs_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vabs.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabs.v4f32")
    __vector(f32[4]) neon_vabs_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vabs.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vabs.v2f64")
    __vector(f64[2]) neon_vabs_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vqabs.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqabs.v16i8")
    __vector(i8[16]) neon_vqabs_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vqabs.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqabs.v8i16")
    __vector(i16[8]) neon_vqabs_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vqabs.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqabs.v4i32")
    __vector(i32[4]) neon_vqabs_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vqabs.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqabs.v2i64")
    __vector(i64[2]) neon_vqabs_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vqabs.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqabs.v4f32")
    __vector(f32[4]) neon_vqabs_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vqabs.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqabs.v2f64")
    __vector(f64[2]) neon_vqabs_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vqneg.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqneg.v16i8")
    __vector(i8[16]) neon_vqneg_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vqneg.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqneg.v8i16")
    __vector(i16[8]) neon_vqneg_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vqneg.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqneg.v4i32")
    __vector(i32[4]) neon_vqneg_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vqneg.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqneg.v2i64")
    __vector(i64[2]) neon_vqneg_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vqneg.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqneg.v4f32")
    __vector(f32[4]) neon_vqneg_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vqneg.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vqneg.v2f64")
    __vector(f64[2]) neon_vqneg_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcls.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcls.v16i8")
    __vector(i8[16]) neon_vcls_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcls.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcls.v8i16")
    __vector(i16[8]) neon_vcls_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcls.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcls.v4i32")
    __vector(i32[4]) neon_vcls_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcls.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcls.v2i64")
    __vector(i64[2]) neon_vcls_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcls.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcls.v4f32")
    __vector(f32[4]) neon_vcls_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcls.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcls.v2f64")
    __vector(f64[2]) neon_vcls_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vclz.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vclz.v16i8")
    __vector(i8[16]) neon_vclz_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vclz.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vclz.v8i16")
    __vector(i16[8]) neon_vclz_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vclz.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vclz.v4i32")
    __vector(i32[4]) neon_vclz_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vclz.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vclz.v2i64")
    __vector(i64[2]) neon_vclz_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vclz.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vclz.v4f32")
    __vector(f32[4]) neon_vclz_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vclz.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vclz.v2f64")
    __vector(f64[2]) neon_vclz_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcnt.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcnt.v16i8")
    __vector(i8[16]) neon_vcnt_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcnt.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcnt.v8i16")
    __vector(i16[8]) neon_vcnt_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcnt.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcnt.v4i32")
    __vector(i32[4]) neon_vcnt_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcnt.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcnt.v2i64")
    __vector(i64[2]) neon_vcnt_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcnt.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcnt.v4f32")
    __vector(f32[4]) neon_vcnt_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcnt.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcnt.v2f64")
    __vector(f64[2]) neon_vcnt_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrecpe.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecpe.v16i8")
    __vector(i8[16]) neon_vrecpe_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrecpe.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecpe.v8i16")
    __vector(i16[8]) neon_vrecpe_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrecpe.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecpe.v4i32")
    __vector(i32[4]) neon_vrecpe_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrecpe.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecpe.v2i64")
    __vector(i64[2]) neon_vrecpe_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrecpe.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecpe.v4f32")
    __vector(f32[4]) neon_vrecpe_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrecpe.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrecpe.v2f64")
    __vector(f64[2]) neon_vrecpe_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrsqrte.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrte.v16i8")
    __vector(i8[16]) neon_vrsqrte_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrsqrte.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrte.v8i16")
    __vector(i16[8]) neon_vrsqrte_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrsqrte.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrte.v4i32")
    __vector(i32[4]) neon_vrsqrte_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrsqrte.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrte.v2i64")
    __vector(i64[2]) neon_vrsqrte_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrsqrte.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrte.v4f32")
    __vector(f32[4]) neon_vrsqrte_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrsqrte.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrsqrte.v2f64")
    __vector(f64[2]) neon_vrsqrte_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtau.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtau_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtau.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtau_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtau.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtau_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtau.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtau_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtau.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtau_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtau.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtau_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtau.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtau_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtau.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtau_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtau.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtau_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtau.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtau_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtau.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtau_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtau.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtau_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtau.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtau_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtau.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtau_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtau.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtau_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtau.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtau_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtau.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtau_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtau.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtau_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtau.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtau_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtau.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtau_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtau.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtau_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtau.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtau_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtau.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtau_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtau.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtau_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtau.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtau_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtau.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtau_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtau.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtau_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtau.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtau_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtau.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtau_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtau.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtau_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtau.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtau_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtau.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtau_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtau.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtau_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtau.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtau_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtau.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtau_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtau.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtau.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtau_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtas.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtas_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtas.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtas_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtas.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtas_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtas.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtas_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtas.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtas_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtas.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtas_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtas.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtas_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtas.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtas_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtas.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtas_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtas.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtas_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtas.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtas_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtas.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtas_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtas.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtas_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtas.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtas_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtas.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtas_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtas.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtas_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtas.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtas_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtas.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtas_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtas.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtas_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtas.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtas_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtas.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtas_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtas.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtas_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtas.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtas_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtas.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtas_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtas.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtas_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtas.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtas_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtas.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtas_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtas.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtas_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtas.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtas_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtas.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtas_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtas.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtas_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtas.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtas_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtas.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtas_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtas.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtas_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtas.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtas_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtas.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtas.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtas_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtnu.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtnu_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtnu.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtnu_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtnu.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtnu_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtnu.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtnu_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtnu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtnu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtnu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtnu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtnu.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtnu_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtnu.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtnu_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtnu.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtnu_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtnu.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtnu_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtnu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtnu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtnu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtnu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtnu.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtnu_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtnu.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtnu_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtnu.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtnu_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtnu.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtnu_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtnu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtnu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtnu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtnu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtnu.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtnu_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtnu.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtnu_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtnu.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtnu_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtnu.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtnu_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtnu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtnu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtnu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtnu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtnu.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtnu_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtnu.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtnu_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtnu.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtnu_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtnu.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtnu_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtnu.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtnu_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtnu.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtnu_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtnu.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtnu_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtnu.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtnu_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtnu.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtnu_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtnu.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtnu_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtnu.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtnu_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtnu.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtnu.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtnu_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtns.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtns_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtns.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtns_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtns.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtns_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtns.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtns_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtns.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtns_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtns.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtns_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtns.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtns_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtns.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtns_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtns.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtns_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtns.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtns_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtns.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtns_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtns.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtns_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtns.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtns_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtns.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtns_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtns.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtns_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtns.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtns_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtns.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtns_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtns.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtns_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtns.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtns_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtns.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtns_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtns.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtns_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtns.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtns_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtns.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtns_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtns.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtns_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtns.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtns_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtns.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtns_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtns.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtns_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtns.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtns_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtns.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtns_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtns.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtns_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtns.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtns_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtns.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtns_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtns.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtns_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtns.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtns_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtns.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtns_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtns.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtns.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtns_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtpu.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtpu_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtpu.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtpu_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtpu.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtpu_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtpu.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtpu_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtpu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtpu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtpu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtpu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtpu.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtpu_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtpu.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtpu_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtpu.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtpu_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtpu.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtpu_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtpu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtpu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtpu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtpu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtpu.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtpu_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtpu.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtpu_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtpu.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtpu_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtpu.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtpu_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtpu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtpu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtpu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtpu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtpu.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtpu_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtpu.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtpu_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtpu.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtpu_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtpu.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtpu_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtpu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtpu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtpu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtpu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtpu.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtpu_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtpu.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtpu_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtpu.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtpu_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtpu.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtpu_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtpu.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtpu_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtpu.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtpu_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtpu.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtpu_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtpu.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtpu_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtpu.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtpu_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtpu.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtpu_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtpu.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtpu_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtpu.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtpu.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtpu_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtps.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtps_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtps.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtps_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtps.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtps_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtps.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtps_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtps.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtps_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtps.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtps_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtps.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtps_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtps.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtps_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtps.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtps_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtps.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtps_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtps.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtps_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtps.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtps_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtps.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtps_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtps.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtps_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtps.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtps_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtps.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtps_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtps.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtps_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtps.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtps_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtps.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtps_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtps.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtps_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtps.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtps_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtps.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtps_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtps.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtps_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtps.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtps_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtps.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtps_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtps.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtps_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtps.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtps_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtps.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtps_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtps.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtps_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtps.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtps_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtps.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtps_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtps.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtps_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtps.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtps_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtps.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtps_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtps.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtps_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtps.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtps.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtps_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtmu.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtmu_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtmu.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtmu_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtmu.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtmu_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtmu.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtmu_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtmu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtmu_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtmu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtmu_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtmu.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtmu_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtmu.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtmu_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtmu.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtmu_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtmu.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtmu_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtmu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtmu_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtmu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtmu_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtmu.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtmu_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtmu.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtmu_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtmu.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtmu_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtmu.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtmu_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtmu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtmu_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtmu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtmu_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtmu.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtmu_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtmu.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtmu_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtmu.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtmu_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtmu.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtmu_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtmu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtmu_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtmu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtmu_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtmu.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtmu_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtmu.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtmu_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtmu.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtmu_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtmu.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtmu_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtmu.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtmu_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtmu.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtmu_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtmu.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtmu_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtmu.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtmu_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtmu.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtmu_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtmu.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtmu_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtmu.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtmu_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtmu.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtmu.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtmu_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtms.v16i8.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v16i8.v16i8")
    __vector(i8[16]) neon_vcvtms_v16i8_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtms.v16i8.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v16i8.v8i16")
    __vector(i8[16]) neon_vcvtms_v16i8_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtms.v16i8.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v16i8.v4i32")
    __vector(i8[16]) neon_vcvtms_v16i8_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtms.v16i8.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v16i8.v2i64")
    __vector(i8[16]) neon_vcvtms_v16i8_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtms.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtms_v16i8_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtms.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtms_v16i8_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtms.v8i16.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v8i16.v16i8")
    __vector(i16[8]) neon_vcvtms_v8i16_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtms.v8i16.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v8i16.v8i16")
    __vector(i16[8]) neon_vcvtms_v8i16_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtms.v8i16.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v8i16.v4i32")
    __vector(i16[8]) neon_vcvtms_v8i16_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtms.v8i16.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v8i16.v2i64")
    __vector(i16[8]) neon_vcvtms_v8i16_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtms.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtms_v8i16_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtms.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtms_v8i16_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtms.v4i32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4i32.v16i8")
    __vector(i32[4]) neon_vcvtms_v4i32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtms.v4i32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4i32.v8i16")
    __vector(i32[4]) neon_vcvtms_v4i32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtms.v4i32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4i32.v4i32")
    __vector(i32[4]) neon_vcvtms_v4i32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtms.v4i32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4i32.v2i64")
    __vector(i32[4]) neon_vcvtms_v4i32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtms.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtms_v4i32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtms.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtms_v4i32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtms.v2i64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2i64.v16i8")
    __vector(i64[2]) neon_vcvtms_v2i64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtms.v2i64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2i64.v8i16")
    __vector(i64[2]) neon_vcvtms_v2i64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtms.v2i64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2i64.v4i32")
    __vector(i64[2]) neon_vcvtms_v2i64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtms.v2i64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2i64.v2i64")
    __vector(i64[2]) neon_vcvtms_v2i64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtms.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtms_v2i64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtms.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtms_v2i64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtms.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtms_v4f32_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtms.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtms_v4f32_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtms.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtms_v4f32_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtms.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtms_v4f32_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtms.v4f32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4f32.v4f32")
    __vector(f32[4]) neon_vcvtms_v4f32_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtms.v4f32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v4f32.v2f64")
    __vector(f32[4]) neon_vcvtms_v4f32_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtms.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtms_v2f64_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vcvtms.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtms_v2f64_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vcvtms.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtms_v2f64_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vcvtms.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtms_v2f64_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vcvtms.v2f64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2f64.v4f32")
    __vector(f64[2]) neon_vcvtms_v2f64_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvtms.v2f64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtms.v2f64.v2f64")
    __vector(f64[2]) neon_vcvtms_v2f64_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vcvtfp2fxs.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v16i8.f32")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v16i8.f64")
    __vector(i8[16]) neon_vcvtfp2fxs_v16i8_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i8.v4f32")
    i8 neon_vcvtfp2fxs_i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i8.f32")
    i8 neon_vcvtfp2fxs_i8_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i8.v2f64")
    i8 neon_vcvtfp2fxs_i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i8.f64")
    i8 neon_vcvtfp2fxs_i8_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v8i16.f32")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v8i16.f64")
    __vector(i16[8]) neon_vcvtfp2fxs_v8i16_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i16.v4f32")
    i16 neon_vcvtfp2fxs_i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i16.f32")
    i16 neon_vcvtfp2fxs_i16_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i16.v2f64")
    i16 neon_vcvtfp2fxs_i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i16.f64")
    i16 neon_vcvtfp2fxs_i16_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v4i32.f32")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v4i32.f64")
    __vector(i32[4]) neon_vcvtfp2fxs_v4i32_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i32.v4f32")
    i32 neon_vcvtfp2fxs_i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i32.f32")
    i32 neon_vcvtfp2fxs_i32_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i32.v2f64")
    i32 neon_vcvtfp2fxs_i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i32.f64")
    i32 neon_vcvtfp2fxs_i32_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v2i64.f32")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.v2i64.f64")
    __vector(i64[2]) neon_vcvtfp2fxs_v2i64_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i64.v4f32")
    i64 neon_vcvtfp2fxs_i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i64.f32")
    i64 neon_vcvtfp2fxs_i64_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i64.v2f64")
    i64 neon_vcvtfp2fxs_i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxs.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxs.i64.f64")
    i64 neon_vcvtfp2fxs_i64_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v16i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v16i8.v4f32")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v16i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v16i8.f32")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v16i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v16i8.v2f64")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v16i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v16i8.f64")
    __vector(i8[16]) neon_vcvtfp2fxu_v16i8_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i8.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i8.v4f32")
    i8 neon_vcvtfp2fxu_i8_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i8.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i8.f32")
    i8 neon_vcvtfp2fxu_i8_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i8.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i8.v2f64")
    i8 neon_vcvtfp2fxu_i8_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i8.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i8.f64")
    i8 neon_vcvtfp2fxu_i8_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v8i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v8i16.v4f32")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v8i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v8i16.f32")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v8i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v8i16.v2f64")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v8i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v8i16.f64")
    __vector(i16[8]) neon_vcvtfp2fxu_v8i16_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i16.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i16.v4f32")
    i16 neon_vcvtfp2fxu_i16_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i16.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i16.f32")
    i16 neon_vcvtfp2fxu_i16_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i16.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i16.v2f64")
    i16 neon_vcvtfp2fxu_i16_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i16.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i16.f64")
    i16 neon_vcvtfp2fxu_i16_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v4i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v4i32.v4f32")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v4i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v4i32.f32")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v4i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v4i32.v2f64")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v4i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v4i32.f64")
    __vector(i32[4]) neon_vcvtfp2fxu_v4i32_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i32.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i32.v4f32")
    i32 neon_vcvtfp2fxu_i32_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i32.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i32.f32")
    i32 neon_vcvtfp2fxu_i32_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i32.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i32.v2f64")
    i32 neon_vcvtfp2fxu_i32_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i32.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i32.f64")
    i32 neon_vcvtfp2fxu_i32_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v2i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v2i64.v4f32")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v2i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v2i64.f32")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v2i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v2i64.v2f64")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.v2i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.v2i64.f64")
    __vector(i64[2]) neon_vcvtfp2fxu_v2i64_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i64.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i64.v4f32")
    i64 neon_vcvtfp2fxu_i64_v4f32(__vector(f32[4]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i64.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i64.f32")
    i64 neon_vcvtfp2fxu_i64_f32(f32, i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i64.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i64.v2f64")
    i64 neon_vcvtfp2fxu_i64_v2f64(__vector(f64[2]), i32);
    /// The `llvm.arm.neon.vcvtfp2fxu.i64.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2fxu.i64.f64")
    i64 neon_vcvtfp2fxu_i64_f64(f64, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.i8")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.i16")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.i32")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v4f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v4f32.i64")
    __vector(f32[4]) neon_vcvtfxs2fp_v4f32_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.v16i8")
    f32 neon_vcvtfxs2fp_f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.i8")
    f32 neon_vcvtfxs2fp_f32_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.v8i16")
    f32 neon_vcvtfxs2fp_f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.i16")
    f32 neon_vcvtfxs2fp_f32_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.v4i32")
    f32 neon_vcvtfxs2fp_f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.i32")
    f32 neon_vcvtfxs2fp_f32_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.v2i64")
    f32 neon_vcvtfxs2fp_f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f32.i64")
    f32 neon_vcvtfxs2fp_f32_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.i8")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.i16")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.i32")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.v2f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.v2f64.i64")
    __vector(f64[2]) neon_vcvtfxs2fp_v2f64_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.v16i8")
    f64 neon_vcvtfxs2fp_f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.i8")
    f64 neon_vcvtfxs2fp_f64_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.v8i16")
    f64 neon_vcvtfxs2fp_f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.i16")
    f64 neon_vcvtfxs2fp_f64_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.v4i32")
    f64 neon_vcvtfxs2fp_f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.i32")
    f64 neon_vcvtfxs2fp_f64_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.v2i64")
    f64 neon_vcvtfxs2fp_f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxs2fp.f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxs2fp.f64.i64")
    f64 neon_vcvtfxs2fp_f64_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.v16i8")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.i8")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.v8i16")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.i16")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.v4i32")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.i32")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.v2i64")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v4f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v4f32.i64")
    __vector(f32[4]) neon_vcvtfxu2fp_v4f32_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.v16i8")
    f32 neon_vcvtfxu2fp_f32_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.i8")
    f32 neon_vcvtfxu2fp_f32_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.v8i16")
    f32 neon_vcvtfxu2fp_f32_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.i16")
    f32 neon_vcvtfxu2fp_f32_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.v4i32")
    f32 neon_vcvtfxu2fp_f32_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.i32")
    f32 neon_vcvtfxu2fp_f32_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.v2i64")
    f32 neon_vcvtfxu2fp_f32_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f32.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f32.i64")
    f32 neon_vcvtfxu2fp_f32_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.v16i8")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.i8")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.v8i16")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.i16")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.v4i32")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.i32")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.v2i64")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.v2f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.v2f64.i64")
    __vector(f64[2]) neon_vcvtfxu2fp_v2f64_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.v16i8")
    f64 neon_vcvtfxu2fp_f64_v16i8(__vector(i8[16]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.i8")
    f64 neon_vcvtfxu2fp_f64_i8(i8, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.v8i16")
    f64 neon_vcvtfxu2fp_f64_v8i16(__vector(i16[8]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.i16")
    f64 neon_vcvtfxu2fp_f64_i16(i16, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.v4i32")
    f64 neon_vcvtfxu2fp_f64_v4i32(__vector(i32[4]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.i32")
    f64 neon_vcvtfxu2fp_f64_i32(i32, i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.v2i64")
    f64 neon_vcvtfxu2fp_f64_v2i64(__vector(i64[2]), i32);
    /// The `llvm.arm.neon.vcvtfxu2fp.f64.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfxu2fp.f64.i64")
    f64 neon_vcvtfxu2fp_f64_i64(i64, i32);
    /// The `llvm.arm.neon.vcvtfp2hf` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvtfp2hf")
    __vector(i16[4]) neon_vcvtfp2hf(__vector(f32[4]));
    /// The `llvm.arm.neon.vcvthf2fp` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vcvthf2fp")
    __vector(f32[4]) neon_vcvthf2fp(__vector(i16[4]));
    /// The `llvm.arm.neon.vtbl1` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbl1")
    __vector(i8[8]) neon_vtbl1(__vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbl2` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbl2")
    __vector(i8[8]) neon_vtbl2(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbl3` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbl3")
    __vector(i8[8]) neon_vtbl3(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbl4` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbl4")
    __vector(i8[8]) neon_vtbl4(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbx1` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbx1")
    __vector(i8[8]) neon_vtbx1(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbx2` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbx2")
    __vector(i8[8]) neon_vtbx2(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbx3` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbx3")
    __vector(i8[8]) neon_vtbx3(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vtbx4` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vtbx4")
    __vector(i8[8]) neon_vtbx4(__vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]), __vector(i8[8]));
    /// The `llvm.arm.neon.vrintn.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintn.v16i8")
    __vector(i8[16]) neon_vrintn_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrintn.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintn.v8i16")
    __vector(i16[8]) neon_vrintn_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrintn.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintn.v4i32")
    __vector(i32[4]) neon_vrintn_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrintn.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintn.v2i64")
    __vector(i64[2]) neon_vrintn_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrintn.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintn.v4f32")
    __vector(f32[4]) neon_vrintn_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrintn.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintn.v2f64")
    __vector(f64[2]) neon_vrintn_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrintx.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintx.v16i8")
    __vector(i8[16]) neon_vrintx_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrintx.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintx.v8i16")
    __vector(i16[8]) neon_vrintx_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrintx.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintx.v4i32")
    __vector(i32[4]) neon_vrintx_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrintx.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintx.v2i64")
    __vector(i64[2]) neon_vrintx_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrintx.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintx.v4f32")
    __vector(f32[4]) neon_vrintx_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrintx.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintx.v2f64")
    __vector(f64[2]) neon_vrintx_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrinta.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrinta.v16i8")
    __vector(i8[16]) neon_vrinta_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrinta.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrinta.v8i16")
    __vector(i16[8]) neon_vrinta_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrinta.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrinta.v4i32")
    __vector(i32[4]) neon_vrinta_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrinta.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrinta.v2i64")
    __vector(i64[2]) neon_vrinta_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrinta.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrinta.v4f32")
    __vector(f32[4]) neon_vrinta_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrinta.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrinta.v2f64")
    __vector(f64[2]) neon_vrinta_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrintz.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintz.v16i8")
    __vector(i8[16]) neon_vrintz_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrintz.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintz.v8i16")
    __vector(i16[8]) neon_vrintz_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrintz.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintz.v4i32")
    __vector(i32[4]) neon_vrintz_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrintz.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintz.v2i64")
    __vector(i64[2]) neon_vrintz_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrintz.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintz.v4f32")
    __vector(f32[4]) neon_vrintz_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrintz.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintz.v2f64")
    __vector(f64[2]) neon_vrintz_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrintm.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintm.v16i8")
    __vector(i8[16]) neon_vrintm_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrintm.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintm.v8i16")
    __vector(i16[8]) neon_vrintm_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrintm.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintm.v4i32")
    __vector(i32[4]) neon_vrintm_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrintm.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintm.v2i64")
    __vector(i64[2]) neon_vrintm_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrintm.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintm.v4f32")
    __vector(f32[4]) neon_vrintm_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrintm.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintm.v2f64")
    __vector(f64[2]) neon_vrintm_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vrintp.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintp.v16i8")
    __vector(i8[16]) neon_vrintp_v16i8(__vector(i8[16]));
    /// The `llvm.arm.neon.vrintp.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintp.v8i16")
    __vector(i16[8]) neon_vrintp_v8i16(__vector(i16[8]));
    /// The `llvm.arm.neon.vrintp.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintp.v4i32")
    __vector(i32[4]) neon_vrintp_v4i32(__vector(i32[4]));
    /// The `llvm.arm.neon.vrintp.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintp.v2i64")
    __vector(i64[2]) neon_vrintp_v2i64(__vector(i64[2]));
    /// The `llvm.arm.neon.vrintp.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintp.v4f32")
    __vector(f32[4]) neon_vrintp_v4f32(__vector(f32[4]));
    /// The `llvm.arm.neon.vrintp.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vrintp.v2f64")
    __vector(f64[2]) neon_vrintp_v2f64(__vector(f64[2]));
    /// The `llvm.arm.neon.vld1.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vld1.v16i8")
    __vector(i8[16]) neon_vld1_v16i8(i8*, i32);
    /// The `llvm.arm.neon.vld1.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vld1.v8i16")
    __vector(i16[8]) neon_vld1_v8i16(i8*, i32);
    /// The `llvm.arm.neon.vld1.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vld1.v4i32")
    __vector(i32[4]) neon_vld1_v4i32(i8*, i32);
    /// The `llvm.arm.neon.vld1.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vld1.v2i64")
    __vector(i64[2]) neon_vld1_v2i64(i8*, i32);
    /// The `llvm.arm.neon.vld1.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vld1.v4f32")
    __vector(f32[4]) neon_vld1_v4f32(i8*, i32);
    /// The `llvm.arm.neon.vld1.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vld1.v2f64")
    __vector(f64[2]) neon_vld1_v2f64(i8*, i32);
    /// The `llvm.arm.neon.vst1.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst1.v16i8")
    void neon_vst1_v16i8(i8*, __vector(i8[16]), i32);
    /// The `llvm.arm.neon.vst1.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst1.v8i16")
    void neon_vst1_v8i16(i8*, __vector(i16[8]), i32);
    /// The `llvm.arm.neon.vst1.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst1.v4i32")
    void neon_vst1_v4i32(i8*, __vector(i32[4]), i32);
    /// The `llvm.arm.neon.vst1.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst1.v2i64")
    void neon_vst1_v2i64(i8*, __vector(i64[2]), i32);
    /// The `llvm.arm.neon.vst1.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst1.v4f32")
    void neon_vst1_v4f32(i8*, __vector(f32[4]), i32);
    /// The `llvm.arm.neon.vst1.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst1.v2f64")
    void neon_vst1_v2f64(i8*, __vector(f64[2]), i32);
    /// The `llvm.arm.neon.vst2.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2.v16i8")
    void neon_vst2_v16i8(i8*, __vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.arm.neon.vst2.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2.v8i16")
    void neon_vst2_v8i16(i8*, __vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.arm.neon.vst2.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2.v4i32")
    void neon_vst2_v4i32(i8*, __vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.arm.neon.vst2.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2.v2i64")
    void neon_vst2_v2i64(i8*, __vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.arm.neon.vst2.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2.v4f32")
    void neon_vst2_v4f32(i8*, __vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.arm.neon.vst2.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2.v2f64")
    void neon_vst2_v2f64(i8*, __vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.arm.neon.vst3.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3.v16i8")
    void neon_vst3_v16i8(i8*, __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.arm.neon.vst3.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3.v8i16")
    void neon_vst3_v8i16(i8*, __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.arm.neon.vst3.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3.v4i32")
    void neon_vst3_v4i32(i8*, __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.arm.neon.vst3.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3.v2i64")
    void neon_vst3_v2i64(i8*, __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.arm.neon.vst3.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3.v4f32")
    void neon_vst3_v4f32(i8*, __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.arm.neon.vst3.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3.v2f64")
    void neon_vst3_v2f64(i8*, __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.arm.neon.vst4.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4.v16i8")
    void neon_vst4_v16i8(i8*, __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.arm.neon.vst4.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4.v8i16")
    void neon_vst4_v8i16(i8*, __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.arm.neon.vst4.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4.v4i32")
    void neon_vst4_v4i32(i8*, __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.arm.neon.vst4.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4.v2i64")
    void neon_vst4_v2i64(i8*, __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.arm.neon.vst4.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4.v4f32")
    void neon_vst4_v4f32(i8*, __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i32);
    /// The `llvm.arm.neon.vst4.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4.v2f64")
    void neon_vst4_v2f64(i8*, __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i32);
    /// The `llvm.arm.neon.vst2lane.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2lane.v16i8")
    void neon_vst2lane_v16i8(i8*, __vector(i8[16]), __vector(i8[16]), i32, i32);
    /// The `llvm.arm.neon.vst2lane.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2lane.v8i16")
    void neon_vst2lane_v8i16(i8*, __vector(i16[8]), __vector(i16[8]), i32, i32);
    /// The `llvm.arm.neon.vst2lane.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2lane.v4i32")
    void neon_vst2lane_v4i32(i8*, __vector(i32[4]), __vector(i32[4]), i32, i32);
    /// The `llvm.arm.neon.vst2lane.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2lane.v2i64")
    void neon_vst2lane_v2i64(i8*, __vector(i64[2]), __vector(i64[2]), i32, i32);
    /// The `llvm.arm.neon.vst2lane.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2lane.v4f32")
    void neon_vst2lane_v4f32(i8*, __vector(f32[4]), __vector(f32[4]), i32, i32);
    /// The `llvm.arm.neon.vst2lane.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst2lane.v2f64")
    void neon_vst2lane_v2f64(i8*, __vector(f64[2]), __vector(f64[2]), i32, i32);
    /// The `llvm.arm.neon.vst3lane.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3lane.v16i8")
    void neon_vst3lane_v16i8(i8*, __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), i32, i32);
    /// The `llvm.arm.neon.vst3lane.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3lane.v8i16")
    void neon_vst3lane_v8i16(i8*, __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), i32, i32);
    /// The `llvm.arm.neon.vst3lane.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3lane.v4i32")
    void neon_vst3lane_v4i32(i8*, __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), i32, i32);
    /// The `llvm.arm.neon.vst3lane.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3lane.v2i64")
    void neon_vst3lane_v2i64(i8*, __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), i32, i32);
    /// The `llvm.arm.neon.vst3lane.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3lane.v4f32")
    void neon_vst3lane_v4f32(i8*, __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i32, i32);
    /// The `llvm.arm.neon.vst3lane.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst3lane.v2f64")
    void neon_vst3lane_v2f64(i8*, __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i32, i32);
    /// The `llvm.arm.neon.vst4lane.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4lane.v16i8")
    void neon_vst4lane_v16i8(i8*, __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), __vector(i8[16]), i32, i32);
    /// The `llvm.arm.neon.vst4lane.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4lane.v8i16")
    void neon_vst4lane_v8i16(i8*, __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), __vector(i16[8]), i32, i32);
    /// The `llvm.arm.neon.vst4lane.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4lane.v4i32")
    void neon_vst4lane_v4i32(i8*, __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), __vector(i32[4]), i32, i32);
    /// The `llvm.arm.neon.vst4lane.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4lane.v2i64")
    void neon_vst4lane_v2i64(i8*, __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), __vector(i64[2]), i32, i32);
    /// The `llvm.arm.neon.vst4lane.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4lane.v4f32")
    void neon_vst4lane_v4f32(i8*, __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), __vector(f32[4]), i32, i32);
    /// The `llvm.arm.neon.vst4lane.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vst4lane.v2f64")
    void neon_vst4lane_v2f64(i8*, __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), __vector(f64[2]), i32, i32);
    /// The `llvm.arm.neon.vbsl.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vbsl.v16i8")
    __vector(i8[16]) neon_vbsl_v16i8(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.vbsl.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vbsl.v8i16")
    __vector(i16[8]) neon_vbsl_v8i16(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.arm.neon.vbsl.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vbsl.v4i32")
    __vector(i32[4]) neon_vbsl_v4i32(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.vbsl.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vbsl.v2i64")
    __vector(i64[2]) neon_vbsl_v2i64(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.arm.neon.vbsl.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vbsl.v4f32")
    __vector(f32[4]) neon_vbsl_v4f32(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.arm.neon.vbsl.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.vbsl.v2f64")
    __vector(f64[2]) neon_vbsl_v2f64(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.arm.neon.aesd` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.aesd")
    __vector(i8[16]) neon_aesd(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.aese` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.aese")
    __vector(i8[16]) neon_aese(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.arm.neon.aesimc` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.aesimc")
    __vector(i8[16]) neon_aesimc(__vector(i8[16]));
    /// The `llvm.arm.neon.aesmc` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.aesmc")
    __vector(i8[16]) neon_aesmc(__vector(i8[16]));
    /// The `llvm.arm.neon.sha1h` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha1h")
    i32 neon_sha1h(i32);
    /// The `llvm.arm.neon.sha1su1` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha1su1")
    __vector(i32[4]) neon_sha1su1(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.sha256su0` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha256su0")
    __vector(i32[4]) neon_sha256su0(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.sha1c` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha1c")
    __vector(i32[4]) neon_sha1c(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.arm.neon.sha1m` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha1m")
    __vector(i32[4]) neon_sha1m(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.arm.neon.sha1p` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha1p")
    __vector(i32[4]) neon_sha1p(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.arm.neon.sha1su0` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha1su0")
    __vector(i32[4]) neon_sha1su0(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.sha256h` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha256h")
    __vector(i32[4]) neon_sha256h(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.sha256h2` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha256h2")
    __vector(i32[4]) neon_sha256h2(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.arm.neon.sha256su1` intrinsic.
    pragma(LDC_intrinsic, "llvm.arm.neon.sha256su1")
    __vector(i32[4]) neon_sha256su1(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
}

/// LLVM intrinsics for the cuda architecture.
version (cuda) {
    /// The `llvm.cuda.syncthreads` intrinsic; known as `__syncthreads` in GCC.
    pragma(LDC_intrinsic, "llvm.cuda.syncthreads")
    void syncthreads();
}

/// LLVM intrinsics for the hexagon architecture.
version (hexagon) {
    /// The `llvm.hexagon.SI.to.SXTHI.asrh` intrinsic; known as `__builtin_SI_to_SXTHI_asrh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.SI.to.SXTHI.asrh")
    i32 SI_to_SXTHI_asrh(i32);
    /// The `llvm.hexagon.circ.ldd` intrinsic; known as `__builtin_circ_ldd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.circ.ldd")
    i8* circ_ldd(i8*, i8*, i32, i32);
    /// The `llvm.hexagon.C2.cmpeq` intrinsic; known as `__builtin_HEXAGON_C2_cmpeq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpeq")
    bool C2_cmpeq(i32, i32);
    /// The `llvm.hexagon.C2.cmpgt` intrinsic; known as `__builtin_HEXAGON_C2_cmpgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgt")
    bool C2_cmpgt(i32, i32);
    /// The `llvm.hexagon.C2.cmpgtu` intrinsic; known as `__builtin_HEXAGON_C2_cmpgtu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgtu")
    bool C2_cmpgtu(i32, i32);
    /// The `llvm.hexagon.C2.cmpeqp` intrinsic; known as `__builtin_HEXAGON_C2_cmpeqp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpeqp")
    bool C2_cmpeqp(i64, i64);
    /// The `llvm.hexagon.C2.cmpgtp` intrinsic; known as `__builtin_HEXAGON_C2_cmpgtp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgtp")
    bool C2_cmpgtp(i64, i64);
    /// The `llvm.hexagon.C2.cmpgtup` intrinsic; known as `__builtin_HEXAGON_C2_cmpgtup` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgtup")
    bool C2_cmpgtup(i64, i64);
    /// The `llvm.hexagon.A4.rcmpeqi` intrinsic; known as `__builtin_HEXAGON_A4_rcmpeqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.rcmpeqi")
    i32 A4_rcmpeqi(i32, i32);
    /// The `llvm.hexagon.A4.rcmpneqi` intrinsic; known as `__builtin_HEXAGON_A4_rcmpneqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.rcmpneqi")
    i32 A4_rcmpneqi(i32, i32);
    /// The `llvm.hexagon.A4.rcmpeq` intrinsic; known as `__builtin_HEXAGON_A4_rcmpeq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.rcmpeq")
    i32 A4_rcmpeq(i32, i32);
    /// The `llvm.hexagon.A4.rcmpneq` intrinsic; known as `__builtin_HEXAGON_A4_rcmpneq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.rcmpneq")
    i32 A4_rcmpneq(i32, i32);
    /// The `llvm.hexagon.C2.bitsset` intrinsic; known as `__builtin_HEXAGON_C2_bitsset` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.bitsset")
    bool C2_bitsset(i32, i32);
    /// The `llvm.hexagon.C2.bitsclr` intrinsic; known as `__builtin_HEXAGON_C2_bitsclr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.bitsclr")
    bool C2_bitsclr(i32, i32);
    /// The `llvm.hexagon.C4.nbitsset` intrinsic; known as `__builtin_HEXAGON_C4_nbitsset` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.nbitsset")
    bool C4_nbitsset(i32, i32);
    /// The `llvm.hexagon.C4.nbitsclr` intrinsic; known as `__builtin_HEXAGON_C4_nbitsclr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.nbitsclr")
    bool C4_nbitsclr(i32, i32);
    /// The `llvm.hexagon.C2.cmpeqi` intrinsic; known as `__builtin_HEXAGON_C2_cmpeqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpeqi")
    bool C2_cmpeqi(i32, i32);
    /// The `llvm.hexagon.C2.cmpgti` intrinsic; known as `__builtin_HEXAGON_C2_cmpgti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgti")
    bool C2_cmpgti(i32, i32);
    /// The `llvm.hexagon.C2.cmpgtui` intrinsic; known as `__builtin_HEXAGON_C2_cmpgtui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgtui")
    bool C2_cmpgtui(i32, i32);
    /// The `llvm.hexagon.C2.cmpgei` intrinsic; known as `__builtin_HEXAGON_C2_cmpgei` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgei")
    bool C2_cmpgei(i32, i32);
    /// The `llvm.hexagon.C2.cmpgeui` intrinsic; known as `__builtin_HEXAGON_C2_cmpgeui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpgeui")
    bool C2_cmpgeui(i32, i32);
    /// The `llvm.hexagon.C2.cmplt` intrinsic; known as `__builtin_HEXAGON_C2_cmplt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmplt")
    bool C2_cmplt(i32, i32);
    /// The `llvm.hexagon.C2.cmpltu` intrinsic; known as `__builtin_HEXAGON_C2_cmpltu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.cmpltu")
    bool C2_cmpltu(i32, i32);
    /// The `llvm.hexagon.C2.bitsclri` intrinsic; known as `__builtin_HEXAGON_C2_bitsclri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.bitsclri")
    bool C2_bitsclri(i32, i32);
    /// The `llvm.hexagon.C4.nbitsclri` intrinsic; known as `__builtin_HEXAGON_C4_nbitsclri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.nbitsclri")
    bool C4_nbitsclri(i32, i32);
    /// The `llvm.hexagon.C4.cmpneqi` intrinsic; known as `__builtin_HEXAGON_C4_cmpneqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.cmpneqi")
    bool C4_cmpneqi(i32, i32);
    /// The `llvm.hexagon.C4.cmpltei` intrinsic; known as `__builtin_HEXAGON_C4_cmpltei` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.cmpltei")
    bool C4_cmpltei(i32, i32);
    /// The `llvm.hexagon.C4.cmplteui` intrinsic; known as `__builtin_HEXAGON_C4_cmplteui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.cmplteui")
    bool C4_cmplteui(i32, i32);
    /// The `llvm.hexagon.C4.cmpneq` intrinsic; known as `__builtin_HEXAGON_C4_cmpneq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.cmpneq")
    bool C4_cmpneq(i32, i32);
    /// The `llvm.hexagon.C4.cmplte` intrinsic; known as `__builtin_HEXAGON_C4_cmplte` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.cmplte")
    bool C4_cmplte(i32, i32);
    /// The `llvm.hexagon.C4.cmplteu` intrinsic; known as `__builtin_HEXAGON_C4_cmplteu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.cmplteu")
    bool C4_cmplteu(i32, i32);
    /// The `llvm.hexagon.C2.and` intrinsic; known as `__builtin_HEXAGON_C2_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.and")
    bool C2_and(i32, i32);
    /// The `llvm.hexagon.C2.or` intrinsic; known as `__builtin_HEXAGON_C2_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.or")
    bool C2_or(i32, i32);
    /// The `llvm.hexagon.C2.xor` intrinsic; known as `__builtin_HEXAGON_C2_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.xor")
    bool C2_xor(i32, i32);
    /// The `llvm.hexagon.C2.andn` intrinsic; known as `__builtin_HEXAGON_C2_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.andn")
    bool C2_andn(i32, i32);
    /// The `llvm.hexagon.C2.not` intrinsic; known as `__builtin_HEXAGON_C2_not` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.not")
    bool C2_not(i32);
    /// The `llvm.hexagon.C2.orn` intrinsic; known as `__builtin_HEXAGON_C2_orn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.orn")
    bool C2_orn(i32, i32);
    /// The `llvm.hexagon.C4.and.and` intrinsic; known as `__builtin_HEXAGON_C4_and_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.and.and")
    bool C4_and_and(i32, i32, i32);
    /// The `llvm.hexagon.C4.and.or` intrinsic; known as `__builtin_HEXAGON_C4_and_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.and.or")
    bool C4_and_or(i32, i32, i32);
    /// The `llvm.hexagon.C4.or.and` intrinsic; known as `__builtin_HEXAGON_C4_or_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.or.and")
    bool C4_or_and(i32, i32, i32);
    /// The `llvm.hexagon.C4.or.or` intrinsic; known as `__builtin_HEXAGON_C4_or_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.or.or")
    bool C4_or_or(i32, i32, i32);
    /// The `llvm.hexagon.C4.and.andn` intrinsic; known as `__builtin_HEXAGON_C4_and_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.and.andn")
    bool C4_and_andn(i32, i32, i32);
    /// The `llvm.hexagon.C4.and.orn` intrinsic; known as `__builtin_HEXAGON_C4_and_orn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.and.orn")
    bool C4_and_orn(i32, i32, i32);
    /// The `llvm.hexagon.C4.or.andn` intrinsic; known as `__builtin_HEXAGON_C4_or_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.or.andn")
    bool C4_or_andn(i32, i32, i32);
    /// The `llvm.hexagon.C4.or.orn` intrinsic; known as `__builtin_HEXAGON_C4_or_orn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.or.orn")
    bool C4_or_orn(i32, i32, i32);
    /// The `llvm.hexagon.C2.pxfer.map` intrinsic; known as `__builtin_HEXAGON_C2_pxfer_map` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.pxfer.map")
    bool C2_pxfer_map(i32);
    /// The `llvm.hexagon.C2.any8` intrinsic; known as `__builtin_HEXAGON_C2_any8` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.any8")
    bool C2_any8(i32);
    /// The `llvm.hexagon.C2.all8` intrinsic; known as `__builtin_HEXAGON_C2_all8` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.all8")
    bool C2_all8(i32);
    /// The `llvm.hexagon.C2.vitpack` intrinsic; known as `__builtin_HEXAGON_C2_vitpack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.vitpack")
    i32 C2_vitpack(i32, i32);
    /// The `llvm.hexagon.C2.mux` intrinsic; known as `__builtin_HEXAGON_C2_mux` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.mux")
    i32 C2_mux(i32, i32, i32);
    /// The `llvm.hexagon.C2.muxii` intrinsic; known as `__builtin_HEXAGON_C2_muxii` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.muxii")
    i32 C2_muxii(i32, i32, i32);
    /// The `llvm.hexagon.C2.muxir` intrinsic; known as `__builtin_HEXAGON_C2_muxir` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.muxir")
    i32 C2_muxir(i32, i32, i32);
    /// The `llvm.hexagon.C2.muxri` intrinsic; known as `__builtin_HEXAGON_C2_muxri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.muxri")
    i32 C2_muxri(i32, i32, i32);
    /// The `llvm.hexagon.C2.vmux` intrinsic; known as `__builtin_HEXAGON_C2_vmux` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.vmux")
    i64 C2_vmux(i32, i64, i64);
    /// The `llvm.hexagon.C2.mask` intrinsic; known as `__builtin_HEXAGON_C2_mask` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.mask")
    i64 C2_mask(i32);
    /// The `llvm.hexagon.A2.vcmpbeq` intrinsic; known as `__builtin_HEXAGON_A2_vcmpbeq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmpbeq")
    bool A2_vcmpbeq(i64, i64);
    /// The `llvm.hexagon.A4.vcmpbeqi` intrinsic; known as `__builtin_HEXAGON_A4_vcmpbeqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpbeqi")
    bool A4_vcmpbeqi(i64, i32);
    /// The `llvm.hexagon.A4.vcmpbeq.any` intrinsic; known as `__builtin_HEXAGON_A4_vcmpbeq_any` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpbeq.any")
    bool A4_vcmpbeq_any(i64, i64);
    /// The `llvm.hexagon.A2.vcmpbgtu` intrinsic; known as `__builtin_HEXAGON_A2_vcmpbgtu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmpbgtu")
    bool A2_vcmpbgtu(i64, i64);
    /// The `llvm.hexagon.A4.vcmpbgtui` intrinsic; known as `__builtin_HEXAGON_A4_vcmpbgtui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpbgtui")
    bool A4_vcmpbgtui(i64, i32);
    /// The `llvm.hexagon.A4.vcmpbgt` intrinsic; known as `__builtin_HEXAGON_A4_vcmpbgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpbgt")
    bool A4_vcmpbgt(i64, i64);
    /// The `llvm.hexagon.A4.vcmpbgti` intrinsic; known as `__builtin_HEXAGON_A4_vcmpbgti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpbgti")
    bool A4_vcmpbgti(i64, i32);
    /// The `llvm.hexagon.A4.cmpbeq` intrinsic; known as `__builtin_HEXAGON_A4_cmpbeq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpbeq")
    bool A4_cmpbeq(i32, i32);
    /// The `llvm.hexagon.A4.cmpbeqi` intrinsic; known as `__builtin_HEXAGON_A4_cmpbeqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpbeqi")
    bool A4_cmpbeqi(i32, i32);
    /// The `llvm.hexagon.A4.cmpbgtu` intrinsic; known as `__builtin_HEXAGON_A4_cmpbgtu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpbgtu")
    bool A4_cmpbgtu(i32, i32);
    /// The `llvm.hexagon.A4.cmpbgtui` intrinsic; known as `__builtin_HEXAGON_A4_cmpbgtui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpbgtui")
    bool A4_cmpbgtui(i32, i32);
    /// The `llvm.hexagon.A4.cmpbgt` intrinsic; known as `__builtin_HEXAGON_A4_cmpbgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpbgt")
    bool A4_cmpbgt(i32, i32);
    /// The `llvm.hexagon.A4.cmpbgti` intrinsic; known as `__builtin_HEXAGON_A4_cmpbgti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpbgti")
    bool A4_cmpbgti(i32, i32);
    /// The `llvm.hexagon.A2.vcmpheq` intrinsic; known as `__builtin_HEXAGON_A2_vcmpheq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmpheq")
    bool A2_vcmpheq(i64, i64);
    /// The `llvm.hexagon.A2.vcmphgt` intrinsic; known as `__builtin_HEXAGON_A2_vcmphgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmphgt")
    bool A2_vcmphgt(i64, i64);
    /// The `llvm.hexagon.A2.vcmphgtu` intrinsic; known as `__builtin_HEXAGON_A2_vcmphgtu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmphgtu")
    bool A2_vcmphgtu(i64, i64);
    /// The `llvm.hexagon.A4.vcmpheqi` intrinsic; known as `__builtin_HEXAGON_A4_vcmpheqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpheqi")
    bool A4_vcmpheqi(i64, i32);
    /// The `llvm.hexagon.A4.vcmphgti` intrinsic; known as `__builtin_HEXAGON_A4_vcmphgti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmphgti")
    bool A4_vcmphgti(i64, i32);
    /// The `llvm.hexagon.A4.vcmphgtui` intrinsic; known as `__builtin_HEXAGON_A4_vcmphgtui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmphgtui")
    bool A4_vcmphgtui(i64, i32);
    /// The `llvm.hexagon.A4.cmpheq` intrinsic; known as `__builtin_HEXAGON_A4_cmpheq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpheq")
    bool A4_cmpheq(i32, i32);
    /// The `llvm.hexagon.A4.cmphgt` intrinsic; known as `__builtin_HEXAGON_A4_cmphgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmphgt")
    bool A4_cmphgt(i32, i32);
    /// The `llvm.hexagon.A4.cmphgtu` intrinsic; known as `__builtin_HEXAGON_A4_cmphgtu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmphgtu")
    bool A4_cmphgtu(i32, i32);
    /// The `llvm.hexagon.A4.cmpheqi` intrinsic; known as `__builtin_HEXAGON_A4_cmpheqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmpheqi")
    bool A4_cmpheqi(i32, i32);
    /// The `llvm.hexagon.A4.cmphgti` intrinsic; known as `__builtin_HEXAGON_A4_cmphgti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmphgti")
    bool A4_cmphgti(i32, i32);
    /// The `llvm.hexagon.A4.cmphgtui` intrinsic; known as `__builtin_HEXAGON_A4_cmphgtui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cmphgtui")
    bool A4_cmphgtui(i32, i32);
    /// The `llvm.hexagon.A2.vcmpweq` intrinsic; known as `__builtin_HEXAGON_A2_vcmpweq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmpweq")
    bool A2_vcmpweq(i64, i64);
    /// The `llvm.hexagon.A2.vcmpwgt` intrinsic; known as `__builtin_HEXAGON_A2_vcmpwgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmpwgt")
    bool A2_vcmpwgt(i64, i64);
    /// The `llvm.hexagon.A2.vcmpwgtu` intrinsic; known as `__builtin_HEXAGON_A2_vcmpwgtu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vcmpwgtu")
    bool A2_vcmpwgtu(i64, i64);
    /// The `llvm.hexagon.A4.vcmpweqi` intrinsic; known as `__builtin_HEXAGON_A4_vcmpweqi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpweqi")
    bool A4_vcmpweqi(i64, i32);
    /// The `llvm.hexagon.A4.vcmpwgti` intrinsic; known as `__builtin_HEXAGON_A4_vcmpwgti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpwgti")
    bool A4_vcmpwgti(i64, i32);
    /// The `llvm.hexagon.A4.vcmpwgtui` intrinsic; known as `__builtin_HEXAGON_A4_vcmpwgtui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vcmpwgtui")
    bool A4_vcmpwgtui(i64, i32);
    /// The `llvm.hexagon.A4.boundscheck` intrinsic; known as `__builtin_HEXAGON_A4_boundscheck` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.boundscheck")
    bool A4_boundscheck(i32, i64);
    /// The `llvm.hexagon.A4.tlbmatch` intrinsic; known as `__builtin_HEXAGON_A4_tlbmatch` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.tlbmatch")
    bool A4_tlbmatch(i64, i32);
    /// The `llvm.hexagon.C2.tfrpr` intrinsic; known as `__builtin_HEXAGON_C2_tfrpr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.tfrpr")
    i32 C2_tfrpr(i32);
    /// The `llvm.hexagon.C2.tfrrp` intrinsic; known as `__builtin_HEXAGON_C2_tfrrp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C2.tfrrp")
    bool C2_tfrrp(i32);
    /// The `llvm.hexagon.C4.fastcorner9` intrinsic; known as `__builtin_HEXAGON_C4_fastcorner9` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.fastcorner9")
    bool C4_fastcorner9(i32, i32);
    /// The `llvm.hexagon.C4.fastcorner9.not` intrinsic; known as `__builtin_HEXAGON_C4_fastcorner9_not` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.C4.fastcorner9.not")
    bool C4_fastcorner9_not(i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.hh.s0")
    i32 M2_mpy_acc_hh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.hh.s1")
    i32 M2_mpy_acc_hh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.hl.s0")
    i32 M2_mpy_acc_hl_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.hl.s1")
    i32 M2_mpy_acc_hl_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.lh.s0")
    i32 M2_mpy_acc_lh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.lh.s1")
    i32 M2_mpy_acc_lh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.ll.s0")
    i32 M2_mpy_acc_ll_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.ll.s1")
    i32 M2_mpy_acc_ll_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.hh.s0")
    i32 M2_mpy_nac_hh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.hh.s1")
    i32 M2_mpy_nac_hh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.hl.s0")
    i32 M2_mpy_nac_hl_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.hl.s1")
    i32 M2_mpy_nac_hl_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.lh.s0")
    i32 M2_mpy_nac_lh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.lh.s1")
    i32 M2_mpy_nac_lh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.ll.s0")
    i32 M2_mpy_nac_ll_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.ll.s1")
    i32 M2_mpy_nac_ll_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.hh.s0")
    i32 M2_mpy_acc_sat_hh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.hh.s1")
    i32 M2_mpy_acc_sat_hh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.hl.s0")
    i32 M2_mpy_acc_sat_hl_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.hl.s1")
    i32 M2_mpy_acc_sat_hl_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.lh.s0")
    i32 M2_mpy_acc_sat_lh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.lh.s1")
    i32 M2_mpy_acc_sat_lh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.ll.s0")
    i32 M2_mpy_acc_sat_ll_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.acc.sat.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_acc_sat_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.acc.sat.ll.s1")
    i32 M2_mpy_acc_sat_ll_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.hh.s0")
    i32 M2_mpy_nac_sat_hh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.hh.s1")
    i32 M2_mpy_nac_sat_hh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.hl.s0")
    i32 M2_mpy_nac_sat_hl_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.hl.s1")
    i32 M2_mpy_nac_sat_hl_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.lh.s0")
    i32 M2_mpy_nac_sat_lh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.lh.s1")
    i32 M2_mpy_nac_sat_lh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.ll.s0")
    i32 M2_mpy_nac_sat_ll_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.nac.sat.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_nac_sat_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.nac.sat.ll.s1")
    i32 M2_mpy_nac_sat_ll_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpy.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.hh.s0")
    i32 M2_mpy_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.hh.s1")
    i32 M2_mpy_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.hl.s0")
    i32 M2_mpy_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.hl.s1")
    i32 M2_mpy_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.lh.s0")
    i32 M2_mpy_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.lh.s1")
    i32 M2_mpy_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.ll.s0")
    i32 M2_mpy_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.ll.s1")
    i32 M2_mpy_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.hh.s0")
    i32 M2_mpy_sat_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.hh.s1")
    i32 M2_mpy_sat_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.hl.s0")
    i32 M2_mpy_sat_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.hl.s1")
    i32 M2_mpy_sat_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.lh.s0")
    i32 M2_mpy_sat_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.lh.s1")
    i32 M2_mpy_sat_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.ll.s0")
    i32 M2_mpy_sat_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.ll.s1")
    i32 M2_mpy_sat_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.hh.s0")
    i32 M2_mpy_rnd_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.hh.s1")
    i32 M2_mpy_rnd_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.hl.s0")
    i32 M2_mpy_rnd_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.hl.s1")
    i32 M2_mpy_rnd_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.lh.s0")
    i32 M2_mpy_rnd_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.lh.s1")
    i32 M2_mpy_rnd_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.ll.s0")
    i32 M2_mpy_rnd_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.rnd.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_rnd_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.rnd.ll.s1")
    i32 M2_mpy_rnd_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.hh.s0")
    i32 M2_mpy_sat_rnd_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.hh.s1")
    i32 M2_mpy_sat_rnd_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.hl.s0")
    i32 M2_mpy_sat_rnd_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.hl.s1")
    i32 M2_mpy_sat_rnd_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.lh.s0")
    i32 M2_mpy_sat_rnd_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.lh.s1")
    i32 M2_mpy_sat_rnd_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.ll.s0")
    i32 M2_mpy_sat_rnd_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpy.sat.rnd.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_sat_rnd_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.sat.rnd.ll.s1")
    i32 M2_mpy_sat_rnd_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.hh.s0")
    i64 M2_mpyd_acc_hh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.hh.s1")
    i64 M2_mpyd_acc_hh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.hl.s0")
    i64 M2_mpyd_acc_hl_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.hl.s1")
    i64 M2_mpyd_acc_hl_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.lh.s0")
    i64 M2_mpyd_acc_lh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.lh.s1")
    i64 M2_mpyd_acc_lh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.ll.s0")
    i64 M2_mpyd_acc_ll_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.acc.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_acc_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.acc.ll.s1")
    i64 M2_mpyd_acc_ll_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.hh.s0")
    i64 M2_mpyd_nac_hh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.hh.s1")
    i64 M2_mpyd_nac_hh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.hl.s0")
    i64 M2_mpyd_nac_hl_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.hl.s1")
    i64 M2_mpyd_nac_hl_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.lh.s0")
    i64 M2_mpyd_nac_lh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.lh.s1")
    i64 M2_mpyd_nac_lh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.ll.s0")
    i64 M2_mpyd_nac_ll_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.nac.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_nac_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.nac.ll.s1")
    i64 M2_mpyd_nac_ll_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyd.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.hh.s0")
    i64 M2_mpyd_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.hh.s1")
    i64 M2_mpyd_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.hl.s0")
    i64 M2_mpyd_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.hl.s1")
    i64 M2_mpyd_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.lh.s0")
    i64 M2_mpyd_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.lh.s1")
    i64 M2_mpyd_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.ll.s0")
    i64 M2_mpyd_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.ll.s1")
    i64 M2_mpyd_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.hh.s0")
    i64 M2_mpyd_rnd_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.hh.s1")
    i64 M2_mpyd_rnd_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.hl.s0")
    i64 M2_mpyd_rnd_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.hl.s1")
    i64 M2_mpyd_rnd_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.lh.s0")
    i64 M2_mpyd_rnd_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.lh.s1")
    i64 M2_mpyd_rnd_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.ll.s0")
    i64 M2_mpyd_rnd_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyd.rnd.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyd_rnd_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyd.rnd.ll.s1")
    i64 M2_mpyd_rnd_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.hh.s0")
    i32 M2_mpyu_acc_hh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.hh.s1")
    i32 M2_mpyu_acc_hh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.hl.s0")
    i32 M2_mpyu_acc_hl_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.hl.s1")
    i32 M2_mpyu_acc_hl_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.lh.s0")
    i32 M2_mpyu_acc_lh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.lh.s1")
    i32 M2_mpyu_acc_lh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.ll.s0")
    i32 M2_mpyu_acc_ll_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.acc.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_acc_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.acc.ll.s1")
    i32 M2_mpyu_acc_ll_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.hh.s0")
    i32 M2_mpyu_nac_hh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.hh.s1")
    i32 M2_mpyu_nac_hh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.hl.s0")
    i32 M2_mpyu_nac_hl_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.hl.s1")
    i32 M2_mpyu_nac_hl_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.lh.s0")
    i32 M2_mpyu_nac_lh_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.lh.s1")
    i32 M2_mpyu_nac_lh_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.ll.s0")
    i32 M2_mpyu_nac_ll_s0(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.nac.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_nac_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.nac.ll.s1")
    i32 M2_mpyu_nac_ll_s1(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyu.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.hh.s0")
    i32 M2_mpyu_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.hh.s1")
    i32 M2_mpyu_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.hl.s0")
    i32 M2_mpyu_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.hl.s1")
    i32 M2_mpyu_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.lh.s0")
    i32 M2_mpyu_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.lh.s1")
    i32 M2_mpyu_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.ll.s0")
    i32 M2_mpyu_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.ll.s1")
    i32 M2_mpyu_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.hh.s0")
    i64 M2_mpyud_acc_hh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.hh.s1")
    i64 M2_mpyud_acc_hh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.hl.s0")
    i64 M2_mpyud_acc_hl_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.hl.s1")
    i64 M2_mpyud_acc_hl_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.lh.s0")
    i64 M2_mpyud_acc_lh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.lh.s1")
    i64 M2_mpyud_acc_lh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.ll.s0")
    i64 M2_mpyud_acc_ll_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.acc.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_acc_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.acc.ll.s1")
    i64 M2_mpyud_acc_ll_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.hh.s0")
    i64 M2_mpyud_nac_hh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.hh.s1")
    i64 M2_mpyud_nac_hh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.hl.s0")
    i64 M2_mpyud_nac_hl_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.hl.s1")
    i64 M2_mpyud_nac_hl_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.lh.s0")
    i64 M2_mpyud_nac_lh_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.lh.s1")
    i64 M2_mpyud_nac_lh_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.ll.s0")
    i64 M2_mpyud_nac_ll_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.nac.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_nac_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.nac.ll.s1")
    i64 M2_mpyud_nac_ll_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpyud.hh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_hh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.hh.s0")
    i64 M2_mpyud_hh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.hh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_hh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.hh.s1")
    i64 M2_mpyud_hh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.hl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_hl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.hl.s0")
    i64 M2_mpyud_hl_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.hl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_hl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.hl.s1")
    i64 M2_mpyud_hl_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.lh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_lh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.lh.s0")
    i64 M2_mpyud_lh_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.lh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_lh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.lh.s1")
    i64 M2_mpyud_lh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.ll.s0` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_ll_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.ll.s0")
    i64 M2_mpyud_ll_s0(i32, i32);
    /// The `llvm.hexagon.M2.mpyud.ll.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpyud_ll_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyud.ll.s1")
    i64 M2_mpyud_ll_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpysmi` intrinsic; known as `__builtin_HEXAGON_M2_mpysmi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpysmi")
    i32 M2_mpysmi(i32, i32);
    /// The `llvm.hexagon.M2.macsip` intrinsic; known as `__builtin_HEXAGON_M2_macsip` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.macsip")
    i32 M2_macsip(i32, i32, i32);
    /// The `llvm.hexagon.M2.macsin` intrinsic; known as `__builtin_HEXAGON_M2_macsin` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.macsin")
    i32 M2_macsin(i32, i32, i32);
    /// The `llvm.hexagon.M2.dpmpyss.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyss_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyss.s0")
    i64 M2_dpmpyss_s0(i32, i32);
    /// The `llvm.hexagon.M2.dpmpyss.acc.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyss_acc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyss.acc.s0")
    i64 M2_dpmpyss_acc_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.dpmpyss.nac.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyss_nac_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyss.nac.s0")
    i64 M2_dpmpyss_nac_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.dpmpyuu.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyuu_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyuu.s0")
    i64 M2_dpmpyuu_s0(i32, i32);
    /// The `llvm.hexagon.M2.dpmpyuu.acc.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyuu_acc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyuu.acc.s0")
    i64 M2_dpmpyuu_acc_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.dpmpyuu.nac.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyuu_nac_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyuu.nac.s0")
    i64 M2_dpmpyuu_nac_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.mpy.up` intrinsic; known as `__builtin_HEXAGON_M2_mpy_up` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.up")
    i32 M2_mpy_up(i32, i32);
    /// The `llvm.hexagon.M2.mpy.up.s1` intrinsic; known as `__builtin_HEXAGON_M2_mpy_up_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.up.s1")
    i32 M2_mpy_up_s1(i32, i32);
    /// The `llvm.hexagon.M2.mpy.up.s1.sat` intrinsic; known as `__builtin_HEXAGON_M2_mpy_up_s1_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpy.up.s1.sat")
    i32 M2_mpy_up_s1_sat(i32, i32);
    /// The `llvm.hexagon.M2.mpyu.up` intrinsic; known as `__builtin_HEXAGON_M2_mpyu_up` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyu.up")
    i32 M2_mpyu_up(i32, i32);
    /// The `llvm.hexagon.M2.mpysu.up` intrinsic; known as `__builtin_HEXAGON_M2_mpysu_up` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpysu.up")
    i32 M2_mpysu_up(i32, i32);
    /// The `llvm.hexagon.M2.dpmpyss.rnd.s0` intrinsic; known as `__builtin_HEXAGON_M2_dpmpyss_rnd_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.dpmpyss.rnd.s0")
    i32 M2_dpmpyss_rnd_s0(i32, i32);
    /// The `llvm.hexagon.M4.mac.up.s1.sat` intrinsic; known as `__builtin_HEXAGON_M4_mac_up_s1_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.mac.up.s1.sat")
    i32 M4_mac_up_s1_sat(i32, i32, i32);
    /// The `llvm.hexagon.M4.nac.up.s1.sat` intrinsic; known as `__builtin_HEXAGON_M4_nac_up_s1_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.nac.up.s1.sat")
    i32 M4_nac_up_s1_sat(i32, i32, i32);
    /// The `llvm.hexagon.M2.mpyi` intrinsic; known as `__builtin_HEXAGON_M2_mpyi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyi")
    i32 M2_mpyi(i32, i32);
    /// The `llvm.hexagon.M2.mpyui` intrinsic; known as `__builtin_HEXAGON_M2_mpyui` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mpyui")
    i32 M2_mpyui(i32, i32);
    /// The `llvm.hexagon.M2.maci` intrinsic; known as `__builtin_HEXAGON_M2_maci` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.maci")
    i32 M2_maci(i32, i32, i32);
    /// The `llvm.hexagon.M2.acci` intrinsic; known as `__builtin_HEXAGON_M2_acci` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.acci")
    i32 M2_acci(i32, i32, i32);
    /// The `llvm.hexagon.M2.accii` intrinsic; known as `__builtin_HEXAGON_M2_accii` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.accii")
    i32 M2_accii(i32, i32, i32);
    /// The `llvm.hexagon.M2.nacci` intrinsic; known as `__builtin_HEXAGON_M2_nacci` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.nacci")
    i32 M2_nacci(i32, i32, i32);
    /// The `llvm.hexagon.M2.naccii` intrinsic; known as `__builtin_HEXAGON_M2_naccii` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.naccii")
    i32 M2_naccii(i32, i32, i32);
    /// The `llvm.hexagon.M2.subacc` intrinsic; known as `__builtin_HEXAGON_M2_subacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.subacc")
    i32 M2_subacc(i32, i32, i32);
    /// The `llvm.hexagon.M4.mpyrr.addr` intrinsic; known as `__builtin_HEXAGON_M4_mpyrr_addr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.mpyrr.addr")
    i32 M4_mpyrr_addr(i32, i32, i32);
    /// The `llvm.hexagon.M4.mpyri.addr.u2` intrinsic; known as `__builtin_HEXAGON_M4_mpyri_addr_u2` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.mpyri.addr.u2")
    i32 M4_mpyri_addr_u2(i32, i32, i32);
    /// The `llvm.hexagon.M4.mpyri.addr` intrinsic; known as `__builtin_HEXAGON_M4_mpyri_addr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.mpyri.addr")
    i32 M4_mpyri_addr(i32, i32, i32);
    /// The `llvm.hexagon.M4.mpyri.addi` intrinsic; known as `__builtin_HEXAGON_M4_mpyri_addi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.mpyri.addi")
    i32 M4_mpyri_addi(i32, i32, i32);
    /// The `llvm.hexagon.M4.mpyrr.addi` intrinsic; known as `__builtin_HEXAGON_M4_mpyrr_addi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.mpyrr.addi")
    i32 M4_mpyrr_addi(i32, i32, i32);
    /// The `llvm.hexagon.M2.vmpy2s.s0` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2s_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2s.s0")
    i64 M2_vmpy2s_s0(i32, i32);
    /// The `llvm.hexagon.M2.vmpy2s.s1` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2s_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2s.s1")
    i64 M2_vmpy2s_s1(i32, i32);
    /// The `llvm.hexagon.M2.vmac2s.s0` intrinsic; known as `__builtin_HEXAGON_M2_vmac2s_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2s.s0")
    i64 M2_vmac2s_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.vmac2s.s1` intrinsic; known as `__builtin_HEXAGON_M2_vmac2s_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2s.s1")
    i64 M2_vmac2s_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.vmpy2su.s0` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2su_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2su.s0")
    i64 M2_vmpy2su_s0(i32, i32);
    /// The `llvm.hexagon.M2.vmpy2su.s1` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2su_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2su.s1")
    i64 M2_vmpy2su_s1(i32, i32);
    /// The `llvm.hexagon.M2.vmac2su.s0` intrinsic; known as `__builtin_HEXAGON_M2_vmac2su_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2su.s0")
    i64 M2_vmac2su_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.vmac2su.s1` intrinsic; known as `__builtin_HEXAGON_M2_vmac2su_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2su.s1")
    i64 M2_vmac2su_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.vmpy2s.s0pack` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2s_s0pack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2s.s0pack")
    i32 M2_vmpy2s_s0pack(i32, i32);
    /// The `llvm.hexagon.M2.vmpy2s.s1pack` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2s_s1pack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2s.s1pack")
    i32 M2_vmpy2s_s1pack(i32, i32);
    /// The `llvm.hexagon.M2.vmac2` intrinsic; known as `__builtin_HEXAGON_M2_vmac2` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2")
    i64 M2_vmac2(i64, i32, i32);
    /// The `llvm.hexagon.M2.vmpy2es.s0` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2es_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2es.s0")
    i64 M2_vmpy2es_s0(i64, i64);
    /// The `llvm.hexagon.M2.vmpy2es.s1` intrinsic; known as `__builtin_HEXAGON_M2_vmpy2es_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmpy2es.s1")
    i64 M2_vmpy2es_s1(i64, i64);
    /// The `llvm.hexagon.M2.vmac2es.s0` intrinsic; known as `__builtin_HEXAGON_M2_vmac2es_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2es.s0")
    i64 M2_vmac2es_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.vmac2es.s1` intrinsic; known as `__builtin_HEXAGON_M2_vmac2es_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2es.s1")
    i64 M2_vmac2es_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.vmac2es` intrinsic; known as `__builtin_HEXAGON_M2_vmac2es` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vmac2es")
    i64 M2_vmac2es(i64, i64, i64);
    /// The `llvm.hexagon.M2.vrmac.s0` intrinsic; known as `__builtin_HEXAGON_M2_vrmac_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrmac.s0")
    i64 M2_vrmac_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.vrmpy.s0` intrinsic; known as `__builtin_HEXAGON_M2_vrmpy_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrmpy.s0")
    i64 M2_vrmpy_s0(i64, i64);
    /// The `llvm.hexagon.M2.vdmpyrs.s0` intrinsic; known as `__builtin_HEXAGON_M2_vdmpyrs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vdmpyrs.s0")
    i32 M2_vdmpyrs_s0(i64, i64);
    /// The `llvm.hexagon.M2.vdmpyrs.s1` intrinsic; known as `__builtin_HEXAGON_M2_vdmpyrs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vdmpyrs.s1")
    i32 M2_vdmpyrs_s1(i64, i64);
    /// The `llvm.hexagon.M5.vrmpybuu` intrinsic; known as `__builtin_HEXAGON_M5_vrmpybuu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vrmpybuu")
    i64 M5_vrmpybuu(i64, i64);
    /// The `llvm.hexagon.M5.vrmacbuu` intrinsic; known as `__builtin_HEXAGON_M5_vrmacbuu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vrmacbuu")
    i64 M5_vrmacbuu(i64, i64, i64);
    /// The `llvm.hexagon.M5.vrmpybsu` intrinsic; known as `__builtin_HEXAGON_M5_vrmpybsu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vrmpybsu")
    i64 M5_vrmpybsu(i64, i64);
    /// The `llvm.hexagon.M5.vrmacbsu` intrinsic; known as `__builtin_HEXAGON_M5_vrmacbsu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vrmacbsu")
    i64 M5_vrmacbsu(i64, i64, i64);
    /// The `llvm.hexagon.M5.vmpybuu` intrinsic; known as `__builtin_HEXAGON_M5_vmpybuu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vmpybuu")
    i64 M5_vmpybuu(i32, i32);
    /// The `llvm.hexagon.M5.vmpybsu` intrinsic; known as `__builtin_HEXAGON_M5_vmpybsu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vmpybsu")
    i64 M5_vmpybsu(i32, i32);
    /// The `llvm.hexagon.M5.vmacbuu` intrinsic; known as `__builtin_HEXAGON_M5_vmacbuu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vmacbuu")
    i64 M5_vmacbuu(i64, i32, i32);
    /// The `llvm.hexagon.M5.vmacbsu` intrinsic; known as `__builtin_HEXAGON_M5_vmacbsu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vmacbsu")
    i64 M5_vmacbsu(i64, i32, i32);
    /// The `llvm.hexagon.M5.vdmpybsu` intrinsic; known as `__builtin_HEXAGON_M5_vdmpybsu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vdmpybsu")
    i64 M5_vdmpybsu(i64, i64);
    /// The `llvm.hexagon.M5.vdmacbsu` intrinsic; known as `__builtin_HEXAGON_M5_vdmacbsu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M5.vdmacbsu")
    i64 M5_vdmacbsu(i64, i64, i64);
    /// The `llvm.hexagon.M2.vdmacs.s0` intrinsic; known as `__builtin_HEXAGON_M2_vdmacs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vdmacs.s0")
    i64 M2_vdmacs_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.vdmacs.s1` intrinsic; known as `__builtin_HEXAGON_M2_vdmacs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vdmacs.s1")
    i64 M2_vdmacs_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.vdmpys.s0` intrinsic; known as `__builtin_HEXAGON_M2_vdmpys_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vdmpys.s0")
    i64 M2_vdmpys_s0(i64, i64);
    /// The `llvm.hexagon.M2.vdmpys.s1` intrinsic; known as `__builtin_HEXAGON_M2_vdmpys_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vdmpys.s1")
    i64 M2_vdmpys_s1(i64, i64);
    /// The `llvm.hexagon.M2.cmpyrs.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmpyrs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpyrs.s0")
    i32 M2_cmpyrs_s0(i32, i32);
    /// The `llvm.hexagon.M2.cmpyrs.s1` intrinsic; known as `__builtin_HEXAGON_M2_cmpyrs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpyrs.s1")
    i32 M2_cmpyrs_s1(i32, i32);
    /// The `llvm.hexagon.M2.cmpyrsc.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmpyrsc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpyrsc.s0")
    i32 M2_cmpyrsc_s0(i32, i32);
    /// The `llvm.hexagon.M2.cmpyrsc.s1` intrinsic; known as `__builtin_HEXAGON_M2_cmpyrsc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpyrsc.s1")
    i32 M2_cmpyrsc_s1(i32, i32);
    /// The `llvm.hexagon.M2.cmacs.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmacs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmacs.s0")
    i64 M2_cmacs_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.cmacs.s1` intrinsic; known as `__builtin_HEXAGON_M2_cmacs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmacs.s1")
    i64 M2_cmacs_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.cmacsc.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmacsc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmacsc.s0")
    i64 M2_cmacsc_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.cmacsc.s1` intrinsic; known as `__builtin_HEXAGON_M2_cmacsc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmacsc.s1")
    i64 M2_cmacsc_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.cmpys.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmpys_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpys.s0")
    i64 M2_cmpys_s0(i32, i32);
    /// The `llvm.hexagon.M2.cmpys.s1` intrinsic; known as `__builtin_HEXAGON_M2_cmpys_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpys.s1")
    i64 M2_cmpys_s1(i32, i32);
    /// The `llvm.hexagon.M2.cmpysc.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmpysc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpysc.s0")
    i64 M2_cmpysc_s0(i32, i32);
    /// The `llvm.hexagon.M2.cmpysc.s1` intrinsic; known as `__builtin_HEXAGON_M2_cmpysc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpysc.s1")
    i64 M2_cmpysc_s1(i32, i32);
    /// The `llvm.hexagon.M2.cnacs.s0` intrinsic; known as `__builtin_HEXAGON_M2_cnacs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cnacs.s0")
    i64 M2_cnacs_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.cnacs.s1` intrinsic; known as `__builtin_HEXAGON_M2_cnacs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cnacs.s1")
    i64 M2_cnacs_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.cnacsc.s0` intrinsic; known as `__builtin_HEXAGON_M2_cnacsc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cnacsc.s0")
    i64 M2_cnacsc_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.cnacsc.s1` intrinsic; known as `__builtin_HEXAGON_M2_cnacsc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cnacsc.s1")
    i64 M2_cnacsc_s1(i64, i32, i32);
    /// The `llvm.hexagon.M2.vrcmpys.s1` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpys_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpys.s1")
    i64 M2_vrcmpys_s1(i64, i32);
    /// The `llvm.hexagon.M2.vrcmpys.acc.s1` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpys_acc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpys.acc.s1")
    i64 M2_vrcmpys_acc_s1(i64, i64, i32);
    /// The `llvm.hexagon.M2.vrcmpys.s1rp` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpys_s1rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpys.s1rp")
    i32 M2_vrcmpys_s1rp(i64, i32);
    /// The `llvm.hexagon.M2.mmacls.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmacls_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacls.s0")
    i64 M2_mmacls_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmacls.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmacls_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacls.s1")
    i64 M2_mmacls_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmachs.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmachs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmachs.s0")
    i64 M2_mmachs_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmachs.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmachs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmachs.s1")
    i64 M2_mmachs_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmpyl.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyl_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyl.s0")
    i64 M2_mmpyl_s0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyl.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyl.s1")
    i64 M2_mmpyl_s1(i64, i64);
    /// The `llvm.hexagon.M2.mmpyh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyh.s0")
    i64 M2_mmpyh_s0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyh.s1")
    i64 M2_mmpyh_s1(i64, i64);
    /// The `llvm.hexagon.M2.mmacls.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmacls_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacls.rs0")
    i64 M2_mmacls_rs0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmacls.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmacls_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacls.rs1")
    i64 M2_mmacls_rs1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmachs.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmachs_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmachs.rs0")
    i64 M2_mmachs_rs0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmachs.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmachs_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmachs.rs1")
    i64 M2_mmachs_rs1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmpyl.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyl_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyl.rs0")
    i64 M2_mmpyl_rs0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyl.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyl_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyl.rs1")
    i64 M2_mmpyl_rs1(i64, i64);
    /// The `llvm.hexagon.M2.mmpyh.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyh_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyh.rs0")
    i64 M2_mmpyh_rs0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyh.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyh_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyh.rs1")
    i64 M2_mmpyh_rs1(i64, i64);
    /// The `llvm.hexagon.M4.vrmpyeh.s0` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyeh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyeh.s0")
    i64 M4_vrmpyeh_s0(i64, i64);
    /// The `llvm.hexagon.M4.vrmpyeh.s1` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyeh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyeh.s1")
    i64 M4_vrmpyeh_s1(i64, i64);
    /// The `llvm.hexagon.M4.vrmpyeh.acc.s0` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyeh_acc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyeh.acc.s0")
    i64 M4_vrmpyeh_acc_s0(i64, i64, i64);
    /// The `llvm.hexagon.M4.vrmpyeh.acc.s1` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyeh_acc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyeh.acc.s1")
    i64 M4_vrmpyeh_acc_s1(i64, i64, i64);
    /// The `llvm.hexagon.M4.vrmpyoh.s0` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyoh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyoh.s0")
    i64 M4_vrmpyoh_s0(i64, i64);
    /// The `llvm.hexagon.M4.vrmpyoh.s1` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyoh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyoh.s1")
    i64 M4_vrmpyoh_s1(i64, i64);
    /// The `llvm.hexagon.M4.vrmpyoh.acc.s0` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyoh_acc_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyoh.acc.s0")
    i64 M4_vrmpyoh_acc_s0(i64, i64, i64);
    /// The `llvm.hexagon.M4.vrmpyoh.acc.s1` intrinsic; known as `__builtin_HEXAGON_M4_vrmpyoh_acc_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vrmpyoh.acc.s1")
    i64 M4_vrmpyoh_acc_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.hmmpyl.rs1` intrinsic; known as `__builtin_HEXAGON_M2_hmmpyl_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.hmmpyl.rs1")
    i32 M2_hmmpyl_rs1(i32, i32);
    /// The `llvm.hexagon.M2.hmmpyh.rs1` intrinsic; known as `__builtin_HEXAGON_M2_hmmpyh_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.hmmpyh.rs1")
    i32 M2_hmmpyh_rs1(i32, i32);
    /// The `llvm.hexagon.M2.hmmpyl.s1` intrinsic; known as `__builtin_HEXAGON_M2_hmmpyl_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.hmmpyl.s1")
    i32 M2_hmmpyl_s1(i32, i32);
    /// The `llvm.hexagon.M2.hmmpyh.s1` intrinsic; known as `__builtin_HEXAGON_M2_hmmpyh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.hmmpyh.s1")
    i32 M2_hmmpyh_s1(i32, i32);
    /// The `llvm.hexagon.M2.mmaculs.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmaculs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmaculs.s0")
    i64 M2_mmaculs_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmaculs.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmaculs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmaculs.s1")
    i64 M2_mmaculs_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmacuhs.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmacuhs_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacuhs.s0")
    i64 M2_mmacuhs_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmacuhs.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmacuhs_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacuhs.s1")
    i64 M2_mmacuhs_s1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmpyul.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyul_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyul.s0")
    i64 M2_mmpyul_s0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyul.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyul_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyul.s1")
    i64 M2_mmpyul_s1(i64, i64);
    /// The `llvm.hexagon.M2.mmpyuh.s0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyuh_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyuh.s0")
    i64 M2_mmpyuh_s0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyuh.s1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyuh_s1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyuh.s1")
    i64 M2_mmpyuh_s1(i64, i64);
    /// The `llvm.hexagon.M2.mmaculs.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmaculs_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmaculs.rs0")
    i64 M2_mmaculs_rs0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmaculs.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmaculs_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmaculs.rs1")
    i64 M2_mmaculs_rs1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmacuhs.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmacuhs_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacuhs.rs0")
    i64 M2_mmacuhs_rs0(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmacuhs.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmacuhs_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmacuhs.rs1")
    i64 M2_mmacuhs_rs1(i64, i64, i64);
    /// The `llvm.hexagon.M2.mmpyul.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyul_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyul.rs0")
    i64 M2_mmpyul_rs0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyul.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyul_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyul.rs1")
    i64 M2_mmpyul_rs1(i64, i64);
    /// The `llvm.hexagon.M2.mmpyuh.rs0` intrinsic; known as `__builtin_HEXAGON_M2_mmpyuh_rs0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyuh.rs0")
    i64 M2_mmpyuh_rs0(i64, i64);
    /// The `llvm.hexagon.M2.mmpyuh.rs1` intrinsic; known as `__builtin_HEXAGON_M2_mmpyuh_rs1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.mmpyuh.rs1")
    i64 M2_mmpyuh_rs1(i64, i64);
    /// The `llvm.hexagon.M2.vrcmaci.s0` intrinsic; known as `__builtin_HEXAGON_M2_vrcmaci_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmaci.s0")
    i64 M2_vrcmaci_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.vrcmacr.s0` intrinsic; known as `__builtin_HEXAGON_M2_vrcmacr_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmacr.s0")
    i64 M2_vrcmacr_s0(i64, i64, i64);
    /// The `llvm.hexagon.M2.vrcmaci.s0c` intrinsic; known as `__builtin_HEXAGON_M2_vrcmaci_s0c` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmaci.s0c")
    i64 M2_vrcmaci_s0c(i64, i64, i64);
    /// The `llvm.hexagon.M2.vrcmacr.s0c` intrinsic; known as `__builtin_HEXAGON_M2_vrcmacr_s0c` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmacr.s0c")
    i64 M2_vrcmacr_s0c(i64, i64, i64);
    /// The `llvm.hexagon.M2.cmaci.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmaci_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmaci.s0")
    i64 M2_cmaci_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.cmacr.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmacr_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmacr.s0")
    i64 M2_cmacr_s0(i64, i32, i32);
    /// The `llvm.hexagon.M2.vrcmpyi.s0` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpyi_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpyi.s0")
    i64 M2_vrcmpyi_s0(i64, i64);
    /// The `llvm.hexagon.M2.vrcmpyr.s0` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpyr_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpyr.s0")
    i64 M2_vrcmpyr_s0(i64, i64);
    /// The `llvm.hexagon.M2.vrcmpyi.s0c` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpyi_s0c` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpyi.s0c")
    i64 M2_vrcmpyi_s0c(i64, i64);
    /// The `llvm.hexagon.M2.vrcmpyr.s0c` intrinsic; known as `__builtin_HEXAGON_M2_vrcmpyr_s0c` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vrcmpyr.s0c")
    i64 M2_vrcmpyr_s0c(i64, i64);
    /// The `llvm.hexagon.M2.cmpyi.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmpyi_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpyi.s0")
    i64 M2_cmpyi_s0(i32, i32);
    /// The `llvm.hexagon.M2.cmpyr.s0` intrinsic; known as `__builtin_HEXAGON_M2_cmpyr_s0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.cmpyr.s0")
    i64 M2_cmpyr_s0(i32, i32);
    /// The `llvm.hexagon.M4.cmpyi.wh` intrinsic; known as `__builtin_HEXAGON_M4_cmpyi_wh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.cmpyi.wh")
    i32 M4_cmpyi_wh(i64, i32);
    /// The `llvm.hexagon.M4.cmpyr.wh` intrinsic; known as `__builtin_HEXAGON_M4_cmpyr_wh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.cmpyr.wh")
    i32 M4_cmpyr_wh(i64, i32);
    /// The `llvm.hexagon.M4.cmpyi.whc` intrinsic; known as `__builtin_HEXAGON_M4_cmpyi_whc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.cmpyi.whc")
    i32 M4_cmpyi_whc(i64, i32);
    /// The `llvm.hexagon.M4.cmpyr.whc` intrinsic; known as `__builtin_HEXAGON_M4_cmpyr_whc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.cmpyr.whc")
    i32 M4_cmpyr_whc(i64, i32);
    /// The `llvm.hexagon.M2.vcmpy.s0.sat.i` intrinsic; known as `__builtin_HEXAGON_M2_vcmpy_s0_sat_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vcmpy.s0.sat.i")
    i64 M2_vcmpy_s0_sat_i(i64, i64);
    /// The `llvm.hexagon.M2.vcmpy.s0.sat.r` intrinsic; known as `__builtin_HEXAGON_M2_vcmpy_s0_sat_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vcmpy.s0.sat.r")
    i64 M2_vcmpy_s0_sat_r(i64, i64);
    /// The `llvm.hexagon.M2.vcmpy.s1.sat.i` intrinsic; known as `__builtin_HEXAGON_M2_vcmpy_s1_sat_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vcmpy.s1.sat.i")
    i64 M2_vcmpy_s1_sat_i(i64, i64);
    /// The `llvm.hexagon.M2.vcmpy.s1.sat.r` intrinsic; known as `__builtin_HEXAGON_M2_vcmpy_s1_sat_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vcmpy.s1.sat.r")
    i64 M2_vcmpy_s1_sat_r(i64, i64);
    /// The `llvm.hexagon.M2.vcmac.s0.sat.i` intrinsic; known as `__builtin_HEXAGON_M2_vcmac_s0_sat_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vcmac.s0.sat.i")
    i64 M2_vcmac_s0_sat_i(i64, i64, i64);
    /// The `llvm.hexagon.M2.vcmac.s0.sat.r` intrinsic; known as `__builtin_HEXAGON_M2_vcmac_s0_sat_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vcmac.s0.sat.r")
    i64 M2_vcmac_s0_sat_r(i64, i64, i64);
    /// The `llvm.hexagon.S2.vcrotate` intrinsic; known as `__builtin_HEXAGON_S2_vcrotate` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vcrotate")
    i64 S2_vcrotate(i64, i32);
    /// The `llvm.hexagon.S4.vrcrotate.acc` intrinsic; known as `__builtin_HEXAGON_S4_vrcrotate_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vrcrotate.acc")
    i64 S4_vrcrotate_acc(i64, i64, i32, i32);
    /// The `llvm.hexagon.S4.vrcrotate` intrinsic; known as `__builtin_HEXAGON_S4_vrcrotate` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vrcrotate")
    i64 S4_vrcrotate(i64, i32, i32);
    /// The `llvm.hexagon.S2.vcnegh` intrinsic; known as `__builtin_HEXAGON_S2_vcnegh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vcnegh")
    i64 S2_vcnegh(i64, i32);
    /// The `llvm.hexagon.S2.vrcnegh` intrinsic; known as `__builtin_HEXAGON_S2_vrcnegh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vrcnegh")
    i64 S2_vrcnegh(i64, i64, i32);
    /// The `llvm.hexagon.M4.pmpyw` intrinsic; known as `__builtin_HEXAGON_M4_pmpyw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.pmpyw")
    i64 M4_pmpyw(i32, i32);
    /// The `llvm.hexagon.M4.vpmpyh` intrinsic; known as `__builtin_HEXAGON_M4_vpmpyh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vpmpyh")
    i64 M4_vpmpyh(i32, i32);
    /// The `llvm.hexagon.M4.pmpyw.acc` intrinsic; known as `__builtin_HEXAGON_M4_pmpyw_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.pmpyw.acc")
    i64 M4_pmpyw_acc(i64, i32, i32);
    /// The `llvm.hexagon.M4.vpmpyh.acc` intrinsic; known as `__builtin_HEXAGON_M4_vpmpyh_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.vpmpyh.acc")
    i64 M4_vpmpyh_acc(i64, i32, i32);
    /// The `llvm.hexagon.A2.add` intrinsic; known as `__builtin_HEXAGON_A2_add` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.add")
    i32 A2_add(i32, i32);
    /// The `llvm.hexagon.A2.sub` intrinsic; known as `__builtin_HEXAGON_A2_sub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.sub")
    i32 A2_sub(i32, i32);
    /// The `llvm.hexagon.A2.addsat` intrinsic; known as `__builtin_HEXAGON_A2_addsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addsat")
    i32 A2_addsat(i32, i32);
    /// The `llvm.hexagon.A2.subsat` intrinsic; known as `__builtin_HEXAGON_A2_subsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subsat")
    i32 A2_subsat(i32, i32);
    /// The `llvm.hexagon.A2.addi` intrinsic; known as `__builtin_HEXAGON_A2_addi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addi")
    i32 A2_addi(i32, i32);
    /// The `llvm.hexagon.A2.addh.l16.ll` intrinsic; known as `__builtin_HEXAGON_A2_addh_l16_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.l16.ll")
    i32 A2_addh_l16_ll(i32, i32);
    /// The `llvm.hexagon.A2.addh.l16.hl` intrinsic; known as `__builtin_HEXAGON_A2_addh_l16_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.l16.hl")
    i32 A2_addh_l16_hl(i32, i32);
    /// The `llvm.hexagon.A2.addh.l16.sat.ll` intrinsic; known as `__builtin_HEXAGON_A2_addh_l16_sat_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.l16.sat.ll")
    i32 A2_addh_l16_sat_ll(i32, i32);
    /// The `llvm.hexagon.A2.addh.l16.sat.hl` intrinsic; known as `__builtin_HEXAGON_A2_addh_l16_sat_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.l16.sat.hl")
    i32 A2_addh_l16_sat_hl(i32, i32);
    /// The `llvm.hexagon.A2.subh.l16.ll` intrinsic; known as `__builtin_HEXAGON_A2_subh_l16_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.l16.ll")
    i32 A2_subh_l16_ll(i32, i32);
    /// The `llvm.hexagon.A2.subh.l16.hl` intrinsic; known as `__builtin_HEXAGON_A2_subh_l16_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.l16.hl")
    i32 A2_subh_l16_hl(i32, i32);
    /// The `llvm.hexagon.A2.subh.l16.sat.ll` intrinsic; known as `__builtin_HEXAGON_A2_subh_l16_sat_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.l16.sat.ll")
    i32 A2_subh_l16_sat_ll(i32, i32);
    /// The `llvm.hexagon.A2.subh.l16.sat.hl` intrinsic; known as `__builtin_HEXAGON_A2_subh_l16_sat_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.l16.sat.hl")
    i32 A2_subh_l16_sat_hl(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.ll` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.ll")
    i32 A2_addh_h16_ll(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.lh` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_lh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.lh")
    i32 A2_addh_h16_lh(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.hl` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.hl")
    i32 A2_addh_h16_hl(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.hh` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_hh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.hh")
    i32 A2_addh_h16_hh(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.sat.ll` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_sat_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.sat.ll")
    i32 A2_addh_h16_sat_ll(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.sat.lh` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_sat_lh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.sat.lh")
    i32 A2_addh_h16_sat_lh(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.sat.hl` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_sat_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.sat.hl")
    i32 A2_addh_h16_sat_hl(i32, i32);
    /// The `llvm.hexagon.A2.addh.h16.sat.hh` intrinsic; known as `__builtin_HEXAGON_A2_addh_h16_sat_hh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addh.h16.sat.hh")
    i32 A2_addh_h16_sat_hh(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.ll` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.ll")
    i32 A2_subh_h16_ll(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.lh` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_lh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.lh")
    i32 A2_subh_h16_lh(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.hl` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.hl")
    i32 A2_subh_h16_hl(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.hh` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_hh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.hh")
    i32 A2_subh_h16_hh(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.sat.ll` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_sat_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.sat.ll")
    i32 A2_subh_h16_sat_ll(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.sat.lh` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_sat_lh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.sat.lh")
    i32 A2_subh_h16_sat_lh(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.sat.hl` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_sat_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.sat.hl")
    i32 A2_subh_h16_sat_hl(i32, i32);
    /// The `llvm.hexagon.A2.subh.h16.sat.hh` intrinsic; known as `__builtin_HEXAGON_A2_subh_h16_sat_hh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subh.h16.sat.hh")
    i32 A2_subh_h16_sat_hh(i32, i32);
    /// The `llvm.hexagon.A2.aslh` intrinsic; known as `__builtin_HEXAGON_A2_aslh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.aslh")
    i32 A2_aslh(i32);
    /// The `llvm.hexagon.A2.asrh` intrinsic; known as `__builtin_HEXAGON_A2_asrh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.asrh")
    i32 A2_asrh(i32);
    /// The `llvm.hexagon.A2.addp` intrinsic; known as `__builtin_HEXAGON_A2_addp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addp")
    i64 A2_addp(i64, i64);
    /// The `llvm.hexagon.A2.addpsat` intrinsic; known as `__builtin_HEXAGON_A2_addpsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addpsat")
    i64 A2_addpsat(i64, i64);
    /// The `llvm.hexagon.A2.addsp` intrinsic; known as `__builtin_HEXAGON_A2_addsp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.addsp")
    i64 A2_addsp(i32, i64);
    /// The `llvm.hexagon.A2.subp` intrinsic; known as `__builtin_HEXAGON_A2_subp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subp")
    i64 A2_subp(i64, i64);
    /// The `llvm.hexagon.A2.neg` intrinsic; known as `__builtin_HEXAGON_A2_neg` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.neg")
    i32 A2_neg(i32);
    /// The `llvm.hexagon.A2.negsat` intrinsic; known as `__builtin_HEXAGON_A2_negsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.negsat")
    i32 A2_negsat(i32);
    /// The `llvm.hexagon.A2.abs` intrinsic; known as `__builtin_HEXAGON_A2_abs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.abs")
    i32 A2_abs(i32);
    /// The `llvm.hexagon.A2.abssat` intrinsic; known as `__builtin_HEXAGON_A2_abssat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.abssat")
    i32 A2_abssat(i32);
    /// The `llvm.hexagon.A2.vconj` intrinsic; known as `__builtin_HEXAGON_A2_vconj` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vconj")
    i64 A2_vconj(i64);
    /// The `llvm.hexagon.A2.negp` intrinsic; known as `__builtin_HEXAGON_A2_negp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.negp")
    i64 A2_negp(i64);
    /// The `llvm.hexagon.A2.absp` intrinsic; known as `__builtin_HEXAGON_A2_absp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.absp")
    i64 A2_absp(i64);
    /// The `llvm.hexagon.A2.max` intrinsic; known as `__builtin_HEXAGON_A2_max` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.max")
    i32 A2_max(i32, i32);
    /// The `llvm.hexagon.A2.maxu` intrinsic; known as `__builtin_HEXAGON_A2_maxu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.maxu")
    i32 A2_maxu(i32, i32);
    /// The `llvm.hexagon.A2.min` intrinsic; known as `__builtin_HEXAGON_A2_min` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.min")
    i32 A2_min(i32, i32);
    /// The `llvm.hexagon.A2.minu` intrinsic; known as `__builtin_HEXAGON_A2_minu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.minu")
    i32 A2_minu(i32, i32);
    /// The `llvm.hexagon.A2.maxp` intrinsic; known as `__builtin_HEXAGON_A2_maxp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.maxp")
    i64 A2_maxp(i64, i64);
    /// The `llvm.hexagon.A2.maxup` intrinsic; known as `__builtin_HEXAGON_A2_maxup` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.maxup")
    i64 A2_maxup(i64, i64);
    /// The `llvm.hexagon.A2.minp` intrinsic; known as `__builtin_HEXAGON_A2_minp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.minp")
    i64 A2_minp(i64, i64);
    /// The `llvm.hexagon.A2.minup` intrinsic; known as `__builtin_HEXAGON_A2_minup` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.minup")
    i64 A2_minup(i64, i64);
    /// The `llvm.hexagon.A2.tfr` intrinsic; known as `__builtin_HEXAGON_A2_tfr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.tfr")
    i32 A2_tfr(i32);
    /// The `llvm.hexagon.A2.tfrsi` intrinsic; known as `__builtin_HEXAGON_A2_tfrsi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.tfrsi")
    i32 A2_tfrsi(i32);
    /// The `llvm.hexagon.A2.tfrp` intrinsic; known as `__builtin_HEXAGON_A2_tfrp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.tfrp")
    i64 A2_tfrp(i64);
    /// The `llvm.hexagon.A2.tfrpi` intrinsic; known as `__builtin_HEXAGON_A2_tfrpi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.tfrpi")
    i64 A2_tfrpi(i32);
    /// The `llvm.hexagon.A2.zxtb` intrinsic; known as `__builtin_HEXAGON_A2_zxtb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.zxtb")
    i32 A2_zxtb(i32);
    /// The `llvm.hexagon.A2.sxtb` intrinsic; known as `__builtin_HEXAGON_A2_sxtb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.sxtb")
    i32 A2_sxtb(i32);
    /// The `llvm.hexagon.A2.zxth` intrinsic; known as `__builtin_HEXAGON_A2_zxth` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.zxth")
    i32 A2_zxth(i32);
    /// The `llvm.hexagon.A2.sxth` intrinsic; known as `__builtin_HEXAGON_A2_sxth` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.sxth")
    i32 A2_sxth(i32);
    /// The `llvm.hexagon.A2.combinew` intrinsic; known as `__builtin_HEXAGON_A2_combinew` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.combinew")
    i64 A2_combinew(i32, i32);
    /// The `llvm.hexagon.A4.combineri` intrinsic; known as `__builtin_HEXAGON_A4_combineri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.combineri")
    i64 A4_combineri(i32, i32);
    /// The `llvm.hexagon.A4.combineir` intrinsic; known as `__builtin_HEXAGON_A4_combineir` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.combineir")
    i64 A4_combineir(i32, i32);
    /// The `llvm.hexagon.A2.combineii` intrinsic; known as `__builtin_HEXAGON_A2_combineii` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.combineii")
    i64 A2_combineii(i32, i32);
    /// The `llvm.hexagon.A2.combine.hh` intrinsic; known as `__builtin_HEXAGON_A2_combine_hh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.combine.hh")
    i32 A2_combine_hh(i32, i32);
    /// The `llvm.hexagon.A2.combine.hl` intrinsic; known as `__builtin_HEXAGON_A2_combine_hl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.combine.hl")
    i32 A2_combine_hl(i32, i32);
    /// The `llvm.hexagon.A2.combine.lh` intrinsic; known as `__builtin_HEXAGON_A2_combine_lh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.combine.lh")
    i32 A2_combine_lh(i32, i32);
    /// The `llvm.hexagon.A2.combine.ll` intrinsic; known as `__builtin_HEXAGON_A2_combine_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.combine.ll")
    i32 A2_combine_ll(i32, i32);
    /// The `llvm.hexagon.A2.tfril` intrinsic; known as `__builtin_HEXAGON_A2_tfril` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.tfril")
    i32 A2_tfril(i32, i32);
    /// The `llvm.hexagon.A2.tfrih` intrinsic; known as `__builtin_HEXAGON_A2_tfrih` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.tfrih")
    i32 A2_tfrih(i32, i32);
    /// The `llvm.hexagon.A2.and` intrinsic; known as `__builtin_HEXAGON_A2_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.and")
    i32 A2_and(i32, i32);
    /// The `llvm.hexagon.A2.or` intrinsic; known as `__builtin_HEXAGON_A2_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.or")
    i32 A2_or(i32, i32);
    /// The `llvm.hexagon.A2.xor` intrinsic; known as `__builtin_HEXAGON_A2_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.xor")
    i32 A2_xor(i32, i32);
    /// The `llvm.hexagon.A2.not` intrinsic; known as `__builtin_HEXAGON_A2_not` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.not")
    i32 A2_not(i32);
    /// The `llvm.hexagon.M2.xor.xacc` intrinsic; known as `__builtin_HEXAGON_M2_xor_xacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.xor.xacc")
    i32 M2_xor_xacc(i32, i32, i32);
    /// The `llvm.hexagon.M4.xor.xacc` intrinsic; known as `__builtin_HEXAGON_M4_xor_xacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.xor.xacc")
    i64 M4_xor_xacc(i64, i64, i64);
    /// The `llvm.hexagon.A4.andn` intrinsic; known as `__builtin_HEXAGON_A4_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.andn")
    i32 A4_andn(i32, i32);
    /// The `llvm.hexagon.A4.orn` intrinsic; known as `__builtin_HEXAGON_A4_orn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.orn")
    i32 A4_orn(i32, i32);
    /// The `llvm.hexagon.A4.andnp` intrinsic; known as `__builtin_HEXAGON_A4_andnp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.andnp")
    i64 A4_andnp(i64, i64);
    /// The `llvm.hexagon.A4.ornp` intrinsic; known as `__builtin_HEXAGON_A4_ornp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.ornp")
    i64 A4_ornp(i64, i64);
    /// The `llvm.hexagon.S4.addaddi` intrinsic; known as `__builtin_HEXAGON_S4_addaddi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.addaddi")
    i32 S4_addaddi(i32, i32, i32);
    /// The `llvm.hexagon.S4.subaddi` intrinsic; known as `__builtin_HEXAGON_S4_subaddi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.subaddi")
    i32 S4_subaddi(i32, i32, i32);
    /// The `llvm.hexagon.M4.and.and` intrinsic; known as `__builtin_HEXAGON_M4_and_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.and.and")
    i32 M4_and_and(i32, i32, i32);
    /// The `llvm.hexagon.M4.and.andn` intrinsic; known as `__builtin_HEXAGON_M4_and_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.and.andn")
    i32 M4_and_andn(i32, i32, i32);
    /// The `llvm.hexagon.M4.and.or` intrinsic; known as `__builtin_HEXAGON_M4_and_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.and.or")
    i32 M4_and_or(i32, i32, i32);
    /// The `llvm.hexagon.M4.and.xor` intrinsic; known as `__builtin_HEXAGON_M4_and_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.and.xor")
    i32 M4_and_xor(i32, i32, i32);
    /// The `llvm.hexagon.M4.or.and` intrinsic; known as `__builtin_HEXAGON_M4_or_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.or.and")
    i32 M4_or_and(i32, i32, i32);
    /// The `llvm.hexagon.M4.or.andn` intrinsic; known as `__builtin_HEXAGON_M4_or_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.or.andn")
    i32 M4_or_andn(i32, i32, i32);
    /// The `llvm.hexagon.M4.or.or` intrinsic; known as `__builtin_HEXAGON_M4_or_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.or.or")
    i32 M4_or_or(i32, i32, i32);
    /// The `llvm.hexagon.M4.or.xor` intrinsic; known as `__builtin_HEXAGON_M4_or_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.or.xor")
    i32 M4_or_xor(i32, i32, i32);
    /// The `llvm.hexagon.S4.or.andix` intrinsic; known as `__builtin_HEXAGON_S4_or_andix` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.or.andix")
    i32 S4_or_andix(i32, i32, i32);
    /// The `llvm.hexagon.S4.or.andi` intrinsic; known as `__builtin_HEXAGON_S4_or_andi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.or.andi")
    i32 S4_or_andi(i32, i32, i32);
    /// The `llvm.hexagon.S4.or.ori` intrinsic; known as `__builtin_HEXAGON_S4_or_ori` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.or.ori")
    i32 S4_or_ori(i32, i32, i32);
    /// The `llvm.hexagon.M4.xor.and` intrinsic; known as `__builtin_HEXAGON_M4_xor_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.xor.and")
    i32 M4_xor_and(i32, i32, i32);
    /// The `llvm.hexagon.M4.xor.or` intrinsic; known as `__builtin_HEXAGON_M4_xor_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.xor.or")
    i32 M4_xor_or(i32, i32, i32);
    /// The `llvm.hexagon.M4.xor.andn` intrinsic; known as `__builtin_HEXAGON_M4_xor_andn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M4.xor.andn")
    i32 M4_xor_andn(i32, i32, i32);
    /// The `llvm.hexagon.A2.subri` intrinsic; known as `__builtin_HEXAGON_A2_subri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.subri")
    i32 A2_subri(i32, i32);
    /// The `llvm.hexagon.A2.andir` intrinsic; known as `__builtin_HEXAGON_A2_andir` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.andir")
    i32 A2_andir(i32, i32);
    /// The `llvm.hexagon.A2.orir` intrinsic; known as `__builtin_HEXAGON_A2_orir` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.orir")
    i32 A2_orir(i32, i32);
    /// The `llvm.hexagon.A2.andp` intrinsic; known as `__builtin_HEXAGON_A2_andp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.andp")
    i64 A2_andp(i64, i64);
    /// The `llvm.hexagon.A2.orp` intrinsic; known as `__builtin_HEXAGON_A2_orp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.orp")
    i64 A2_orp(i64, i64);
    /// The `llvm.hexagon.A2.xorp` intrinsic; known as `__builtin_HEXAGON_A2_xorp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.xorp")
    i64 A2_xorp(i64, i64);
    /// The `llvm.hexagon.A2.notp` intrinsic; known as `__builtin_HEXAGON_A2_notp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.notp")
    i64 A2_notp(i64);
    /// The `llvm.hexagon.A2.sxtw` intrinsic; known as `__builtin_HEXAGON_A2_sxtw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.sxtw")
    i64 A2_sxtw(i32);
    /// The `llvm.hexagon.A2.sat` intrinsic; known as `__builtin_HEXAGON_A2_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.sat")
    i32 A2_sat(i64);
    /// The `llvm.hexagon.A2.roundsat` intrinsic; known as `__builtin_HEXAGON_A2_roundsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.roundsat")
    i32 A2_roundsat(i64);
    /// The `llvm.hexagon.A2.sath` intrinsic; known as `__builtin_HEXAGON_A2_sath` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.sath")
    i32 A2_sath(i32);
    /// The `llvm.hexagon.A2.satuh` intrinsic; known as `__builtin_HEXAGON_A2_satuh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.satuh")
    i32 A2_satuh(i32);
    /// The `llvm.hexagon.A2.satub` intrinsic; known as `__builtin_HEXAGON_A2_satub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.satub")
    i32 A2_satub(i32);
    /// The `llvm.hexagon.A2.satb` intrinsic; known as `__builtin_HEXAGON_A2_satb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.satb")
    i32 A2_satb(i32);
    /// The `llvm.hexagon.A2.vaddub` intrinsic; known as `__builtin_HEXAGON_A2_vaddub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddub")
    i64 A2_vaddub(i64, i64);
    /// The `llvm.hexagon.A2.vaddb.map` intrinsic; known as `__builtin_HEXAGON_A2_vaddb_map` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddb.map")
    i64 A2_vaddb_map(i64, i64);
    /// The `llvm.hexagon.A2.vaddubs` intrinsic; known as `__builtin_HEXAGON_A2_vaddubs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddubs")
    i64 A2_vaddubs(i64, i64);
    /// The `llvm.hexagon.A2.vaddh` intrinsic; known as `__builtin_HEXAGON_A2_vaddh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddh")
    i64 A2_vaddh(i64, i64);
    /// The `llvm.hexagon.A2.vaddhs` intrinsic; known as `__builtin_HEXAGON_A2_vaddhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddhs")
    i64 A2_vaddhs(i64, i64);
    /// The `llvm.hexagon.A2.vadduhs` intrinsic; known as `__builtin_HEXAGON_A2_vadduhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vadduhs")
    i64 A2_vadduhs(i64, i64);
    /// The `llvm.hexagon.A5.vaddhubs` intrinsic; known as `__builtin_HEXAGON_A5_vaddhubs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A5.vaddhubs")
    i32 A5_vaddhubs(i64, i64);
    /// The `llvm.hexagon.A2.vaddw` intrinsic; known as `__builtin_HEXAGON_A2_vaddw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddw")
    i64 A2_vaddw(i64, i64);
    /// The `llvm.hexagon.A2.vaddws` intrinsic; known as `__builtin_HEXAGON_A2_vaddws` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vaddws")
    i64 A2_vaddws(i64, i64);
    /// The `llvm.hexagon.S4.vxaddsubw` intrinsic; known as `__builtin_HEXAGON_S4_vxaddsubw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vxaddsubw")
    i64 S4_vxaddsubw(i64, i64);
    /// The `llvm.hexagon.S4.vxsubaddw` intrinsic; known as `__builtin_HEXAGON_S4_vxsubaddw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vxsubaddw")
    i64 S4_vxsubaddw(i64, i64);
    /// The `llvm.hexagon.S4.vxaddsubh` intrinsic; known as `__builtin_HEXAGON_S4_vxaddsubh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vxaddsubh")
    i64 S4_vxaddsubh(i64, i64);
    /// The `llvm.hexagon.S4.vxsubaddh` intrinsic; known as `__builtin_HEXAGON_S4_vxsubaddh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vxsubaddh")
    i64 S4_vxsubaddh(i64, i64);
    /// The `llvm.hexagon.S4.vxaddsubhr` intrinsic; known as `__builtin_HEXAGON_S4_vxaddsubhr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vxaddsubhr")
    i64 S4_vxaddsubhr(i64, i64);
    /// The `llvm.hexagon.S4.vxsubaddhr` intrinsic; known as `__builtin_HEXAGON_S4_vxsubaddhr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.vxsubaddhr")
    i64 S4_vxsubaddhr(i64, i64);
    /// The `llvm.hexagon.A2.svavgh` intrinsic; known as `__builtin_HEXAGON_A2_svavgh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svavgh")
    i32 A2_svavgh(i32, i32);
    /// The `llvm.hexagon.A2.svavghs` intrinsic; known as `__builtin_HEXAGON_A2_svavghs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svavghs")
    i32 A2_svavghs(i32, i32);
    /// The `llvm.hexagon.A2.svnavgh` intrinsic; known as `__builtin_HEXAGON_A2_svnavgh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svnavgh")
    i32 A2_svnavgh(i32, i32);
    /// The `llvm.hexagon.A2.svaddh` intrinsic; known as `__builtin_HEXAGON_A2_svaddh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svaddh")
    i32 A2_svaddh(i32, i32);
    /// The `llvm.hexagon.A2.svaddhs` intrinsic; known as `__builtin_HEXAGON_A2_svaddhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svaddhs")
    i32 A2_svaddhs(i32, i32);
    /// The `llvm.hexagon.A2.svadduhs` intrinsic; known as `__builtin_HEXAGON_A2_svadduhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svadduhs")
    i32 A2_svadduhs(i32, i32);
    /// The `llvm.hexagon.A2.svsubh` intrinsic; known as `__builtin_HEXAGON_A2_svsubh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svsubh")
    i32 A2_svsubh(i32, i32);
    /// The `llvm.hexagon.A2.svsubhs` intrinsic; known as `__builtin_HEXAGON_A2_svsubhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svsubhs")
    i32 A2_svsubhs(i32, i32);
    /// The `llvm.hexagon.A2.svsubuhs` intrinsic; known as `__builtin_HEXAGON_A2_svsubuhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.svsubuhs")
    i32 A2_svsubuhs(i32, i32);
    /// The `llvm.hexagon.A2.vraddub` intrinsic; known as `__builtin_HEXAGON_A2_vraddub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vraddub")
    i64 A2_vraddub(i64, i64);
    /// The `llvm.hexagon.A2.vraddub.acc` intrinsic; known as `__builtin_HEXAGON_A2_vraddub_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vraddub.acc")
    i64 A2_vraddub_acc(i64, i64, i64);
    /// The `llvm.hexagon.M2.vraddh` intrinsic; known as `__builtin_HEXAGON_M2_vraddh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vraddh")
    i32 M2_vraddh(i64, i64);
    /// The `llvm.hexagon.M2.vradduh` intrinsic; known as `__builtin_HEXAGON_M2_vradduh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vradduh")
    i32 M2_vradduh(i64, i64);
    /// The `llvm.hexagon.A2.vsubub` intrinsic; known as `__builtin_HEXAGON_A2_vsubub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubub")
    i64 A2_vsubub(i64, i64);
    /// The `llvm.hexagon.A2.vsubb.map` intrinsic; known as `__builtin_HEXAGON_A2_vsubb_map` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubb.map")
    i64 A2_vsubb_map(i64, i64);
    /// The `llvm.hexagon.A2.vsububs` intrinsic; known as `__builtin_HEXAGON_A2_vsububs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsububs")
    i64 A2_vsububs(i64, i64);
    /// The `llvm.hexagon.A2.vsubh` intrinsic; known as `__builtin_HEXAGON_A2_vsubh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubh")
    i64 A2_vsubh(i64, i64);
    /// The `llvm.hexagon.A2.vsubhs` intrinsic; known as `__builtin_HEXAGON_A2_vsubhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubhs")
    i64 A2_vsubhs(i64, i64);
    /// The `llvm.hexagon.A2.vsubuhs` intrinsic; known as `__builtin_HEXAGON_A2_vsubuhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubuhs")
    i64 A2_vsubuhs(i64, i64);
    /// The `llvm.hexagon.A2.vsubw` intrinsic; known as `__builtin_HEXAGON_A2_vsubw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubw")
    i64 A2_vsubw(i64, i64);
    /// The `llvm.hexagon.A2.vsubws` intrinsic; known as `__builtin_HEXAGON_A2_vsubws` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vsubws")
    i64 A2_vsubws(i64, i64);
    /// The `llvm.hexagon.A2.vabsh` intrinsic; known as `__builtin_HEXAGON_A2_vabsh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vabsh")
    i64 A2_vabsh(i64);
    /// The `llvm.hexagon.A2.vabshsat` intrinsic; known as `__builtin_HEXAGON_A2_vabshsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vabshsat")
    i64 A2_vabshsat(i64);
    /// The `llvm.hexagon.A2.vabsw` intrinsic; known as `__builtin_HEXAGON_A2_vabsw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vabsw")
    i64 A2_vabsw(i64);
    /// The `llvm.hexagon.A2.vabswsat` intrinsic; known as `__builtin_HEXAGON_A2_vabswsat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vabswsat")
    i64 A2_vabswsat(i64);
    /// The `llvm.hexagon.M2.vabsdiffw` intrinsic; known as `__builtin_HEXAGON_M2_vabsdiffw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vabsdiffw")
    i64 M2_vabsdiffw(i64, i64);
    /// The `llvm.hexagon.M2.vabsdiffh` intrinsic; known as `__builtin_HEXAGON_M2_vabsdiffh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.M2.vabsdiffh")
    i64 M2_vabsdiffh(i64, i64);
    /// The `llvm.hexagon.A2.vrsadub` intrinsic; known as `__builtin_HEXAGON_A2_vrsadub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vrsadub")
    i64 A2_vrsadub(i64, i64);
    /// The `llvm.hexagon.A2.vrsadub.acc` intrinsic; known as `__builtin_HEXAGON_A2_vrsadub_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vrsadub.acc")
    i64 A2_vrsadub_acc(i64, i64, i64);
    /// The `llvm.hexagon.A2.vavgub` intrinsic; known as `__builtin_HEXAGON_A2_vavgub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavgub")
    i64 A2_vavgub(i64, i64);
    /// The `llvm.hexagon.A2.vavguh` intrinsic; known as `__builtin_HEXAGON_A2_vavguh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavguh")
    i64 A2_vavguh(i64, i64);
    /// The `llvm.hexagon.A2.vavgh` intrinsic; known as `__builtin_HEXAGON_A2_vavgh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavgh")
    i64 A2_vavgh(i64, i64);
    /// The `llvm.hexagon.A2.vnavgh` intrinsic; known as `__builtin_HEXAGON_A2_vnavgh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vnavgh")
    i64 A2_vnavgh(i64, i64);
    /// The `llvm.hexagon.A2.vavgw` intrinsic; known as `__builtin_HEXAGON_A2_vavgw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavgw")
    i64 A2_vavgw(i64, i64);
    /// The `llvm.hexagon.A2.vnavgw` intrinsic; known as `__builtin_HEXAGON_A2_vnavgw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vnavgw")
    i64 A2_vnavgw(i64, i64);
    /// The `llvm.hexagon.A2.vavgwr` intrinsic; known as `__builtin_HEXAGON_A2_vavgwr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavgwr")
    i64 A2_vavgwr(i64, i64);
    /// The `llvm.hexagon.A2.vnavgwr` intrinsic; known as `__builtin_HEXAGON_A2_vnavgwr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vnavgwr")
    i64 A2_vnavgwr(i64, i64);
    /// The `llvm.hexagon.A2.vavgwcr` intrinsic; known as `__builtin_HEXAGON_A2_vavgwcr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavgwcr")
    i64 A2_vavgwcr(i64, i64);
    /// The `llvm.hexagon.A2.vnavgwcr` intrinsic; known as `__builtin_HEXAGON_A2_vnavgwcr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vnavgwcr")
    i64 A2_vnavgwcr(i64, i64);
    /// The `llvm.hexagon.A2.vavghcr` intrinsic; known as `__builtin_HEXAGON_A2_vavghcr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavghcr")
    i64 A2_vavghcr(i64, i64);
    /// The `llvm.hexagon.A2.vnavghcr` intrinsic; known as `__builtin_HEXAGON_A2_vnavghcr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vnavghcr")
    i64 A2_vnavghcr(i64, i64);
    /// The `llvm.hexagon.A2.vavguw` intrinsic; known as `__builtin_HEXAGON_A2_vavguw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavguw")
    i64 A2_vavguw(i64, i64);
    /// The `llvm.hexagon.A2.vavguwr` intrinsic; known as `__builtin_HEXAGON_A2_vavguwr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavguwr")
    i64 A2_vavguwr(i64, i64);
    /// The `llvm.hexagon.A2.vavgubr` intrinsic; known as `__builtin_HEXAGON_A2_vavgubr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavgubr")
    i64 A2_vavgubr(i64, i64);
    /// The `llvm.hexagon.A2.vavguhr` intrinsic; known as `__builtin_HEXAGON_A2_vavguhr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavguhr")
    i64 A2_vavguhr(i64, i64);
    /// The `llvm.hexagon.A2.vavghr` intrinsic; known as `__builtin_HEXAGON_A2_vavghr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vavghr")
    i64 A2_vavghr(i64, i64);
    /// The `llvm.hexagon.A2.vnavghr` intrinsic; known as `__builtin_HEXAGON_A2_vnavghr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vnavghr")
    i64 A2_vnavghr(i64, i64);
    /// The `llvm.hexagon.A4.round.ri` intrinsic; known as `__builtin_HEXAGON_A4_round_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.round.ri")
    i32 A4_round_ri(i32, i32);
    /// The `llvm.hexagon.A4.round.rr` intrinsic; known as `__builtin_HEXAGON_A4_round_rr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.round.rr")
    i32 A4_round_rr(i32, i32);
    /// The `llvm.hexagon.A4.round.ri.sat` intrinsic; known as `__builtin_HEXAGON_A4_round_ri_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.round.ri.sat")
    i32 A4_round_ri_sat(i32, i32);
    /// The `llvm.hexagon.A4.round.rr.sat` intrinsic; known as `__builtin_HEXAGON_A4_round_rr_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.round.rr.sat")
    i32 A4_round_rr_sat(i32, i32);
    /// The `llvm.hexagon.A4.cround.ri` intrinsic; known as `__builtin_HEXAGON_A4_cround_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cround.ri")
    i32 A4_cround_ri(i32, i32);
    /// The `llvm.hexagon.A4.cround.rr` intrinsic; known as `__builtin_HEXAGON_A4_cround_rr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.cround.rr")
    i32 A4_cround_rr(i32, i32);
    /// The `llvm.hexagon.A4.vrminh` intrinsic; known as `__builtin_HEXAGON_A4_vrminh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrminh")
    i64 A4_vrminh(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrmaxh` intrinsic; known as `__builtin_HEXAGON_A4_vrmaxh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrmaxh")
    i64 A4_vrmaxh(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrminuh` intrinsic; known as `__builtin_HEXAGON_A4_vrminuh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrminuh")
    i64 A4_vrminuh(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrmaxuh` intrinsic; known as `__builtin_HEXAGON_A4_vrmaxuh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrmaxuh")
    i64 A4_vrmaxuh(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrminw` intrinsic; known as `__builtin_HEXAGON_A4_vrminw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrminw")
    i64 A4_vrminw(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrmaxw` intrinsic; known as `__builtin_HEXAGON_A4_vrmaxw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrmaxw")
    i64 A4_vrmaxw(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrminuw` intrinsic; known as `__builtin_HEXAGON_A4_vrminuw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrminuw")
    i64 A4_vrminuw(i64, i64, i32);
    /// The `llvm.hexagon.A4.vrmaxuw` intrinsic; known as `__builtin_HEXAGON_A4_vrmaxuw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.vrmaxuw")
    i64 A4_vrmaxuw(i64, i64, i32);
    /// The `llvm.hexagon.A2.vminb` intrinsic; known as `__builtin_HEXAGON_A2_vminb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vminb")
    i64 A2_vminb(i64, i64);
    /// The `llvm.hexagon.A2.vmaxb` intrinsic; known as `__builtin_HEXAGON_A2_vmaxb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vmaxb")
    i64 A2_vmaxb(i64, i64);
    /// The `llvm.hexagon.A2.vminub` intrinsic; known as `__builtin_HEXAGON_A2_vminub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vminub")
    i64 A2_vminub(i64, i64);
    /// The `llvm.hexagon.A2.vmaxub` intrinsic; known as `__builtin_HEXAGON_A2_vmaxub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vmaxub")
    i64 A2_vmaxub(i64, i64);
    /// The `llvm.hexagon.A2.vminh` intrinsic; known as `__builtin_HEXAGON_A2_vminh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vminh")
    i64 A2_vminh(i64, i64);
    /// The `llvm.hexagon.A2.vmaxh` intrinsic; known as `__builtin_HEXAGON_A2_vmaxh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vmaxh")
    i64 A2_vmaxh(i64, i64);
    /// The `llvm.hexagon.A2.vminuh` intrinsic; known as `__builtin_HEXAGON_A2_vminuh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vminuh")
    i64 A2_vminuh(i64, i64);
    /// The `llvm.hexagon.A2.vmaxuh` intrinsic; known as `__builtin_HEXAGON_A2_vmaxuh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vmaxuh")
    i64 A2_vmaxuh(i64, i64);
    /// The `llvm.hexagon.A2.vminw` intrinsic; known as `__builtin_HEXAGON_A2_vminw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vminw")
    i64 A2_vminw(i64, i64);
    /// The `llvm.hexagon.A2.vmaxw` intrinsic; known as `__builtin_HEXAGON_A2_vmaxw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vmaxw")
    i64 A2_vmaxw(i64, i64);
    /// The `llvm.hexagon.A2.vminuw` intrinsic; known as `__builtin_HEXAGON_A2_vminuw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vminuw")
    i64 A2_vminuw(i64, i64);
    /// The `llvm.hexagon.A2.vmaxuw` intrinsic; known as `__builtin_HEXAGON_A2_vmaxuw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.vmaxuw")
    i64 A2_vmaxuw(i64, i64);
    /// The `llvm.hexagon.A4.modwrapu` intrinsic; known as `__builtin_HEXAGON_A4_modwrapu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.modwrapu")
    i32 A4_modwrapu(i32, i32);
    /// The `llvm.hexagon.F2.sfadd` intrinsic; known as `__builtin_HEXAGON_F2_sfadd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfadd")
    f32 F2_sfadd(f32, f32);
    /// The `llvm.hexagon.F2.sfsub` intrinsic; known as `__builtin_HEXAGON_F2_sfsub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfsub")
    f32 F2_sfsub(f32, f32);
    /// The `llvm.hexagon.F2.sfmpy` intrinsic; known as `__builtin_HEXAGON_F2_sfmpy` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfmpy")
    f32 F2_sfmpy(f32, f32);
    /// The `llvm.hexagon.F2.sffma` intrinsic; known as `__builtin_HEXAGON_F2_sffma` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffma")
    f32 F2_sffma(f32, f32, f32);
    /// The `llvm.hexagon.F2.sffma.sc` intrinsic; known as `__builtin_HEXAGON_F2_sffma_sc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffma.sc")
    f32 F2_sffma_sc(f32, f32, f32, i32);
    /// The `llvm.hexagon.F2.sffms` intrinsic; known as `__builtin_HEXAGON_F2_sffms` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffms")
    f32 F2_sffms(f32, f32, f32);
    /// The `llvm.hexagon.F2.sffma.lib` intrinsic; known as `__builtin_HEXAGON_F2_sffma_lib` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffma.lib")
    f32 F2_sffma_lib(f32, f32, f32);
    /// The `llvm.hexagon.F2.sffms.lib` intrinsic; known as `__builtin_HEXAGON_F2_sffms_lib` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffms.lib")
    f32 F2_sffms_lib(f32, f32, f32);
    /// The `llvm.hexagon.F2.sfcmpeq` intrinsic; known as `__builtin_HEXAGON_F2_sfcmpeq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfcmpeq")
    bool F2_sfcmpeq(f32, f32);
    /// The `llvm.hexagon.F2.sfcmpgt` intrinsic; known as `__builtin_HEXAGON_F2_sfcmpgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfcmpgt")
    bool F2_sfcmpgt(f32, f32);
    /// The `llvm.hexagon.F2.sfcmpge` intrinsic; known as `__builtin_HEXAGON_F2_sfcmpge` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfcmpge")
    bool F2_sfcmpge(f32, f32);
    /// The `llvm.hexagon.F2.sfcmpuo` intrinsic; known as `__builtin_HEXAGON_F2_sfcmpuo` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfcmpuo")
    bool F2_sfcmpuo(f32, f32);
    /// The `llvm.hexagon.F2.sfmax` intrinsic; known as `__builtin_HEXAGON_F2_sfmax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfmax")
    f32 F2_sfmax(f32, f32);
    /// The `llvm.hexagon.F2.sfmin` intrinsic; known as `__builtin_HEXAGON_F2_sfmin` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfmin")
    f32 F2_sfmin(f32, f32);
    /// The `llvm.hexagon.F2.sfclass` intrinsic; known as `__builtin_HEXAGON_F2_sfclass` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfclass")
    bool F2_sfclass(f32, i32);
    /// The `llvm.hexagon.F2.sfimm.p` intrinsic; known as `__builtin_HEXAGON_F2_sfimm_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfimm.p")
    f32 F2_sfimm_p(i32);
    /// The `llvm.hexagon.F2.sfimm.n` intrinsic; known as `__builtin_HEXAGON_F2_sfimm_n` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sfimm.n")
    f32 F2_sfimm_n(i32);
    /// The `llvm.hexagon.F2.sffixupn` intrinsic; known as `__builtin_HEXAGON_F2_sffixupn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffixupn")
    f32 F2_sffixupn(f32, f32);
    /// The `llvm.hexagon.F2.sffixupd` intrinsic; known as `__builtin_HEXAGON_F2_sffixupd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffixupd")
    f32 F2_sffixupd(f32, f32);
    /// The `llvm.hexagon.F2.sffixupr` intrinsic; known as `__builtin_HEXAGON_F2_sffixupr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.sffixupr")
    f32 F2_sffixupr(f32);
    /// The `llvm.hexagon.F2.dfadd` intrinsic; known as `__builtin_HEXAGON_F2_dfadd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfadd")
    f64 F2_dfadd(f64, f64);
    /// The `llvm.hexagon.F2.dfsub` intrinsic; known as `__builtin_HEXAGON_F2_dfsub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfsub")
    f64 F2_dfsub(f64, f64);
    /// The `llvm.hexagon.F2.dfmpy` intrinsic; known as `__builtin_HEXAGON_F2_dfmpy` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfmpy")
    f64 F2_dfmpy(f64, f64);
    /// The `llvm.hexagon.F2.dffma` intrinsic; known as `__builtin_HEXAGON_F2_dffma` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffma")
    f64 F2_dffma(f64, f64, f64);
    /// The `llvm.hexagon.F2.dffms` intrinsic; known as `__builtin_HEXAGON_F2_dffms` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffms")
    f64 F2_dffms(f64, f64, f64);
    /// The `llvm.hexagon.F2.dffma.lib` intrinsic; known as `__builtin_HEXAGON_F2_dffma_lib` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffma.lib")
    f64 F2_dffma_lib(f64, f64, f64);
    /// The `llvm.hexagon.F2.dffms.lib` intrinsic; known as `__builtin_HEXAGON_F2_dffms_lib` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffms.lib")
    f64 F2_dffms_lib(f64, f64, f64);
    /// The `llvm.hexagon.F2.dffma.sc` intrinsic; known as `__builtin_HEXAGON_F2_dffma_sc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffma.sc")
    f64 F2_dffma_sc(f64, f64, f64, i32);
    /// The `llvm.hexagon.F2.dfmax` intrinsic; known as `__builtin_HEXAGON_F2_dfmax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfmax")
    f64 F2_dfmax(f64, f64);
    /// The `llvm.hexagon.F2.dfmin` intrinsic; known as `__builtin_HEXAGON_F2_dfmin` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfmin")
    f64 F2_dfmin(f64, f64);
    /// The `llvm.hexagon.F2.dfcmpeq` intrinsic; known as `__builtin_HEXAGON_F2_dfcmpeq` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfcmpeq")
    bool F2_dfcmpeq(f64, f64);
    /// The `llvm.hexagon.F2.dfcmpgt` intrinsic; known as `__builtin_HEXAGON_F2_dfcmpgt` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfcmpgt")
    bool F2_dfcmpgt(f64, f64);
    /// The `llvm.hexagon.F2.dfcmpge` intrinsic; known as `__builtin_HEXAGON_F2_dfcmpge` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfcmpge")
    bool F2_dfcmpge(f64, f64);
    /// The `llvm.hexagon.F2.dfcmpuo` intrinsic; known as `__builtin_HEXAGON_F2_dfcmpuo` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfcmpuo")
    bool F2_dfcmpuo(f64, f64);
    /// The `llvm.hexagon.F2.dfclass` intrinsic; known as `__builtin_HEXAGON_F2_dfclass` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfclass")
    bool F2_dfclass(f64, i32);
    /// The `llvm.hexagon.F2.dfimm.p` intrinsic; known as `__builtin_HEXAGON_F2_dfimm_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfimm.p")
    f64 F2_dfimm_p(i32);
    /// The `llvm.hexagon.F2.dfimm.n` intrinsic; known as `__builtin_HEXAGON_F2_dfimm_n` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dfimm.n")
    f64 F2_dfimm_n(i32);
    /// The `llvm.hexagon.F2.dffixupn` intrinsic; known as `__builtin_HEXAGON_F2_dffixupn` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffixupn")
    f64 F2_dffixupn(f64, f64);
    /// The `llvm.hexagon.F2.dffixupd` intrinsic; known as `__builtin_HEXAGON_F2_dffixupd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffixupd")
    f64 F2_dffixupd(f64, f64);
    /// The `llvm.hexagon.F2.dffixupr` intrinsic; known as `__builtin_HEXAGON_F2_dffixupr` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.dffixupr")
    f64 F2_dffixupr(f64);
    /// The `llvm.hexagon.F2.conv.sf2df` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2df` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2df")
    f64 F2_conv_sf2df(f32);
    /// The `llvm.hexagon.F2.conv.df2sf` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2sf` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2sf")
    f32 F2_conv_df2sf(f64);
    /// The `llvm.hexagon.F2.conv.uw2sf` intrinsic; known as `__builtin_HEXAGON_F2_conv_uw2sf` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.uw2sf")
    f32 F2_conv_uw2sf(i32);
    /// The `llvm.hexagon.F2.conv.uw2df` intrinsic; known as `__builtin_HEXAGON_F2_conv_uw2df` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.uw2df")
    f64 F2_conv_uw2df(i32);
    /// The `llvm.hexagon.F2.conv.w2sf` intrinsic; known as `__builtin_HEXAGON_F2_conv_w2sf` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.w2sf")
    f32 F2_conv_w2sf(i32);
    /// The `llvm.hexagon.F2.conv.w2df` intrinsic; known as `__builtin_HEXAGON_F2_conv_w2df` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.w2df")
    f64 F2_conv_w2df(i32);
    /// The `llvm.hexagon.F2.conv.ud2sf` intrinsic; known as `__builtin_HEXAGON_F2_conv_ud2sf` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.ud2sf")
    f32 F2_conv_ud2sf(i64);
    /// The `llvm.hexagon.F2.conv.ud2df` intrinsic; known as `__builtin_HEXAGON_F2_conv_ud2df` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.ud2df")
    f64 F2_conv_ud2df(i64);
    /// The `llvm.hexagon.F2.conv.d2sf` intrinsic; known as `__builtin_HEXAGON_F2_conv_d2sf` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.d2sf")
    f32 F2_conv_d2sf(i64);
    /// The `llvm.hexagon.F2.conv.d2df` intrinsic; known as `__builtin_HEXAGON_F2_conv_d2df` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.d2df")
    f64 F2_conv_d2df(i64);
    /// The `llvm.hexagon.F2.conv.sf2uw` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2uw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2uw")
    i32 F2_conv_sf2uw(f32);
    /// The `llvm.hexagon.F2.conv.sf2w` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2w` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2w")
    i32 F2_conv_sf2w(f32);
    /// The `llvm.hexagon.F2.conv.sf2ud` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2ud` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2ud")
    i64 F2_conv_sf2ud(f32);
    /// The `llvm.hexagon.F2.conv.sf2d` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2d` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2d")
    i64 F2_conv_sf2d(f32);
    /// The `llvm.hexagon.F2.conv.df2uw` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2uw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2uw")
    i32 F2_conv_df2uw(f64);
    /// The `llvm.hexagon.F2.conv.df2w` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2w` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2w")
    i32 F2_conv_df2w(f64);
    /// The `llvm.hexagon.F2.conv.df2ud` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2ud` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2ud")
    i64 F2_conv_df2ud(f64);
    /// The `llvm.hexagon.F2.conv.df2d` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2d` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2d")
    i64 F2_conv_df2d(f64);
    /// The `llvm.hexagon.F2.conv.sf2uw.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2uw_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2uw.chop")
    i32 F2_conv_sf2uw_chop(f32);
    /// The `llvm.hexagon.F2.conv.sf2w.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2w_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2w.chop")
    i32 F2_conv_sf2w_chop(f32);
    /// The `llvm.hexagon.F2.conv.sf2ud.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2ud_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2ud.chop")
    i64 F2_conv_sf2ud_chop(f32);
    /// The `llvm.hexagon.F2.conv.sf2d.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_sf2d_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.sf2d.chop")
    i64 F2_conv_sf2d_chop(f32);
    /// The `llvm.hexagon.F2.conv.df2uw.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2uw_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2uw.chop")
    i32 F2_conv_df2uw_chop(f64);
    /// The `llvm.hexagon.F2.conv.df2w.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2w_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2w.chop")
    i32 F2_conv_df2w_chop(f64);
    /// The `llvm.hexagon.F2.conv.df2ud.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2ud_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2ud.chop")
    i64 F2_conv_df2ud_chop(f64);
    /// The `llvm.hexagon.F2.conv.df2d.chop` intrinsic; known as `__builtin_HEXAGON_F2_conv_df2d_chop` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.F2.conv.df2d.chop")
    i64 F2_conv_df2d_chop(f64);
    /// The `llvm.hexagon.S2.asr.r.r` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.r")
    i32 S2_asr_r_r(i32, i32);
    /// The `llvm.hexagon.S2.asl.r.r` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.r")
    i32 S2_asl_r_r(i32, i32);
    /// The `llvm.hexagon.S2.lsr.r.r` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.r")
    i32 S2_lsr_r_r(i32, i32);
    /// The `llvm.hexagon.S2.lsl.r.r` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.r")
    i32 S2_lsl_r_r(i32, i32);
    /// The `llvm.hexagon.S2.asr.r.p` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.p")
    i64 S2_asr_r_p(i64, i32);
    /// The `llvm.hexagon.S2.asl.r.p` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.p")
    i64 S2_asl_r_p(i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.p` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.p")
    i64 S2_lsr_r_p(i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.p` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.p")
    i64 S2_lsl_r_p(i64, i32);
    /// The `llvm.hexagon.S2.asr.r.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.r.acc")
    i32 S2_asr_r_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.r.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.r.acc")
    i32 S2_asl_r_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.r.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.r.acc")
    i32 S2_lsr_r_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsl.r.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.r.acc")
    i32 S2_lsl_r_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.r.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.p.acc")
    i64 S2_asr_r_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.r.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.p.acc")
    i64 S2_asl_r_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.p.acc")
    i64 S2_lsr_r_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.p.acc")
    i64 S2_lsl_r_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.r.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.r.nac")
    i32 S2_asr_r_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.r.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.r.nac")
    i32 S2_asl_r_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.r.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.r.nac")
    i32 S2_lsr_r_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsl.r.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.r.nac")
    i32 S2_lsl_r_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.r.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.p.nac")
    i64 S2_asr_r_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.r.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.p.nac")
    i64 S2_asl_r_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.p.nac")
    i64 S2_lsr_r_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.p.nac")
    i64 S2_lsl_r_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.r.r.and` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.r.and")
    i32 S2_asr_r_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.r.r.and` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.r.and")
    i32 S2_asl_r_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.r.r.and` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.r.and")
    i32 S2_lsr_r_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsl.r.r.and` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.r.and")
    i32 S2_lsl_r_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.r.r.or` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.r.or")
    i32 S2_asr_r_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.r.r.or` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.r.or")
    i32 S2_asl_r_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.r.r.or` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.r.or")
    i32 S2_lsr_r_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsl.r.r.or` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.r.or")
    i32 S2_lsl_r_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.r.p.and` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.p.and")
    i64 S2_asr_r_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.r.p.and` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.p.and")
    i64 S2_asl_r_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.p.and` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.p.and")
    i64 S2_lsr_r_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.p.and` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.p.and")
    i64 S2_lsl_r_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.r.p.or` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.p.or")
    i64 S2_asr_r_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.r.p.or` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.p.or")
    i64 S2_asl_r_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.p.or` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.p.or")
    i64 S2_lsr_r_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.p.or` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.p.or")
    i64 S2_lsl_r_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.r.p.xor` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_p_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.p.xor")
    i64 S2_asr_r_p_xor(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.r.p.xor` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_p_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.p.xor")
    i64 S2_asl_r_p_xor(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.p.xor` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_p_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.p.xor")
    i64 S2_lsr_r_p_xor(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.p.xor` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_p_xor` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.p.xor")
    i64 S2_lsl_r_p_xor(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.r.r.sat` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_r_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.r.sat")
    i32 S2_asr_r_r_sat(i32, i32);
    /// The `llvm.hexagon.S2.asl.r.r.sat` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_r_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.r.sat")
    i32 S2_asl_r_r_sat(i32, i32);
    /// The `llvm.hexagon.S2.asr.i.r` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r")
    i32 S2_asr_i_r(i32, i32);
    /// The `llvm.hexagon.S2.lsr.i.r` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.r")
    i32 S2_lsr_i_r(i32, i32);
    /// The `llvm.hexagon.S2.asl.i.r` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r")
    i32 S2_asl_i_r(i32, i32);
    /// The `llvm.hexagon.S2.asr.i.p` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p")
    i64 S2_asr_i_p(i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.p` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.p")
    i64 S2_lsr_i_p(i64, i32);
    /// The `llvm.hexagon.S2.asl.i.p` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.p")
    i64 S2_asl_i_p(i64, i32);
    /// The `llvm.hexagon.S2.asr.i.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r.acc")
    i32 S2_asr_i_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.i.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.r.acc")
    i32 S2_lsr_i_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.i.r.acc` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r.acc")
    i32 S2_asl_i_r_acc(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.i.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p.acc")
    i64 S2_asr_i_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.p.acc")
    i64 S2_lsr_i_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.i.p.acc` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_p_acc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.p.acc")
    i64 S2_asl_i_p_acc(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.i.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r.nac")
    i32 S2_asr_i_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.i.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.r.nac")
    i32 S2_lsr_i_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.i.r.nac` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r.nac")
    i32 S2_asl_i_r_nac(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.i.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p.nac")
    i64 S2_asr_i_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.p.nac")
    i64 S2_lsr_i_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.i.p.nac` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_p_nac` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.p.nac")
    i64 S2_asl_i_p_nac(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.r.xacc` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_r_xacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.r.xacc")
    i32 S2_lsr_i_r_xacc(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.i.r.xacc` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r_xacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r.xacc")
    i32 S2_asl_i_r_xacc(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.i.p.xacc` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_p_xacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.p.xacc")
    i64 S2_lsr_i_p_xacc(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.i.p.xacc` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_p_xacc` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.p.xacc")
    i64 S2_asl_i_p_xacc(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.i.r.and` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r.and")
    i32 S2_asr_i_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.i.r.and` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.r.and")
    i32 S2_lsr_i_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.i.r.and` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r.and")
    i32 S2_asl_i_r_and(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.i.r.or` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r.or")
    i32 S2_asr_i_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.lsr.i.r.or` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.r.or")
    i32 S2_lsr_i_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.asl.i.r.or` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r.or")
    i32 S2_asl_i_r_or(i32, i32, i32);
    /// The `llvm.hexagon.S2.asr.i.p.and` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p.and")
    i64 S2_asr_i_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.p.and` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.p.and")
    i64 S2_lsr_i_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.i.p.and` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_p_and` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.p.and")
    i64 S2_asl_i_p_and(i64, i64, i32);
    /// The `llvm.hexagon.S2.asr.i.p.or` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p.or")
    i64 S2_asr_i_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.p.or` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.p.or")
    i64 S2_lsr_i_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.i.p.or` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_p_or` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.p.or")
    i64 S2_asl_i_p_or(i64, i64, i32);
    /// The `llvm.hexagon.S2.asl.i.r.sat` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_r_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.r.sat")
    i32 S2_asl_i_r_sat(i32, i32);
    /// The `llvm.hexagon.S2.asr.i.r.rnd` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r_rnd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r.rnd")
    i32 S2_asr_i_r_rnd(i32, i32);
    /// The `llvm.hexagon.S2.asr.i.r.rnd.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_r_rnd_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.r.rnd.goodsyntax")
    i32 S2_asr_i_r_rnd_goodsyntax(i32, i32);
    /// The `llvm.hexagon.S2.asr.i.p.rnd` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p_rnd` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p.rnd")
    i64 S2_asr_i_p_rnd(i64, i32);
    /// The `llvm.hexagon.S2.asr.i.p.rnd.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_p_rnd_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.p.rnd.goodsyntax")
    i64 S2_asr_i_p_rnd_goodsyntax(i64, i32);
    /// The `llvm.hexagon.S4.lsli` intrinsic; known as `__builtin_HEXAGON_S4_lsli` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.lsli")
    i32 S4_lsli(i32, i32);
    /// The `llvm.hexagon.S2.addasl.rrri` intrinsic; known as `__builtin_HEXAGON_S2_addasl_rrri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.addasl.rrri")
    i32 S2_addasl_rrri(i32, i32, i32);
    /// The `llvm.hexagon.S4.andi.asl.ri` intrinsic; known as `__builtin_HEXAGON_S4_andi_asl_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.andi.asl.ri")
    i32 S4_andi_asl_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.ori.asl.ri` intrinsic; known as `__builtin_HEXAGON_S4_ori_asl_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.ori.asl.ri")
    i32 S4_ori_asl_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.addi.asl.ri` intrinsic; known as `__builtin_HEXAGON_S4_addi_asl_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.addi.asl.ri")
    i32 S4_addi_asl_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.subi.asl.ri` intrinsic; known as `__builtin_HEXAGON_S4_subi_asl_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.subi.asl.ri")
    i32 S4_subi_asl_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.andi.lsr.ri` intrinsic; known as `__builtin_HEXAGON_S4_andi_lsr_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.andi.lsr.ri")
    i32 S4_andi_lsr_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.ori.lsr.ri` intrinsic; known as `__builtin_HEXAGON_S4_ori_lsr_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.ori.lsr.ri")
    i32 S4_ori_lsr_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.addi.lsr.ri` intrinsic; known as `__builtin_HEXAGON_S4_addi_lsr_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.addi.lsr.ri")
    i32 S4_addi_lsr_ri(i32, i32, i32);
    /// The `llvm.hexagon.S4.subi.lsr.ri` intrinsic; known as `__builtin_HEXAGON_S4_subi_lsr_ri` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.subi.lsr.ri")
    i32 S4_subi_lsr_ri(i32, i32, i32);
    /// The `llvm.hexagon.S2.valignib` intrinsic; known as `__builtin_HEXAGON_S2_valignib` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.valignib")
    i64 S2_valignib(i64, i64, i32);
    /// The `llvm.hexagon.S2.valignrb` intrinsic; known as `__builtin_HEXAGON_S2_valignrb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.valignrb")
    i64 S2_valignrb(i64, i64, i32);
    /// The `llvm.hexagon.S2.vspliceib` intrinsic; known as `__builtin_HEXAGON_S2_vspliceib` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vspliceib")
    i64 S2_vspliceib(i64, i64, i32);
    /// The `llvm.hexagon.S2.vsplicerb` intrinsic; known as `__builtin_HEXAGON_S2_vsplicerb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsplicerb")
    i64 S2_vsplicerb(i64, i64, i32);
    /// The `llvm.hexagon.S2.vsplatrh` intrinsic; known as `__builtin_HEXAGON_S2_vsplatrh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsplatrh")
    i64 S2_vsplatrh(i32);
    /// The `llvm.hexagon.S2.vsplatrb` intrinsic; known as `__builtin_HEXAGON_S2_vsplatrb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsplatrb")
    i32 S2_vsplatrb(i32);
    /// The `llvm.hexagon.S2.insert` intrinsic; known as `__builtin_HEXAGON_S2_insert` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.insert")
    i32 S2_insert(i32, i32, i32, i32);
    /// The `llvm.hexagon.S2.tableidxb.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S2_tableidxb_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.tableidxb.goodsyntax")
    i32 S2_tableidxb_goodsyntax(i32, i32, i32, i32);
    /// The `llvm.hexagon.S2.tableidxh.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S2_tableidxh_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.tableidxh.goodsyntax")
    i32 S2_tableidxh_goodsyntax(i32, i32, i32, i32);
    /// The `llvm.hexagon.S2.tableidxw.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S2_tableidxw_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.tableidxw.goodsyntax")
    i32 S2_tableidxw_goodsyntax(i32, i32, i32, i32);
    /// The `llvm.hexagon.S2.tableidxd.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S2_tableidxd_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.tableidxd.goodsyntax")
    i32 S2_tableidxd_goodsyntax(i32, i32, i32, i32);
    /// The `llvm.hexagon.A4.bitspliti` intrinsic; known as `__builtin_HEXAGON_A4_bitspliti` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.bitspliti")
    i64 A4_bitspliti(i32, i32);
    /// The `llvm.hexagon.A4.bitsplit` intrinsic; known as `__builtin_HEXAGON_A4_bitsplit` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A4.bitsplit")
    i64 A4_bitsplit(i32, i32);
    /// The `llvm.hexagon.S4.extract` intrinsic; known as `__builtin_HEXAGON_S4_extract` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.extract")
    i32 S4_extract(i32, i32, i32);
    /// The `llvm.hexagon.S2.extractu` intrinsic; known as `__builtin_HEXAGON_S2_extractu` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.extractu")
    i32 S2_extractu(i32, i32, i32);
    /// The `llvm.hexagon.S2.insertp` intrinsic; known as `__builtin_HEXAGON_S2_insertp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.insertp")
    i64 S2_insertp(i64, i64, i32, i32);
    /// The `llvm.hexagon.S4.extractp` intrinsic; known as `__builtin_HEXAGON_S4_extractp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.extractp")
    i64 S4_extractp(i64, i32, i32);
    /// The `llvm.hexagon.S2.extractup` intrinsic; known as `__builtin_HEXAGON_S2_extractup` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.extractup")
    i64 S2_extractup(i64, i32, i32);
    /// The `llvm.hexagon.S2.insert.rp` intrinsic; known as `__builtin_HEXAGON_S2_insert_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.insert.rp")
    i32 S2_insert_rp(i32, i32, i64);
    /// The `llvm.hexagon.S4.extract.rp` intrinsic; known as `__builtin_HEXAGON_S4_extract_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.extract.rp")
    i32 S4_extract_rp(i32, i64);
    /// The `llvm.hexagon.S2.extractu.rp` intrinsic; known as `__builtin_HEXAGON_S2_extractu_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.extractu.rp")
    i32 S2_extractu_rp(i32, i64);
    /// The `llvm.hexagon.S2.insertp.rp` intrinsic; known as `__builtin_HEXAGON_S2_insertp_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.insertp.rp")
    i64 S2_insertp_rp(i64, i64, i64);
    /// The `llvm.hexagon.S4.extractp.rp` intrinsic; known as `__builtin_HEXAGON_S4_extractp_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.extractp.rp")
    i64 S4_extractp_rp(i64, i64);
    /// The `llvm.hexagon.S2.extractup.rp` intrinsic; known as `__builtin_HEXAGON_S2_extractup_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.extractup.rp")
    i64 S2_extractup_rp(i64, i64);
    /// The `llvm.hexagon.S2.tstbit.i` intrinsic; known as `__builtin_HEXAGON_S2_tstbit_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.tstbit.i")
    bool S2_tstbit_i(i32, i32);
    /// The `llvm.hexagon.S4.ntstbit.i` intrinsic; known as `__builtin_HEXAGON_S4_ntstbit_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.ntstbit.i")
    bool S4_ntstbit_i(i32, i32);
    /// The `llvm.hexagon.S2.setbit.i` intrinsic; known as `__builtin_HEXAGON_S2_setbit_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.setbit.i")
    i32 S2_setbit_i(i32, i32);
    /// The `llvm.hexagon.S2.togglebit.i` intrinsic; known as `__builtin_HEXAGON_S2_togglebit_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.togglebit.i")
    i32 S2_togglebit_i(i32, i32);
    /// The `llvm.hexagon.S2.clrbit.i` intrinsic; known as `__builtin_HEXAGON_S2_clrbit_i` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.clrbit.i")
    i32 S2_clrbit_i(i32, i32);
    /// The `llvm.hexagon.S2.tstbit.r` intrinsic; known as `__builtin_HEXAGON_S2_tstbit_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.tstbit.r")
    bool S2_tstbit_r(i32, i32);
    /// The `llvm.hexagon.S4.ntstbit.r` intrinsic; known as `__builtin_HEXAGON_S4_ntstbit_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.ntstbit.r")
    bool S4_ntstbit_r(i32, i32);
    /// The `llvm.hexagon.S2.setbit.r` intrinsic; known as `__builtin_HEXAGON_S2_setbit_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.setbit.r")
    i32 S2_setbit_r(i32, i32);
    /// The `llvm.hexagon.S2.togglebit.r` intrinsic; known as `__builtin_HEXAGON_S2_togglebit_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.togglebit.r")
    i32 S2_togglebit_r(i32, i32);
    /// The `llvm.hexagon.S2.clrbit.r` intrinsic; known as `__builtin_HEXAGON_S2_clrbit_r` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.clrbit.r")
    i32 S2_clrbit_r(i32, i32);
    /// The `llvm.hexagon.S2.asr.i.vh` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.vh")
    i64 S2_asr_i_vh(i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.vh` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.vh")
    i64 S2_lsr_i_vh(i64, i32);
    /// The `llvm.hexagon.S2.asl.i.vh` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.vh")
    i64 S2_asl_i_vh(i64, i32);
    /// The `llvm.hexagon.S2.asr.r.vh` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.vh")
    i64 S2_asr_r_vh(i64, i32);
    /// The `llvm.hexagon.S5.asrhub.rnd.sat.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S5_asrhub_rnd_sat_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S5.asrhub.rnd.sat.goodsyntax")
    i32 S5_asrhub_rnd_sat_goodsyntax(i64, i32);
    /// The `llvm.hexagon.S5.asrhub.sat` intrinsic; known as `__builtin_HEXAGON_S5_asrhub_sat` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S5.asrhub.sat")
    i32 S5_asrhub_sat(i64, i32);
    /// The `llvm.hexagon.S5.vasrhrnd.goodsyntax` intrinsic; known as `__builtin_HEXAGON_S5_vasrhrnd_goodsyntax` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S5.vasrhrnd.goodsyntax")
    i64 S5_vasrhrnd_goodsyntax(i64, i32);
    /// The `llvm.hexagon.S2.asl.r.vh` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.vh")
    i64 S2_asl_r_vh(i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.vh` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.vh")
    i64 S2_lsr_r_vh(i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.vh` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_vh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.vh")
    i64 S2_lsl_r_vh(i64, i32);
    /// The `llvm.hexagon.S2.asr.i.vw` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.vw")
    i64 S2_asr_i_vw(i64, i32);
    /// The `llvm.hexagon.S2.asr.i.svw.trun` intrinsic; known as `__builtin_HEXAGON_S2_asr_i_svw_trun` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.i.svw.trun")
    i32 S2_asr_i_svw_trun(i64, i32);
    /// The `llvm.hexagon.S2.asr.r.svw.trun` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_svw_trun` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.svw.trun")
    i32 S2_asr_r_svw_trun(i64, i32);
    /// The `llvm.hexagon.S2.lsr.i.vw` intrinsic; known as `__builtin_HEXAGON_S2_lsr_i_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.i.vw")
    i64 S2_lsr_i_vw(i64, i32);
    /// The `llvm.hexagon.S2.asl.i.vw` intrinsic; known as `__builtin_HEXAGON_S2_asl_i_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.i.vw")
    i64 S2_asl_i_vw(i64, i32);
    /// The `llvm.hexagon.S2.asr.r.vw` intrinsic; known as `__builtin_HEXAGON_S2_asr_r_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asr.r.vw")
    i64 S2_asr_r_vw(i64, i32);
    /// The `llvm.hexagon.S2.asl.r.vw` intrinsic; known as `__builtin_HEXAGON_S2_asl_r_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.asl.r.vw")
    i64 S2_asl_r_vw(i64, i32);
    /// The `llvm.hexagon.S2.lsr.r.vw` intrinsic; known as `__builtin_HEXAGON_S2_lsr_r_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsr.r.vw")
    i64 S2_lsr_r_vw(i64, i32);
    /// The `llvm.hexagon.S2.lsl.r.vw` intrinsic; known as `__builtin_HEXAGON_S2_lsl_r_vw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lsl.r.vw")
    i64 S2_lsl_r_vw(i64, i32);
    /// The `llvm.hexagon.S2.vrndpackwh` intrinsic; known as `__builtin_HEXAGON_S2_vrndpackwh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vrndpackwh")
    i32 S2_vrndpackwh(i64);
    /// The `llvm.hexagon.S2.vrndpackwhs` intrinsic; known as `__builtin_HEXAGON_S2_vrndpackwhs` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vrndpackwhs")
    i32 S2_vrndpackwhs(i64);
    /// The `llvm.hexagon.S2.vsxtbh` intrinsic; known as `__builtin_HEXAGON_S2_vsxtbh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsxtbh")
    i64 S2_vsxtbh(i32);
    /// The `llvm.hexagon.S2.vzxtbh` intrinsic; known as `__builtin_HEXAGON_S2_vzxtbh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vzxtbh")
    i64 S2_vzxtbh(i32);
    /// The `llvm.hexagon.S2.vsathub` intrinsic; known as `__builtin_HEXAGON_S2_vsathub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsathub")
    i32 S2_vsathub(i64);
    /// The `llvm.hexagon.S2.svsathub` intrinsic; known as `__builtin_HEXAGON_S2_svsathub` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.svsathub")
    i32 S2_svsathub(i32);
    /// The `llvm.hexagon.S2.svsathb` intrinsic; known as `__builtin_HEXAGON_S2_svsathb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.svsathb")
    i32 S2_svsathb(i32);
    /// The `llvm.hexagon.S2.vsathb` intrinsic; known as `__builtin_HEXAGON_S2_vsathb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsathb")
    i32 S2_vsathb(i64);
    /// The `llvm.hexagon.S2.vtrunohb` intrinsic; known as `__builtin_HEXAGON_S2_vtrunohb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vtrunohb")
    i32 S2_vtrunohb(i64);
    /// The `llvm.hexagon.S2.vtrunewh` intrinsic; known as `__builtin_HEXAGON_S2_vtrunewh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vtrunewh")
    i64 S2_vtrunewh(i64, i64);
    /// The `llvm.hexagon.S2.vtrunowh` intrinsic; known as `__builtin_HEXAGON_S2_vtrunowh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vtrunowh")
    i64 S2_vtrunowh(i64, i64);
    /// The `llvm.hexagon.S2.vtrunehb` intrinsic; known as `__builtin_HEXAGON_S2_vtrunehb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vtrunehb")
    i32 S2_vtrunehb(i64);
    /// The `llvm.hexagon.S2.vsxthw` intrinsic; known as `__builtin_HEXAGON_S2_vsxthw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsxthw")
    i64 S2_vsxthw(i32);
    /// The `llvm.hexagon.S2.vzxthw` intrinsic; known as `__builtin_HEXAGON_S2_vzxthw` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vzxthw")
    i64 S2_vzxthw(i32);
    /// The `llvm.hexagon.S2.vsatwh` intrinsic; known as `__builtin_HEXAGON_S2_vsatwh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsatwh")
    i32 S2_vsatwh(i64);
    /// The `llvm.hexagon.S2.vsatwuh` intrinsic; known as `__builtin_HEXAGON_S2_vsatwuh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsatwuh")
    i32 S2_vsatwuh(i64);
    /// The `llvm.hexagon.S2.packhl` intrinsic; known as `__builtin_HEXAGON_S2_packhl` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.packhl")
    i64 S2_packhl(i32, i32);
    /// The `llvm.hexagon.A2.swiz` intrinsic; known as `__builtin_HEXAGON_A2_swiz` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.A2.swiz")
    i32 A2_swiz(i32);
    /// The `llvm.hexagon.S2.vsathub.nopack` intrinsic; known as `__builtin_HEXAGON_S2_vsathub_nopack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsathub.nopack")
    i64 S2_vsathub_nopack(i64);
    /// The `llvm.hexagon.S2.vsathb.nopack` intrinsic; known as `__builtin_HEXAGON_S2_vsathb_nopack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsathb.nopack")
    i64 S2_vsathb_nopack(i64);
    /// The `llvm.hexagon.S2.vsatwh.nopack` intrinsic; known as `__builtin_HEXAGON_S2_vsatwh_nopack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsatwh.nopack")
    i64 S2_vsatwh_nopack(i64);
    /// The `llvm.hexagon.S2.vsatwuh.nopack` intrinsic; known as `__builtin_HEXAGON_S2_vsatwuh_nopack` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.vsatwuh.nopack")
    i64 S2_vsatwuh_nopack(i64);
    /// The `llvm.hexagon.S2.shuffob` intrinsic; known as `__builtin_HEXAGON_S2_shuffob` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.shuffob")
    i64 S2_shuffob(i64, i64);
    /// The `llvm.hexagon.S2.shuffeb` intrinsic; known as `__builtin_HEXAGON_S2_shuffeb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.shuffeb")
    i64 S2_shuffeb(i64, i64);
    /// The `llvm.hexagon.S2.shuffoh` intrinsic; known as `__builtin_HEXAGON_S2_shuffoh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.shuffoh")
    i64 S2_shuffoh(i64, i64);
    /// The `llvm.hexagon.S2.shuffeh` intrinsic; known as `__builtin_HEXAGON_S2_shuffeh` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.shuffeh")
    i64 S2_shuffeh(i64, i64);
    /// The `llvm.hexagon.S5.popcountp` intrinsic; known as `__builtin_HEXAGON_S5_popcountp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S5.popcountp")
    i32 S5_popcountp(i64);
    /// The `llvm.hexagon.S4.parity` intrinsic; known as `__builtin_HEXAGON_S4_parity` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.parity")
    i32 S4_parity(i32, i32);
    /// The `llvm.hexagon.S2.parityp` intrinsic; known as `__builtin_HEXAGON_S2_parityp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.parityp")
    i32 S2_parityp(i64, i64);
    /// The `llvm.hexagon.S2.lfsp` intrinsic; known as `__builtin_HEXAGON_S2_lfsp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.lfsp")
    i64 S2_lfsp(i64, i64);
    /// The `llvm.hexagon.S2.clbnorm` intrinsic; known as `__builtin_HEXAGON_S2_clbnorm` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.clbnorm")
    i32 S2_clbnorm(i32);
    /// The `llvm.hexagon.S4.clbaddi` intrinsic; known as `__builtin_HEXAGON_S4_clbaddi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.clbaddi")
    i32 S4_clbaddi(i32, i32);
    /// The `llvm.hexagon.S4.clbpnorm` intrinsic; known as `__builtin_HEXAGON_S4_clbpnorm` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.clbpnorm")
    i32 S4_clbpnorm(i64);
    /// The `llvm.hexagon.S4.clbpaddi` intrinsic; known as `__builtin_HEXAGON_S4_clbpaddi` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S4.clbpaddi")
    i32 S4_clbpaddi(i64, i32);
    /// The `llvm.hexagon.S2.clb` intrinsic; known as `__builtin_HEXAGON_S2_clb` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.clb")
    i32 S2_clb(i32);
    /// The `llvm.hexagon.S2.cl0` intrinsic; known as `__builtin_HEXAGON_S2_cl0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.cl0")
    i32 S2_cl0(i32);
    /// The `llvm.hexagon.S2.cl1` intrinsic; known as `__builtin_HEXAGON_S2_cl1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.cl1")
    i32 S2_cl1(i32);
    /// The `llvm.hexagon.S2.clbp` intrinsic; known as `__builtin_HEXAGON_S2_clbp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.clbp")
    i32 S2_clbp(i64);
    /// The `llvm.hexagon.S2.cl0p` intrinsic; known as `__builtin_HEXAGON_S2_cl0p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.cl0p")
    i32 S2_cl0p(i64);
    /// The `llvm.hexagon.S2.cl1p` intrinsic; known as `__builtin_HEXAGON_S2_cl1p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.cl1p")
    i32 S2_cl1p(i64);
    /// The `llvm.hexagon.S2.brev` intrinsic; known as `__builtin_HEXAGON_S2_brev` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.brev")
    i32 S2_brev(i32);
    /// The `llvm.hexagon.S2.brevp` intrinsic; known as `__builtin_HEXAGON_S2_brevp` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.brevp")
    i64 S2_brevp(i64);
    /// The `llvm.hexagon.S2.ct0` intrinsic; known as `__builtin_HEXAGON_S2_ct0` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.ct0")
    i32 S2_ct0(i32);
    /// The `llvm.hexagon.S2.ct1` intrinsic; known as `__builtin_HEXAGON_S2_ct1` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.ct1")
    i32 S2_ct1(i32);
    /// The `llvm.hexagon.S2.ct0p` intrinsic; known as `__builtin_HEXAGON_S2_ct0p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.ct0p")
    i32 S2_ct0p(i64);
    /// The `llvm.hexagon.S2.ct1p` intrinsic; known as `__builtin_HEXAGON_S2_ct1p` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.ct1p")
    i32 S2_ct1p(i64);
    /// The `llvm.hexagon.S2.interleave` intrinsic; known as `__builtin_HEXAGON_S2_interleave` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.interleave")
    i64 S2_interleave(i64);
    /// The `llvm.hexagon.S2.deinterleave` intrinsic; known as `__builtin_HEXAGON_S2_deinterleave` in GCC.
    pragma(LDC_intrinsic, "llvm.hexagon.S2.deinterleave")
    i64 S2_deinterleave(i64);
}

/// LLVM intrinsics for the mips architecture.
version (mips) {
    /// The `llvm.mips.addu.qb` intrinsic; known as `__builtin_mips_addu_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addu.qb")
    __vector(i8[4]) addu_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.addu.s.qb` intrinsic; known as `__builtin_mips_addu_s_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addu.s.qb")
    __vector(i8[4]) addu_s_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.subu.qb` intrinsic; known as `__builtin_mips_subu_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subu.qb")
    __vector(i8[4]) subu_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.subu.s.qb` intrinsic; known as `__builtin_mips_subu_s_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subu.s.qb")
    __vector(i8[4]) subu_s_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.madd` intrinsic; known as `__builtin_mips_madd` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.madd")
    i64 madd(i64, i32, i32);
    /// The `llvm.mips.maddu` intrinsic; known as `__builtin_mips_maddu` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddu")
    i64 maddu(i64, i32, i32);
    /// The `llvm.mips.msub` intrinsic; known as `__builtin_mips_msub` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msub")
    i64 msub(i64, i32, i32);
    /// The `llvm.mips.msubu` intrinsic; known as `__builtin_mips_msubu` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubu")
    i64 msubu(i64, i32, i32);
    /// The `llvm.mips.addsc` intrinsic; known as `__builtin_mips_addsc` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addsc")
    i32 addsc(i32, i32);
    /// The `llvm.mips.addwc` intrinsic; known as `__builtin_mips_addwc` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addwc")
    i32 addwc(i32, i32);
    /// The `llvm.mips.modsub` intrinsic; known as `__builtin_mips_modsub` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.modsub")
    i32 modsub(i32, i32);
    /// The `llvm.mips.raddu.w.qb` intrinsic; known as `__builtin_mips_raddu_w_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.raddu.w.qb")
    i32 raddu_w_qb(__vector(i8[4]));
    /// The `llvm.mips.shll.qb` intrinsic; known as `__builtin_mips_shll_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shll.qb")
    __vector(i8[4]) shll_qb(__vector(i8[4]), i32);
    /// The `llvm.mips.shrl.qb` intrinsic; known as `__builtin_mips_shrl_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shrl.qb")
    __vector(i8[4]) shrl_qb(__vector(i8[4]), i32);
    /// The `llvm.mips.shilo` intrinsic; known as `__builtin_mips_shilo` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shilo")
    i64 shilo(i64, i32);
    /// The `llvm.mips.mult` intrinsic; known as `__builtin_mips_mult` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mult")
    i64 mult(i32, i32);
    /// The `llvm.mips.multu` intrinsic; known as `__builtin_mips_multu` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.multu")
    i64 multu(i32, i32);
    /// The `llvm.mips.dpau.h.qbl` intrinsic; known as `__builtin_mips_dpau_h_qbl` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpau.h.qbl")
    i64 dpau_h_qbl(i64, __vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.dpau.h.qbr` intrinsic; known as `__builtin_mips_dpau_h_qbr` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpau.h.qbr")
    i64 dpau_h_qbr(i64, __vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.dpsu.h.qbl` intrinsic; known as `__builtin_mips_dpsu_h_qbl` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsu.h.qbl")
    i64 dpsu_h_qbl(i64, __vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.dpsu.h.qbr` intrinsic; known as `__builtin_mips_dpsu_h_qbr` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsu.h.qbr")
    i64 dpsu_h_qbr(i64, __vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpu.eq.qb` intrinsic; known as `__builtin_mips_cmpu_eq_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpu.eq.qb")
    void cmpu_eq_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpu.lt.qb` intrinsic; known as `__builtin_mips_cmpu_lt_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpu.lt.qb")
    void cmpu_lt_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpu.le.qb` intrinsic; known as `__builtin_mips_cmpu_le_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpu.le.qb")
    void cmpu_le_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpgu.eq.qb` intrinsic; known as `__builtin_mips_cmpgu_eq_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpgu.eq.qb")
    i32 cmpgu_eq_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpgu.lt.qb` intrinsic; known as `__builtin_mips_cmpgu_lt_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpgu.lt.qb")
    i32 cmpgu_lt_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpgu.le.qb` intrinsic; known as `__builtin_mips_cmpgu_le_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpgu.le.qb")
    i32 cmpgu_le_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.extr.s.h` intrinsic; known as `__builtin_mips_extr_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.extr.s.h")
    i32 extr_s_h(i64, i32);
    /// The `llvm.mips.extr.w` intrinsic; known as `__builtin_mips_extr_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.extr.w")
    i32 extr_w(i64, i32);
    /// The `llvm.mips.extr.rs.w` intrinsic; known as `__builtin_mips_extr_rs_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.extr.rs.w")
    i32 extr_rs_w(i64, i32);
    /// The `llvm.mips.extr.r.w` intrinsic; known as `__builtin_mips_extr_r_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.extr.r.w")
    i32 extr_r_w(i64, i32);
    /// The `llvm.mips.extp` intrinsic; known as `__builtin_mips_extp` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.extp")
    i32 extp(i64, i32);
    /// The `llvm.mips.extpdp` intrinsic; known as `__builtin_mips_extpdp` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.extpdp")
    i32 extpdp(i64, i32);
    /// The `llvm.mips.wrdsp` intrinsic; known as `__builtin_mips_wrdsp` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.wrdsp")
    void wrdsp(i32, i32);
    /// The `llvm.mips.rddsp` intrinsic; known as `__builtin_mips_rddsp` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.rddsp")
    i32 rddsp(i32);
    /// The `llvm.mips.insv` intrinsic; known as `__builtin_mips_insv` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insv")
    i32 insv(i32, i32);
    /// The `llvm.mips.bitrev` intrinsic; known as `__builtin_mips_bitrev` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bitrev")
    i32 bitrev(i32);
    /// The `llvm.mips.repl.qb` intrinsic; known as `__builtin_mips_repl_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.repl.qb")
    __vector(i8[4]) repl_qb(i32);
    /// The `llvm.mips.pick.qb` intrinsic; known as `__builtin_mips_pick_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pick.qb")
    __vector(i8[4]) pick_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.mthlip` intrinsic; known as `__builtin_mips_mthlip` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mthlip")
    i64 mthlip(i64, i32);
    /// The `llvm.mips.bposge32` intrinsic; known as `__builtin_mips_bposge32` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bposge32")
    i32 bposge32();
    /// The `llvm.mips.lbux` intrinsic; known as `__builtin_mips_lbux` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.lbux")
    i32 lbux(i8*, i32);
    /// The `llvm.mips.lhx` intrinsic; known as `__builtin_mips_lhx` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.lhx")
    i32 lhx(i8*, i32);
    /// The `llvm.mips.lwx` intrinsic; known as `__builtin_mips_lwx` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.lwx")
    i32 lwx(i8*, i32);
    /// The `llvm.mips.addu.ph` intrinsic; known as `__builtin_mips_addu_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addu.ph")
    __vector(i16[2]) addu_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.addu.s.ph` intrinsic; known as `__builtin_mips_addu_s_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addu.s.ph")
    __vector(i16[2]) addu_s_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.adduh.qb` intrinsic; known as `__builtin_mips_adduh_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adduh.qb")
    __vector(i8[4]) adduh_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.adduh.r.qb` intrinsic; known as `__builtin_mips_adduh_r_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adduh.r.qb")
    __vector(i8[4]) adduh_r_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.append` intrinsic; known as `__builtin_mips_append` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.append")
    i32 append(i32, i32, i32);
    /// The `llvm.mips.balign` intrinsic; known as `__builtin_mips_balign` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.balign")
    i32 balign(i32, i32, i32);
    /// The `llvm.mips.cmpgdu.eq.qb` intrinsic; known as `__builtin_mips_cmpgdu_eq_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpgdu.eq.qb")
    i32 cmpgdu_eq_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpgdu.lt.qb` intrinsic; known as `__builtin_mips_cmpgdu_lt_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpgdu.lt.qb")
    i32 cmpgdu_lt_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.cmpgdu.le.qb` intrinsic; known as `__builtin_mips_cmpgdu_le_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cmpgdu.le.qb")
    i32 cmpgdu_le_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.dpa.w.ph` intrinsic; known as `__builtin_mips_dpa_w_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpa.w.ph")
    i64 dpa_w_ph(i64, __vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.dps.w.ph` intrinsic; known as `__builtin_mips_dps_w_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dps.w.ph")
    i64 dps_w_ph(i64, __vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.dpax.w.ph` intrinsic; known as `__builtin_mips_dpax_w_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpax.w.ph")
    i64 dpax_w_ph(i64, __vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.dpsx.w.ph` intrinsic; known as `__builtin_mips_dpsx_w_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsx.w.ph")
    i64 dpsx_w_ph(i64, __vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.mul.ph` intrinsic; known as `__builtin_mips_mul_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mul.ph")
    __vector(i16[2]) mul_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.mul.s.ph` intrinsic; known as `__builtin_mips_mul_s_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mul.s.ph")
    __vector(i16[2]) mul_s_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.mulsa.w.ph` intrinsic; known as `__builtin_mips_mulsa_w_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulsa.w.ph")
    i64 mulsa_w_ph(i64, __vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.precr.qb.ph` intrinsic; known as `__builtin_mips_precr_qb_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.precr.qb.ph")
    __vector(i8[4]) precr_qb_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.precr.sra.ph.w` intrinsic; known as `__builtin_mips_precr_sra_ph_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.precr.sra.ph.w")
    __vector(i16[2]) precr_sra_ph_w(i32, i32, i32);
    /// The `llvm.mips.precr.sra.r.ph.w` intrinsic; known as `__builtin_mips_precr_sra_r_ph_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.precr.sra.r.ph.w")
    __vector(i16[2]) precr_sra_r_ph_w(i32, i32, i32);
    /// The `llvm.mips.prepend` intrinsic; known as `__builtin_mips_prepend` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.prepend")
    i32 prepend(i32, i32, i32);
    /// The `llvm.mips.shra.qb` intrinsic; known as `__builtin_mips_shra_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shra.qb")
    __vector(i8[4]) shra_qb(__vector(i8[4]), i32);
    /// The `llvm.mips.shra.r.qb` intrinsic; known as `__builtin_mips_shra_r_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shra.r.qb")
    __vector(i8[4]) shra_r_qb(__vector(i8[4]), i32);
    /// The `llvm.mips.shrl.ph` intrinsic; known as `__builtin_mips_shrl_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shrl.ph")
    __vector(i16[2]) shrl_ph(__vector(i16[2]), i32);
    /// The `llvm.mips.subu.ph` intrinsic; known as `__builtin_mips_subu_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subu.ph")
    __vector(i16[2]) subu_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.subu.s.ph` intrinsic; known as `__builtin_mips_subu_s_ph` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subu.s.ph")
    __vector(i16[2]) subu_s_ph(__vector(i16[2]), __vector(i16[2]));
    /// The `llvm.mips.subuh.qb` intrinsic; known as `__builtin_mips_subuh_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subuh.qb")
    __vector(i8[4]) subuh_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.subuh.r.qb` intrinsic; known as `__builtin_mips_subuh_r_qb` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subuh.r.qb")
    __vector(i8[4]) subuh_r_qb(__vector(i8[4]), __vector(i8[4]));
    /// The `llvm.mips.add.a.b` intrinsic; known as `__builtin_msa_add_a_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.add.a.b")
    __vector(i8[16]) add_a_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.add.a.h` intrinsic; known as `__builtin_msa_add_a_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.add.a.h")
    __vector(i16[8]) add_a_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.add.a.w` intrinsic; known as `__builtin_msa_add_a_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.add.a.w")
    __vector(i32[4]) add_a_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.add.a.d` intrinsic; known as `__builtin_msa_add_a_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.add.a.d")
    __vector(i64[2]) add_a_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.adds.a.b` intrinsic; known as `__builtin_msa_adds_a_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.a.b")
    __vector(i8[16]) adds_a_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.adds.a.h` intrinsic; known as `__builtin_msa_adds_a_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.a.h")
    __vector(i16[8]) adds_a_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.adds.a.w` intrinsic; known as `__builtin_msa_adds_a_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.a.w")
    __vector(i32[4]) adds_a_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.adds.a.d` intrinsic; known as `__builtin_msa_adds_a_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.a.d")
    __vector(i64[2]) adds_a_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.adds.s.b` intrinsic; known as `__builtin_msa_adds_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.s.b")
    __vector(i8[16]) adds_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.adds.s.h` intrinsic; known as `__builtin_msa_adds_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.s.h")
    __vector(i16[8]) adds_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.adds.s.w` intrinsic; known as `__builtin_msa_adds_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.s.w")
    __vector(i32[4]) adds_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.adds.s.d` intrinsic; known as `__builtin_msa_adds_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.s.d")
    __vector(i64[2]) adds_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.adds.u.b` intrinsic; known as `__builtin_msa_adds_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.u.b")
    __vector(i8[16]) adds_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.adds.u.h` intrinsic; known as `__builtin_msa_adds_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.u.h")
    __vector(i16[8]) adds_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.adds.u.w` intrinsic; known as `__builtin_msa_adds_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.u.w")
    __vector(i32[4]) adds_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.adds.u.d` intrinsic; known as `__builtin_msa_adds_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.adds.u.d")
    __vector(i64[2]) adds_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.addv.b` intrinsic; known as `__builtin_msa_addv_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addv.b")
    __vector(i8[16]) addv_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.addv.h` intrinsic; known as `__builtin_msa_addv_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addv.h")
    __vector(i16[8]) addv_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.addv.w` intrinsic; known as `__builtin_msa_addv_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addv.w")
    __vector(i32[4]) addv_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.addv.d` intrinsic; known as `__builtin_msa_addv_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addv.d")
    __vector(i64[2]) addv_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.addvi.b` intrinsic; known as `__builtin_msa_addvi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addvi.b")
    __vector(i8[16]) addvi_b(__vector(i8[16]), i32);
    /// The `llvm.mips.addvi.h` intrinsic; known as `__builtin_msa_addvi_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addvi.h")
    __vector(i16[8]) addvi_h(__vector(i16[8]), i32);
    /// The `llvm.mips.addvi.w` intrinsic; known as `__builtin_msa_addvi_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addvi.w")
    __vector(i32[4]) addvi_w(__vector(i32[4]), i32);
    /// The `llvm.mips.addvi.d` intrinsic; known as `__builtin_msa_addvi_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.addvi.d")
    __vector(i64[2]) addvi_d(__vector(i64[2]), i32);
    /// The `llvm.mips.and.v` intrinsic; known as `__builtin_msa_and_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.and.v")
    __vector(i8[16]) and_v(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.andi.b` intrinsic; known as `__builtin_msa_andi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.andi.b")
    __vector(i8[16]) andi_b(__vector(i8[16]), i32);
    /// The `llvm.mips.asub.s.b` intrinsic; known as `__builtin_msa_asub_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.s.b")
    __vector(i8[16]) asub_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.asub.s.h` intrinsic; known as `__builtin_msa_asub_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.s.h")
    __vector(i16[8]) asub_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.asub.s.w` intrinsic; known as `__builtin_msa_asub_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.s.w")
    __vector(i32[4]) asub_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.asub.s.d` intrinsic; known as `__builtin_msa_asub_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.s.d")
    __vector(i64[2]) asub_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.asub.u.b` intrinsic; known as `__builtin_msa_asub_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.u.b")
    __vector(i8[16]) asub_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.asub.u.h` intrinsic; known as `__builtin_msa_asub_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.u.h")
    __vector(i16[8]) asub_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.asub.u.w` intrinsic; known as `__builtin_msa_asub_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.u.w")
    __vector(i32[4]) asub_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.asub.u.d` intrinsic; known as `__builtin_msa_asub_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.asub.u.d")
    __vector(i64[2]) asub_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.ave.s.b` intrinsic; known as `__builtin_msa_ave_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.s.b")
    __vector(i8[16]) ave_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ave.s.h` intrinsic; known as `__builtin_msa_ave_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.s.h")
    __vector(i16[8]) ave_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ave.s.w` intrinsic; known as `__builtin_msa_ave_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.s.w")
    __vector(i32[4]) ave_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ave.s.d` intrinsic; known as `__builtin_msa_ave_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.s.d")
    __vector(i64[2]) ave_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.ave.u.b` intrinsic; known as `__builtin_msa_ave_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.u.b")
    __vector(i8[16]) ave_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ave.u.h` intrinsic; known as `__builtin_msa_ave_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.u.h")
    __vector(i16[8]) ave_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ave.u.w` intrinsic; known as `__builtin_msa_ave_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.u.w")
    __vector(i32[4]) ave_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ave.u.d` intrinsic; known as `__builtin_msa_ave_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ave.u.d")
    __vector(i64[2]) ave_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.aver.s.b` intrinsic; known as `__builtin_msa_aver_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.s.b")
    __vector(i8[16]) aver_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.aver.s.h` intrinsic; known as `__builtin_msa_aver_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.s.h")
    __vector(i16[8]) aver_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.aver.s.w` intrinsic; known as `__builtin_msa_aver_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.s.w")
    __vector(i32[4]) aver_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.aver.s.d` intrinsic; known as `__builtin_msa_aver_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.s.d")
    __vector(i64[2]) aver_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.aver.u.b` intrinsic; known as `__builtin_msa_aver_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.u.b")
    __vector(i8[16]) aver_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.aver.u.h` intrinsic; known as `__builtin_msa_aver_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.u.h")
    __vector(i16[8]) aver_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.aver.u.w` intrinsic; known as `__builtin_msa_aver_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.u.w")
    __vector(i32[4]) aver_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.aver.u.d` intrinsic; known as `__builtin_msa_aver_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.aver.u.d")
    __vector(i64[2]) aver_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.bclr.b` intrinsic; known as `__builtin_msa_bclr_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclr.b")
    __vector(i8[16]) bclr_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.bclr.h` intrinsic; known as `__builtin_msa_bclr_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclr.h")
    __vector(i16[8]) bclr_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.bclr.w` intrinsic; known as `__builtin_msa_bclr_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclr.w")
    __vector(i32[4]) bclr_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.bclr.d` intrinsic; known as `__builtin_msa_bclr_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclr.d")
    __vector(i64[2]) bclr_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.bclri.b` intrinsic; known as `__builtin_msa_bclri_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclri.b")
    __vector(i8[16]) bclri_b(__vector(i8[16]), i32);
    /// The `llvm.mips.bclri.h` intrinsic; known as `__builtin_msa_bclri_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclri.h")
    __vector(i16[8]) bclri_h(__vector(i16[8]), i32);
    /// The `llvm.mips.bclri.w` intrinsic; known as `__builtin_msa_bclri_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclri.w")
    __vector(i32[4]) bclri_w(__vector(i32[4]), i32);
    /// The `llvm.mips.bclri.d` intrinsic; known as `__builtin_msa_bclri_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bclri.d")
    __vector(i64[2]) bclri_d(__vector(i64[2]), i32);
    /// The `llvm.mips.binsl.b` intrinsic; known as `__builtin_msa_binsl_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsl.b")
    __vector(i8[16]) binsl_b(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.binsl.h` intrinsic; known as `__builtin_msa_binsl_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsl.h")
    __vector(i16[8]) binsl_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.binsl.w` intrinsic; known as `__builtin_msa_binsl_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsl.w")
    __vector(i32[4]) binsl_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.binsl.d` intrinsic; known as `__builtin_msa_binsl_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsl.d")
    __vector(i64[2]) binsl_d(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.binsli.b` intrinsic; known as `__builtin_msa_binsli_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsli.b")
    __vector(i8[16]) binsli_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.binsli.h` intrinsic; known as `__builtin_msa_binsli_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsli.h")
    __vector(i16[8]) binsli_h(__vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.mips.binsli.w` intrinsic; known as `__builtin_msa_binsli_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsli.w")
    __vector(i32[4]) binsli_w(__vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.mips.binsli.d` intrinsic; known as `__builtin_msa_binsli_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsli.d")
    __vector(i64[2]) binsli_d(__vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.mips.binsr.b` intrinsic; known as `__builtin_msa_binsr_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsr.b")
    __vector(i8[16]) binsr_b(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.binsr.h` intrinsic; known as `__builtin_msa_binsr_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsr.h")
    __vector(i16[8]) binsr_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.binsr.w` intrinsic; known as `__builtin_msa_binsr_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsr.w")
    __vector(i32[4]) binsr_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.binsr.d` intrinsic; known as `__builtin_msa_binsr_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsr.d")
    __vector(i64[2]) binsr_d(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.binsri.b` intrinsic; known as `__builtin_msa_binsri_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsri.b")
    __vector(i8[16]) binsri_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.binsri.h` intrinsic; known as `__builtin_msa_binsri_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsri.h")
    __vector(i16[8]) binsri_h(__vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.mips.binsri.w` intrinsic; known as `__builtin_msa_binsri_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsri.w")
    __vector(i32[4]) binsri_w(__vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.mips.binsri.d` intrinsic; known as `__builtin_msa_binsri_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.binsri.d")
    __vector(i64[2]) binsri_d(__vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.mips.bmnz.v` intrinsic; known as `__builtin_msa_bmnz_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bmnz.v")
    __vector(i8[16]) bmnz_v(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.bmnzi.b` intrinsic; known as `__builtin_msa_bmnzi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bmnzi.b")
    __vector(i8[16]) bmnzi_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.bmz.v` intrinsic; known as `__builtin_msa_bmz_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bmz.v")
    __vector(i8[16]) bmz_v(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.bmzi.b` intrinsic; known as `__builtin_msa_bmzi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bmzi.b")
    __vector(i8[16]) bmzi_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.bneg.b` intrinsic; known as `__builtin_msa_bneg_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bneg.b")
    __vector(i8[16]) bneg_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.bneg.h` intrinsic; known as `__builtin_msa_bneg_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bneg.h")
    __vector(i16[8]) bneg_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.bneg.w` intrinsic; known as `__builtin_msa_bneg_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bneg.w")
    __vector(i32[4]) bneg_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.bneg.d` intrinsic; known as `__builtin_msa_bneg_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bneg.d")
    __vector(i64[2]) bneg_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.bnegi.b` intrinsic; known as `__builtin_msa_bnegi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnegi.b")
    __vector(i8[16]) bnegi_b(__vector(i8[16]), i32);
    /// The `llvm.mips.bnegi.h` intrinsic; known as `__builtin_msa_bnegi_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnegi.h")
    __vector(i16[8]) bnegi_h(__vector(i16[8]), i32);
    /// The `llvm.mips.bnegi.w` intrinsic; known as `__builtin_msa_bnegi_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnegi.w")
    __vector(i32[4]) bnegi_w(__vector(i32[4]), i32);
    /// The `llvm.mips.bnegi.d` intrinsic; known as `__builtin_msa_bnegi_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnegi.d")
    __vector(i64[2]) bnegi_d(__vector(i64[2]), i32);
    /// The `llvm.mips.bnz.b` intrinsic; known as `__builtin_msa_bnz_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnz.b")
    i32 bnz_b(__vector(i8[16]));
    /// The `llvm.mips.bnz.h` intrinsic; known as `__builtin_msa_bnz_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnz.h")
    i32 bnz_h(__vector(i16[8]));
    /// The `llvm.mips.bnz.w` intrinsic; known as `__builtin_msa_bnz_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnz.w")
    i32 bnz_w(__vector(i32[4]));
    /// The `llvm.mips.bnz.d` intrinsic; known as `__builtin_msa_bnz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnz.d")
    i32 bnz_d(__vector(i64[2]));
    /// The `llvm.mips.bnz.v` intrinsic; known as `__builtin_msa_bnz_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bnz.v")
    i32 bnz_v(__vector(i8[16]));
    /// The `llvm.mips.bsel.v` intrinsic; known as `__builtin_msa_bsel_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bsel.v")
    __vector(i8[16]) bsel_v(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.bseli.b` intrinsic; known as `__builtin_msa_bseli_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bseli.b")
    __vector(i8[16]) bseli_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.bset.b` intrinsic; known as `__builtin_msa_bset_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bset.b")
    __vector(i8[16]) bset_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.bset.h` intrinsic; known as `__builtin_msa_bset_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bset.h")
    __vector(i16[8]) bset_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.bset.w` intrinsic; known as `__builtin_msa_bset_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bset.w")
    __vector(i32[4]) bset_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.bset.d` intrinsic; known as `__builtin_msa_bset_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bset.d")
    __vector(i64[2]) bset_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.bseti.b` intrinsic; known as `__builtin_msa_bseti_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bseti.b")
    __vector(i8[16]) bseti_b(__vector(i8[16]), i32);
    /// The `llvm.mips.bseti.h` intrinsic; known as `__builtin_msa_bseti_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bseti.h")
    __vector(i16[8]) bseti_h(__vector(i16[8]), i32);
    /// The `llvm.mips.bseti.w` intrinsic; known as `__builtin_msa_bseti_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bseti.w")
    __vector(i32[4]) bseti_w(__vector(i32[4]), i32);
    /// The `llvm.mips.bseti.d` intrinsic; known as `__builtin_msa_bseti_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bseti.d")
    __vector(i64[2]) bseti_d(__vector(i64[2]), i32);
    /// The `llvm.mips.bz.b` intrinsic; known as `__builtin_msa_bz_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bz.b")
    i32 bz_b(__vector(i8[16]));
    /// The `llvm.mips.bz.h` intrinsic; known as `__builtin_msa_bz_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bz.h")
    i32 bz_h(__vector(i16[8]));
    /// The `llvm.mips.bz.w` intrinsic; known as `__builtin_msa_bz_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bz.w")
    i32 bz_w(__vector(i32[4]));
    /// The `llvm.mips.bz.d` intrinsic; known as `__builtin_msa_bz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bz.d")
    i32 bz_d(__vector(i64[2]));
    /// The `llvm.mips.bz.v` intrinsic; known as `__builtin_msa_bz_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.bz.v")
    i32 bz_v(__vector(i8[16]));
    /// The `llvm.mips.ceq.b` intrinsic; known as `__builtin_msa_ceq_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceq.b")
    __vector(i8[16]) ceq_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ceq.h` intrinsic; known as `__builtin_msa_ceq_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceq.h")
    __vector(i16[8]) ceq_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ceq.w` intrinsic; known as `__builtin_msa_ceq_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceq.w")
    __vector(i32[4]) ceq_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ceq.d` intrinsic; known as `__builtin_msa_ceq_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceq.d")
    __vector(i64[2]) ceq_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.ceqi.b` intrinsic; known as `__builtin_msa_ceqi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceqi.b")
    __vector(i8[16]) ceqi_b(__vector(i8[16]), i32);
    /// The `llvm.mips.ceqi.h` intrinsic; known as `__builtin_msa_ceqi_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceqi.h")
    __vector(i16[8]) ceqi_h(__vector(i16[8]), i32);
    /// The `llvm.mips.ceqi.w` intrinsic; known as `__builtin_msa_ceqi_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceqi.w")
    __vector(i32[4]) ceqi_w(__vector(i32[4]), i32);
    /// The `llvm.mips.ceqi.d` intrinsic; known as `__builtin_msa_ceqi_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ceqi.d")
    __vector(i64[2]) ceqi_d(__vector(i64[2]), i32);
    /// The `llvm.mips.cfcmsa` intrinsic; known as `__builtin_msa_cfcmsa` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cfcmsa")
    i32 cfcmsa(i32);
    /// The `llvm.mips.cle.s.b` intrinsic; known as `__builtin_msa_cle_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.s.b")
    __vector(i8[16]) cle_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.cle.s.h` intrinsic; known as `__builtin_msa_cle_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.s.h")
    __vector(i16[8]) cle_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.cle.s.w` intrinsic; known as `__builtin_msa_cle_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.s.w")
    __vector(i32[4]) cle_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.cle.s.d` intrinsic; known as `__builtin_msa_cle_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.s.d")
    __vector(i64[2]) cle_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.cle.u.b` intrinsic; known as `__builtin_msa_cle_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.u.b")
    __vector(i8[16]) cle_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.cle.u.h` intrinsic; known as `__builtin_msa_cle_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.u.h")
    __vector(i16[8]) cle_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.cle.u.w` intrinsic; known as `__builtin_msa_cle_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.u.w")
    __vector(i32[4]) cle_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.cle.u.d` intrinsic; known as `__builtin_msa_cle_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.cle.u.d")
    __vector(i64[2]) cle_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.clei.s.b` intrinsic; known as `__builtin_msa_clei_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.s.b")
    __vector(i8[16]) clei_s_b(__vector(i8[16]), i32);
    /// The `llvm.mips.clei.s.h` intrinsic; known as `__builtin_msa_clei_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.s.h")
    __vector(i16[8]) clei_s_h(__vector(i16[8]), i32);
    /// The `llvm.mips.clei.s.w` intrinsic; known as `__builtin_msa_clei_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.s.w")
    __vector(i32[4]) clei_s_w(__vector(i32[4]), i32);
    /// The `llvm.mips.clei.s.d` intrinsic; known as `__builtin_msa_clei_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.s.d")
    __vector(i64[2]) clei_s_d(__vector(i64[2]), i32);
    /// The `llvm.mips.clei.u.b` intrinsic; known as `__builtin_msa_clei_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.u.b")
    __vector(i8[16]) clei_u_b(__vector(i8[16]), i32);
    /// The `llvm.mips.clei.u.h` intrinsic; known as `__builtin_msa_clei_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.u.h")
    __vector(i16[8]) clei_u_h(__vector(i16[8]), i32);
    /// The `llvm.mips.clei.u.w` intrinsic; known as `__builtin_msa_clei_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.u.w")
    __vector(i32[4]) clei_u_w(__vector(i32[4]), i32);
    /// The `llvm.mips.clei.u.d` intrinsic; known as `__builtin_msa_clei_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clei.u.d")
    __vector(i64[2]) clei_u_d(__vector(i64[2]), i32);
    /// The `llvm.mips.clt.s.b` intrinsic; known as `__builtin_msa_clt_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.s.b")
    __vector(i8[16]) clt_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.clt.s.h` intrinsic; known as `__builtin_msa_clt_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.s.h")
    __vector(i16[8]) clt_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.clt.s.w` intrinsic; known as `__builtin_msa_clt_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.s.w")
    __vector(i32[4]) clt_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.clt.s.d` intrinsic; known as `__builtin_msa_clt_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.s.d")
    __vector(i64[2]) clt_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.clt.u.b` intrinsic; known as `__builtin_msa_clt_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.u.b")
    __vector(i8[16]) clt_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.clt.u.h` intrinsic; known as `__builtin_msa_clt_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.u.h")
    __vector(i16[8]) clt_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.clt.u.w` intrinsic; known as `__builtin_msa_clt_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.u.w")
    __vector(i32[4]) clt_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.clt.u.d` intrinsic; known as `__builtin_msa_clt_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clt.u.d")
    __vector(i64[2]) clt_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.clti.s.b` intrinsic; known as `__builtin_msa_clti_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.s.b")
    __vector(i8[16]) clti_s_b(__vector(i8[16]), i32);
    /// The `llvm.mips.clti.s.h` intrinsic; known as `__builtin_msa_clti_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.s.h")
    __vector(i16[8]) clti_s_h(__vector(i16[8]), i32);
    /// The `llvm.mips.clti.s.w` intrinsic; known as `__builtin_msa_clti_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.s.w")
    __vector(i32[4]) clti_s_w(__vector(i32[4]), i32);
    /// The `llvm.mips.clti.s.d` intrinsic; known as `__builtin_msa_clti_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.s.d")
    __vector(i64[2]) clti_s_d(__vector(i64[2]), i32);
    /// The `llvm.mips.clti.u.b` intrinsic; known as `__builtin_msa_clti_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.u.b")
    __vector(i8[16]) clti_u_b(__vector(i8[16]), i32);
    /// The `llvm.mips.clti.u.h` intrinsic; known as `__builtin_msa_clti_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.u.h")
    __vector(i16[8]) clti_u_h(__vector(i16[8]), i32);
    /// The `llvm.mips.clti.u.w` intrinsic; known as `__builtin_msa_clti_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.u.w")
    __vector(i32[4]) clti_u_w(__vector(i32[4]), i32);
    /// The `llvm.mips.clti.u.d` intrinsic; known as `__builtin_msa_clti_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.clti.u.d")
    __vector(i64[2]) clti_u_d(__vector(i64[2]), i32);
    /// The `llvm.mips.copy.s.b` intrinsic; known as `__builtin_msa_copy_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.s.b")
    i32 copy_s_b(__vector(i8[16]), i32);
    /// The `llvm.mips.copy.s.h` intrinsic; known as `__builtin_msa_copy_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.s.h")
    i32 copy_s_h(__vector(i16[8]), i32);
    /// The `llvm.mips.copy.s.w` intrinsic; known as `__builtin_msa_copy_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.s.w")
    i32 copy_s_w(__vector(i32[4]), i32);
    /// The `llvm.mips.copy.s.d` intrinsic; known as `__builtin_msa_copy_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.s.d")
    i64 copy_s_d(__vector(i64[2]), i32);
    /// The `llvm.mips.copy.u.b` intrinsic; known as `__builtin_msa_copy_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.u.b")
    i32 copy_u_b(__vector(i8[16]), i32);
    /// The `llvm.mips.copy.u.h` intrinsic; known as `__builtin_msa_copy_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.u.h")
    i32 copy_u_h(__vector(i16[8]), i32);
    /// The `llvm.mips.copy.u.w` intrinsic; known as `__builtin_msa_copy_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.u.w")
    i32 copy_u_w(__vector(i32[4]), i32);
    /// The `llvm.mips.copy.u.d` intrinsic; known as `__builtin_msa_copy_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.copy.u.d")
    i64 copy_u_d(__vector(i64[2]), i32);
    /// The `llvm.mips.ctcmsa` intrinsic; known as `__builtin_msa_ctcmsa` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ctcmsa")
    void ctcmsa(i32, i32);
    /// The `llvm.mips.div.s.b` intrinsic; known as `__builtin_msa_div_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.s.b")
    __vector(i8[16]) div_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.div.s.h` intrinsic; known as `__builtin_msa_div_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.s.h")
    __vector(i16[8]) div_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.div.s.w` intrinsic; known as `__builtin_msa_div_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.s.w")
    __vector(i32[4]) div_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.div.s.d` intrinsic; known as `__builtin_msa_div_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.s.d")
    __vector(i64[2]) div_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.div.u.b` intrinsic; known as `__builtin_msa_div_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.u.b")
    __vector(i8[16]) div_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.div.u.h` intrinsic; known as `__builtin_msa_div_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.u.h")
    __vector(i16[8]) div_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.div.u.w` intrinsic; known as `__builtin_msa_div_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.u.w")
    __vector(i32[4]) div_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.div.u.d` intrinsic; known as `__builtin_msa_div_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.div.u.d")
    __vector(i64[2]) div_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.dlsa` intrinsic; known as `__builtin_mips_dlsa` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dlsa")
    i64 dlsa(i64, i64, i32);
    /// The `llvm.mips.dotp.s.h` intrinsic; known as `__builtin_msa_dotp_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dotp.s.h")
    __vector(i16[8]) dotp_s_h(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.dotp.s.w` intrinsic; known as `__builtin_msa_dotp_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dotp.s.w")
    __vector(i32[4]) dotp_s_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.dotp.s.d` intrinsic; known as `__builtin_msa_dotp_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dotp.s.d")
    __vector(i64[2]) dotp_s_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.dotp.u.h` intrinsic; known as `__builtin_msa_dotp_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dotp.u.h")
    __vector(i16[8]) dotp_u_h(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.dotp.u.w` intrinsic; known as `__builtin_msa_dotp_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dotp.u.w")
    __vector(i32[4]) dotp_u_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.dotp.u.d` intrinsic; known as `__builtin_msa_dotp_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dotp.u.d")
    __vector(i64[2]) dotp_u_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.dpadd.s.h` intrinsic; known as `__builtin_msa_dpadd_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpadd.s.h")
    __vector(i16[8]) dpadd_s_h(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.dpadd.s.w` intrinsic; known as `__builtin_msa_dpadd_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpadd.s.w")
    __vector(i32[4]) dpadd_s_w(__vector(i32[4]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.dpadd.s.d` intrinsic; known as `__builtin_msa_dpadd_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpadd.s.d")
    __vector(i64[2]) dpadd_s_d(__vector(i64[2]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.dpadd.u.h` intrinsic; known as `__builtin_msa_dpadd_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpadd.u.h")
    __vector(i16[8]) dpadd_u_h(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.dpadd.u.w` intrinsic; known as `__builtin_msa_dpadd_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpadd.u.w")
    __vector(i32[4]) dpadd_u_w(__vector(i32[4]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.dpadd.u.d` intrinsic; known as `__builtin_msa_dpadd_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpadd.u.d")
    __vector(i64[2]) dpadd_u_d(__vector(i64[2]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.dpsub.s.h` intrinsic; known as `__builtin_msa_dpsub_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsub.s.h")
    __vector(i16[8]) dpsub_s_h(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.dpsub.s.w` intrinsic; known as `__builtin_msa_dpsub_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsub.s.w")
    __vector(i32[4]) dpsub_s_w(__vector(i32[4]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.dpsub.s.d` intrinsic; known as `__builtin_msa_dpsub_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsub.s.d")
    __vector(i64[2]) dpsub_s_d(__vector(i64[2]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.dpsub.u.h` intrinsic; known as `__builtin_msa_dpsub_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsub.u.h")
    __vector(i16[8]) dpsub_u_h(__vector(i16[8]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.dpsub.u.w` intrinsic; known as `__builtin_msa_dpsub_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsub.u.w")
    __vector(i32[4]) dpsub_u_w(__vector(i32[4]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.dpsub.u.d` intrinsic; known as `__builtin_msa_dpsub_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.dpsub.u.d")
    __vector(i64[2]) dpsub_u_d(__vector(i64[2]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.fadd.w` intrinsic; known as `__builtin_msa_fadd_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fadd.w")
    __vector(f32[4]) fadd_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fadd.d` intrinsic; known as `__builtin_msa_fadd_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fadd.d")
    __vector(f64[2]) fadd_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcaf.w` intrinsic; known as `__builtin_msa_fcaf_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcaf.w")
    __vector(i32[4]) fcaf_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcaf.d` intrinsic; known as `__builtin_msa_fcaf_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcaf.d")
    __vector(i64[2]) fcaf_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fceq.w` intrinsic; known as `__builtin_msa_fceq_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fceq.w")
    __vector(i32[4]) fceq_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fceq.d` intrinsic; known as `__builtin_msa_fceq_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fceq.d")
    __vector(i64[2]) fceq_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcle.w` intrinsic; known as `__builtin_msa_fcle_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcle.w")
    __vector(i32[4]) fcle_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcle.d` intrinsic; known as `__builtin_msa_fcle_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcle.d")
    __vector(i64[2]) fcle_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fclt.w` intrinsic; known as `__builtin_msa_fclt_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fclt.w")
    __vector(i32[4]) fclt_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fclt.d` intrinsic; known as `__builtin_msa_fclt_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fclt.d")
    __vector(i64[2]) fclt_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fclass.w` intrinsic; known as `__builtin_msa_fclass_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fclass.w")
    __vector(i32[4]) fclass_w(__vector(f32[4]));
    /// The `llvm.mips.fclass.d` intrinsic; known as `__builtin_msa_fclass_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fclass.d")
    __vector(i64[2]) fclass_d(__vector(f64[2]));
    /// The `llvm.mips.fcne.w` intrinsic; known as `__builtin_msa_fcne_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcne.w")
    __vector(i32[4]) fcne_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcne.d` intrinsic; known as `__builtin_msa_fcne_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcne.d")
    __vector(i64[2]) fcne_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcor.w` intrinsic; known as `__builtin_msa_fcor_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcor.w")
    __vector(i32[4]) fcor_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcor.d` intrinsic; known as `__builtin_msa_fcor_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcor.d")
    __vector(i64[2]) fcor_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcueq.w` intrinsic; known as `__builtin_msa_fcueq_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcueq.w")
    __vector(i32[4]) fcueq_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcueq.d` intrinsic; known as `__builtin_msa_fcueq_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcueq.d")
    __vector(i64[2]) fcueq_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcule.w` intrinsic; known as `__builtin_msa_fcule_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcule.w")
    __vector(i32[4]) fcule_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcule.d` intrinsic; known as `__builtin_msa_fcule_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcule.d")
    __vector(i64[2]) fcule_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcult.w` intrinsic; known as `__builtin_msa_fcult_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcult.w")
    __vector(i32[4]) fcult_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcult.d` intrinsic; known as `__builtin_msa_fcult_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcult.d")
    __vector(i64[2]) fcult_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcun.w` intrinsic; known as `__builtin_msa_fcun_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcun.w")
    __vector(i32[4]) fcun_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcun.d` intrinsic; known as `__builtin_msa_fcun_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcun.d")
    __vector(i64[2]) fcun_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fcune.w` intrinsic; known as `__builtin_msa_fcune_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcune.w")
    __vector(i32[4]) fcune_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fcune.d` intrinsic; known as `__builtin_msa_fcune_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fcune.d")
    __vector(i64[2]) fcune_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fdiv.w` intrinsic; known as `__builtin_msa_fdiv_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fdiv.w")
    __vector(f32[4]) fdiv_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fdiv.d` intrinsic; known as `__builtin_msa_fdiv_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fdiv.d")
    __vector(f64[2]) fdiv_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fexdo.w` intrinsic; known as `__builtin_msa_fexdo_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fexdo.w")
    __vector(f32[4]) fexdo_w(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fexp2.w` intrinsic; known as `__builtin_msa_fexp2_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fexp2.w")
    __vector(f32[4]) fexp2_w(__vector(f32[4]), __vector(i32[4]));
    /// The `llvm.mips.fexp2.d` intrinsic; known as `__builtin_msa_fexp2_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fexp2.d")
    __vector(f64[2]) fexp2_d(__vector(f64[2]), __vector(i64[2]));
    /// The `llvm.mips.fexupl.d` intrinsic; known as `__builtin_msa_fexupl_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fexupl.d")
    __vector(f64[2]) fexupl_d(__vector(f32[4]));
    /// The `llvm.mips.fexupr.d` intrinsic; known as `__builtin_msa_fexupr_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fexupr.d")
    __vector(f64[2]) fexupr_d(__vector(f32[4]));
    /// The `llvm.mips.ffint.s.w` intrinsic; known as `__builtin_msa_ffint_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffint.s.w")
    __vector(f32[4]) ffint_s_w(__vector(i32[4]));
    /// The `llvm.mips.ffint.s.d` intrinsic; known as `__builtin_msa_ffint_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffint.s.d")
    __vector(f64[2]) ffint_s_d(__vector(i64[2]));
    /// The `llvm.mips.ffint.u.w` intrinsic; known as `__builtin_msa_ffint_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffint.u.w")
    __vector(f32[4]) ffint_u_w(__vector(i32[4]));
    /// The `llvm.mips.ffint.u.d` intrinsic; known as `__builtin_msa_ffint_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffint.u.d")
    __vector(f64[2]) ffint_u_d(__vector(i64[2]));
    /// The `llvm.mips.ffql.w` intrinsic; known as `__builtin_msa_ffql_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffql.w")
    __vector(f32[4]) ffql_w(__vector(i16[8]));
    /// The `llvm.mips.ffql.d` intrinsic; known as `__builtin_msa_ffql_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffql.d")
    __vector(f64[2]) ffql_d(__vector(i32[4]));
    /// The `llvm.mips.ffqr.w` intrinsic; known as `__builtin_msa_ffqr_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffqr.w")
    __vector(f32[4]) ffqr_w(__vector(i16[8]));
    /// The `llvm.mips.ffqr.d` intrinsic; known as `__builtin_msa_ffqr_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ffqr.d")
    __vector(f64[2]) ffqr_d(__vector(i32[4]));
    /// The `llvm.mips.fill.b` intrinsic; known as `__builtin_msa_fill_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fill.b")
    __vector(i8[16]) fill_b(i32);
    /// The `llvm.mips.fill.h` intrinsic; known as `__builtin_msa_fill_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fill.h")
    __vector(i16[8]) fill_h(i32);
    /// The `llvm.mips.fill.w` intrinsic; known as `__builtin_msa_fill_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fill.w")
    __vector(i32[4]) fill_w(i32);
    /// The `llvm.mips.fill.d` intrinsic; known as `__builtin_msa_fill_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fill.d")
    __vector(i64[2]) fill_d(i64);
    /// The `llvm.mips.flog2.w` intrinsic; known as `__builtin_msa_flog2_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.flog2.w")
    __vector(f32[4]) flog2_w(__vector(f32[4]));
    /// The `llvm.mips.flog2.d` intrinsic; known as `__builtin_msa_flog2_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.flog2.d")
    __vector(f64[2]) flog2_d(__vector(f64[2]));
    /// The `llvm.mips.fmadd.w` intrinsic; known as `__builtin_msa_fmadd_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmadd.w")
    __vector(f32[4]) fmadd_w(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmadd.d` intrinsic; known as `__builtin_msa_fmadd_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmadd.d")
    __vector(f64[2]) fmadd_d(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fmax.w` intrinsic; known as `__builtin_msa_fmax_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmax.w")
    __vector(f32[4]) fmax_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmax.d` intrinsic; known as `__builtin_msa_fmax_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmax.d")
    __vector(f64[2]) fmax_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fmax.a.w` intrinsic; known as `__builtin_msa_fmax_a_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmax.a.w")
    __vector(f32[4]) fmax_a_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmax.a.d` intrinsic; known as `__builtin_msa_fmax_a_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmax.a.d")
    __vector(f64[2]) fmax_a_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fmin.w` intrinsic; known as `__builtin_msa_fmin_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmin.w")
    __vector(f32[4]) fmin_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmin.d` intrinsic; known as `__builtin_msa_fmin_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmin.d")
    __vector(f64[2]) fmin_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fmin.a.w` intrinsic; known as `__builtin_msa_fmin_a_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmin.a.w")
    __vector(f32[4]) fmin_a_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmin.a.d` intrinsic; known as `__builtin_msa_fmin_a_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmin.a.d")
    __vector(f64[2]) fmin_a_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fmsub.w` intrinsic; known as `__builtin_msa_fmsub_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmsub.w")
    __vector(f32[4]) fmsub_w(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmsub.d` intrinsic; known as `__builtin_msa_fmsub_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmsub.d")
    __vector(f64[2]) fmsub_d(__vector(f64[2]), __vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fmul.w` intrinsic; known as `__builtin_msa_fmul_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmul.w")
    __vector(f32[4]) fmul_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fmul.d` intrinsic; known as `__builtin_msa_fmul_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fmul.d")
    __vector(f64[2]) fmul_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.frint.w` intrinsic; known as `__builtin_msa_frint_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.frint.w")
    __vector(f32[4]) frint_w(__vector(f32[4]));
    /// The `llvm.mips.frint.d` intrinsic; known as `__builtin_msa_frint_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.frint.d")
    __vector(f64[2]) frint_d(__vector(f64[2]));
    /// The `llvm.mips.frcp.w` intrinsic; known as `__builtin_msa_frcp_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.frcp.w")
    __vector(f32[4]) frcp_w(__vector(f32[4]));
    /// The `llvm.mips.frcp.d` intrinsic; known as `__builtin_msa_frcp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.frcp.d")
    __vector(f64[2]) frcp_d(__vector(f64[2]));
    /// The `llvm.mips.frsqrt.w` intrinsic; known as `__builtin_msa_frsqrt_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.frsqrt.w")
    __vector(f32[4]) frsqrt_w(__vector(f32[4]));
    /// The `llvm.mips.frsqrt.d` intrinsic; known as `__builtin_msa_frsqrt_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.frsqrt.d")
    __vector(f64[2]) frsqrt_d(__vector(f64[2]));
    /// The `llvm.mips.fsaf.w` intrinsic; known as `__builtin_msa_fsaf_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsaf.w")
    __vector(i32[4]) fsaf_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsaf.d` intrinsic; known as `__builtin_msa_fsaf_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsaf.d")
    __vector(i64[2]) fsaf_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fseq.w` intrinsic; known as `__builtin_msa_fseq_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fseq.w")
    __vector(i32[4]) fseq_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fseq.d` intrinsic; known as `__builtin_msa_fseq_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fseq.d")
    __vector(i64[2]) fseq_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsle.w` intrinsic; known as `__builtin_msa_fsle_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsle.w")
    __vector(i32[4]) fsle_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsle.d` intrinsic; known as `__builtin_msa_fsle_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsle.d")
    __vector(i64[2]) fsle_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fslt.w` intrinsic; known as `__builtin_msa_fslt_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fslt.w")
    __vector(i32[4]) fslt_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fslt.d` intrinsic; known as `__builtin_msa_fslt_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fslt.d")
    __vector(i64[2]) fslt_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsne.w` intrinsic; known as `__builtin_msa_fsne_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsne.w")
    __vector(i32[4]) fsne_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsne.d` intrinsic; known as `__builtin_msa_fsne_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsne.d")
    __vector(i64[2]) fsne_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsor.w` intrinsic; known as `__builtin_msa_fsor_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsor.w")
    __vector(i32[4]) fsor_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsor.d` intrinsic; known as `__builtin_msa_fsor_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsor.d")
    __vector(i64[2]) fsor_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsqrt.w` intrinsic; known as `__builtin_msa_fsqrt_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsqrt.w")
    __vector(f32[4]) fsqrt_w(__vector(f32[4]));
    /// The `llvm.mips.fsqrt.d` intrinsic; known as `__builtin_msa_fsqrt_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsqrt.d")
    __vector(f64[2]) fsqrt_d(__vector(f64[2]));
    /// The `llvm.mips.fsub.w` intrinsic; known as `__builtin_msa_fsub_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsub.w")
    __vector(f32[4]) fsub_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsub.d` intrinsic; known as `__builtin_msa_fsub_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsub.d")
    __vector(f64[2]) fsub_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsueq.w` intrinsic; known as `__builtin_msa_fsueq_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsueq.w")
    __vector(i32[4]) fsueq_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsueq.d` intrinsic; known as `__builtin_msa_fsueq_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsueq.d")
    __vector(i64[2]) fsueq_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsule.w` intrinsic; known as `__builtin_msa_fsule_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsule.w")
    __vector(i32[4]) fsule_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsule.d` intrinsic; known as `__builtin_msa_fsule_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsule.d")
    __vector(i64[2]) fsule_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsult.w` intrinsic; known as `__builtin_msa_fsult_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsult.w")
    __vector(i32[4]) fsult_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsult.d` intrinsic; known as `__builtin_msa_fsult_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsult.d")
    __vector(i64[2]) fsult_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsun.w` intrinsic; known as `__builtin_msa_fsun_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsun.w")
    __vector(i32[4]) fsun_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsun.d` intrinsic; known as `__builtin_msa_fsun_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsun.d")
    __vector(i64[2]) fsun_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.fsune.w` intrinsic; known as `__builtin_msa_fsune_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsune.w")
    __vector(i32[4]) fsune_w(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.fsune.d` intrinsic; known as `__builtin_msa_fsune_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.fsune.d")
    __vector(i64[2]) fsune_d(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.ftint.s.w` intrinsic; known as `__builtin_msa_ftint_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftint.s.w")
    __vector(i32[4]) ftint_s_w(__vector(f32[4]));
    /// The `llvm.mips.ftint.s.d` intrinsic; known as `__builtin_msa_ftint_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftint.s.d")
    __vector(i64[2]) ftint_s_d(__vector(f64[2]));
    /// The `llvm.mips.ftint.u.w` intrinsic; known as `__builtin_msa_ftint_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftint.u.w")
    __vector(i32[4]) ftint_u_w(__vector(f32[4]));
    /// The `llvm.mips.ftint.u.d` intrinsic; known as `__builtin_msa_ftint_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftint.u.d")
    __vector(i64[2]) ftint_u_d(__vector(f64[2]));
    /// The `llvm.mips.ftq.h` intrinsic; known as `__builtin_msa_ftq_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftq.h")
    __vector(i16[8]) ftq_h(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.mips.ftq.w` intrinsic; known as `__builtin_msa_ftq_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftq.w")
    __vector(i32[4]) ftq_w(__vector(f64[2]), __vector(f64[2]));
    /// The `llvm.mips.ftrunc.s.w` intrinsic; known as `__builtin_msa_ftrunc_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftrunc.s.w")
    __vector(i32[4]) ftrunc_s_w(__vector(f32[4]));
    /// The `llvm.mips.ftrunc.s.d` intrinsic; known as `__builtin_msa_ftrunc_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftrunc.s.d")
    __vector(i64[2]) ftrunc_s_d(__vector(f64[2]));
    /// The `llvm.mips.ftrunc.u.w` intrinsic; known as `__builtin_msa_ftrunc_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftrunc.u.w")
    __vector(i32[4]) ftrunc_u_w(__vector(f32[4]));
    /// The `llvm.mips.ftrunc.u.d` intrinsic; known as `__builtin_msa_ftrunc_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ftrunc.u.d")
    __vector(i64[2]) ftrunc_u_d(__vector(f64[2]));
    /// The `llvm.mips.hadd.s.h` intrinsic; known as `__builtin_msa_hadd_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hadd.s.h")
    __vector(i16[8]) hadd_s_h(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.hadd.s.w` intrinsic; known as `__builtin_msa_hadd_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hadd.s.w")
    __vector(i32[4]) hadd_s_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.hadd.s.d` intrinsic; known as `__builtin_msa_hadd_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hadd.s.d")
    __vector(i64[2]) hadd_s_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.hadd.u.h` intrinsic; known as `__builtin_msa_hadd_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hadd.u.h")
    __vector(i16[8]) hadd_u_h(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.hadd.u.w` intrinsic; known as `__builtin_msa_hadd_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hadd.u.w")
    __vector(i32[4]) hadd_u_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.hadd.u.d` intrinsic; known as `__builtin_msa_hadd_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hadd.u.d")
    __vector(i64[2]) hadd_u_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.hsub.s.h` intrinsic; known as `__builtin_msa_hsub_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hsub.s.h")
    __vector(i16[8]) hsub_s_h(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.hsub.s.w` intrinsic; known as `__builtin_msa_hsub_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hsub.s.w")
    __vector(i32[4]) hsub_s_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.hsub.s.d` intrinsic; known as `__builtin_msa_hsub_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hsub.s.d")
    __vector(i64[2]) hsub_s_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.hsub.u.h` intrinsic; known as `__builtin_msa_hsub_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hsub.u.h")
    __vector(i16[8]) hsub_u_h(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.hsub.u.w` intrinsic; known as `__builtin_msa_hsub_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hsub.u.w")
    __vector(i32[4]) hsub_u_w(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.hsub.u.d` intrinsic; known as `__builtin_msa_hsub_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.hsub.u.d")
    __vector(i64[2]) hsub_u_d(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ilvev.b` intrinsic; known as `__builtin_msa_ilvev_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvev.b")
    __vector(i8[16]) ilvev_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ilvev.h` intrinsic; known as `__builtin_msa_ilvev_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvev.h")
    __vector(i16[8]) ilvev_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ilvev.w` intrinsic; known as `__builtin_msa_ilvev_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvev.w")
    __vector(i32[4]) ilvev_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ilvev.d` intrinsic; known as `__builtin_msa_ilvev_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvev.d")
    __vector(i64[2]) ilvev_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.ilvl.b` intrinsic; known as `__builtin_msa_ilvl_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvl.b")
    __vector(i8[16]) ilvl_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ilvl.h` intrinsic; known as `__builtin_msa_ilvl_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvl.h")
    __vector(i16[8]) ilvl_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ilvl.w` intrinsic; known as `__builtin_msa_ilvl_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvl.w")
    __vector(i32[4]) ilvl_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ilvl.d` intrinsic; known as `__builtin_msa_ilvl_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvl.d")
    __vector(i64[2]) ilvl_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.ilvod.b` intrinsic; known as `__builtin_msa_ilvod_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvod.b")
    __vector(i8[16]) ilvod_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ilvod.h` intrinsic; known as `__builtin_msa_ilvod_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvod.h")
    __vector(i16[8]) ilvod_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ilvod.w` intrinsic; known as `__builtin_msa_ilvod_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvod.w")
    __vector(i32[4]) ilvod_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ilvod.d` intrinsic; known as `__builtin_msa_ilvod_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvod.d")
    __vector(i64[2]) ilvod_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.ilvr.b` intrinsic; known as `__builtin_msa_ilvr_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvr.b")
    __vector(i8[16]) ilvr_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ilvr.h` intrinsic; known as `__builtin_msa_ilvr_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvr.h")
    __vector(i16[8]) ilvr_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.ilvr.w` intrinsic; known as `__builtin_msa_ilvr_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvr.w")
    __vector(i32[4]) ilvr_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.ilvr.d` intrinsic; known as `__builtin_msa_ilvr_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ilvr.d")
    __vector(i64[2]) ilvr_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.insert.b` intrinsic; known as `__builtin_msa_insert_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insert.b")
    __vector(i8[16]) insert_b(__vector(i8[16]), i32, i32);
    /// The `llvm.mips.insert.h` intrinsic; known as `__builtin_msa_insert_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insert.h")
    __vector(i16[8]) insert_h(__vector(i16[8]), i32, i32);
    /// The `llvm.mips.insert.w` intrinsic; known as `__builtin_msa_insert_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insert.w")
    __vector(i32[4]) insert_w(__vector(i32[4]), i32, i32);
    /// The `llvm.mips.insert.d` intrinsic; known as `__builtin_msa_insert_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insert.d")
    __vector(i64[2]) insert_d(__vector(i64[2]), i32, i64);
    /// The `llvm.mips.insve.b` intrinsic; known as `__builtin_msa_insve_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insve.b")
    __vector(i8[16]) insve_b(__vector(i8[16]), i32, __vector(i8[16]));
    /// The `llvm.mips.insve.h` intrinsic; known as `__builtin_msa_insve_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insve.h")
    __vector(i16[8]) insve_h(__vector(i16[8]), i32, __vector(i16[8]));
    /// The `llvm.mips.insve.w` intrinsic; known as `__builtin_msa_insve_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insve.w")
    __vector(i32[4]) insve_w(__vector(i32[4]), i32, __vector(i32[4]));
    /// The `llvm.mips.insve.d` intrinsic; known as `__builtin_msa_insve_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.insve.d")
    __vector(i64[2]) insve_d(__vector(i64[2]), i32, __vector(i64[2]));
    /// The `llvm.mips.ld.b` intrinsic; known as `__builtin_msa_ld_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ld.b")
    __vector(i8[16]) ld_b(i8*, i32);
    /// The `llvm.mips.ld.h` intrinsic; known as `__builtin_msa_ld_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ld.h")
    __vector(i16[8]) ld_h(i8*, i32);
    /// The `llvm.mips.ld.w` intrinsic; known as `__builtin_msa_ld_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ld.w")
    __vector(i32[4]) ld_w(i8*, i32);
    /// The `llvm.mips.ld.d` intrinsic; known as `__builtin_msa_ld_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ld.d")
    __vector(i64[2]) ld_d(i8*, i32);
    /// The `llvm.mips.ldi.b` intrinsic; known as `__builtin_msa_ldi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ldi.b")
    __vector(i8[16]) ldi_b(i32);
    /// The `llvm.mips.ldi.h` intrinsic; known as `__builtin_msa_ldi_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ldi.h")
    __vector(i16[8]) ldi_h(i32);
    /// The `llvm.mips.ldi.w` intrinsic; known as `__builtin_msa_ldi_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ldi.w")
    __vector(i32[4]) ldi_w(i32);
    /// The `llvm.mips.ldi.d` intrinsic; known as `__builtin_msa_ldi_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ldi.d")
    __vector(i64[2]) ldi_d(i32);
    /// The `llvm.mips.lsa` intrinsic; known as `__builtin_mips_lsa` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.lsa")
    i32 lsa(i32, i32, i32);
    /// The `llvm.mips.madd.q.h` intrinsic; known as `__builtin_msa_madd_q_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.madd.q.h")
    __vector(i16[8]) madd_q_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.madd.q.w` intrinsic; known as `__builtin_msa_madd_q_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.madd.q.w")
    __vector(i32[4]) madd_q_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.maddr.q.h` intrinsic; known as `__builtin_msa_maddr_q_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddr.q.h")
    __vector(i16[8]) maddr_q_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.maddr.q.w` intrinsic; known as `__builtin_msa_maddr_q_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddr.q.w")
    __vector(i32[4]) maddr_q_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.maddv.b` intrinsic; known as `__builtin_msa_maddv_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddv.b")
    __vector(i8[16]) maddv_b(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.maddv.h` intrinsic; known as `__builtin_msa_maddv_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddv.h")
    __vector(i16[8]) maddv_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.maddv.w` intrinsic; known as `__builtin_msa_maddv_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddv.w")
    __vector(i32[4]) maddv_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.maddv.d` intrinsic; known as `__builtin_msa_maddv_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maddv.d")
    __vector(i64[2]) maddv_d(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.max.a.b` intrinsic; known as `__builtin_msa_max_a_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.a.b")
    __vector(i8[16]) max_a_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.max.a.h` intrinsic; known as `__builtin_msa_max_a_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.a.h")
    __vector(i16[8]) max_a_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.max.a.w` intrinsic; known as `__builtin_msa_max_a_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.a.w")
    __vector(i32[4]) max_a_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.max.a.d` intrinsic; known as `__builtin_msa_max_a_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.a.d")
    __vector(i64[2]) max_a_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.max.s.b` intrinsic; known as `__builtin_msa_max_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.s.b")
    __vector(i8[16]) max_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.max.s.h` intrinsic; known as `__builtin_msa_max_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.s.h")
    __vector(i16[8]) max_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.max.s.w` intrinsic; known as `__builtin_msa_max_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.s.w")
    __vector(i32[4]) max_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.max.s.d` intrinsic; known as `__builtin_msa_max_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.s.d")
    __vector(i64[2]) max_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.max.u.b` intrinsic; known as `__builtin_msa_max_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.u.b")
    __vector(i8[16]) max_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.max.u.h` intrinsic; known as `__builtin_msa_max_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.u.h")
    __vector(i16[8]) max_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.max.u.w` intrinsic; known as `__builtin_msa_max_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.u.w")
    __vector(i32[4]) max_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.max.u.d` intrinsic; known as `__builtin_msa_max_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.max.u.d")
    __vector(i64[2]) max_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.maxi.s.b` intrinsic; known as `__builtin_msa_maxi_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.s.b")
    __vector(i8[16]) maxi_s_b(__vector(i8[16]), i32);
    /// The `llvm.mips.maxi.s.h` intrinsic; known as `__builtin_msa_maxi_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.s.h")
    __vector(i16[8]) maxi_s_h(__vector(i16[8]), i32);
    /// The `llvm.mips.maxi.s.w` intrinsic; known as `__builtin_msa_maxi_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.s.w")
    __vector(i32[4]) maxi_s_w(__vector(i32[4]), i32);
    /// The `llvm.mips.maxi.s.d` intrinsic; known as `__builtin_msa_maxi_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.s.d")
    __vector(i64[2]) maxi_s_d(__vector(i64[2]), i32);
    /// The `llvm.mips.maxi.u.b` intrinsic; known as `__builtin_msa_maxi_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.u.b")
    __vector(i8[16]) maxi_u_b(__vector(i8[16]), i32);
    /// The `llvm.mips.maxi.u.h` intrinsic; known as `__builtin_msa_maxi_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.u.h")
    __vector(i16[8]) maxi_u_h(__vector(i16[8]), i32);
    /// The `llvm.mips.maxi.u.w` intrinsic; known as `__builtin_msa_maxi_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.u.w")
    __vector(i32[4]) maxi_u_w(__vector(i32[4]), i32);
    /// The `llvm.mips.maxi.u.d` intrinsic; known as `__builtin_msa_maxi_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.maxi.u.d")
    __vector(i64[2]) maxi_u_d(__vector(i64[2]), i32);
    /// The `llvm.mips.min.a.b` intrinsic; known as `__builtin_msa_min_a_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.a.b")
    __vector(i8[16]) min_a_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.min.a.h` intrinsic; known as `__builtin_msa_min_a_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.a.h")
    __vector(i16[8]) min_a_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.min.a.w` intrinsic; known as `__builtin_msa_min_a_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.a.w")
    __vector(i32[4]) min_a_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.min.a.d` intrinsic; known as `__builtin_msa_min_a_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.a.d")
    __vector(i64[2]) min_a_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.min.s.b` intrinsic; known as `__builtin_msa_min_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.s.b")
    __vector(i8[16]) min_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.min.s.h` intrinsic; known as `__builtin_msa_min_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.s.h")
    __vector(i16[8]) min_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.min.s.w` intrinsic; known as `__builtin_msa_min_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.s.w")
    __vector(i32[4]) min_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.min.s.d` intrinsic; known as `__builtin_msa_min_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.s.d")
    __vector(i64[2]) min_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.min.u.b` intrinsic; known as `__builtin_msa_min_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.u.b")
    __vector(i8[16]) min_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.min.u.h` intrinsic; known as `__builtin_msa_min_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.u.h")
    __vector(i16[8]) min_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.min.u.w` intrinsic; known as `__builtin_msa_min_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.u.w")
    __vector(i32[4]) min_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.min.u.d` intrinsic; known as `__builtin_msa_min_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.min.u.d")
    __vector(i64[2]) min_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.mini.s.b` intrinsic; known as `__builtin_msa_mini_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.s.b")
    __vector(i8[16]) mini_s_b(__vector(i8[16]), i32);
    /// The `llvm.mips.mini.s.h` intrinsic; known as `__builtin_msa_mini_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.s.h")
    __vector(i16[8]) mini_s_h(__vector(i16[8]), i32);
    /// The `llvm.mips.mini.s.w` intrinsic; known as `__builtin_msa_mini_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.s.w")
    __vector(i32[4]) mini_s_w(__vector(i32[4]), i32);
    /// The `llvm.mips.mini.s.d` intrinsic; known as `__builtin_msa_mini_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.s.d")
    __vector(i64[2]) mini_s_d(__vector(i64[2]), i32);
    /// The `llvm.mips.mini.u.b` intrinsic; known as `__builtin_msa_mini_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.u.b")
    __vector(i8[16]) mini_u_b(__vector(i8[16]), i32);
    /// The `llvm.mips.mini.u.h` intrinsic; known as `__builtin_msa_mini_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.u.h")
    __vector(i16[8]) mini_u_h(__vector(i16[8]), i32);
    /// The `llvm.mips.mini.u.w` intrinsic; known as `__builtin_msa_mini_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.u.w")
    __vector(i32[4]) mini_u_w(__vector(i32[4]), i32);
    /// The `llvm.mips.mini.u.d` intrinsic; known as `__builtin_msa_mini_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mini.u.d")
    __vector(i64[2]) mini_u_d(__vector(i64[2]), i32);
    /// The `llvm.mips.mod.s.b` intrinsic; known as `__builtin_msa_mod_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.s.b")
    __vector(i8[16]) mod_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.mod.s.h` intrinsic; known as `__builtin_msa_mod_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.s.h")
    __vector(i16[8]) mod_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.mod.s.w` intrinsic; known as `__builtin_msa_mod_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.s.w")
    __vector(i32[4]) mod_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.mod.s.d` intrinsic; known as `__builtin_msa_mod_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.s.d")
    __vector(i64[2]) mod_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.mod.u.b` intrinsic; known as `__builtin_msa_mod_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.u.b")
    __vector(i8[16]) mod_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.mod.u.h` intrinsic; known as `__builtin_msa_mod_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.u.h")
    __vector(i16[8]) mod_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.mod.u.w` intrinsic; known as `__builtin_msa_mod_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.u.w")
    __vector(i32[4]) mod_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.mod.u.d` intrinsic; known as `__builtin_msa_mod_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mod.u.d")
    __vector(i64[2]) mod_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.move.v` intrinsic; known as `__builtin_msa_move_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.move.v")
    __vector(i8[16]) move_v(__vector(i8[16]));
    /// The `llvm.mips.msub.q.h` intrinsic; known as `__builtin_msa_msub_q_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msub.q.h")
    __vector(i16[8]) msub_q_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.msub.q.w` intrinsic; known as `__builtin_msa_msub_q_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msub.q.w")
    __vector(i32[4]) msub_q_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.msubr.q.h` intrinsic; known as `__builtin_msa_msubr_q_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubr.q.h")
    __vector(i16[8]) msubr_q_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.msubr.q.w` intrinsic; known as `__builtin_msa_msubr_q_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubr.q.w")
    __vector(i32[4]) msubr_q_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.msubv.b` intrinsic; known as `__builtin_msa_msubv_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubv.b")
    __vector(i8[16]) msubv_b(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.msubv.h` intrinsic; known as `__builtin_msa_msubv_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubv.h")
    __vector(i16[8]) msubv_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.msubv.w` intrinsic; known as `__builtin_msa_msubv_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubv.w")
    __vector(i32[4]) msubv_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.msubv.d` intrinsic; known as `__builtin_msa_msubv_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.msubv.d")
    __vector(i64[2]) msubv_d(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.mul.q.h` intrinsic; known as `__builtin_msa_mul_q_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mul.q.h")
    __vector(i16[8]) mul_q_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.mul.q.w` intrinsic; known as `__builtin_msa_mul_q_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mul.q.w")
    __vector(i32[4]) mul_q_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.mulr.q.h` intrinsic; known as `__builtin_msa_mulr_q_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulr.q.h")
    __vector(i16[8]) mulr_q_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.mulr.q.w` intrinsic; known as `__builtin_msa_mulr_q_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulr.q.w")
    __vector(i32[4]) mulr_q_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.mulv.b` intrinsic; known as `__builtin_msa_mulv_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulv.b")
    __vector(i8[16]) mulv_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.mulv.h` intrinsic; known as `__builtin_msa_mulv_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulv.h")
    __vector(i16[8]) mulv_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.mulv.w` intrinsic; known as `__builtin_msa_mulv_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulv.w")
    __vector(i32[4]) mulv_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.mulv.d` intrinsic; known as `__builtin_msa_mulv_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.mulv.d")
    __vector(i64[2]) mulv_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.nloc.b` intrinsic; known as `__builtin_msa_nloc_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nloc.b")
    __vector(i8[16]) nloc_b(__vector(i8[16]));
    /// The `llvm.mips.nloc.h` intrinsic; known as `__builtin_msa_nloc_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nloc.h")
    __vector(i16[8]) nloc_h(__vector(i16[8]));
    /// The `llvm.mips.nloc.w` intrinsic; known as `__builtin_msa_nloc_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nloc.w")
    __vector(i32[4]) nloc_w(__vector(i32[4]));
    /// The `llvm.mips.nloc.d` intrinsic; known as `__builtin_msa_nloc_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nloc.d")
    __vector(i64[2]) nloc_d(__vector(i64[2]));
    /// The `llvm.mips.nlzc.b` intrinsic; known as `__builtin_msa_nlzc_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nlzc.b")
    __vector(i8[16]) nlzc_b(__vector(i8[16]));
    /// The `llvm.mips.nlzc.h` intrinsic; known as `__builtin_msa_nlzc_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nlzc.h")
    __vector(i16[8]) nlzc_h(__vector(i16[8]));
    /// The `llvm.mips.nlzc.w` intrinsic; known as `__builtin_msa_nlzc_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nlzc.w")
    __vector(i32[4]) nlzc_w(__vector(i32[4]));
    /// The `llvm.mips.nlzc.d` intrinsic; known as `__builtin_msa_nlzc_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nlzc.d")
    __vector(i64[2]) nlzc_d(__vector(i64[2]));
    /// The `llvm.mips.nor.v` intrinsic; known as `__builtin_msa_nor_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nor.v")
    __vector(i8[16]) nor_v(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.nori.b` intrinsic; known as `__builtin_msa_nori_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.nori.b")
    __vector(i8[16]) nori_b(__vector(i8[16]), i32);
    /// The `llvm.mips.or.v` intrinsic; known as `__builtin_msa_or_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.or.v")
    __vector(i8[16]) or_v(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.ori.b` intrinsic; known as `__builtin_msa_ori_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.ori.b")
    __vector(i8[16]) ori_b(__vector(i8[16]), i32);
    /// The `llvm.mips.pckev.b` intrinsic; known as `__builtin_msa_pckev_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckev.b")
    __vector(i8[16]) pckev_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.pckev.h` intrinsic; known as `__builtin_msa_pckev_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckev.h")
    __vector(i16[8]) pckev_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.pckev.w` intrinsic; known as `__builtin_msa_pckev_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckev.w")
    __vector(i32[4]) pckev_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.pckev.d` intrinsic; known as `__builtin_msa_pckev_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckev.d")
    __vector(i64[2]) pckev_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.pckod.b` intrinsic; known as `__builtin_msa_pckod_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckod.b")
    __vector(i8[16]) pckod_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.pckod.h` intrinsic; known as `__builtin_msa_pckod_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckod.h")
    __vector(i16[8]) pckod_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.pckod.w` intrinsic; known as `__builtin_msa_pckod_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckod.w")
    __vector(i32[4]) pckod_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.pckod.d` intrinsic; known as `__builtin_msa_pckod_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pckod.d")
    __vector(i64[2]) pckod_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.pcnt.b` intrinsic; known as `__builtin_msa_pcnt_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pcnt.b")
    __vector(i8[16]) pcnt_b(__vector(i8[16]));
    /// The `llvm.mips.pcnt.h` intrinsic; known as `__builtin_msa_pcnt_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pcnt.h")
    __vector(i16[8]) pcnt_h(__vector(i16[8]));
    /// The `llvm.mips.pcnt.w` intrinsic; known as `__builtin_msa_pcnt_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pcnt.w")
    __vector(i32[4]) pcnt_w(__vector(i32[4]));
    /// The `llvm.mips.pcnt.d` intrinsic; known as `__builtin_msa_pcnt_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.pcnt.d")
    __vector(i64[2]) pcnt_d(__vector(i64[2]));
    /// The `llvm.mips.sat.s.b` intrinsic; known as `__builtin_msa_sat_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.s.b")
    __vector(i8[16]) sat_s_b(__vector(i8[16]), i32);
    /// The `llvm.mips.sat.s.h` intrinsic; known as `__builtin_msa_sat_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.s.h")
    __vector(i16[8]) sat_s_h(__vector(i16[8]), i32);
    /// The `llvm.mips.sat.s.w` intrinsic; known as `__builtin_msa_sat_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.s.w")
    __vector(i32[4]) sat_s_w(__vector(i32[4]), i32);
    /// The `llvm.mips.sat.s.d` intrinsic; known as `__builtin_msa_sat_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.s.d")
    __vector(i64[2]) sat_s_d(__vector(i64[2]), i32);
    /// The `llvm.mips.sat.u.b` intrinsic; known as `__builtin_msa_sat_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.u.b")
    __vector(i8[16]) sat_u_b(__vector(i8[16]), i32);
    /// The `llvm.mips.sat.u.h` intrinsic; known as `__builtin_msa_sat_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.u.h")
    __vector(i16[8]) sat_u_h(__vector(i16[8]), i32);
    /// The `llvm.mips.sat.u.w` intrinsic; known as `__builtin_msa_sat_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.u.w")
    __vector(i32[4]) sat_u_w(__vector(i32[4]), i32);
    /// The `llvm.mips.sat.u.d` intrinsic; known as `__builtin_msa_sat_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sat.u.d")
    __vector(i64[2]) sat_u_d(__vector(i64[2]), i32);
    /// The `llvm.mips.shf.b` intrinsic; known as `__builtin_msa_shf_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shf.b")
    __vector(i8[16]) shf_b(__vector(i8[16]), i32);
    /// The `llvm.mips.shf.h` intrinsic; known as `__builtin_msa_shf_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shf.h")
    __vector(i16[8]) shf_h(__vector(i16[8]), i32);
    /// The `llvm.mips.shf.w` intrinsic; known as `__builtin_msa_shf_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.shf.w")
    __vector(i32[4]) shf_w(__vector(i32[4]), i32);
    /// The `llvm.mips.sld.b` intrinsic; known as `__builtin_msa_sld_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sld.b")
    __vector(i8[16]) sld_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.sld.h` intrinsic; known as `__builtin_msa_sld_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sld.h")
    __vector(i16[8]) sld_h(__vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.mips.sld.w` intrinsic; known as `__builtin_msa_sld_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sld.w")
    __vector(i32[4]) sld_w(__vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.mips.sld.d` intrinsic; known as `__builtin_msa_sld_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sld.d")
    __vector(i64[2]) sld_d(__vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.mips.sldi.b` intrinsic; known as `__builtin_msa_sldi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sldi.b")
    __vector(i8[16]) sldi_b(__vector(i8[16]), __vector(i8[16]), i32);
    /// The `llvm.mips.sldi.h` intrinsic; known as `__builtin_msa_sldi_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sldi.h")
    __vector(i16[8]) sldi_h(__vector(i16[8]), __vector(i16[8]), i32);
    /// The `llvm.mips.sldi.w` intrinsic; known as `__builtin_msa_sldi_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sldi.w")
    __vector(i32[4]) sldi_w(__vector(i32[4]), __vector(i32[4]), i32);
    /// The `llvm.mips.sldi.d` intrinsic; known as `__builtin_msa_sldi_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sldi.d")
    __vector(i64[2]) sldi_d(__vector(i64[2]), __vector(i64[2]), i32);
    /// The `llvm.mips.sll.b` intrinsic; known as `__builtin_msa_sll_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sll.b")
    __vector(i8[16]) sll_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.sll.h` intrinsic; known as `__builtin_msa_sll_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sll.h")
    __vector(i16[8]) sll_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.sll.w` intrinsic; known as `__builtin_msa_sll_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sll.w")
    __vector(i32[4]) sll_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.sll.d` intrinsic; known as `__builtin_msa_sll_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sll.d")
    __vector(i64[2]) sll_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.slli.b` intrinsic; known as `__builtin_msa_slli_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.slli.b")
    __vector(i8[16]) slli_b(__vector(i8[16]), i32);
    /// The `llvm.mips.slli.h` intrinsic; known as `__builtin_msa_slli_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.slli.h")
    __vector(i16[8]) slli_h(__vector(i16[8]), i32);
    /// The `llvm.mips.slli.w` intrinsic; known as `__builtin_msa_slli_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.slli.w")
    __vector(i32[4]) slli_w(__vector(i32[4]), i32);
    /// The `llvm.mips.slli.d` intrinsic; known as `__builtin_msa_slli_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.slli.d")
    __vector(i64[2]) slli_d(__vector(i64[2]), i32);
    /// The `llvm.mips.splat.b` intrinsic; known as `__builtin_msa_splat_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splat.b")
    __vector(i8[16]) splat_b(__vector(i8[16]), i32);
    /// The `llvm.mips.splat.h` intrinsic; known as `__builtin_msa_splat_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splat.h")
    __vector(i16[8]) splat_h(__vector(i16[8]), i32);
    /// The `llvm.mips.splat.w` intrinsic; known as `__builtin_msa_splat_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splat.w")
    __vector(i32[4]) splat_w(__vector(i32[4]), i32);
    /// The `llvm.mips.splat.d` intrinsic; known as `__builtin_msa_splat_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splat.d")
    __vector(i64[2]) splat_d(__vector(i64[2]), i32);
    /// The `llvm.mips.splati.b` intrinsic; known as `__builtin_msa_splati_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splati.b")
    __vector(i8[16]) splati_b(__vector(i8[16]), i32);
    /// The `llvm.mips.splati.h` intrinsic; known as `__builtin_msa_splati_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splati.h")
    __vector(i16[8]) splati_h(__vector(i16[8]), i32);
    /// The `llvm.mips.splati.w` intrinsic; known as `__builtin_msa_splati_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splati.w")
    __vector(i32[4]) splati_w(__vector(i32[4]), i32);
    /// The `llvm.mips.splati.d` intrinsic; known as `__builtin_msa_splati_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.splati.d")
    __vector(i64[2]) splati_d(__vector(i64[2]), i32);
    /// The `llvm.mips.sra.b` intrinsic; known as `__builtin_msa_sra_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sra.b")
    __vector(i8[16]) sra_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.sra.h` intrinsic; known as `__builtin_msa_sra_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sra.h")
    __vector(i16[8]) sra_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.sra.w` intrinsic; known as `__builtin_msa_sra_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sra.w")
    __vector(i32[4]) sra_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.sra.d` intrinsic; known as `__builtin_msa_sra_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.sra.d")
    __vector(i64[2]) sra_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.srai.b` intrinsic; known as `__builtin_msa_srai_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srai.b")
    __vector(i8[16]) srai_b(__vector(i8[16]), i32);
    /// The `llvm.mips.srai.h` intrinsic; known as `__builtin_msa_srai_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srai.h")
    __vector(i16[8]) srai_h(__vector(i16[8]), i32);
    /// The `llvm.mips.srai.w` intrinsic; known as `__builtin_msa_srai_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srai.w")
    __vector(i32[4]) srai_w(__vector(i32[4]), i32);
    /// The `llvm.mips.srai.d` intrinsic; known as `__builtin_msa_srai_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srai.d")
    __vector(i64[2]) srai_d(__vector(i64[2]), i32);
    /// The `llvm.mips.srar.b` intrinsic; known as `__builtin_msa_srar_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srar.b")
    __vector(i8[16]) srar_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.srar.h` intrinsic; known as `__builtin_msa_srar_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srar.h")
    __vector(i16[8]) srar_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.srar.w` intrinsic; known as `__builtin_msa_srar_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srar.w")
    __vector(i32[4]) srar_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.srar.d` intrinsic; known as `__builtin_msa_srar_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srar.d")
    __vector(i64[2]) srar_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.srari.b` intrinsic; known as `__builtin_msa_srari_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srari.b")
    __vector(i8[16]) srari_b(__vector(i8[16]), i32);
    /// The `llvm.mips.srari.h` intrinsic; known as `__builtin_msa_srari_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srari.h")
    __vector(i16[8]) srari_h(__vector(i16[8]), i32);
    /// The `llvm.mips.srari.w` intrinsic; known as `__builtin_msa_srari_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srari.w")
    __vector(i32[4]) srari_w(__vector(i32[4]), i32);
    /// The `llvm.mips.srari.d` intrinsic; known as `__builtin_msa_srari_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srari.d")
    __vector(i64[2]) srari_d(__vector(i64[2]), i32);
    /// The `llvm.mips.srl.b` intrinsic; known as `__builtin_msa_srl_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srl.b")
    __vector(i8[16]) srl_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.srl.h` intrinsic; known as `__builtin_msa_srl_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srl.h")
    __vector(i16[8]) srl_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.srl.w` intrinsic; known as `__builtin_msa_srl_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srl.w")
    __vector(i32[4]) srl_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.srl.d` intrinsic; known as `__builtin_msa_srl_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srl.d")
    __vector(i64[2]) srl_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.srli.b` intrinsic; known as `__builtin_msa_srli_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srli.b")
    __vector(i8[16]) srli_b(__vector(i8[16]), i32);
    /// The `llvm.mips.srli.h` intrinsic; known as `__builtin_msa_srli_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srli.h")
    __vector(i16[8]) srli_h(__vector(i16[8]), i32);
    /// The `llvm.mips.srli.w` intrinsic; known as `__builtin_msa_srli_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srli.w")
    __vector(i32[4]) srli_w(__vector(i32[4]), i32);
    /// The `llvm.mips.srli.d` intrinsic; known as `__builtin_msa_srli_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srli.d")
    __vector(i64[2]) srli_d(__vector(i64[2]), i32);
    /// The `llvm.mips.srlr.b` intrinsic; known as `__builtin_msa_srlr_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlr.b")
    __vector(i8[16]) srlr_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.srlr.h` intrinsic; known as `__builtin_msa_srlr_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlr.h")
    __vector(i16[8]) srlr_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.srlr.w` intrinsic; known as `__builtin_msa_srlr_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlr.w")
    __vector(i32[4]) srlr_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.srlr.d` intrinsic; known as `__builtin_msa_srlr_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlr.d")
    __vector(i64[2]) srlr_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.srlri.b` intrinsic; known as `__builtin_msa_srlri_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlri.b")
    __vector(i8[16]) srlri_b(__vector(i8[16]), i32);
    /// The `llvm.mips.srlri.h` intrinsic; known as `__builtin_msa_srlri_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlri.h")
    __vector(i16[8]) srlri_h(__vector(i16[8]), i32);
    /// The `llvm.mips.srlri.w` intrinsic; known as `__builtin_msa_srlri_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlri.w")
    __vector(i32[4]) srlri_w(__vector(i32[4]), i32);
    /// The `llvm.mips.srlri.d` intrinsic; known as `__builtin_msa_srlri_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.srlri.d")
    __vector(i64[2]) srlri_d(__vector(i64[2]), i32);
    /// The `llvm.mips.st.b` intrinsic; known as `__builtin_msa_st_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.st.b")
    void st_b(__vector(i8[16]), i8*, i32);
    /// The `llvm.mips.st.h` intrinsic; known as `__builtin_msa_st_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.st.h")
    void st_h(__vector(i16[8]), i8*, i32);
    /// The `llvm.mips.st.w` intrinsic; known as `__builtin_msa_st_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.st.w")
    void st_w(__vector(i32[4]), i8*, i32);
    /// The `llvm.mips.st.d` intrinsic; known as `__builtin_msa_st_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.st.d")
    void st_d(__vector(i64[2]), i8*, i32);
    /// The `llvm.mips.subs.s.b` intrinsic; known as `__builtin_msa_subs_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.s.b")
    __vector(i8[16]) subs_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.subs.s.h` intrinsic; known as `__builtin_msa_subs_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.s.h")
    __vector(i16[8]) subs_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.subs.s.w` intrinsic; known as `__builtin_msa_subs_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.s.w")
    __vector(i32[4]) subs_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.subs.s.d` intrinsic; known as `__builtin_msa_subs_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.s.d")
    __vector(i64[2]) subs_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.subs.u.b` intrinsic; known as `__builtin_msa_subs_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.u.b")
    __vector(i8[16]) subs_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.subs.u.h` intrinsic; known as `__builtin_msa_subs_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.u.h")
    __vector(i16[8]) subs_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.subs.u.w` intrinsic; known as `__builtin_msa_subs_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.u.w")
    __vector(i32[4]) subs_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.subs.u.d` intrinsic; known as `__builtin_msa_subs_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subs.u.d")
    __vector(i64[2]) subs_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.subsus.u.b` intrinsic; known as `__builtin_msa_subsus_u_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsus.u.b")
    __vector(i8[16]) subsus_u_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.subsus.u.h` intrinsic; known as `__builtin_msa_subsus_u_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsus.u.h")
    __vector(i16[8]) subsus_u_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.subsus.u.w` intrinsic; known as `__builtin_msa_subsus_u_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsus.u.w")
    __vector(i32[4]) subsus_u_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.subsus.u.d` intrinsic; known as `__builtin_msa_subsus_u_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsus.u.d")
    __vector(i64[2]) subsus_u_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.subsuu.s.b` intrinsic; known as `__builtin_msa_subsuu_s_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsuu.s.b")
    __vector(i8[16]) subsuu_s_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.subsuu.s.h` intrinsic; known as `__builtin_msa_subsuu_s_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsuu.s.h")
    __vector(i16[8]) subsuu_s_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.subsuu.s.w` intrinsic; known as `__builtin_msa_subsuu_s_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsuu.s.w")
    __vector(i32[4]) subsuu_s_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.subsuu.s.d` intrinsic; known as `__builtin_msa_subsuu_s_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subsuu.s.d")
    __vector(i64[2]) subsuu_s_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.subv.b` intrinsic; known as `__builtin_msa_subv_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subv.b")
    __vector(i8[16]) subv_b(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.subv.h` intrinsic; known as `__builtin_msa_subv_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subv.h")
    __vector(i16[8]) subv_h(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.subv.w` intrinsic; known as `__builtin_msa_subv_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subv.w")
    __vector(i32[4]) subv_w(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.subv.d` intrinsic; known as `__builtin_msa_subv_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subv.d")
    __vector(i64[2]) subv_d(__vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.subvi.b` intrinsic; known as `__builtin_msa_subvi_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subvi.b")
    __vector(i8[16]) subvi_b(__vector(i8[16]), i32);
    /// The `llvm.mips.subvi.h` intrinsic; known as `__builtin_msa_subvi_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subvi.h")
    __vector(i16[8]) subvi_h(__vector(i16[8]), i32);
    /// The `llvm.mips.subvi.w` intrinsic; known as `__builtin_msa_subvi_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subvi.w")
    __vector(i32[4]) subvi_w(__vector(i32[4]), i32);
    /// The `llvm.mips.subvi.d` intrinsic; known as `__builtin_msa_subvi_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.subvi.d")
    __vector(i64[2]) subvi_d(__vector(i64[2]), i32);
    /// The `llvm.mips.vshf.b` intrinsic; known as `__builtin_msa_vshf_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.vshf.b")
    __vector(i8[16]) vshf_b(__vector(i8[16]), __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.vshf.h` intrinsic; known as `__builtin_msa_vshf_h` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.vshf.h")
    __vector(i16[8]) vshf_h(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.mips.vshf.w` intrinsic; known as `__builtin_msa_vshf_w` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.vshf.w")
    __vector(i32[4]) vshf_w(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.mips.vshf.d` intrinsic; known as `__builtin_msa_vshf_d` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.vshf.d")
    __vector(i64[2]) vshf_d(__vector(i64[2]), __vector(i64[2]), __vector(i64[2]));
    /// The `llvm.mips.xor.v` intrinsic; known as `__builtin_msa_xor_v` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.xor.v")
    __vector(i8[16]) xor_v(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.mips.xori.b` intrinsic; known as `__builtin_msa_xori_b` in GCC.
    pragma(LDC_intrinsic, "llvm.mips.xori.b")
    __vector(i8[16]) xori_b(__vector(i8[16]), i32);
}

/// LLVM intrinsics for the nvvm architecture.
version (nvvm) {
    /// The `llvm.nvvm.clz.i` intrinsic; known as `__nvvm_clz_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.clz.i")
    i32 clz_i(i32);
    /// The `llvm.nvvm.clz.ll` intrinsic; known as `__nvvm_clz_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.clz.ll")
    i32 clz_ll(i64);
    /// The `llvm.nvvm.popc.i` intrinsic; known as `__nvvm_popc_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.popc.i")
    i32 popc_i(i32);
    /// The `llvm.nvvm.popc.ll` intrinsic; known as `__nvvm_popc_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.popc.ll")
    i32 popc_ll(i64);
    /// The `llvm.nvvm.prmt` intrinsic; known as `__nvvm_prmt` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.prmt")
    i32 prmt(i32, i32, i32);
    /// The `llvm.nvvm.min.i` intrinsic; known as `__nvvm_min_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.min.i")
    i32 min_i(i32, i32);
    /// The `llvm.nvvm.min.ui` intrinsic; known as `__nvvm_min_ui` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.min.ui")
    i32 min_ui(i32, i32);
    /// The `llvm.nvvm.min.ll` intrinsic; known as `__nvvm_min_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.min.ll")
    i64 min_ll(i64, i64);
    /// The `llvm.nvvm.min.ull` intrinsic; known as `__nvvm_min_ull` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.min.ull")
    i64 min_ull(i64, i64);
    /// The `llvm.nvvm.max.i` intrinsic; known as `__nvvm_max_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.max.i")
    i32 max_i(i32, i32);
    /// The `llvm.nvvm.max.ui` intrinsic; known as `__nvvm_max_ui` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.max.ui")
    i32 max_ui(i32, i32);
    /// The `llvm.nvvm.max.ll` intrinsic; known as `__nvvm_max_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.max.ll")
    i64 max_ll(i64, i64);
    /// The `llvm.nvvm.max.ull` intrinsic; known as `__nvvm_max_ull` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.max.ull")
    i64 max_ull(i64, i64);
    /// The `llvm.nvvm.fmin.f` intrinsic; known as `__nvvm_fmin_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fmin.f")
    f32 fmin_f(f32, f32);
    /// The `llvm.nvvm.fmin.ftz.f` intrinsic; known as `__nvvm_fmin_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fmin.ftz.f")
    f32 fmin_ftz_f(f32, f32);
    /// The `llvm.nvvm.fmax.f` intrinsic; known as `__nvvm_fmax_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fmax.f")
    f32 fmax_f(f32, f32);
    /// The `llvm.nvvm.fmax.ftz.f` intrinsic; known as `__nvvm_fmax_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fmax.ftz.f")
    f32 fmax_ftz_f(f32, f32);
    /// The `llvm.nvvm.fmin.d` intrinsic; known as `__nvvm_fmin_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fmin.d")
    f64 fmin_d(f64, f64);
    /// The `llvm.nvvm.fmax.d` intrinsic; known as `__nvvm_fmax_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fmax.d")
    f64 fmax_d(f64, f64);
    /// The `llvm.nvvm.mulhi.i` intrinsic; known as `__nvvm_mulhi_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mulhi.i")
    i32 mulhi_i(i32, i32);
    /// The `llvm.nvvm.mulhi.ui` intrinsic; known as `__nvvm_mulhi_ui` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mulhi.ui")
    i32 mulhi_ui(i32, i32);
    /// The `llvm.nvvm.mulhi.ll` intrinsic; known as `__nvvm_mulhi_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mulhi.ll")
    i64 mulhi_ll(i64, i64);
    /// The `llvm.nvvm.mulhi.ull` intrinsic; known as `__nvvm_mulhi_ull` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mulhi.ull")
    i64 mulhi_ull(i64, i64);
    /// The `llvm.nvvm.mul.rn.ftz.f` intrinsic; known as `__nvvm_mul_rn_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rn.ftz.f")
    f32 mul_rn_ftz_f(f32, f32);
    /// The `llvm.nvvm.mul.rn.f` intrinsic; known as `__nvvm_mul_rn_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rn.f")
    f32 mul_rn_f(f32, f32);
    /// The `llvm.nvvm.mul.rz.ftz.f` intrinsic; known as `__nvvm_mul_rz_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rz.ftz.f")
    f32 mul_rz_ftz_f(f32, f32);
    /// The `llvm.nvvm.mul.rz.f` intrinsic; known as `__nvvm_mul_rz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rz.f")
    f32 mul_rz_f(f32, f32);
    /// The `llvm.nvvm.mul.rm.ftz.f` intrinsic; known as `__nvvm_mul_rm_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rm.ftz.f")
    f32 mul_rm_ftz_f(f32, f32);
    /// The `llvm.nvvm.mul.rm.f` intrinsic; known as `__nvvm_mul_rm_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rm.f")
    f32 mul_rm_f(f32, f32);
    /// The `llvm.nvvm.mul.rp.ftz.f` intrinsic; known as `__nvvm_mul_rp_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rp.ftz.f")
    f32 mul_rp_ftz_f(f32, f32);
    /// The `llvm.nvvm.mul.rp.f` intrinsic; known as `__nvvm_mul_rp_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rp.f")
    f32 mul_rp_f(f32, f32);
    /// The `llvm.nvvm.mul.rn.d` intrinsic; known as `__nvvm_mul_rn_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rn.d")
    f64 mul_rn_d(f64, f64);
    /// The `llvm.nvvm.mul.rz.d` intrinsic; known as `__nvvm_mul_rz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rz.d")
    f64 mul_rz_d(f64, f64);
    /// The `llvm.nvvm.mul.rm.d` intrinsic; known as `__nvvm_mul_rm_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rm.d")
    f64 mul_rm_d(f64, f64);
    /// The `llvm.nvvm.mul.rp.d` intrinsic; known as `__nvvm_mul_rp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul.rp.d")
    f64 mul_rp_d(f64, f64);
    /// The `llvm.nvvm.mul24.i` intrinsic; known as `__nvvm_mul24_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul24.i")
    i32 mul24_i(i32, i32);
    /// The `llvm.nvvm.mul24.ui` intrinsic; known as `__nvvm_mul24_ui` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.mul24.ui")
    i32 mul24_ui(i32, i32);
    /// The `llvm.nvvm.div.approx.ftz.f` intrinsic; known as `__nvvm_div_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.approx.ftz.f")
    f32 div_approx_ftz_f(f32, f32);
    /// The `llvm.nvvm.div.approx.f` intrinsic; known as `__nvvm_div_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.approx.f")
    f32 div_approx_f(f32, f32);
    /// The `llvm.nvvm.div.rn.ftz.f` intrinsic; known as `__nvvm_div_rn_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rn.ftz.f")
    f32 div_rn_ftz_f(f32, f32);
    /// The `llvm.nvvm.div.rn.f` intrinsic; known as `__nvvm_div_rn_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rn.f")
    f32 div_rn_f(f32, f32);
    /// The `llvm.nvvm.div.rz.ftz.f` intrinsic; known as `__nvvm_div_rz_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rz.ftz.f")
    f32 div_rz_ftz_f(f32, f32);
    /// The `llvm.nvvm.div.rz.f` intrinsic; known as `__nvvm_div_rz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rz.f")
    f32 div_rz_f(f32, f32);
    /// The `llvm.nvvm.div.rm.ftz.f` intrinsic; known as `__nvvm_div_rm_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rm.ftz.f")
    f32 div_rm_ftz_f(f32, f32);
    /// The `llvm.nvvm.div.rm.f` intrinsic; known as `__nvvm_div_rm_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rm.f")
    f32 div_rm_f(f32, f32);
    /// The `llvm.nvvm.div.rp.ftz.f` intrinsic; known as `__nvvm_div_rp_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rp.ftz.f")
    f32 div_rp_ftz_f(f32, f32);
    /// The `llvm.nvvm.div.rp.f` intrinsic; known as `__nvvm_div_rp_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rp.f")
    f32 div_rp_f(f32, f32);
    /// The `llvm.nvvm.div.rn.d` intrinsic; known as `__nvvm_div_rn_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rn.d")
    f64 div_rn_d(f64, f64);
    /// The `llvm.nvvm.div.rz.d` intrinsic; known as `__nvvm_div_rz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rz.d")
    f64 div_rz_d(f64, f64);
    /// The `llvm.nvvm.div.rm.d` intrinsic; known as `__nvvm_div_rm_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rm.d")
    f64 div_rm_d(f64, f64);
    /// The `llvm.nvvm.div.rp.d` intrinsic; known as `__nvvm_div_rp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.div.rp.d")
    f64 div_rp_d(f64, f64);
    /// The `llvm.nvvm.brev32` intrinsic; known as `__nvvm_brev32` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.brev32")
    i32 brev32(i32);
    /// The `llvm.nvvm.brev64` intrinsic; known as `__nvvm_brev64` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.brev64")
    i64 brev64(i64);
    /// The `llvm.nvvm.sad.i` intrinsic; known as `__nvvm_sad_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sad.i")
    i32 sad_i(i32, i32, i32);
    /// The `llvm.nvvm.sad.ui` intrinsic; known as `__nvvm_sad_ui` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sad.ui")
    i32 sad_ui(i32, i32, i32);
    /// The `llvm.nvvm.floor.ftz.f` intrinsic; known as `__nvvm_floor_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.floor.ftz.f")
    f32 floor_ftz_f(f32);
    /// The `llvm.nvvm.floor.f` intrinsic; known as `__nvvm_floor_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.floor.f")
    f32 floor_f(f32);
    /// The `llvm.nvvm.floor.d` intrinsic; known as `__nvvm_floor_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.floor.d")
    f64 floor_d(f64);
    /// The `llvm.nvvm.ceil.ftz.f` intrinsic; known as `__nvvm_ceil_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ceil.ftz.f")
    f32 ceil_ftz_f(f32);
    /// The `llvm.nvvm.ceil.f` intrinsic; known as `__nvvm_ceil_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ceil.f")
    f32 ceil_f(f32);
    /// The `llvm.nvvm.ceil.d` intrinsic; known as `__nvvm_ceil_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ceil.d")
    f64 ceil_d(f64);
    /// The `llvm.nvvm.abs.i` intrinsic; known as `__nvvm_abs_i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.abs.i")
    i32 abs_i(i32);
    /// The `llvm.nvvm.abs.ll` intrinsic; known as `__nvvm_abs_ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.abs.ll")
    i64 abs_ll(i64);
    /// The `llvm.nvvm.fabs.ftz.f` intrinsic; known as `__nvvm_fabs_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fabs.ftz.f")
    f32 fabs_ftz_f(f32);
    /// The `llvm.nvvm.fabs.f` intrinsic; known as `__nvvm_fabs_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fabs.f")
    f32 fabs_f(f32);
    /// The `llvm.nvvm.fabs.d` intrinsic; known as `__nvvm_fabs_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fabs.d")
    f64 fabs_d(f64);
    /// The `llvm.nvvm.round.ftz.f` intrinsic; known as `__nvvm_round_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.round.ftz.f")
    f32 round_ftz_f(f32);
    /// The `llvm.nvvm.round.f` intrinsic; known as `__nvvm_round_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.round.f")
    f32 round_f(f32);
    /// The `llvm.nvvm.round.d` intrinsic; known as `__nvvm_round_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.round.d")
    f64 round_d(f64);
    /// The `llvm.nvvm.trunc.ftz.f` intrinsic; known as `__nvvm_trunc_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.trunc.ftz.f")
    f32 trunc_ftz_f(f32);
    /// The `llvm.nvvm.trunc.f` intrinsic; known as `__nvvm_trunc_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.trunc.f")
    f32 trunc_f(f32);
    /// The `llvm.nvvm.trunc.d` intrinsic; known as `__nvvm_trunc_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.trunc.d")
    f64 trunc_d(f64);
    /// The `llvm.nvvm.saturate.ftz.f` intrinsic; known as `__nvvm_saturate_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.saturate.ftz.f")
    f32 saturate_ftz_f(f32);
    /// The `llvm.nvvm.saturate.f` intrinsic; known as `__nvvm_saturate_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.saturate.f")
    f32 saturate_f(f32);
    /// The `llvm.nvvm.saturate.d` intrinsic; known as `__nvvm_saturate_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.saturate.d")
    f64 saturate_d(f64);
    /// The `llvm.nvvm.ex2.approx.ftz.f` intrinsic; known as `__nvvm_ex2_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ex2.approx.ftz.f")
    f32 ex2_approx_ftz_f(f32);
    /// The `llvm.nvvm.ex2.approx.f` intrinsic; known as `__nvvm_ex2_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ex2.approx.f")
    f32 ex2_approx_f(f32);
    /// The `llvm.nvvm.ex2.approx.d` intrinsic; known as `__nvvm_ex2_approx_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ex2.approx.d")
    f64 ex2_approx_d(f64);
    /// The `llvm.nvvm.lg2.approx.ftz.f` intrinsic; known as `__nvvm_lg2_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.lg2.approx.ftz.f")
    f32 lg2_approx_ftz_f(f32);
    /// The `llvm.nvvm.lg2.approx.f` intrinsic; known as `__nvvm_lg2_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.lg2.approx.f")
    f32 lg2_approx_f(f32);
    /// The `llvm.nvvm.lg2.approx.d` intrinsic; known as `__nvvm_lg2_approx_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.lg2.approx.d")
    f64 lg2_approx_d(f64);
    /// The `llvm.nvvm.sin.approx.ftz.f` intrinsic; known as `__nvvm_sin_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sin.approx.ftz.f")
    f32 sin_approx_ftz_f(f32);
    /// The `llvm.nvvm.sin.approx.f` intrinsic; known as `__nvvm_sin_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sin.approx.f")
    f32 sin_approx_f(f32);
    /// The `llvm.nvvm.cos.approx.ftz.f` intrinsic; known as `__nvvm_cos_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.cos.approx.ftz.f")
    f32 cos_approx_ftz_f(f32);
    /// The `llvm.nvvm.cos.approx.f` intrinsic; known as `__nvvm_cos_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.cos.approx.f")
    f32 cos_approx_f(f32);
    /// The `llvm.nvvm.fma.rn.ftz.f` intrinsic; known as `__nvvm_fma_rn_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rn.ftz.f")
    f32 fma_rn_ftz_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rn.f` intrinsic; known as `__nvvm_fma_rn_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rn.f")
    f32 fma_rn_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rz.ftz.f` intrinsic; known as `__nvvm_fma_rz_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rz.ftz.f")
    f32 fma_rz_ftz_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rz.f` intrinsic; known as `__nvvm_fma_rz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rz.f")
    f32 fma_rz_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rm.ftz.f` intrinsic; known as `__nvvm_fma_rm_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rm.ftz.f")
    f32 fma_rm_ftz_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rm.f` intrinsic; known as `__nvvm_fma_rm_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rm.f")
    f32 fma_rm_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rp.ftz.f` intrinsic; known as `__nvvm_fma_rp_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rp.ftz.f")
    f32 fma_rp_ftz_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rp.f` intrinsic; known as `__nvvm_fma_rp_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rp.f")
    f32 fma_rp_f(f32, f32, f32);
    /// The `llvm.nvvm.fma.rn.d` intrinsic; known as `__nvvm_fma_rn_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rn.d")
    f64 fma_rn_d(f64, f64, f64);
    /// The `llvm.nvvm.fma.rz.d` intrinsic; known as `__nvvm_fma_rz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rz.d")
    f64 fma_rz_d(f64, f64, f64);
    /// The `llvm.nvvm.fma.rm.d` intrinsic; known as `__nvvm_fma_rm_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rm.d")
    f64 fma_rm_d(f64, f64, f64);
    /// The `llvm.nvvm.fma.rp.d` intrinsic; known as `__nvvm_fma_rp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.fma.rp.d")
    f64 fma_rp_d(f64, f64, f64);
    /// The `llvm.nvvm.rcp.rn.ftz.f` intrinsic; known as `__nvvm_rcp_rn_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rn.ftz.f")
    f32 rcp_rn_ftz_f(f32);
    /// The `llvm.nvvm.rcp.rn.f` intrinsic; known as `__nvvm_rcp_rn_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rn.f")
    f32 rcp_rn_f(f32);
    /// The `llvm.nvvm.rcp.rz.ftz.f` intrinsic; known as `__nvvm_rcp_rz_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rz.ftz.f")
    f32 rcp_rz_ftz_f(f32);
    /// The `llvm.nvvm.rcp.rz.f` intrinsic; known as `__nvvm_rcp_rz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rz.f")
    f32 rcp_rz_f(f32);
    /// The `llvm.nvvm.rcp.rm.ftz.f` intrinsic; known as `__nvvm_rcp_rm_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rm.ftz.f")
    f32 rcp_rm_ftz_f(f32);
    /// The `llvm.nvvm.rcp.rm.f` intrinsic; known as `__nvvm_rcp_rm_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rm.f")
    f32 rcp_rm_f(f32);
    /// The `llvm.nvvm.rcp.rp.ftz.f` intrinsic; known as `__nvvm_rcp_rp_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rp.ftz.f")
    f32 rcp_rp_ftz_f(f32);
    /// The `llvm.nvvm.rcp.rp.f` intrinsic; known as `__nvvm_rcp_rp_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rp.f")
    f32 rcp_rp_f(f32);
    /// The `llvm.nvvm.rcp.rn.d` intrinsic; known as `__nvvm_rcp_rn_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rn.d")
    f64 rcp_rn_d(f64);
    /// The `llvm.nvvm.rcp.rz.d` intrinsic; known as `__nvvm_rcp_rz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rz.d")
    f64 rcp_rz_d(f64);
    /// The `llvm.nvvm.rcp.rm.d` intrinsic; known as `__nvvm_rcp_rm_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rm.d")
    f64 rcp_rm_d(f64);
    /// The `llvm.nvvm.rcp.rp.d` intrinsic; known as `__nvvm_rcp_rp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.rp.d")
    f64 rcp_rp_d(f64);
    /// The `llvm.nvvm.rcp.approx.ftz.d` intrinsic; known as `__nvvm_rcp_approx_ftz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rcp.approx.ftz.d")
    f64 rcp_approx_ftz_d(f64);
    /// The `llvm.nvvm.sqrt.f` intrinsic; known as `__nvvm_sqrt_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.f")
    f32 sqrt_f(f32);
    /// The `llvm.nvvm.sqrt.rn.ftz.f` intrinsic; known as `__nvvm_sqrt_rn_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rn.ftz.f")
    f32 sqrt_rn_ftz_f(f32);
    /// The `llvm.nvvm.sqrt.rn.f` intrinsic; known as `__nvvm_sqrt_rn_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rn.f")
    f32 sqrt_rn_f(f32);
    /// The `llvm.nvvm.sqrt.rz.ftz.f` intrinsic; known as `__nvvm_sqrt_rz_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rz.ftz.f")
    f32 sqrt_rz_ftz_f(f32);
    /// The `llvm.nvvm.sqrt.rz.f` intrinsic; known as `__nvvm_sqrt_rz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rz.f")
    f32 sqrt_rz_f(f32);
    /// The `llvm.nvvm.sqrt.rm.ftz.f` intrinsic; known as `__nvvm_sqrt_rm_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rm.ftz.f")
    f32 sqrt_rm_ftz_f(f32);
    /// The `llvm.nvvm.sqrt.rm.f` intrinsic; known as `__nvvm_sqrt_rm_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rm.f")
    f32 sqrt_rm_f(f32);
    /// The `llvm.nvvm.sqrt.rp.ftz.f` intrinsic; known as `__nvvm_sqrt_rp_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rp.ftz.f")
    f32 sqrt_rp_ftz_f(f32);
    /// The `llvm.nvvm.sqrt.rp.f` intrinsic; known as `__nvvm_sqrt_rp_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rp.f")
    f32 sqrt_rp_f(f32);
    /// The `llvm.nvvm.sqrt.approx.ftz.f` intrinsic; known as `__nvvm_sqrt_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.approx.ftz.f")
    f32 sqrt_approx_ftz_f(f32);
    /// The `llvm.nvvm.sqrt.approx.f` intrinsic; known as `__nvvm_sqrt_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.approx.f")
    f32 sqrt_approx_f(f32);
    /// The `llvm.nvvm.sqrt.rn.d` intrinsic; known as `__nvvm_sqrt_rn_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rn.d")
    f64 sqrt_rn_d(f64);
    /// The `llvm.nvvm.sqrt.rz.d` intrinsic; known as `__nvvm_sqrt_rz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rz.d")
    f64 sqrt_rz_d(f64);
    /// The `llvm.nvvm.sqrt.rm.d` intrinsic; known as `__nvvm_sqrt_rm_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rm.d")
    f64 sqrt_rm_d(f64);
    /// The `llvm.nvvm.sqrt.rp.d` intrinsic; known as `__nvvm_sqrt_rp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sqrt.rp.d")
    f64 sqrt_rp_d(f64);
    /// The `llvm.nvvm.rsqrt.approx.ftz.f` intrinsic; known as `__nvvm_rsqrt_approx_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rsqrt.approx.ftz.f")
    f32 rsqrt_approx_ftz_f(f32);
    /// The `llvm.nvvm.rsqrt.approx.f` intrinsic; known as `__nvvm_rsqrt_approx_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rsqrt.approx.f")
    f32 rsqrt_approx_f(f32);
    /// The `llvm.nvvm.rsqrt.approx.d` intrinsic; known as `__nvvm_rsqrt_approx_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rsqrt.approx.d")
    f64 rsqrt_approx_d(f64);
    /// The `llvm.nvvm.add.rn.ftz.f` intrinsic; known as `__nvvm_add_rn_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rn.ftz.f")
    f32 add_rn_ftz_f(f32, f32);
    /// The `llvm.nvvm.add.rn.f` intrinsic; known as `__nvvm_add_rn_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rn.f")
    f32 add_rn_f(f32, f32);
    /// The `llvm.nvvm.add.rz.ftz.f` intrinsic; known as `__nvvm_add_rz_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rz.ftz.f")
    f32 add_rz_ftz_f(f32, f32);
    /// The `llvm.nvvm.add.rz.f` intrinsic; known as `__nvvm_add_rz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rz.f")
    f32 add_rz_f(f32, f32);
    /// The `llvm.nvvm.add.rm.ftz.f` intrinsic; known as `__nvvm_add_rm_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rm.ftz.f")
    f32 add_rm_ftz_f(f32, f32);
    /// The `llvm.nvvm.add.rm.f` intrinsic; known as `__nvvm_add_rm_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rm.f")
    f32 add_rm_f(f32, f32);
    /// The `llvm.nvvm.add.rp.ftz.f` intrinsic; known as `__nvvm_add_rp_ftz_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rp.ftz.f")
    f32 add_rp_ftz_f(f32, f32);
    /// The `llvm.nvvm.add.rp.f` intrinsic; known as `__nvvm_add_rp_f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rp.f")
    f32 add_rp_f(f32, f32);
    /// The `llvm.nvvm.add.rn.d` intrinsic; known as `__nvvm_add_rn_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rn.d")
    f64 add_rn_d(f64, f64);
    /// The `llvm.nvvm.add.rz.d` intrinsic; known as `__nvvm_add_rz_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rz.d")
    f64 add_rz_d(f64, f64);
    /// The `llvm.nvvm.add.rm.d` intrinsic; known as `__nvvm_add_rm_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rm.d")
    f64 add_rm_d(f64, f64);
    /// The `llvm.nvvm.add.rp.d` intrinsic; known as `__nvvm_add_rp_d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.add.rp.d")
    f64 add_rp_d(f64, f64);
    /// The `llvm.nvvm.d2f.rn.ftz` intrinsic; known as `__nvvm_d2f_rn_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rn.ftz")
    f32 d2f_rn_ftz(f64);
    /// The `llvm.nvvm.d2f.rn` intrinsic; known as `__nvvm_d2f_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rn")
    f32 d2f_rn(f64);
    /// The `llvm.nvvm.d2f.rz.ftz` intrinsic; known as `__nvvm_d2f_rz_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rz.ftz")
    f32 d2f_rz_ftz(f64);
    /// The `llvm.nvvm.d2f.rz` intrinsic; known as `__nvvm_d2f_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rz")
    f32 d2f_rz(f64);
    /// The `llvm.nvvm.d2f.rm.ftz` intrinsic; known as `__nvvm_d2f_rm_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rm.ftz")
    f32 d2f_rm_ftz(f64);
    /// The `llvm.nvvm.d2f.rm` intrinsic; known as `__nvvm_d2f_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rm")
    f32 d2f_rm(f64);
    /// The `llvm.nvvm.d2f.rp.ftz` intrinsic; known as `__nvvm_d2f_rp_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rp.ftz")
    f32 d2f_rp_ftz(f64);
    /// The `llvm.nvvm.d2f.rp` intrinsic; known as `__nvvm_d2f_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2f.rp")
    f32 d2f_rp(f64);
    /// The `llvm.nvvm.d2i.rn` intrinsic; known as `__nvvm_d2i_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2i.rn")
    i32 d2i_rn(f64);
    /// The `llvm.nvvm.d2i.rz` intrinsic; known as `__nvvm_d2i_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2i.rz")
    i32 d2i_rz(f64);
    /// The `llvm.nvvm.d2i.rm` intrinsic; known as `__nvvm_d2i_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2i.rm")
    i32 d2i_rm(f64);
    /// The `llvm.nvvm.d2i.rp` intrinsic; known as `__nvvm_d2i_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2i.rp")
    i32 d2i_rp(f64);
    /// The `llvm.nvvm.d2ui.rn` intrinsic; known as `__nvvm_d2ui_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ui.rn")
    i32 d2ui_rn(f64);
    /// The `llvm.nvvm.d2ui.rz` intrinsic; known as `__nvvm_d2ui_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ui.rz")
    i32 d2ui_rz(f64);
    /// The `llvm.nvvm.d2ui.rm` intrinsic; known as `__nvvm_d2ui_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ui.rm")
    i32 d2ui_rm(f64);
    /// The `llvm.nvvm.d2ui.rp` intrinsic; known as `__nvvm_d2ui_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ui.rp")
    i32 d2ui_rp(f64);
    /// The `llvm.nvvm.i2d.rn` intrinsic; known as `__nvvm_i2d_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2d.rn")
    f64 i2d_rn(i32);
    /// The `llvm.nvvm.i2d.rz` intrinsic; known as `__nvvm_i2d_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2d.rz")
    f64 i2d_rz(i32);
    /// The `llvm.nvvm.i2d.rm` intrinsic; known as `__nvvm_i2d_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2d.rm")
    f64 i2d_rm(i32);
    /// The `llvm.nvvm.i2d.rp` intrinsic; known as `__nvvm_i2d_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2d.rp")
    f64 i2d_rp(i32);
    /// The `llvm.nvvm.ui2d.rn` intrinsic; known as `__nvvm_ui2d_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2d.rn")
    f64 ui2d_rn(i32);
    /// The `llvm.nvvm.ui2d.rz` intrinsic; known as `__nvvm_ui2d_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2d.rz")
    f64 ui2d_rz(i32);
    /// The `llvm.nvvm.ui2d.rm` intrinsic; known as `__nvvm_ui2d_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2d.rm")
    f64 ui2d_rm(i32);
    /// The `llvm.nvvm.ui2d.rp` intrinsic; known as `__nvvm_ui2d_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2d.rp")
    f64 ui2d_rp(i32);
    /// The `llvm.nvvm.f2i.rn.ftz` intrinsic; known as `__nvvm_f2i_rn_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rn.ftz")
    i32 f2i_rn_ftz(f32);
    /// The `llvm.nvvm.f2i.rn` intrinsic; known as `__nvvm_f2i_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rn")
    i32 f2i_rn(f32);
    /// The `llvm.nvvm.f2i.rz.ftz` intrinsic; known as `__nvvm_f2i_rz_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rz.ftz")
    i32 f2i_rz_ftz(f32);
    /// The `llvm.nvvm.f2i.rz` intrinsic; known as `__nvvm_f2i_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rz")
    i32 f2i_rz(f32);
    /// The `llvm.nvvm.f2i.rm.ftz` intrinsic; known as `__nvvm_f2i_rm_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rm.ftz")
    i32 f2i_rm_ftz(f32);
    /// The `llvm.nvvm.f2i.rm` intrinsic; known as `__nvvm_f2i_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rm")
    i32 f2i_rm(f32);
    /// The `llvm.nvvm.f2i.rp.ftz` intrinsic; known as `__nvvm_f2i_rp_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rp.ftz")
    i32 f2i_rp_ftz(f32);
    /// The `llvm.nvvm.f2i.rp` intrinsic; known as `__nvvm_f2i_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2i.rp")
    i32 f2i_rp(f32);
    /// The `llvm.nvvm.f2ui.rn.ftz` intrinsic; known as `__nvvm_f2ui_rn_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rn.ftz")
    i32 f2ui_rn_ftz(f32);
    /// The `llvm.nvvm.f2ui.rn` intrinsic; known as `__nvvm_f2ui_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rn")
    i32 f2ui_rn(f32);
    /// The `llvm.nvvm.f2ui.rz.ftz` intrinsic; known as `__nvvm_f2ui_rz_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rz.ftz")
    i32 f2ui_rz_ftz(f32);
    /// The `llvm.nvvm.f2ui.rz` intrinsic; known as `__nvvm_f2ui_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rz")
    i32 f2ui_rz(f32);
    /// The `llvm.nvvm.f2ui.rm.ftz` intrinsic; known as `__nvvm_f2ui_rm_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rm.ftz")
    i32 f2ui_rm_ftz(f32);
    /// The `llvm.nvvm.f2ui.rm` intrinsic; known as `__nvvm_f2ui_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rm")
    i32 f2ui_rm(f32);
    /// The `llvm.nvvm.f2ui.rp.ftz` intrinsic; known as `__nvvm_f2ui_rp_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rp.ftz")
    i32 f2ui_rp_ftz(f32);
    /// The `llvm.nvvm.f2ui.rp` intrinsic; known as `__nvvm_f2ui_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ui.rp")
    i32 f2ui_rp(f32);
    /// The `llvm.nvvm.i2f.rn` intrinsic; known as `__nvvm_i2f_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2f.rn")
    f32 i2f_rn(i32);
    /// The `llvm.nvvm.i2f.rz` intrinsic; known as `__nvvm_i2f_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2f.rz")
    f32 i2f_rz(i32);
    /// The `llvm.nvvm.i2f.rm` intrinsic; known as `__nvvm_i2f_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2f.rm")
    f32 i2f_rm(i32);
    /// The `llvm.nvvm.i2f.rp` intrinsic; known as `__nvvm_i2f_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.i2f.rp")
    f32 i2f_rp(i32);
    /// The `llvm.nvvm.ui2f.rn` intrinsic; known as `__nvvm_ui2f_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2f.rn")
    f32 ui2f_rn(i32);
    /// The `llvm.nvvm.ui2f.rz` intrinsic; known as `__nvvm_ui2f_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2f.rz")
    f32 ui2f_rz(i32);
    /// The `llvm.nvvm.ui2f.rm` intrinsic; known as `__nvvm_ui2f_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2f.rm")
    f32 ui2f_rm(i32);
    /// The `llvm.nvvm.ui2f.rp` intrinsic; known as `__nvvm_ui2f_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ui2f.rp")
    f32 ui2f_rp(i32);
    /// The `llvm.nvvm.lohi.i2d` intrinsic; known as `__nvvm_lohi_i2d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.lohi.i2d")
    f64 lohi_i2d(i32, i32);
    /// The `llvm.nvvm.d2i.lo` intrinsic; known as `__nvvm_d2i_lo` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2i.lo")
    i32 d2i_lo(f64);
    /// The `llvm.nvvm.d2i.hi` intrinsic; known as `__nvvm_d2i_hi` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2i.hi")
    i32 d2i_hi(f64);
    /// The `llvm.nvvm.f2ll.rn.ftz` intrinsic; known as `__nvvm_f2ll_rn_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rn.ftz")
    i64 f2ll_rn_ftz(f32);
    /// The `llvm.nvvm.f2ll.rn` intrinsic; known as `__nvvm_f2ll_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rn")
    i64 f2ll_rn(f32);
    /// The `llvm.nvvm.f2ll.rz.ftz` intrinsic; known as `__nvvm_f2ll_rz_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rz.ftz")
    i64 f2ll_rz_ftz(f32);
    /// The `llvm.nvvm.f2ll.rz` intrinsic; known as `__nvvm_f2ll_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rz")
    i64 f2ll_rz(f32);
    /// The `llvm.nvvm.f2ll.rm.ftz` intrinsic; known as `__nvvm_f2ll_rm_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rm.ftz")
    i64 f2ll_rm_ftz(f32);
    /// The `llvm.nvvm.f2ll.rm` intrinsic; known as `__nvvm_f2ll_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rm")
    i64 f2ll_rm(f32);
    /// The `llvm.nvvm.f2ll.rp.ftz` intrinsic; known as `__nvvm_f2ll_rp_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rp.ftz")
    i64 f2ll_rp_ftz(f32);
    /// The `llvm.nvvm.f2ll.rp` intrinsic; known as `__nvvm_f2ll_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ll.rp")
    i64 f2ll_rp(f32);
    /// The `llvm.nvvm.f2ull.rn.ftz` intrinsic; known as `__nvvm_f2ull_rn_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rn.ftz")
    i64 f2ull_rn_ftz(f32);
    /// The `llvm.nvvm.f2ull.rn` intrinsic; known as `__nvvm_f2ull_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rn")
    i64 f2ull_rn(f32);
    /// The `llvm.nvvm.f2ull.rz.ftz` intrinsic; known as `__nvvm_f2ull_rz_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rz.ftz")
    i64 f2ull_rz_ftz(f32);
    /// The `llvm.nvvm.f2ull.rz` intrinsic; known as `__nvvm_f2ull_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rz")
    i64 f2ull_rz(f32);
    /// The `llvm.nvvm.f2ull.rm.ftz` intrinsic; known as `__nvvm_f2ull_rm_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rm.ftz")
    i64 f2ull_rm_ftz(f32);
    /// The `llvm.nvvm.f2ull.rm` intrinsic; known as `__nvvm_f2ull_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rm")
    i64 f2ull_rm(f32);
    /// The `llvm.nvvm.f2ull.rp.ftz` intrinsic; known as `__nvvm_f2ull_rp_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rp.ftz")
    i64 f2ull_rp_ftz(f32);
    /// The `llvm.nvvm.f2ull.rp` intrinsic; known as `__nvvm_f2ull_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2ull.rp")
    i64 f2ull_rp(f32);
    /// The `llvm.nvvm.d2ll.rn` intrinsic; known as `__nvvm_d2ll_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ll.rn")
    i64 d2ll_rn(f64);
    /// The `llvm.nvvm.d2ll.rz` intrinsic; known as `__nvvm_d2ll_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ll.rz")
    i64 d2ll_rz(f64);
    /// The `llvm.nvvm.d2ll.rm` intrinsic; known as `__nvvm_d2ll_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ll.rm")
    i64 d2ll_rm(f64);
    /// The `llvm.nvvm.d2ll.rp` intrinsic; known as `__nvvm_d2ll_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ll.rp")
    i64 d2ll_rp(f64);
    /// The `llvm.nvvm.d2ull.rn` intrinsic; known as `__nvvm_d2ull_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ull.rn")
    i64 d2ull_rn(f64);
    /// The `llvm.nvvm.d2ull.rz` intrinsic; known as `__nvvm_d2ull_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ull.rz")
    i64 d2ull_rz(f64);
    /// The `llvm.nvvm.d2ull.rm` intrinsic; known as `__nvvm_d2ull_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ull.rm")
    i64 d2ull_rm(f64);
    /// The `llvm.nvvm.d2ull.rp` intrinsic; known as `__nvvm_d2ull_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.d2ull.rp")
    i64 d2ull_rp(f64);
    /// The `llvm.nvvm.ll2f.rn` intrinsic; known as `__nvvm_ll2f_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2f.rn")
    f32 ll2f_rn(i64);
    /// The `llvm.nvvm.ll2f.rz` intrinsic; known as `__nvvm_ll2f_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2f.rz")
    f32 ll2f_rz(i64);
    /// The `llvm.nvvm.ll2f.rm` intrinsic; known as `__nvvm_ll2f_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2f.rm")
    f32 ll2f_rm(i64);
    /// The `llvm.nvvm.ll2f.rp` intrinsic; known as `__nvvm_ll2f_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2f.rp")
    f32 ll2f_rp(i64);
    /// The `llvm.nvvm.ull2f.rn` intrinsic; known as `__nvvm_ull2f_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2f.rn")
    f32 ull2f_rn(i64);
    /// The `llvm.nvvm.ull2f.rz` intrinsic; known as `__nvvm_ull2f_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2f.rz")
    f32 ull2f_rz(i64);
    /// The `llvm.nvvm.ull2f.rm` intrinsic; known as `__nvvm_ull2f_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2f.rm")
    f32 ull2f_rm(i64);
    /// The `llvm.nvvm.ull2f.rp` intrinsic; known as `__nvvm_ull2f_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2f.rp")
    f32 ull2f_rp(i64);
    /// The `llvm.nvvm.ll2d.rn` intrinsic; known as `__nvvm_ll2d_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2d.rn")
    f64 ll2d_rn(i64);
    /// The `llvm.nvvm.ll2d.rz` intrinsic; known as `__nvvm_ll2d_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2d.rz")
    f64 ll2d_rz(i64);
    /// The `llvm.nvvm.ll2d.rm` intrinsic; known as `__nvvm_ll2d_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2d.rm")
    f64 ll2d_rm(i64);
    /// The `llvm.nvvm.ll2d.rp` intrinsic; known as `__nvvm_ll2d_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ll2d.rp")
    f64 ll2d_rp(i64);
    /// The `llvm.nvvm.ull2d.rn` intrinsic; known as `__nvvm_ull2d_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2d.rn")
    f64 ull2d_rn(i64);
    /// The `llvm.nvvm.ull2d.rz` intrinsic; known as `__nvvm_ull2d_rz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2d.rz")
    f64 ull2d_rz(i64);
    /// The `llvm.nvvm.ull2d.rm` intrinsic; known as `__nvvm_ull2d_rm` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2d.rm")
    f64 ull2d_rm(i64);
    /// The `llvm.nvvm.ull2d.rp` intrinsic; known as `__nvvm_ull2d_rp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.ull2d.rp")
    f64 ull2d_rp(i64);
    /// The `llvm.nvvm.f2h.rn.ftz` intrinsic; known as `__nvvm_f2h_rn_ftz` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2h.rn.ftz")
    i16 f2h_rn_ftz(f32);
    /// The `llvm.nvvm.f2h.rn` intrinsic; known as `__nvvm_f2h_rn` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.f2h.rn")
    i16 f2h_rn(f32);
    /// The `llvm.nvvm.h2f` intrinsic; known as `__nvvm_h2f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.h2f")
    f32 h2f(i16);
    /// The `llvm.nvvm.bitcast.f2i` intrinsic; known as `__nvvm_bitcast_f2i` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.bitcast.f2i")
    i32 bitcast_f2i(f32);
    /// The `llvm.nvvm.bitcast.i2f` intrinsic; known as `__nvvm_bitcast_i2f` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.bitcast.i2f")
    f32 bitcast_i2f(i32);
    /// The `llvm.nvvm.bitcast.ll2d` intrinsic; known as `__nvvm_bitcast_ll2d` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.bitcast.ll2d")
    f64 bitcast_ll2d(i64);
    /// The `llvm.nvvm.bitcast.d2ll` intrinsic; known as `__nvvm_bitcast_d2ll` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.bitcast.d2ll")
    i64 bitcast_d2ll(f64);
    /// The `llvm.nvvm.atomic.load.add.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.atomic.load.add.f32")
    f32 atomic_load_add_f32(f32*, f32);
    /// The `llvm.nvvm.atomic.load.inc.32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.atomic.load.inc.32")
    i32 atomic_load_inc_32(i32*, i32);
    /// The `llvm.nvvm.atomic.load.dec.32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.atomic.load.dec.32")
    i32 atomic_load_dec_32(i32*, i32);
    /// The `llvm.nvvm.barrier0` intrinsic; known as `__nvvm_bar0` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.barrier0")
    void barrier0();
    /// The `llvm.nvvm.barrier0.popc` intrinsic; known as `__nvvm_bar0_popc` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.barrier0.popc")
    i32 barrier0_popc(i32);
    /// The `llvm.nvvm.barrier0.and` intrinsic; known as `__nvvm_bar0_and` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.barrier0.and")
    i32 barrier0_and(i32);
    /// The `llvm.nvvm.barrier0.or` intrinsic; known as `__nvvm_bar0_or` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.barrier0.or")
    i32 barrier0_or(i32);
    /// The `llvm.nvvm.membar.cta` intrinsic; known as `__nvvm_membar_cta` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.membar.cta")
    void membar_cta();
    /// The `llvm.nvvm.membar.gl` intrinsic; known as `__nvvm_membar_gl` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.membar.gl")
    void membar_gl();
    /// The `llvm.nvvm.membar.sys` intrinsic; known as `__nvvm_membar_sys` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.membar.sys")
    void membar_sys();
    /// The `llvm.nvvm.read.ptx.sreg.tid.x` intrinsic; known as `__nvvm_read_ptx_sreg_tid_x` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.tid.x")
    i32 read_ptx_sreg_tid_x();
    /// The `llvm.nvvm.read.ptx.sreg.tid.y` intrinsic; known as `__nvvm_read_ptx_sreg_tid_y` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.tid.y")
    i32 read_ptx_sreg_tid_y();
    /// The `llvm.nvvm.read.ptx.sreg.tid.z` intrinsic; known as `__nvvm_read_ptx_sreg_tid_z` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.tid.z")
    i32 read_ptx_sreg_tid_z();
    /// The `llvm.nvvm.read.ptx.sreg.ntid.x` intrinsic; known as `__nvvm_read_ptx_sreg_ntid_x` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.ntid.x")
    i32 read_ptx_sreg_ntid_x();
    /// The `llvm.nvvm.read.ptx.sreg.ntid.y` intrinsic; known as `__nvvm_read_ptx_sreg_ntid_y` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.ntid.y")
    i32 read_ptx_sreg_ntid_y();
    /// The `llvm.nvvm.read.ptx.sreg.ntid.z` intrinsic; known as `__nvvm_read_ptx_sreg_ntid_z` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.ntid.z")
    i32 read_ptx_sreg_ntid_z();
    /// The `llvm.nvvm.read.ptx.sreg.ctaid.x` intrinsic; known as `__nvvm_read_ptx_sreg_ctaid_x` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.ctaid.x")
    i32 read_ptx_sreg_ctaid_x();
    /// The `llvm.nvvm.read.ptx.sreg.ctaid.y` intrinsic; known as `__nvvm_read_ptx_sreg_ctaid_y` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.ctaid.y")
    i32 read_ptx_sreg_ctaid_y();
    /// The `llvm.nvvm.read.ptx.sreg.ctaid.z` intrinsic; known as `__nvvm_read_ptx_sreg_ctaid_z` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.ctaid.z")
    i32 read_ptx_sreg_ctaid_z();
    /// The `llvm.nvvm.read.ptx.sreg.nctaid.x` intrinsic; known as `__nvvm_read_ptx_sreg_nctaid_x` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.nctaid.x")
    i32 read_ptx_sreg_nctaid_x();
    /// The `llvm.nvvm.read.ptx.sreg.nctaid.y` intrinsic; known as `__nvvm_read_ptx_sreg_nctaid_y` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.nctaid.y")
    i32 read_ptx_sreg_nctaid_y();
    /// The `llvm.nvvm.read.ptx.sreg.nctaid.z` intrinsic; known as `__nvvm_read_ptx_sreg_nctaid_z` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.nctaid.z")
    i32 read_ptx_sreg_nctaid_z();
    /// The `llvm.nvvm.read.ptx.sreg.warpsize` intrinsic; known as `__nvvm_read_ptx_sreg_warpsize` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.warpsize")
    i32 read_ptx_sreg_warpsize();
    /// The `llvm.nvvm.ldu.global.i.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.v16i8")
    __vector(i8[16]) ldu_global_i_v16i8(__vector(i8[16])*, i32);
    /// The `llvm.nvvm.ldu.global.i.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.i8")
    i8 ldu_global_i_i8(i8*, i32);
    /// The `llvm.nvvm.ldu.global.i.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.v8i16")
    __vector(i16[8]) ldu_global_i_v8i16(__vector(i16[8])*, i32);
    /// The `llvm.nvvm.ldu.global.i.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.i16")
    i16 ldu_global_i_i16(i16*, i32);
    /// The `llvm.nvvm.ldu.global.i.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.v4i32")
    __vector(i32[4]) ldu_global_i_v4i32(__vector(i32[4])*, i32);
    /// The `llvm.nvvm.ldu.global.i.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.i32")
    i32 ldu_global_i_i32(i32*, i32);
    /// The `llvm.nvvm.ldu.global.i.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.v2i64")
    __vector(i64[2]) ldu_global_i_v2i64(__vector(i64[2])*, i32);
    /// The `llvm.nvvm.ldu.global.i.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.i.i64")
    i64 ldu_global_i_i64(i64*, i32);
    /// The `llvm.nvvm.ldu.global.f.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.f.v4f32")
    __vector(f32[4]) ldu_global_f_v4f32(f32x4*, i32);
    /// The `llvm.nvvm.ldu.global.f.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.f.f32")
    f32 ldu_global_f_f32(f32*, i32);
    /// The `llvm.nvvm.ldu.global.f.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.f.v2f64")
    __vector(f64[2]) ldu_global_f_v2f64(f64x2*, i32);
    /// The `llvm.nvvm.ldu.global.f.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.f.f64")
    f64 ldu_global_f_f64(f64*, i32);
    /// The `llvm.nvvm.ldu.global.p.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldu.global.p.p0i8")
    i8* ldu_global_p_p0i8(i8*, i32);
    /// The `llvm.nvvm.ldg.global.i.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.v16i8")
    __vector(i8[16]) ldg_global_i_v16i8(__vector(i8[16])*, i32);
    /// The `llvm.nvvm.ldg.global.i.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.i8")
    i8 ldg_global_i_i8(i8*, i32);
    /// The `llvm.nvvm.ldg.global.i.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.v8i16")
    __vector(i16[8]) ldg_global_i_v8i16(__vector(i16[8])*, i32);
    /// The `llvm.nvvm.ldg.global.i.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.i16")
    i16 ldg_global_i_i16(i16*, i32);
    /// The `llvm.nvvm.ldg.global.i.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.v4i32")
    __vector(i32[4]) ldg_global_i_v4i32(__vector(i32[4])*, i32);
    /// The `llvm.nvvm.ldg.global.i.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.i32")
    i32 ldg_global_i_i32(i32*, i32);
    /// The `llvm.nvvm.ldg.global.i.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.v2i64")
    __vector(i64[2]) ldg_global_i_v2i64(__vector(i64[2])*, i32);
    /// The `llvm.nvvm.ldg.global.i.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.i.i64")
    i64 ldg_global_i_i64(i64*, i32);
    /// The `llvm.nvvm.ldg.global.f.v4f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.f.v4f32")
    __vector(f32[4]) ldg_global_f_v4f32(f32x4*, i32);
    /// The `llvm.nvvm.ldg.global.f.f32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.f.f32")
    f32 ldg_global_f_f32(f32*, i32);
    /// The `llvm.nvvm.ldg.global.f.v2f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.f.v2f64")
    __vector(f64[2]) ldg_global_f_v2f64(f64x2*, i32);
    /// The `llvm.nvvm.ldg.global.f.f64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.f.f64")
    f64 ldg_global_f_f64(f64*, i32);
    /// The `llvm.nvvm.ldg.global.p.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ldg.global.p.p0i8")
    i8* ldg_global_p_p0i8(i8*, i32);
    /// The `llvm.nvvm.ptr.local.to.gen.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.local.to.gen.p0i8.p0i8")
    i8* ptr_local_to_gen_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.shared.to.gen.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.shared.to.gen.p0i8.p0i8")
    i8* ptr_shared_to_gen_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.global.to.gen.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.global.to.gen.p0i8.p0i8")
    i8* ptr_global_to_gen_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.constant.to.gen.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.constant.to.gen.p0i8.p0i8")
    i8* ptr_constant_to_gen_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.gen.to.global.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.gen.to.global.p0i8.p0i8")
    i8* ptr_gen_to_global_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.gen.to.shared.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.gen.to.shared.p0i8.p0i8")
    i8* ptr_gen_to_shared_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.gen.to.local.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.gen.to.local.p0i8.p0i8")
    i8* ptr_gen_to_local_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.gen.to.constant.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.gen.to.constant.p0i8.p0i8")
    i8* ptr_gen_to_constant_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.ptr.gen.to.param.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.ptr.gen.to.param.p0i8.p0i8")
    i8* ptr_gen_to_param_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.move.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.move.i16")
    i16 move_i16(i16);
    /// The `llvm.nvvm.move.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.move.i32")
    i32 move_i32(i32);
    /// The `llvm.nvvm.move.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.move.i64")
    i64 move_i64(i64);
    /// The `llvm.nvvm.move.float` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.move.float")
    f32 move_float(f32);
    /// The `llvm.nvvm.move.double` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.move.double")
    f64 move_double(f64);
    /// The `llvm.nvvm.move.ptr.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.move.ptr.p0i8.p0i8")
    i8* move_ptr_p0i8_p0i8(i8*);
    /// The `llvm.nvvm.texsurf.handle.internal.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.texsurf.handle.internal.p0i8")
    i64 texsurf_handle_internal_p0i8(i8*);
    /// The `llvm.nvvm.compiler.error.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.compiler.error.p0i8")
    void compiler_error_p0i8(i8*);
    /// The `llvm.nvvm.compiler.warn.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.compiler.warn.p0i8")
    void compiler_warn_p0i8(i8*);
    /// The `llvm.nvvm.reflect.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.reflect.p0i8")
    i32 reflect_p0i8(i8*);
    /// The `llvm.nvvm.isspacep.const` intrinsic; known as `__nvvm_isspacep_const` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.isspacep.const")
    bool isspacep_const(i8*);
    /// The `llvm.nvvm.isspacep.global` intrinsic; known as `__nvvm_isspacep_global` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.isspacep.global")
    bool isspacep_global(i8*);
    /// The `llvm.nvvm.isspacep.local` intrinsic; known as `__nvvm_isspacep_local` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.isspacep.local")
    bool isspacep_local(i8*);
    /// The `llvm.nvvm.isspacep.shared` intrinsic; known as `__nvvm_isspacep_shared` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.isspacep.shared")
    bool isspacep_shared(i8*);
    /// The `llvm.nvvm.read.ptx.sreg.envreg0` intrinsic; known as `__nvvm_read_ptx_sreg_envreg0` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg0")
    i32 read_ptx_sreg_envreg0();
    /// The `llvm.nvvm.read.ptx.sreg.envreg1` intrinsic; known as `__nvvm_read_ptx_sreg_envreg1` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg1")
    i32 read_ptx_sreg_envreg1();
    /// The `llvm.nvvm.read.ptx.sreg.envreg2` intrinsic; known as `__nvvm_read_ptx_sreg_envreg2` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg2")
    i32 read_ptx_sreg_envreg2();
    /// The `llvm.nvvm.read.ptx.sreg.envreg3` intrinsic; known as `__nvvm_read_ptx_sreg_envreg3` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg3")
    i32 read_ptx_sreg_envreg3();
    /// The `llvm.nvvm.read.ptx.sreg.envreg4` intrinsic; known as `__nvvm_read_ptx_sreg_envreg4` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg4")
    i32 read_ptx_sreg_envreg4();
    /// The `llvm.nvvm.read.ptx.sreg.envreg5` intrinsic; known as `__nvvm_read_ptx_sreg_envreg5` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg5")
    i32 read_ptx_sreg_envreg5();
    /// The `llvm.nvvm.read.ptx.sreg.envreg6` intrinsic; known as `__nvvm_read_ptx_sreg_envreg6` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg6")
    i32 read_ptx_sreg_envreg6();
    /// The `llvm.nvvm.read.ptx.sreg.envreg7` intrinsic; known as `__nvvm_read_ptx_sreg_envreg7` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg7")
    i32 read_ptx_sreg_envreg7();
    /// The `llvm.nvvm.read.ptx.sreg.envreg8` intrinsic; known as `__nvvm_read_ptx_sreg_envreg8` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg8")
    i32 read_ptx_sreg_envreg8();
    /// The `llvm.nvvm.read.ptx.sreg.envreg9` intrinsic; known as `__nvvm_read_ptx_sreg_envreg9` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg9")
    i32 read_ptx_sreg_envreg9();
    /// The `llvm.nvvm.read.ptx.sreg.envreg10` intrinsic; known as `__nvvm_read_ptx_sreg_envreg10` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg10")
    i32 read_ptx_sreg_envreg10();
    /// The `llvm.nvvm.read.ptx.sreg.envreg11` intrinsic; known as `__nvvm_read_ptx_sreg_envreg11` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg11")
    i32 read_ptx_sreg_envreg11();
    /// The `llvm.nvvm.read.ptx.sreg.envreg12` intrinsic; known as `__nvvm_read_ptx_sreg_envreg12` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg12")
    i32 read_ptx_sreg_envreg12();
    /// The `llvm.nvvm.read.ptx.sreg.envreg13` intrinsic; known as `__nvvm_read_ptx_sreg_envreg13` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg13")
    i32 read_ptx_sreg_envreg13();
    /// The `llvm.nvvm.read.ptx.sreg.envreg14` intrinsic; known as `__nvvm_read_ptx_sreg_envreg14` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg14")
    i32 read_ptx_sreg_envreg14();
    /// The `llvm.nvvm.read.ptx.sreg.envreg15` intrinsic; known as `__nvvm_read_ptx_sreg_envreg15` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg15")
    i32 read_ptx_sreg_envreg15();
    /// The `llvm.nvvm.read.ptx.sreg.envreg16` intrinsic; known as `__nvvm_read_ptx_sreg_envreg16` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg16")
    i32 read_ptx_sreg_envreg16();
    /// The `llvm.nvvm.read.ptx.sreg.envreg17` intrinsic; known as `__nvvm_read_ptx_sreg_envreg17` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg17")
    i32 read_ptx_sreg_envreg17();
    /// The `llvm.nvvm.read.ptx.sreg.envreg18` intrinsic; known as `__nvvm_read_ptx_sreg_envreg18` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg18")
    i32 read_ptx_sreg_envreg18();
    /// The `llvm.nvvm.read.ptx.sreg.envreg19` intrinsic; known as `__nvvm_read_ptx_sreg_envreg19` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg19")
    i32 read_ptx_sreg_envreg19();
    /// The `llvm.nvvm.read.ptx.sreg.envreg20` intrinsic; known as `__nvvm_read_ptx_sreg_envreg20` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg20")
    i32 read_ptx_sreg_envreg20();
    /// The `llvm.nvvm.read.ptx.sreg.envreg21` intrinsic; known as `__nvvm_read_ptx_sreg_envreg21` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg21")
    i32 read_ptx_sreg_envreg21();
    /// The `llvm.nvvm.read.ptx.sreg.envreg22` intrinsic; known as `__nvvm_read_ptx_sreg_envreg22` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg22")
    i32 read_ptx_sreg_envreg22();
    /// The `llvm.nvvm.read.ptx.sreg.envreg23` intrinsic; known as `__nvvm_read_ptx_sreg_envreg23` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg23")
    i32 read_ptx_sreg_envreg23();
    /// The `llvm.nvvm.read.ptx.sreg.envreg24` intrinsic; known as `__nvvm_read_ptx_sreg_envreg24` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg24")
    i32 read_ptx_sreg_envreg24();
    /// The `llvm.nvvm.read.ptx.sreg.envreg25` intrinsic; known as `__nvvm_read_ptx_sreg_envreg25` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg25")
    i32 read_ptx_sreg_envreg25();
    /// The `llvm.nvvm.read.ptx.sreg.envreg26` intrinsic; known as `__nvvm_read_ptx_sreg_envreg26` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg26")
    i32 read_ptx_sreg_envreg26();
    /// The `llvm.nvvm.read.ptx.sreg.envreg27` intrinsic; known as `__nvvm_read_ptx_sreg_envreg27` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg27")
    i32 read_ptx_sreg_envreg27();
    /// The `llvm.nvvm.read.ptx.sreg.envreg28` intrinsic; known as `__nvvm_read_ptx_sreg_envreg28` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg28")
    i32 read_ptx_sreg_envreg28();
    /// The `llvm.nvvm.read.ptx.sreg.envreg29` intrinsic; known as `__nvvm_read_ptx_sreg_envreg29` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg29")
    i32 read_ptx_sreg_envreg29();
    /// The `llvm.nvvm.read.ptx.sreg.envreg30` intrinsic; known as `__nvvm_read_ptx_sreg_envreg30` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg30")
    i32 read_ptx_sreg_envreg30();
    /// The `llvm.nvvm.read.ptx.sreg.envreg31` intrinsic; known as `__nvvm_read_ptx_sreg_envreg31` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.read.ptx.sreg.envreg31")
    i32 read_ptx_sreg_envreg31();
    /// The `llvm.nvvm.suld.1d.i8.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i8.clamp")
    i16 suld_1d_i8_clamp(i64, i32);
    /// The `llvm.nvvm.suld.1d.i16.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i16.clamp")
    i16 suld_1d_i16_clamp(i64, i32);
    /// The `llvm.nvvm.suld.1d.i32.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i32.clamp")
    i32 suld_1d_i32_clamp(i64, i32);
    /// The `llvm.nvvm.suld.1d.i64.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i64.clamp")
    i64 suld_1d_i64_clamp(i64, i32);
    /// The `llvm.nvvm.suld.1d.array.i8.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i8.clamp")
    i16 suld_1d_array_i8_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i16.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i16.clamp")
    i16 suld_1d_array_i16_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i32.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i32.clamp")
    i32 suld_1d_array_i32_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i64.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i64.clamp")
    i64 suld_1d_array_i64_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i8.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i8.clamp")
    i16 suld_2d_i8_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i16.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i16.clamp")
    i16 suld_2d_i16_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i32.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i32.clamp")
    i32 suld_2d_i32_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i64.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i64.clamp")
    i64 suld_2d_i64_clamp(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i8.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i8.clamp")
    i16 suld_2d_array_i8_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i16.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i16.clamp")
    i16 suld_2d_array_i16_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i32.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i32.clamp")
    i32 suld_2d_array_i32_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i64.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i64.clamp")
    i64 suld_2d_array_i64_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i8.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i8.clamp")
    i16 suld_3d_i8_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i16.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i16.clamp")
    i16 suld_3d_i16_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i32.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i32.clamp")
    i32 suld_3d_i32_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i64.clamp` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i64.clamp")
    i64 suld_3d_i64_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.1d.i8.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i8.trap")
    i16 suld_1d_i8_trap(i64, i32);
    /// The `llvm.nvvm.suld.1d.i16.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i16.trap")
    i16 suld_1d_i16_trap(i64, i32);
    /// The `llvm.nvvm.suld.1d.i32.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i32.trap")
    i32 suld_1d_i32_trap(i64, i32);
    /// The `llvm.nvvm.suld.1d.i64.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i64.trap")
    i64 suld_1d_i64_trap(i64, i32);
    /// The `llvm.nvvm.suld.1d.array.i8.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i8.trap")
    i16 suld_1d_array_i8_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i16.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i16.trap")
    i16 suld_1d_array_i16_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i32.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i32.trap")
    i32 suld_1d_array_i32_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i64.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i64.trap")
    i64 suld_1d_array_i64_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i8.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i8.trap")
    i16 suld_2d_i8_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i16.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i16.trap")
    i16 suld_2d_i16_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i32.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i32.trap")
    i32 suld_2d_i32_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i64.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i64.trap")
    i64 suld_2d_i64_trap(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i8.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i8.trap")
    i16 suld_2d_array_i8_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i16.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i16.trap")
    i16 suld_2d_array_i16_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i32.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i32.trap")
    i32 suld_2d_array_i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i64.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i64.trap")
    i64 suld_2d_array_i64_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i8.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i8.trap")
    i16 suld_3d_i8_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i16.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i16.trap")
    i16 suld_3d_i16_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i32.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i32.trap")
    i32 suld_3d_i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i64.trap` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i64.trap")
    i64 suld_3d_i64_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.1d.i8.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i8.zero")
    i16 suld_1d_i8_zero(i64, i32);
    /// The `llvm.nvvm.suld.1d.i16.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i16.zero")
    i16 suld_1d_i16_zero(i64, i32);
    /// The `llvm.nvvm.suld.1d.i32.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i32.zero")
    i32 suld_1d_i32_zero(i64, i32);
    /// The `llvm.nvvm.suld.1d.i64.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.i64.zero")
    i64 suld_1d_i64_zero(i64, i32);
    /// The `llvm.nvvm.suld.1d.array.i8.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i8.zero")
    i16 suld_1d_array_i8_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i16.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i16.zero")
    i16 suld_1d_array_i16_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i32.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i32.zero")
    i32 suld_1d_array_i32_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.1d.array.i64.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.1d.array.i64.zero")
    i64 suld_1d_array_i64_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i8.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i8.zero")
    i16 suld_2d_i8_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i16.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i16.zero")
    i16 suld_2d_i16_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i32.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i32.zero")
    i32 suld_2d_i32_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.i64.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.i64.zero")
    i64 suld_2d_i64_zero(i64, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i8.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i8.zero")
    i16 suld_2d_array_i8_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i16.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i16.zero")
    i16 suld_2d_array_i16_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i32.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i32.zero")
    i32 suld_2d_array_i32_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.2d.array.i64.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.2d.array.i64.zero")
    i64 suld_2d_array_i64_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i8.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i8.zero")
    i16 suld_3d_i8_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i16.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i16.zero")
    i16 suld_3d_i16_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i32.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i32.zero")
    i32 suld_3d_i32_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.suld.3d.i64.zero` intrinsic.
    pragma(LDC_intrinsic, "llvm.nvvm.suld.3d.i64.zero")
    i64 suld_3d_i64_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.txq.channel.order` intrinsic; known as `__nvvm_txq_channel_order` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.channel.order")
    i32 txq_channel_order(i64);
    /// The `llvm.nvvm.txq.channel.data.type` intrinsic; known as `__nvvm_txq_channel_data_type` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.channel.data.type")
    i32 txq_channel_data_type(i64);
    /// The `llvm.nvvm.txq.width` intrinsic; known as `__nvvm_txq_width` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.width")
    i32 txq_width(i64);
    /// The `llvm.nvvm.txq.height` intrinsic; known as `__nvvm_txq_height` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.height")
    i32 txq_height(i64);
    /// The `llvm.nvvm.txq.depth` intrinsic; known as `__nvvm_txq_depth` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.depth")
    i32 txq_depth(i64);
    /// The `llvm.nvvm.txq.array.size` intrinsic; known as `__nvvm_txq_array_size` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.array.size")
    i32 txq_array_size(i64);
    /// The `llvm.nvvm.txq.num.samples` intrinsic; known as `__nvvm_txq_num_samples` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.num.samples")
    i32 txq_num_samples(i64);
    /// The `llvm.nvvm.txq.num.mipmap.levels` intrinsic; known as `__nvvm_txq_num_mipmap_levels` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.txq.num.mipmap.levels")
    i32 txq_num_mipmap_levels(i64);
    /// The `llvm.nvvm.suq.channel.order` intrinsic; known as `__nvvm_suq_channel_order` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.suq.channel.order")
    i32 suq_channel_order(i64);
    /// The `llvm.nvvm.suq.channel.data.type` intrinsic; known as `__nvvm_suq_channel_data_type` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.suq.channel.data.type")
    i32 suq_channel_data_type(i64);
    /// The `llvm.nvvm.suq.width` intrinsic; known as `__nvvm_suq_width` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.suq.width")
    i32 suq_width(i64);
    /// The `llvm.nvvm.suq.height` intrinsic; known as `__nvvm_suq_height` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.suq.height")
    i32 suq_height(i64);
    /// The `llvm.nvvm.suq.depth` intrinsic; known as `__nvvm_suq_depth` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.suq.depth")
    i32 suq_depth(i64);
    /// The `llvm.nvvm.suq.array.size` intrinsic; known as `__nvvm_suq_array_size` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.suq.array.size")
    i32 suq_array_size(i64);
    /// The `llvm.nvvm.istypep.sampler` intrinsic; known as `__nvvm_istypep_sampler` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.istypep.sampler")
    bool istypep_sampler(i64);
    /// The `llvm.nvvm.istypep.surface` intrinsic; known as `__nvvm_istypep_surface` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.istypep.surface")
    bool istypep_surface(i64);
    /// The `llvm.nvvm.istypep.texture` intrinsic; known as `__nvvm_istypep_texture` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.istypep.texture")
    bool istypep_texture(i64);
    /// The `llvm.nvvm.sust.b.1d.i8.clamp` intrinsic; known as `__nvvm_sust_b_1d_i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i8.clamp")
    void sust_b_1d_i8_clamp(i64, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.i16.clamp` intrinsic; known as `__nvvm_sust_b_1d_i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i16.clamp")
    void sust_b_1d_i16_clamp(i64, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.i32.clamp` intrinsic; known as `__nvvm_sust_b_1d_i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i32.clamp")
    void sust_b_1d_i32_clamp(i64, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.i64.clamp` intrinsic; known as `__nvvm_sust_b_1d_i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i64.clamp")
    void sust_b_1d_i64_clamp(i64, i32, i64);
    /// The `llvm.nvvm.sust.b.1d.v2i8.clamp` intrinsic; known as `__nvvm_sust_b_1d_v2i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i8.clamp")
    void sust_b_1d_v2i8_clamp(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v2i16.clamp` intrinsic; known as `__nvvm_sust_b_1d_v2i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i16.clamp")
    void sust_b_1d_v2i16_clamp(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v2i32.clamp` intrinsic; known as `__nvvm_sust_b_1d_v2i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i32.clamp")
    void sust_b_1d_v2i32_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.v2i64.clamp` intrinsic; known as `__nvvm_sust_b_1d_v2i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i64.clamp")
    void sust_b_1d_v2i64_clamp(i64, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.1d.v4i8.clamp` intrinsic; known as `__nvvm_sust_b_1d_v4i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i8.clamp")
    void sust_b_1d_v4i8_clamp(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v4i16.clamp` intrinsic; known as `__nvvm_sust_b_1d_v4i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i16.clamp")
    void sust_b_1d_v4i16_clamp(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v4i32.clamp` intrinsic; known as `__nvvm_sust_b_1d_v4i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i32.clamp")
    void sust_b_1d_v4i32_clamp(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.i8.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i8.clamp")
    void sust_b_1d_array_i8_clamp(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.array.i16.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i16.clamp")
    void sust_b_1d_array_i16_clamp(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.array.i32.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i32.clamp")
    void sust_b_1d_array_i32_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.i64.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i64.clamp")
    void sust_b_1d_array_i64_clamp(i64, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.1d.array.v2i8.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v2i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i8.clamp")
    void sust_b_1d_array_v2i8_clamp(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v2i16.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v2i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i16.clamp")
    void sust_b_1d_array_v2i16_clamp(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v2i32.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v2i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i32.clamp")
    void sust_b_1d_array_v2i32_clamp(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.v2i64.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v2i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i64.clamp")
    void sust_b_1d_array_v2i64_clamp(i64, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.1d.array.v4i8.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v4i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i8.clamp")
    void sust_b_1d_array_v4i8_clamp(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v4i16.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v4i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i16.clamp")
    void sust_b_1d_array_v4i16_clamp(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v4i32.clamp` intrinsic; known as `__nvvm_sust_b_1d_array_v4i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i32.clamp")
    void sust_b_1d_array_v4i32_clamp(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.i8.clamp` intrinsic; known as `__nvvm_sust_b_2d_i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i8.clamp")
    void sust_b_2d_i8_clamp(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.i16.clamp` intrinsic; known as `__nvvm_sust_b_2d_i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i16.clamp")
    void sust_b_2d_i16_clamp(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.i32.clamp` intrinsic; known as `__nvvm_sust_b_2d_i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i32.clamp")
    void sust_b_2d_i32_clamp(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.i64.clamp` intrinsic; known as `__nvvm_sust_b_2d_i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i64.clamp")
    void sust_b_2d_i64_clamp(i64, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.2d.v2i8.clamp` intrinsic; known as `__nvvm_sust_b_2d_v2i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i8.clamp")
    void sust_b_2d_v2i8_clamp(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v2i16.clamp` intrinsic; known as `__nvvm_sust_b_2d_v2i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i16.clamp")
    void sust_b_2d_v2i16_clamp(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v2i32.clamp` intrinsic; known as `__nvvm_sust_b_2d_v2i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i32.clamp")
    void sust_b_2d_v2i32_clamp(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.v2i64.clamp` intrinsic; known as `__nvvm_sust_b_2d_v2i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i64.clamp")
    void sust_b_2d_v2i64_clamp(i64, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.2d.v4i8.clamp` intrinsic; known as `__nvvm_sust_b_2d_v4i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i8.clamp")
    void sust_b_2d_v4i8_clamp(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v4i16.clamp` intrinsic; known as `__nvvm_sust_b_2d_v4i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i16.clamp")
    void sust_b_2d_v4i16_clamp(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v4i32.clamp` intrinsic; known as `__nvvm_sust_b_2d_v4i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i32.clamp")
    void sust_b_2d_v4i32_clamp(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.i8.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i8.clamp")
    void sust_b_2d_array_i8_clamp(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.array.i16.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i16.clamp")
    void sust_b_2d_array_i16_clamp(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.array.i32.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i32.clamp")
    void sust_b_2d_array_i32_clamp(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.i64.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i64.clamp")
    void sust_b_2d_array_i64_clamp(i64, i32, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.2d.array.v2i8.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v2i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i8.clamp")
    void sust_b_2d_array_v2i8_clamp(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v2i16.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v2i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i16.clamp")
    void sust_b_2d_array_v2i16_clamp(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v2i32.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v2i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i32.clamp")
    void sust_b_2d_array_v2i32_clamp(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.v2i64.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v2i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i64.clamp")
    void sust_b_2d_array_v2i64_clamp(i64, i32, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.2d.array.v4i8.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v4i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i8.clamp")
    void sust_b_2d_array_v4i8_clamp(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v4i16.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v4i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i16.clamp")
    void sust_b_2d_array_v4i16_clamp(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v4i32.clamp` intrinsic; known as `__nvvm_sust_b_2d_array_v4i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i32.clamp")
    void sust_b_2d_array_v4i32_clamp(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.i8.clamp` intrinsic; known as `__nvvm_sust_b_3d_i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i8.clamp")
    void sust_b_3d_i8_clamp(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.3d.i16.clamp` intrinsic; known as `__nvvm_sust_b_3d_i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i16.clamp")
    void sust_b_3d_i16_clamp(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.3d.i32.clamp` intrinsic; known as `__nvvm_sust_b_3d_i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i32.clamp")
    void sust_b_3d_i32_clamp(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.i64.clamp` intrinsic; known as `__nvvm_sust_b_3d_i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i64.clamp")
    void sust_b_3d_i64_clamp(i64, i32, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.3d.v2i8.clamp` intrinsic; known as `__nvvm_sust_b_3d_v2i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i8.clamp")
    void sust_b_3d_v2i8_clamp(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v2i16.clamp` intrinsic; known as `__nvvm_sust_b_3d_v2i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i16.clamp")
    void sust_b_3d_v2i16_clamp(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v2i32.clamp` intrinsic; known as `__nvvm_sust_b_3d_v2i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i32.clamp")
    void sust_b_3d_v2i32_clamp(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.v2i64.clamp` intrinsic; known as `__nvvm_sust_b_3d_v2i64_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i64.clamp")
    void sust_b_3d_v2i64_clamp(i64, i32, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.3d.v4i8.clamp` intrinsic; known as `__nvvm_sust_b_3d_v4i8_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i8.clamp")
    void sust_b_3d_v4i8_clamp(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v4i16.clamp` intrinsic; known as `__nvvm_sust_b_3d_v4i16_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i16.clamp")
    void sust_b_3d_v4i16_clamp(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v4i32.clamp` intrinsic; known as `__nvvm_sust_b_3d_v4i32_clamp` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i32.clamp")
    void sust_b_3d_v4i32_clamp(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.i8.trap` intrinsic; known as `__nvvm_sust_b_1d_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i8.trap")
    void sust_b_1d_i8_trap(i64, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.i16.trap` intrinsic; known as `__nvvm_sust_b_1d_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i16.trap")
    void sust_b_1d_i16_trap(i64, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.i32.trap` intrinsic; known as `__nvvm_sust_b_1d_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i32.trap")
    void sust_b_1d_i32_trap(i64, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.i64.trap` intrinsic; known as `__nvvm_sust_b_1d_i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i64.trap")
    void sust_b_1d_i64_trap(i64, i32, i64);
    /// The `llvm.nvvm.sust.b.1d.v2i8.trap` intrinsic; known as `__nvvm_sust_b_1d_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i8.trap")
    void sust_b_1d_v2i8_trap(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v2i16.trap` intrinsic; known as `__nvvm_sust_b_1d_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i16.trap")
    void sust_b_1d_v2i16_trap(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v2i32.trap` intrinsic; known as `__nvvm_sust_b_1d_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i32.trap")
    void sust_b_1d_v2i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.v2i64.trap` intrinsic; known as `__nvvm_sust_b_1d_v2i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i64.trap")
    void sust_b_1d_v2i64_trap(i64, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.1d.v4i8.trap` intrinsic; known as `__nvvm_sust_b_1d_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i8.trap")
    void sust_b_1d_v4i8_trap(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v4i16.trap` intrinsic; known as `__nvvm_sust_b_1d_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i16.trap")
    void sust_b_1d_v4i16_trap(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v4i32.trap` intrinsic; known as `__nvvm_sust_b_1d_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i32.trap")
    void sust_b_1d_v4i32_trap(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.i8.trap` intrinsic; known as `__nvvm_sust_b_1d_array_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i8.trap")
    void sust_b_1d_array_i8_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.array.i16.trap` intrinsic; known as `__nvvm_sust_b_1d_array_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i16.trap")
    void sust_b_1d_array_i16_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.array.i32.trap` intrinsic; known as `__nvvm_sust_b_1d_array_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i32.trap")
    void sust_b_1d_array_i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.i64.trap` intrinsic; known as `__nvvm_sust_b_1d_array_i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i64.trap")
    void sust_b_1d_array_i64_trap(i64, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.1d.array.v2i8.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i8.trap")
    void sust_b_1d_array_v2i8_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v2i16.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i16.trap")
    void sust_b_1d_array_v2i16_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v2i32.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i32.trap")
    void sust_b_1d_array_v2i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.v2i64.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v2i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i64.trap")
    void sust_b_1d_array_v2i64_trap(i64, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.1d.array.v4i8.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i8.trap")
    void sust_b_1d_array_v4i8_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v4i16.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i16.trap")
    void sust_b_1d_array_v4i16_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v4i32.trap` intrinsic; known as `__nvvm_sust_b_1d_array_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i32.trap")
    void sust_b_1d_array_v4i32_trap(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.i8.trap` intrinsic; known as `__nvvm_sust_b_2d_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i8.trap")
    void sust_b_2d_i8_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.i16.trap` intrinsic; known as `__nvvm_sust_b_2d_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i16.trap")
    void sust_b_2d_i16_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.i32.trap` intrinsic; known as `__nvvm_sust_b_2d_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i32.trap")
    void sust_b_2d_i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.i64.trap` intrinsic; known as `__nvvm_sust_b_2d_i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i64.trap")
    void sust_b_2d_i64_trap(i64, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.2d.v2i8.trap` intrinsic; known as `__nvvm_sust_b_2d_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i8.trap")
    void sust_b_2d_v2i8_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v2i16.trap` intrinsic; known as `__nvvm_sust_b_2d_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i16.trap")
    void sust_b_2d_v2i16_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v2i32.trap` intrinsic; known as `__nvvm_sust_b_2d_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i32.trap")
    void sust_b_2d_v2i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.v2i64.trap` intrinsic; known as `__nvvm_sust_b_2d_v2i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i64.trap")
    void sust_b_2d_v2i64_trap(i64, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.2d.v4i8.trap` intrinsic; known as `__nvvm_sust_b_2d_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i8.trap")
    void sust_b_2d_v4i8_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v4i16.trap` intrinsic; known as `__nvvm_sust_b_2d_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i16.trap")
    void sust_b_2d_v4i16_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v4i32.trap` intrinsic; known as `__nvvm_sust_b_2d_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i32.trap")
    void sust_b_2d_v4i32_trap(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.i8.trap` intrinsic; known as `__nvvm_sust_b_2d_array_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i8.trap")
    void sust_b_2d_array_i8_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.array.i16.trap` intrinsic; known as `__nvvm_sust_b_2d_array_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i16.trap")
    void sust_b_2d_array_i16_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.array.i32.trap` intrinsic; known as `__nvvm_sust_b_2d_array_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i32.trap")
    void sust_b_2d_array_i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.i64.trap` intrinsic; known as `__nvvm_sust_b_2d_array_i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i64.trap")
    void sust_b_2d_array_i64_trap(i64, i32, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.2d.array.v2i8.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i8.trap")
    void sust_b_2d_array_v2i8_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v2i16.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i16.trap")
    void sust_b_2d_array_v2i16_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v2i32.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i32.trap")
    void sust_b_2d_array_v2i32_trap(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.v2i64.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v2i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i64.trap")
    void sust_b_2d_array_v2i64_trap(i64, i32, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.2d.array.v4i8.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i8.trap")
    void sust_b_2d_array_v4i8_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v4i16.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i16.trap")
    void sust_b_2d_array_v4i16_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v4i32.trap` intrinsic; known as `__nvvm_sust_b_2d_array_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i32.trap")
    void sust_b_2d_array_v4i32_trap(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.i8.trap` intrinsic; known as `__nvvm_sust_b_3d_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i8.trap")
    void sust_b_3d_i8_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.3d.i16.trap` intrinsic; known as `__nvvm_sust_b_3d_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i16.trap")
    void sust_b_3d_i16_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.3d.i32.trap` intrinsic; known as `__nvvm_sust_b_3d_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i32.trap")
    void sust_b_3d_i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.i64.trap` intrinsic; known as `__nvvm_sust_b_3d_i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i64.trap")
    void sust_b_3d_i64_trap(i64, i32, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.3d.v2i8.trap` intrinsic; known as `__nvvm_sust_b_3d_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i8.trap")
    void sust_b_3d_v2i8_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v2i16.trap` intrinsic; known as `__nvvm_sust_b_3d_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i16.trap")
    void sust_b_3d_v2i16_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v2i32.trap` intrinsic; known as `__nvvm_sust_b_3d_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i32.trap")
    void sust_b_3d_v2i32_trap(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.v2i64.trap` intrinsic; known as `__nvvm_sust_b_3d_v2i64_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i64.trap")
    void sust_b_3d_v2i64_trap(i64, i32, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.3d.v4i8.trap` intrinsic; known as `__nvvm_sust_b_3d_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i8.trap")
    void sust_b_3d_v4i8_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v4i16.trap` intrinsic; known as `__nvvm_sust_b_3d_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i16.trap")
    void sust_b_3d_v4i16_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v4i32.trap` intrinsic; known as `__nvvm_sust_b_3d_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i32.trap")
    void sust_b_3d_v4i32_trap(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.i8.zero` intrinsic; known as `__nvvm_sust_b_1d_i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i8.zero")
    void sust_b_1d_i8_zero(i64, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.i16.zero` intrinsic; known as `__nvvm_sust_b_1d_i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i16.zero")
    void sust_b_1d_i16_zero(i64, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.i32.zero` intrinsic; known as `__nvvm_sust_b_1d_i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i32.zero")
    void sust_b_1d_i32_zero(i64, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.i64.zero` intrinsic; known as `__nvvm_sust_b_1d_i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.i64.zero")
    void sust_b_1d_i64_zero(i64, i32, i64);
    /// The `llvm.nvvm.sust.b.1d.v2i8.zero` intrinsic; known as `__nvvm_sust_b_1d_v2i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i8.zero")
    void sust_b_1d_v2i8_zero(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v2i16.zero` intrinsic; known as `__nvvm_sust_b_1d_v2i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i16.zero")
    void sust_b_1d_v2i16_zero(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v2i32.zero` intrinsic; known as `__nvvm_sust_b_1d_v2i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i32.zero")
    void sust_b_1d_v2i32_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.v2i64.zero` intrinsic; known as `__nvvm_sust_b_1d_v2i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v2i64.zero")
    void sust_b_1d_v2i64_zero(i64, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.1d.v4i8.zero` intrinsic; known as `__nvvm_sust_b_1d_v4i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i8.zero")
    void sust_b_1d_v4i8_zero(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v4i16.zero` intrinsic; known as `__nvvm_sust_b_1d_v4i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i16.zero")
    void sust_b_1d_v4i16_zero(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.v4i32.zero` intrinsic; known as `__nvvm_sust_b_1d_v4i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.v4i32.zero")
    void sust_b_1d_v4i32_zero(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.i8.zero` intrinsic; known as `__nvvm_sust_b_1d_array_i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i8.zero")
    void sust_b_1d_array_i8_zero(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.array.i16.zero` intrinsic; known as `__nvvm_sust_b_1d_array_i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i16.zero")
    void sust_b_1d_array_i16_zero(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.1d.array.i32.zero` intrinsic; known as `__nvvm_sust_b_1d_array_i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i32.zero")
    void sust_b_1d_array_i32_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.i64.zero` intrinsic; known as `__nvvm_sust_b_1d_array_i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.i64.zero")
    void sust_b_1d_array_i64_zero(i64, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.1d.array.v2i8.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v2i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i8.zero")
    void sust_b_1d_array_v2i8_zero(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v2i16.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v2i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i16.zero")
    void sust_b_1d_array_v2i16_zero(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v2i32.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v2i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i32.zero")
    void sust_b_1d_array_v2i32_zero(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.1d.array.v2i64.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v2i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v2i64.zero")
    void sust_b_1d_array_v2i64_zero(i64, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.1d.array.v4i8.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v4i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i8.zero")
    void sust_b_1d_array_v4i8_zero(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v4i16.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v4i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i16.zero")
    void sust_b_1d_array_v4i16_zero(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.1d.array.v4i32.zero` intrinsic; known as `__nvvm_sust_b_1d_array_v4i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.1d.array.v4i32.zero")
    void sust_b_1d_array_v4i32_zero(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.i8.zero` intrinsic; known as `__nvvm_sust_b_2d_i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i8.zero")
    void sust_b_2d_i8_zero(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.i16.zero` intrinsic; known as `__nvvm_sust_b_2d_i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i16.zero")
    void sust_b_2d_i16_zero(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.i32.zero` intrinsic; known as `__nvvm_sust_b_2d_i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i32.zero")
    void sust_b_2d_i32_zero(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.i64.zero` intrinsic; known as `__nvvm_sust_b_2d_i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.i64.zero")
    void sust_b_2d_i64_zero(i64, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.2d.v2i8.zero` intrinsic; known as `__nvvm_sust_b_2d_v2i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i8.zero")
    void sust_b_2d_v2i8_zero(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v2i16.zero` intrinsic; known as `__nvvm_sust_b_2d_v2i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i16.zero")
    void sust_b_2d_v2i16_zero(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v2i32.zero` intrinsic; known as `__nvvm_sust_b_2d_v2i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i32.zero")
    void sust_b_2d_v2i32_zero(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.v2i64.zero` intrinsic; known as `__nvvm_sust_b_2d_v2i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v2i64.zero")
    void sust_b_2d_v2i64_zero(i64, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.2d.v4i8.zero` intrinsic; known as `__nvvm_sust_b_2d_v4i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i8.zero")
    void sust_b_2d_v4i8_zero(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v4i16.zero` intrinsic; known as `__nvvm_sust_b_2d_v4i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i16.zero")
    void sust_b_2d_v4i16_zero(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.v4i32.zero` intrinsic; known as `__nvvm_sust_b_2d_v4i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.v4i32.zero")
    void sust_b_2d_v4i32_zero(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.i8.zero` intrinsic; known as `__nvvm_sust_b_2d_array_i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i8.zero")
    void sust_b_2d_array_i8_zero(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.array.i16.zero` intrinsic; known as `__nvvm_sust_b_2d_array_i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i16.zero")
    void sust_b_2d_array_i16_zero(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.2d.array.i32.zero` intrinsic; known as `__nvvm_sust_b_2d_array_i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i32.zero")
    void sust_b_2d_array_i32_zero(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.i64.zero` intrinsic; known as `__nvvm_sust_b_2d_array_i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.i64.zero")
    void sust_b_2d_array_i64_zero(i64, i32, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.2d.array.v2i8.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v2i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i8.zero")
    void sust_b_2d_array_v2i8_zero(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v2i16.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v2i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i16.zero")
    void sust_b_2d_array_v2i16_zero(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v2i32.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v2i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i32.zero")
    void sust_b_2d_array_v2i32_zero(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.2d.array.v2i64.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v2i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v2i64.zero")
    void sust_b_2d_array_v2i64_zero(i64, i32, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.2d.array.v4i8.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v4i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i8.zero")
    void sust_b_2d_array_v4i8_zero(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v4i16.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v4i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i16.zero")
    void sust_b_2d_array_v4i16_zero(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.2d.array.v4i32.zero` intrinsic; known as `__nvvm_sust_b_2d_array_v4i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.2d.array.v4i32.zero")
    void sust_b_2d_array_v4i32_zero(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.i8.zero` intrinsic; known as `__nvvm_sust_b_3d_i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i8.zero")
    void sust_b_3d_i8_zero(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.3d.i16.zero` intrinsic; known as `__nvvm_sust_b_3d_i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i16.zero")
    void sust_b_3d_i16_zero(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.b.3d.i32.zero` intrinsic; known as `__nvvm_sust_b_3d_i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i32.zero")
    void sust_b_3d_i32_zero(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.i64.zero` intrinsic; known as `__nvvm_sust_b_3d_i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.i64.zero")
    void sust_b_3d_i64_zero(i64, i32, i32, i32, i64);
    /// The `llvm.nvvm.sust.b.3d.v2i8.zero` intrinsic; known as `__nvvm_sust_b_3d_v2i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i8.zero")
    void sust_b_3d_v2i8_zero(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v2i16.zero` intrinsic; known as `__nvvm_sust_b_3d_v2i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i16.zero")
    void sust_b_3d_v2i16_zero(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v2i32.zero` intrinsic; known as `__nvvm_sust_b_3d_v2i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i32.zero")
    void sust_b_3d_v2i32_zero(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.b.3d.v2i64.zero` intrinsic; known as `__nvvm_sust_b_3d_v2i64_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v2i64.zero")
    void sust_b_3d_v2i64_zero(i64, i32, i32, i32, i64, i64);
    /// The `llvm.nvvm.sust.b.3d.v4i8.zero` intrinsic; known as `__nvvm_sust_b_3d_v4i8_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i8.zero")
    void sust_b_3d_v4i8_zero(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v4i16.zero` intrinsic; known as `__nvvm_sust_b_3d_v4i16_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i16.zero")
    void sust_b_3d_v4i16_zero(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.b.3d.v4i32.zero` intrinsic; known as `__nvvm_sust_b_3d_v4i32_zero` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.b.3d.v4i32.zero")
    void sust_b_3d_v4i32_zero(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.1d.i8.trap` intrinsic; known as `__nvvm_sust_p_1d_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.i8.trap")
    void sust_p_1d_i8_trap(i64, i32, i16);
    /// The `llvm.nvvm.sust.p.1d.i16.trap` intrinsic; known as `__nvvm_sust_p_1d_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.i16.trap")
    void sust_p_1d_i16_trap(i64, i32, i16);
    /// The `llvm.nvvm.sust.p.1d.i32.trap` intrinsic; known as `__nvvm_sust_p_1d_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.i32.trap")
    void sust_p_1d_i32_trap(i64, i32, i32);
    /// The `llvm.nvvm.sust.p.1d.v2i8.trap` intrinsic; known as `__nvvm_sust_p_1d_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.v2i8.trap")
    void sust_p_1d_v2i8_trap(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.v2i16.trap` intrinsic; known as `__nvvm_sust_p_1d_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.v2i16.trap")
    void sust_p_1d_v2i16_trap(i64, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.v2i32.trap` intrinsic; known as `__nvvm_sust_p_1d_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.v2i32.trap")
    void sust_p_1d_v2i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.1d.v4i8.trap` intrinsic; known as `__nvvm_sust_p_1d_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.v4i8.trap")
    void sust_p_1d_v4i8_trap(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.v4i16.trap` intrinsic; known as `__nvvm_sust_p_1d_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.v4i16.trap")
    void sust_p_1d_v4i16_trap(i64, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.v4i32.trap` intrinsic; known as `__nvvm_sust_p_1d_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.v4i32.trap")
    void sust_p_1d_v4i32_trap(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.1d.array.i8.trap` intrinsic; known as `__nvvm_sust_p_1d_array_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.i8.trap")
    void sust_p_1d_array_i8_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.1d.array.i16.trap` intrinsic; known as `__nvvm_sust_p_1d_array_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.i16.trap")
    void sust_p_1d_array_i16_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.1d.array.i32.trap` intrinsic; known as `__nvvm_sust_p_1d_array_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.i32.trap")
    void sust_p_1d_array_i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.1d.array.v2i8.trap` intrinsic; known as `__nvvm_sust_p_1d_array_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.v2i8.trap")
    void sust_p_1d_array_v2i8_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.array.v2i16.trap` intrinsic; known as `__nvvm_sust_p_1d_array_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.v2i16.trap")
    void sust_p_1d_array_v2i16_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.array.v2i32.trap` intrinsic; known as `__nvvm_sust_p_1d_array_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.v2i32.trap")
    void sust_p_1d_array_v2i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.1d.array.v4i8.trap` intrinsic; known as `__nvvm_sust_p_1d_array_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.v4i8.trap")
    void sust_p_1d_array_v4i8_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.array.v4i16.trap` intrinsic; known as `__nvvm_sust_p_1d_array_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.v4i16.trap")
    void sust_p_1d_array_v4i16_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.1d.array.v4i32.trap` intrinsic; known as `__nvvm_sust_p_1d_array_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.1d.array.v4i32.trap")
    void sust_p_1d_array_v4i32_trap(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.2d.i8.trap` intrinsic; known as `__nvvm_sust_p_2d_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.i8.trap")
    void sust_p_2d_i8_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.2d.i16.trap` intrinsic; known as `__nvvm_sust_p_2d_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.i16.trap")
    void sust_p_2d_i16_trap(i64, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.2d.i32.trap` intrinsic; known as `__nvvm_sust_p_2d_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.i32.trap")
    void sust_p_2d_i32_trap(i64, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.2d.v2i8.trap` intrinsic; known as `__nvvm_sust_p_2d_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.v2i8.trap")
    void sust_p_2d_v2i8_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.v2i16.trap` intrinsic; known as `__nvvm_sust_p_2d_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.v2i16.trap")
    void sust_p_2d_v2i16_trap(i64, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.v2i32.trap` intrinsic; known as `__nvvm_sust_p_2d_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.v2i32.trap")
    void sust_p_2d_v2i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.2d.v4i8.trap` intrinsic; known as `__nvvm_sust_p_2d_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.v4i8.trap")
    void sust_p_2d_v4i8_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.v4i16.trap` intrinsic; known as `__nvvm_sust_p_2d_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.v4i16.trap")
    void sust_p_2d_v4i16_trap(i64, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.v4i32.trap` intrinsic; known as `__nvvm_sust_p_2d_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.v4i32.trap")
    void sust_p_2d_v4i32_trap(i64, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.2d.array.i8.trap` intrinsic; known as `__nvvm_sust_p_2d_array_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.i8.trap")
    void sust_p_2d_array_i8_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.2d.array.i16.trap` intrinsic; known as `__nvvm_sust_p_2d_array_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.i16.trap")
    void sust_p_2d_array_i16_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.2d.array.i32.trap` intrinsic; known as `__nvvm_sust_p_2d_array_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.i32.trap")
    void sust_p_2d_array_i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.2d.array.v2i8.trap` intrinsic; known as `__nvvm_sust_p_2d_array_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.v2i8.trap")
    void sust_p_2d_array_v2i8_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.array.v2i16.trap` intrinsic; known as `__nvvm_sust_p_2d_array_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.v2i16.trap")
    void sust_p_2d_array_v2i16_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.array.v2i32.trap` intrinsic; known as `__nvvm_sust_p_2d_array_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.v2i32.trap")
    void sust_p_2d_array_v2i32_trap(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.2d.array.v4i8.trap` intrinsic; known as `__nvvm_sust_p_2d_array_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.v4i8.trap")
    void sust_p_2d_array_v4i8_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.array.v4i16.trap` intrinsic; known as `__nvvm_sust_p_2d_array_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.v4i16.trap")
    void sust_p_2d_array_v4i16_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.2d.array.v4i32.trap` intrinsic; known as `__nvvm_sust_p_2d_array_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.2d.array.v4i32.trap")
    void sust_p_2d_array_v4i32_trap(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.3d.i8.trap` intrinsic; known as `__nvvm_sust_p_3d_i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.i8.trap")
    void sust_p_3d_i8_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.3d.i16.trap` intrinsic; known as `__nvvm_sust_p_3d_i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.i16.trap")
    void sust_p_3d_i16_trap(i64, i32, i32, i32, i16);
    /// The `llvm.nvvm.sust.p.3d.i32.trap` intrinsic; known as `__nvvm_sust_p_3d_i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.i32.trap")
    void sust_p_3d_i32_trap(i64, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.3d.v2i8.trap` intrinsic; known as `__nvvm_sust_p_3d_v2i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.v2i8.trap")
    void sust_p_3d_v2i8_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.3d.v2i16.trap` intrinsic; known as `__nvvm_sust_p_3d_v2i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.v2i16.trap")
    void sust_p_3d_v2i16_trap(i64, i32, i32, i32, i16, i16);
    /// The `llvm.nvvm.sust.p.3d.v2i32.trap` intrinsic; known as `__nvvm_sust_p_3d_v2i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.v2i32.trap")
    void sust_p_3d_v2i32_trap(i64, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.sust.p.3d.v4i8.trap` intrinsic; known as `__nvvm_sust_p_3d_v4i8_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.v4i8.trap")
    void sust_p_3d_v4i8_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.3d.v4i16.trap` intrinsic; known as `__nvvm_sust_p_3d_v4i16_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.v4i16.trap")
    void sust_p_3d_v4i16_trap(i64, i32, i32, i32, i16, i16, i16, i16);
    /// The `llvm.nvvm.sust.p.3d.v4i32.trap` intrinsic; known as `__nvvm_sust_p_3d_v4i32_trap` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.sust.p.3d.v4i32.trap")
    void sust_p_3d_v4i32_trap(i64, i32, i32, i32, i32, i32, i32, i32);
    /// The `llvm.nvvm.rotate.b32` intrinsic; known as `__nvvm_rotate_b32` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rotate.b32")
    i32 rotate_b32(i32, i32);
    /// The `llvm.nvvm.rotate.b64` intrinsic; known as `__nvvm_rotate_b64` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rotate.b64")
    i64 rotate_b64(i64, i32);
    /// The `llvm.nvvm.rotate.right.b64` intrinsic; known as `__nvvm_rotate_right_b64` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.rotate.right.b64")
    i64 rotate_right_b64(i64, i32);
    /// The `llvm.nvvm.swap.lo.hi.b64` intrinsic; known as `__nvvm_swap_lo_hi_b64` in GCC.
    pragma(LDC_intrinsic, "llvm.nvvm.swap.lo.hi.b64")
    i64 swap_lo_hi_b64(i64);
}

/// LLVM intrinsics for the ppc architecture.
version (ppc) {
    /// The `llvm.ppc.dcba` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcba")
    void dcba(i8*);
    /// The `llvm.ppc.dcbf` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbf")
    void dcbf(i8*);
    /// The `llvm.ppc.dcbi` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbi")
    void dcbi(i8*);
    /// The `llvm.ppc.dcbst` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbst")
    void dcbst(i8*);
    /// The `llvm.ppc.dcbt` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbt")
    void dcbt(i8*);
    /// The `llvm.ppc.dcbtst` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbtst")
    void dcbtst(i8*);
    /// The `llvm.ppc.dcbz` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbz")
    void dcbz(i8*);
    /// The `llvm.ppc.dcbzl` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.dcbzl")
    void dcbzl(i8*);
    /// The `llvm.ppc.sync` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.sync")
    void sync();
    /// The `llvm.ppc.lwsync` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.lwsync")
    void lwsync();
    /// The `llvm.ppc.mtctr.v16i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.v16i8")
    void mtctr_v16i8(__vector(i8[16]));
    /// The `llvm.ppc.mtctr.i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.i8")
    void mtctr_i8(i8);
    /// The `llvm.ppc.mtctr.v8i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.v8i16")
    void mtctr_v8i16(__vector(i16[8]));
    /// The `llvm.ppc.mtctr.i16` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.i16")
    void mtctr_i16(i16);
    /// The `llvm.ppc.mtctr.v4i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.v4i32")
    void mtctr_v4i32(__vector(i32[4]));
    /// The `llvm.ppc.mtctr.i32` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.i32")
    void mtctr_i32(i32);
    /// The `llvm.ppc.mtctr.v2i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.v2i64")
    void mtctr_v2i64(__vector(i64[2]));
    /// The `llvm.ppc.mtctr.i64` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.mtctr.i64")
    void mtctr_i64(i64);
    /// The `llvm.ppc.is.decremented.ctr.nonzero` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.is.decremented.ctr.nonzero")
    bool is_decremented_ctr_nonzero();
    /// The `llvm.ppc.altivec.dss` intrinsic; known as `__builtin_altivec_dss` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.dss")
    void altivec_dss(i32);
    /// The `llvm.ppc.altivec.dssall` intrinsic; known as `__builtin_altivec_dssall` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.dssall")
    void altivec_dssall();
    /// The `llvm.ppc.altivec.dst` intrinsic; known as `__builtin_altivec_dst` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.dst")
    void altivec_dst(i8*, i32, i32);
    /// The `llvm.ppc.altivec.dstt` intrinsic; known as `__builtin_altivec_dstt` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.dstt")
    void altivec_dstt(i8*, i32, i32);
    /// The `llvm.ppc.altivec.dstst` intrinsic; known as `__builtin_altivec_dstst` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.dstst")
    void altivec_dstst(i8*, i32, i32);
    /// The `llvm.ppc.altivec.dststt` intrinsic; known as `__builtin_altivec_dststt` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.dststt")
    void altivec_dststt(i8*, i32, i32);
    /// The `llvm.ppc.altivec.mfvscr` intrinsic; known as `__builtin_altivec_mfvscr` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.mfvscr")
    __vector(i16[8]) altivec_mfvscr();
    /// The `llvm.ppc.altivec.mtvscr` intrinsic; known as `__builtin_altivec_mtvscr` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.mtvscr")
    void altivec_mtvscr(__vector(i32[4]));
    /// The `llvm.ppc.altivec.lvx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvx")
    __vector(i32[4]) altivec_lvx(i8*);
    /// The `llvm.ppc.altivec.lvxl` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvxl")
    __vector(i32[4]) altivec_lvxl(i8*);
    /// The `llvm.ppc.altivec.lvebx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvebx")
    __vector(i8[16]) altivec_lvebx(i8*);
    /// The `llvm.ppc.altivec.lvehx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvehx")
    __vector(i16[8]) altivec_lvehx(i8*);
    /// The `llvm.ppc.altivec.lvewx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvewx")
    __vector(i32[4]) altivec_lvewx(i8*);
    /// The `llvm.ppc.altivec.stvx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.stvx")
    void altivec_stvx(__vector(i32[4]), i8*);
    /// The `llvm.ppc.altivec.stvxl` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.stvxl")
    void altivec_stvxl(__vector(i32[4]), i8*);
    /// The `llvm.ppc.altivec.stvebx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.stvebx")
    void altivec_stvebx(__vector(i8[16]), i8*);
    /// The `llvm.ppc.altivec.stvehx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.stvehx")
    void altivec_stvehx(__vector(i16[8]), i8*);
    /// The `llvm.ppc.altivec.stvewx` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.stvewx")
    void altivec_stvewx(__vector(i32[4]), i8*);
    /// The `llvm.ppc.altivec.vcmpbfp` intrinsic; known as `__builtin_altivec_vcmpbfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpbfp")
    __vector(i32[4]) altivec_vcmpbfp(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpeqfp` intrinsic; known as `__builtin_altivec_vcmpeqfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpeqfp")
    __vector(i32[4]) altivec_vcmpeqfp(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpgefp` intrinsic; known as `__builtin_altivec_vcmpgefp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgefp")
    __vector(i32[4]) altivec_vcmpgefp(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpgtfp` intrinsic; known as `__builtin_altivec_vcmpgtfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtfp")
    __vector(i32[4]) altivec_vcmpgtfp(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpequw` intrinsic; known as `__builtin_altivec_vcmpequw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpequw")
    __vector(i32[4]) altivec_vcmpequw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vcmpgtsw` intrinsic; known as `__builtin_altivec_vcmpgtsw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtsw")
    __vector(i32[4]) altivec_vcmpgtsw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vcmpgtuw` intrinsic; known as `__builtin_altivec_vcmpgtuw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtuw")
    __vector(i32[4]) altivec_vcmpgtuw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vcmpequh` intrinsic; known as `__builtin_altivec_vcmpequh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpequh")
    __vector(i16[8]) altivec_vcmpequh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vcmpgtsh` intrinsic; known as `__builtin_altivec_vcmpgtsh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtsh")
    __vector(i16[8]) altivec_vcmpgtsh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vcmpgtuh` intrinsic; known as `__builtin_altivec_vcmpgtuh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtuh")
    __vector(i16[8]) altivec_vcmpgtuh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vcmpequb` intrinsic; known as `__builtin_altivec_vcmpequb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpequb")
    __vector(i8[16]) altivec_vcmpequb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vcmpgtsb` intrinsic; known as `__builtin_altivec_vcmpgtsb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtsb")
    __vector(i8[16]) altivec_vcmpgtsb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vcmpgtub` intrinsic; known as `__builtin_altivec_vcmpgtub` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtub")
    __vector(i8[16]) altivec_vcmpgtub(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vcmpbfp.p` intrinsic; known as `__builtin_altivec_vcmpbfp_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpbfp.p")
    i32 altivec_vcmpbfp_p(i32, __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpeqfp.p` intrinsic; known as `__builtin_altivec_vcmpeqfp_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpeqfp.p")
    i32 altivec_vcmpeqfp_p(i32, __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpgefp.p` intrinsic; known as `__builtin_altivec_vcmpgefp_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgefp.p")
    i32 altivec_vcmpgefp_p(i32, __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpgtfp.p` intrinsic; known as `__builtin_altivec_vcmpgtfp_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtfp.p")
    i32 altivec_vcmpgtfp_p(i32, __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vcmpequw.p` intrinsic; known as `__builtin_altivec_vcmpequw_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpequw.p")
    i32 altivec_vcmpequw_p(i32, __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vcmpgtsw.p` intrinsic; known as `__builtin_altivec_vcmpgtsw_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtsw.p")
    i32 altivec_vcmpgtsw_p(i32, __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vcmpgtuw.p` intrinsic; known as `__builtin_altivec_vcmpgtuw_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtuw.p")
    i32 altivec_vcmpgtuw_p(i32, __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vcmpequh.p` intrinsic; known as `__builtin_altivec_vcmpequh_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpequh.p")
    i32 altivec_vcmpequh_p(i32, __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vcmpgtsh.p` intrinsic; known as `__builtin_altivec_vcmpgtsh_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtsh.p")
    i32 altivec_vcmpgtsh_p(i32, __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vcmpgtuh.p` intrinsic; known as `__builtin_altivec_vcmpgtuh_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtuh.p")
    i32 altivec_vcmpgtuh_p(i32, __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vcmpequb.p` intrinsic; known as `__builtin_altivec_vcmpequb_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpequb.p")
    i32 altivec_vcmpequb_p(i32, __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vcmpgtsb.p` intrinsic; known as `__builtin_altivec_vcmpgtsb_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtsb.p")
    i32 altivec_vcmpgtsb_p(i32, __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vcmpgtub.p` intrinsic; known as `__builtin_altivec_vcmpgtub_p` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcmpgtub.p")
    i32 altivec_vcmpgtub_p(i32, __vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vavgsb` intrinsic; known as `__builtin_altivec_vavgsb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vavgsb")
    __vector(i8[16]) altivec_vavgsb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vavgsh` intrinsic; known as `__builtin_altivec_vavgsh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vavgsh")
    __vector(i16[8]) altivec_vavgsh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vavgsw` intrinsic; known as `__builtin_altivec_vavgsw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vavgsw")
    __vector(i32[4]) altivec_vavgsw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vavgub` intrinsic; known as `__builtin_altivec_vavgub` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vavgub")
    __vector(i8[16]) altivec_vavgub(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vavguh` intrinsic; known as `__builtin_altivec_vavguh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vavguh")
    __vector(i16[8]) altivec_vavguh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vavguw` intrinsic; known as `__builtin_altivec_vavguw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vavguw")
    __vector(i32[4]) altivec_vavguw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmaxfp` intrinsic; known as `__builtin_altivec_vmaxfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxfp")
    __vector(f32[4]) altivec_vmaxfp(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vmaxsb` intrinsic; known as `__builtin_altivec_vmaxsb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxsb")
    __vector(i8[16]) altivec_vmaxsb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vmaxsh` intrinsic; known as `__builtin_altivec_vmaxsh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxsh")
    __vector(i16[8]) altivec_vmaxsh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmaxsw` intrinsic; known as `__builtin_altivec_vmaxsw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxsw")
    __vector(i32[4]) altivec_vmaxsw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmaxub` intrinsic; known as `__builtin_altivec_vmaxub` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxub")
    __vector(i8[16]) altivec_vmaxub(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vmaxuh` intrinsic; known as `__builtin_altivec_vmaxuh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxuh")
    __vector(i16[8]) altivec_vmaxuh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmaxuw` intrinsic; known as `__builtin_altivec_vmaxuw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaxuw")
    __vector(i32[4]) altivec_vmaxuw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vminfp` intrinsic; known as `__builtin_altivec_vminfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminfp")
    __vector(f32[4]) altivec_vminfp(__vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vminsb` intrinsic; known as `__builtin_altivec_vminsb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminsb")
    __vector(i8[16]) altivec_vminsb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vminsh` intrinsic; known as `__builtin_altivec_vminsh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminsh")
    __vector(i16[8]) altivec_vminsh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vminsw` intrinsic; known as `__builtin_altivec_vminsw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminsw")
    __vector(i32[4]) altivec_vminsw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vminub` intrinsic; known as `__builtin_altivec_vminub` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminub")
    __vector(i8[16]) altivec_vminub(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vminuh` intrinsic; known as `__builtin_altivec_vminuh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminuh")
    __vector(i16[8]) altivec_vminuh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vminuw` intrinsic; known as `__builtin_altivec_vminuw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vminuw")
    __vector(i32[4]) altivec_vminuw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vaddubs` intrinsic; known as `__builtin_altivec_vaddubs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vaddubs")
    __vector(i8[16]) altivec_vaddubs(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vaddsbs` intrinsic; known as `__builtin_altivec_vaddsbs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vaddsbs")
    __vector(i8[16]) altivec_vaddsbs(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vadduhs` intrinsic; known as `__builtin_altivec_vadduhs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vadduhs")
    __vector(i16[8]) altivec_vadduhs(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vaddshs` intrinsic; known as `__builtin_altivec_vaddshs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vaddshs")
    __vector(i16[8]) altivec_vaddshs(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vadduws` intrinsic; known as `__builtin_altivec_vadduws` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vadduws")
    __vector(i32[4]) altivec_vadduws(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vaddsws` intrinsic; known as `__builtin_altivec_vaddsws` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vaddsws")
    __vector(i32[4]) altivec_vaddsws(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vaddcuw` intrinsic; known as `__builtin_altivec_vaddcuw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vaddcuw")
    __vector(i32[4]) altivec_vaddcuw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsububs` intrinsic; known as `__builtin_altivec_vsububs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsububs")
    __vector(i8[16]) altivec_vsububs(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vsubsbs` intrinsic; known as `__builtin_altivec_vsubsbs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsubsbs")
    __vector(i8[16]) altivec_vsubsbs(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vsubuhs` intrinsic; known as `__builtin_altivec_vsubuhs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsubuhs")
    __vector(i16[8]) altivec_vsubuhs(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vsubshs` intrinsic; known as `__builtin_altivec_vsubshs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsubshs")
    __vector(i16[8]) altivec_vsubshs(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vsubuws` intrinsic; known as `__builtin_altivec_vsubuws` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsubuws")
    __vector(i32[4]) altivec_vsubuws(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsubsws` intrinsic; known as `__builtin_altivec_vsubsws` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsubsws")
    __vector(i32[4]) altivec_vsubsws(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsubcuw` intrinsic; known as `__builtin_altivec_vsubcuw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsubcuw")
    __vector(i32[4]) altivec_vsubcuw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmhaddshs` intrinsic; known as `__builtin_altivec_vmhaddshs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmhaddshs")
    __vector(i16[8]) altivec_vmhaddshs(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmhraddshs` intrinsic; known as `__builtin_altivec_vmhraddshs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmhraddshs")
    __vector(i16[8]) altivec_vmhraddshs(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmaddfp` intrinsic; known as `__builtin_altivec_vmaddfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmaddfp")
    __vector(f32[4]) altivec_vmaddfp(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vnmsubfp` intrinsic; known as `__builtin_altivec_vnmsubfp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vnmsubfp")
    __vector(f32[4]) altivec_vnmsubfp(__vector(f32[4]), __vector(f32[4]), __vector(f32[4]));
    /// The `llvm.ppc.altivec.vmsummbm` intrinsic; known as `__builtin_altivec_vmsummbm` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmsummbm")
    __vector(i32[4]) altivec_vmsummbm(__vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmsumshm` intrinsic; known as `__builtin_altivec_vmsumshm` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmsumshm")
    __vector(i32[4]) altivec_vmsumshm(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmsumshs` intrinsic; known as `__builtin_altivec_vmsumshs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmsumshs")
    __vector(i32[4]) altivec_vmsumshs(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmsumubm` intrinsic; known as `__builtin_altivec_vmsumubm` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmsumubm")
    __vector(i32[4]) altivec_vmsumubm(__vector(i8[16]), __vector(i8[16]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmsumuhm` intrinsic; known as `__builtin_altivec_vmsumuhm` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmsumuhm")
    __vector(i32[4]) altivec_vmsumuhm(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmsumuhs` intrinsic; known as `__builtin_altivec_vmsumuhs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmsumuhs")
    __vector(i32[4]) altivec_vmsumuhs(__vector(i16[8]), __vector(i16[8]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmulesb` intrinsic; known as `__builtin_altivec_vmulesb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmulesb")
    __vector(i16[8]) altivec_vmulesb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vmulesh` intrinsic; known as `__builtin_altivec_vmulesh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmulesh")
    __vector(i32[4]) altivec_vmulesh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmuleub` intrinsic; known as `__builtin_altivec_vmuleub` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmuleub")
    __vector(i16[8]) altivec_vmuleub(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vmuleuh` intrinsic; known as `__builtin_altivec_vmuleuh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmuleuh")
    __vector(i32[4]) altivec_vmuleuh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmulosb` intrinsic; known as `__builtin_altivec_vmulosb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmulosb")
    __vector(i16[8]) altivec_vmulosb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vmulosh` intrinsic; known as `__builtin_altivec_vmulosh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmulosh")
    __vector(i32[4]) altivec_vmulosh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vmuloub` intrinsic; known as `__builtin_altivec_vmuloub` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmuloub")
    __vector(i16[8]) altivec_vmuloub(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vmulouh` intrinsic; known as `__builtin_altivec_vmulouh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmulouh")
    __vector(i32[4]) altivec_vmulouh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vsumsws` intrinsic; known as `__builtin_altivec_vsumsws` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsumsws")
    __vector(i32[4]) altivec_vsumsws(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsum2sws` intrinsic; known as `__builtin_altivec_vsum2sws` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsum2sws")
    __vector(i32[4]) altivec_vsum2sws(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsum4sbs` intrinsic; known as `__builtin_altivec_vsum4sbs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsum4sbs")
    __vector(i32[4]) altivec_vsum4sbs(__vector(i8[16]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsum4shs` intrinsic; known as `__builtin_altivec_vsum4shs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsum4shs")
    __vector(i32[4]) altivec_vsum4shs(__vector(i16[8]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsum4ubs` intrinsic; known as `__builtin_altivec_vsum4ubs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsum4ubs")
    __vector(i32[4]) altivec_vsum4ubs(__vector(i8[16]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vmladduhm` intrinsic; known as `__builtin_altivec_vmladduhm` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vmladduhm")
    __vector(i16[8]) altivec_vmladduhm(__vector(i16[8]), __vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vpkpx` intrinsic; known as `__builtin_altivec_vpkpx` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkpx")
    __vector(i16[8]) altivec_vpkpx(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vpkshss` intrinsic; known as `__builtin_altivec_vpkshss` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkshss")
    __vector(i8[16]) altivec_vpkshss(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vpkshus` intrinsic; known as `__builtin_altivec_vpkshus` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkshus")
    __vector(i8[16]) altivec_vpkshus(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vpkswss` intrinsic; known as `__builtin_altivec_vpkswss` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkswss")
    __vector(i8[16]) altivec_vpkswss(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vpkswus` intrinsic; known as `__builtin_altivec_vpkswus` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkswus")
    __vector(i16[8]) altivec_vpkswus(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vpkuhus` intrinsic; known as `__builtin_altivec_vpkuhus` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkuhus")
    __vector(i8[16]) altivec_vpkuhus(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vpkuwus` intrinsic; known as `__builtin_altivec_vpkuwus` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vpkuwus")
    __vector(i16[8]) altivec_vpkuwus(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vupkhpx` intrinsic; known as `__builtin_altivec_vupkhpx` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vupkhpx")
    __vector(i32[4]) altivec_vupkhpx(__vector(i16[8]));
    /// The `llvm.ppc.altivec.vupkhsb` intrinsic; known as `__builtin_altivec_vupkhsb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vupkhsb")
    __vector(i16[8]) altivec_vupkhsb(__vector(i8[16]));
    /// The `llvm.ppc.altivec.vupkhsh` intrinsic; known as `__builtin_altivec_vupkhsh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vupkhsh")
    __vector(i32[4]) altivec_vupkhsh(__vector(i16[8]));
    /// The `llvm.ppc.altivec.vupklpx` intrinsic; known as `__builtin_altivec_vupklpx` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vupklpx")
    __vector(i32[4]) altivec_vupklpx(__vector(i16[8]));
    /// The `llvm.ppc.altivec.vupklsb` intrinsic; known as `__builtin_altivec_vupklsb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vupklsb")
    __vector(i16[8]) altivec_vupklsb(__vector(i8[16]));
    /// The `llvm.ppc.altivec.vupklsh` intrinsic; known as `__builtin_altivec_vupklsh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vupklsh")
    __vector(i32[4]) altivec_vupklsh(__vector(i16[8]));
    /// The `llvm.ppc.altivec.vcfsx` intrinsic; known as `__builtin_altivec_vcfsx` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcfsx")
    __vector(f32[4]) altivec_vcfsx(__vector(i32[4]), i32);
    /// The `llvm.ppc.altivec.vcfux` intrinsic; known as `__builtin_altivec_vcfux` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vcfux")
    __vector(f32[4]) altivec_vcfux(__vector(i32[4]), i32);
    /// The `llvm.ppc.altivec.vctsxs` intrinsic; known as `__builtin_altivec_vctsxs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vctsxs")
    __vector(i32[4]) altivec_vctsxs(__vector(f32[4]), i32);
    /// The `llvm.ppc.altivec.vctuxs` intrinsic; known as `__builtin_altivec_vctuxs` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vctuxs")
    __vector(i32[4]) altivec_vctuxs(__vector(f32[4]), i32);
    /// The `llvm.ppc.altivec.vrfim` intrinsic; known as `__builtin_altivec_vrfim` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrfim")
    __vector(f32[4]) altivec_vrfim(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vrfin` intrinsic; known as `__builtin_altivec_vrfin` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrfin")
    __vector(f32[4]) altivec_vrfin(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vrfip` intrinsic; known as `__builtin_altivec_vrfip` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrfip")
    __vector(f32[4]) altivec_vrfip(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vrfiz` intrinsic; known as `__builtin_altivec_vrfiz` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrfiz")
    __vector(f32[4]) altivec_vrfiz(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vsl` intrinsic; known as `__builtin_altivec_vsl` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsl")
    __vector(i32[4]) altivec_vsl(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vslo` intrinsic; known as `__builtin_altivec_vslo` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vslo")
    __vector(i32[4]) altivec_vslo(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vslb` intrinsic; known as `__builtin_altivec_vslb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vslb")
    __vector(i8[16]) altivec_vslb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vslh` intrinsic; known as `__builtin_altivec_vslh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vslh")
    __vector(i16[8]) altivec_vslh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vslw` intrinsic; known as `__builtin_altivec_vslw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vslw")
    __vector(i32[4]) altivec_vslw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsr` intrinsic; known as `__builtin_altivec_vsr` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsr")
    __vector(i32[4]) altivec_vsr(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsro` intrinsic; known as `__builtin_altivec_vsro` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsro")
    __vector(i32[4]) altivec_vsro(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsrb` intrinsic; known as `__builtin_altivec_vsrb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsrb")
    __vector(i8[16]) altivec_vsrb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vsrh` intrinsic; known as `__builtin_altivec_vsrh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsrh")
    __vector(i16[8]) altivec_vsrh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vsrw` intrinsic; known as `__builtin_altivec_vsrw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsrw")
    __vector(i32[4]) altivec_vsrw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vsrab` intrinsic; known as `__builtin_altivec_vsrab` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsrab")
    __vector(i8[16]) altivec_vsrab(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vsrah` intrinsic; known as `__builtin_altivec_vsrah` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsrah")
    __vector(i16[8]) altivec_vsrah(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vsraw` intrinsic; known as `__builtin_altivec_vsraw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsraw")
    __vector(i32[4]) altivec_vsraw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vrlb` intrinsic; known as `__builtin_altivec_vrlb` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrlb")
    __vector(i8[16]) altivec_vrlb(__vector(i8[16]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vrlh` intrinsic; known as `__builtin_altivec_vrlh` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrlh")
    __vector(i16[8]) altivec_vrlh(__vector(i16[8]), __vector(i16[8]));
    /// The `llvm.ppc.altivec.vrlw` intrinsic; known as `__builtin_altivec_vrlw` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrlw")
    __vector(i32[4]) altivec_vrlw(__vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.lvsl` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvsl")
    __vector(i8[16]) altivec_lvsl(i8*);
    /// The `llvm.ppc.altivec.lvsr` intrinsic.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.lvsr")
    __vector(i8[16]) altivec_lvsr(i8*);
    /// The `llvm.ppc.altivec.vperm` intrinsic; known as `__builtin_altivec_vperm_4si` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vperm")
    __vector(i32[4]) altivec_vperm(__vector(i32[4]), __vector(i32[4]), __vector(i8[16]));
    /// The `llvm.ppc.altivec.vsel` intrinsic; known as `__builtin_altivec_vsel_4si` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vsel")
    __vector(i32[4]) altivec_vsel(__vector(i32[4]), __vector(i32[4]), __vector(i32[4]));
    /// The `llvm.ppc.altivec.vexptefp` intrinsic; known as `__builtin_altivec_vexptefp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vexptefp")
    __vector(f32[4]) altivec_vexptefp(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vlogefp` intrinsic; known as `__builtin_altivec_vlogefp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vlogefp")
    __vector(f32[4]) altivec_vlogefp(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vrefp` intrinsic; known as `__builtin_altivec_vrefp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrefp")
    __vector(f32[4]) altivec_vrefp(__vector(f32[4]));
    /// The `llvm.ppc.altivec.vrsqrtefp` intrinsic; known as `__builtin_altivec_vrsqrtefp` in GCC.
    pragma(LDC_intrinsic, "llvm.ppc.altivec.vrsqrtefp")
    __vector(f32[4]) altivec_vrsqrtefp(__vector(f32[4]));
}

/// LLVM intrinsics for the ptx architecture.
version (ptx) {
    /// The `llvm.ptx.read.laneid` intrinsic; known as `__builtin_ptx_read_laneid` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.laneid")
    i32 read_laneid();
    /// The `llvm.ptx.read.warpid` intrinsic; known as `__builtin_ptx_read_warpid` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.warpid")
    i32 read_warpid();
    /// The `llvm.ptx.read.nwarpid` intrinsic; known as `__builtin_ptx_read_nwarpid` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.nwarpid")
    i32 read_nwarpid();
    /// The `llvm.ptx.read.smid` intrinsic; known as `__builtin_ptx_read_smid` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.smid")
    i32 read_smid();
    /// The `llvm.ptx.read.nsmid` intrinsic; known as `__builtin_ptx_read_nsmid` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.nsmid")
    i32 read_nsmid();
    /// The `llvm.ptx.read.gridid` intrinsic; known as `__builtin_ptx_read_gridid` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.gridid")
    i32 read_gridid();
    /// The `llvm.ptx.read.lanemask.eq` intrinsic; known as `__builtin_ptx_read_lanemask_eq` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.lanemask.eq")
    i32 read_lanemask_eq();
    /// The `llvm.ptx.read.lanemask.le` intrinsic; known as `__builtin_ptx_read_lanemask_le` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.lanemask.le")
    i32 read_lanemask_le();
    /// The `llvm.ptx.read.lanemask.lt` intrinsic; known as `__builtin_ptx_read_lanemask_lt` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.lanemask.lt")
    i32 read_lanemask_lt();
    /// The `llvm.ptx.read.lanemask.ge` intrinsic; known as `__builtin_ptx_read_lanemask_ge` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.lanemask.ge")
    i32 read_lanemask_ge();
    /// The `llvm.ptx.read.lanemask.gt` intrinsic; known as `__builtin_ptx_read_lanemask_gt` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.lanemask.gt")
    i32 read_lanemask_gt();
    /// The `llvm.ptx.read.clock` intrinsic; known as `__builtin_ptx_read_clock` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.clock")
    i32 read_clock();
    /// The `llvm.ptx.read.clock64` intrinsic; known as `__builtin_ptx_read_clock64` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.clock64")
    i64 read_clock64();
    /// The `llvm.ptx.read.pm0` intrinsic; known as `__builtin_ptx_read_pm0` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.pm0")
    i32 read_pm0();
    /// The `llvm.ptx.read.pm1` intrinsic; known as `__builtin_ptx_read_pm1` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.pm1")
    i32 read_pm1();
    /// The `llvm.ptx.read.pm2` intrinsic; known as `__builtin_ptx_read_pm2` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.pm2")
    i32 read_pm2();
    /// The `llvm.ptx.read.pm3` intrinsic; known as `__builtin_ptx_read_pm3` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.read.pm3")
    i32 read_pm3();
    /// The `llvm.ptx.bar.sync` intrinsic; known as `__builtin_ptx_bar_sync` in GCC.
    pragma(LDC_intrinsic, "llvm.ptx.bar.sync")
    void bar_sync(i32);
}

/// LLVM intrinsics for the xcore architecture.
version (xcore) {
    /// The `llvm.xcore.bitrev` intrinsic; known as `__builtin_bitrev` in GCC.
    pragma(LDC_intrinsic, "llvm.xcore.bitrev")
    i32 bitrev(i32);
    /// The `llvm.xcore.crc32` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.crc32")
    i32 crc32(i32, i32, i32);
    /// The `llvm.xcore.sext` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.sext")
    i32 sext(i32, i32);
    /// The `llvm.xcore.zext` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.zext")
    i32 zext(i32, i32);
    /// The `llvm.xcore.getid` intrinsic; known as `__builtin_getid` in GCC.
    pragma(LDC_intrinsic, "llvm.xcore.getid")
    i32 getid();
    /// The `llvm.xcore.getps` intrinsic; known as `__builtin_getps` in GCC.
    pragma(LDC_intrinsic, "llvm.xcore.getps")
    i32 getps(i32);
    /// The `llvm.xcore.setps` intrinsic; known as `__builtin_setps` in GCC.
    pragma(LDC_intrinsic, "llvm.xcore.setps")
    void setps(i32, i32);
    /// The `llvm.xcore.geted` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.geted")
    i32 geted();
    /// The `llvm.xcore.getet` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.getet")
    i32 getet();
    /// The `llvm.xcore.setsr` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setsr")
    void setsr(i32);
    /// The `llvm.xcore.clrsr` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.clrsr")
    void clrsr(i32);
    /// The `llvm.xcore.getr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.getr.p0i8")
    i8* getr_p0i8(i32);
    /// The `llvm.xcore.freer.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.freer.p0i8")
    void freer_p0i8(i8*);
    /// The `llvm.xcore.in.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.in.p0i8")
    i32 in_p0i8(i8*);
    /// The `llvm.xcore.int.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.int.p0i8")
    i32 int_p0i8(i8*);
    /// The `llvm.xcore.inct.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.inct.p0i8")
    i32 inct_p0i8(i8*);
    /// The `llvm.xcore.out.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.out.p0i8")
    void out_p0i8(i8*, i32);
    /// The `llvm.xcore.outt.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.outt.p0i8")
    void outt_p0i8(i8*, i32);
    /// The `llvm.xcore.outct.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.outct.p0i8")
    void outct_p0i8(i8*, i32);
    /// The `llvm.xcore.chkct.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.chkct.p0i8")
    void chkct_p0i8(i8*, i32);
    /// The `llvm.xcore.testct.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.testct.p0i8")
    i32 testct_p0i8(i8*);
    /// The `llvm.xcore.testwct.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.testwct.p0i8")
    i32 testwct_p0i8(i8*);
    /// The `llvm.xcore.setd.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setd.p0i8")
    void setd_p0i8(i8*, i32);
    /// The `llvm.xcore.setc.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setc.p0i8")
    void setc_p0i8(i8*, i32);
    /// The `llvm.xcore.inshr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.inshr.p0i8")
    i32 inshr_p0i8(i8*, i32);
    /// The `llvm.xcore.outshr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.outshr.p0i8")
    i32 outshr_p0i8(i8*, i32);
    /// The `llvm.xcore.setpt.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setpt.p0i8")
    void setpt_p0i8(i8*, i32);
    /// The `llvm.xcore.clrpt.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.clrpt.p0i8")
    void clrpt_p0i8(i8*);
    /// The `llvm.xcore.getts.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.getts.p0i8")
    i32 getts_p0i8(i8*);
    /// The `llvm.xcore.syncr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.syncr.p0i8")
    void syncr_p0i8(i8*);
    /// The `llvm.xcore.settw.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.settw.p0i8")
    void settw_p0i8(i8*, i32);
    /// The `llvm.xcore.setv.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setv.p0i8")
    void setv_p0i8(i8*, i8*);
    /// The `llvm.xcore.setev.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setev.p0i8")
    void setev_p0i8(i8*, i8*);
    /// The `llvm.xcore.eeu.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.eeu.p0i8")
    void eeu_p0i8(i8*);
    /// The `llvm.xcore.edu.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.edu.p0i8")
    void edu_p0i8(i8*);
    /// The `llvm.xcore.setclk.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setclk.p0i8.p0i8")
    void setclk_p0i8_p0i8(i8*, i8*);
    /// The `llvm.xcore.setrdy.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setrdy.p0i8.p0i8")
    void setrdy_p0i8_p0i8(i8*, i8*);
    /// The `llvm.xcore.setpsc.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.setpsc.p0i8")
    void setpsc_p0i8(i8*, i32);
    /// The `llvm.xcore.peek.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.peek.p0i8")
    i32 peek_p0i8(i8*);
    /// The `llvm.xcore.endin.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.endin.p0i8")
    i32 endin_p0i8(i8*);
    /// The `llvm.xcore.waitevent` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.waitevent")
    i8* waitevent();
    /// The `llvm.xcore.checkevent` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.checkevent")
    i8* checkevent(i8*);
    /// The `llvm.xcore.clre` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.clre")
    void clre();
    /// The `llvm.xcore.getst.p0i8.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.getst.p0i8.p0i8")
    i8* getst_p0i8_p0i8(i8*);
    /// The `llvm.xcore.msync.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.msync.p0i8")
    void msync_p0i8(i8*);
    /// The `llvm.xcore.ssync` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.ssync")
    void ssync();
    /// The `llvm.xcore.mjoin.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.mjoin.p0i8")
    void mjoin_p0i8(i8*);
    /// The `llvm.xcore.initsp.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.initsp.p0i8")
    void initsp_p0i8(i8*, i8*);
    /// The `llvm.xcore.initpc.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.initpc.p0i8")
    void initpc_p0i8(i8*, i8*);
    /// The `llvm.xcore.initlr.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.initlr.p0i8")
    void initlr_p0i8(i8*, i8*);
    /// The `llvm.xcore.initcp.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.initcp.p0i8")
    void initcp_p0i8(i8*, i8*);
    /// The `llvm.xcore.initdp.p0i8` intrinsic.
    pragma(LDC_intrinsic, "llvm.xcore.initdp.p0i8")
    void initdp_p0i8(i8*, i8*);
}

}
